/**
 * This class represents an HL7 Data Element
 */

package gov.nist.hl7refdb.generator;

// JDK
import java.util.Vector;

// DOM
import org.w3c.dom.Node;
import org.w3c.dom.Element;
import org.w3c.dom.Document;

public class HL7DataElement {
    /** A Vector to store HL7 data values */
    private Vector vDataValue;

    /** Item number in the HL7 Database */
    protected long item_number;
    /** Segment associated with the primitive data element */
    protected String segment;
    /** HL7 name of the primitive data element */
    protected String name;
    /** Name of the parent for this element */
    protected String qualifier;
    /** Item number of the parent for this element */
    protected long qualifier_item_number;
    /** Datatype of the primitive data element */
    protected String datatype;
    /** Length of the element as given in the HL7 DB*/
    protected long length;
    /** Describes the allowable range of the element */
    protected String range;
    /** Describes the match pattern (regular expression) of the element */
    protected String format;
    /** Table ID for this element */
    protected long table_id;
    /** Reference to the specification for this element */
    protected String reference;
    /** Indicates the present of a configuration value and the name of the file it is located in*/
    protected String config_ref;
    /** Source of the data */
    protected String source;
    /** Version of the standard this element applies to */
    protected String version;
    /** Indicates the status of the data entry for this element in an administrative sense */
    protected String status;
    /** Provides additional space for comments */
    protected String notes;

    /**
     * Constructor - init the Vector
     */
    public HL7DataElement() {
	vDataValue = new Vector();
    }

    // Get Methods
    public long getItemNumber() {
	return item_number;
    }

    public String getSegment() {
	return segment;
    }

    public String getName() {
	return name;
    }

    public String getQualifier() {
	return qualifier;
    }

    public long getQualifierItemNumber() {
	return qualifier_item_number;
    }

    public String getDatatype() {
	return datatype;
    }

    public long getLength() {
	return length;
    }

    public String getRange() {
	return range;
    }

    public String getFormat() {
	return format;
    }

    public long getTableId() {
	return table_id;
    }

    public String getReference() {
	return reference;
    }

    public String getConfigRef() {
	return config_ref;
    }

    public String getSource() {
	return source;
    }

    public String getVersion() {
	return version;
    }

    public String getStatus() {
	return status;
    }

    public String getNotes() {
	return notes;
    }

    // Set Methods
    public void setItemNumber(long item_number) {
	this.item_number = item_number;
    }

    public void setSegment(String segment) {
	this.segment = segment;
    }

    public void setName(String name) {
	this.name = name;
    }

    public void setQualifier(String qualifier) {
	this.qualifier = qualifier;
    }

    public void setQualifierItemNumber(long qualifier_item_number) {
	this.qualifier_item_number = qualifier_item_number;
    }

    public void setDatatype(String datatype) {
	this.datatype = datatype;
    }

    public void setLength(long length) {
	this.length = length;
    }

    public void setRange(String range) {
	this.range = range;
    }

    public void setFormat(String format) {
	this.format = format;
    }

    public void setTableId(long table_id) {
	this.table_id = table_id;
    }

    public void setReference(String reference) {
	this.reference = reference;
    }

    public void setConfigRef(String config_ref) {
	this.config_ref = config_ref;
    }

    public void setSource(String source) {
	this.source = source;
    }

    public void setVersion(String version) {
	this.version = version;
    }

    public void setStatus(String status) {
	this.status = status;
    }

    public void setNotes(String notes) {
	this.notes = notes;
    }

    // Methods

    /**
     * Return true if the two objects are equal
     * means same item_number, segment, name and qualifier
     * @param o The object to be compared to
     * @return true if equals
     */
    public boolean equals(Object o) {
	HL7DataElement hl7de = (HL7DataElement)o;
	boolean equals = true;

	// item_number
	if(item_number != hl7de.getItemNumber())
	    equals = false;
	// segment
	if(equals && segment.compareTo(hl7de.getSegment()) != 0)
	    equals = false;
	// name
	if(equals && name.compareTo(hl7de.getName()) != 0)
	    equals = false;
	// qualifier
	if(equals && qualifier.compareTo(hl7de.getQualifier()) != 0)
	    equals = false;
	return equals;
    }

    /**
     * Add a value to this element
     * @param hl7Value The HL7 Data Value to add
     */
    public void addHL7DataValue(HL7DataValue hl7Value) {
	if(hl7Value != null && !vDataValue.contains(hl7Value))
	    vDataValue.add(hl7Value);
    }

    /**
     * Remove a value to this element
     * @param hl7Value The HL7 Data Value to remove
     */
    public void removeHL7DataValue(HL7DataValue hl7Value) {
	if(hl7Value != null && vDataValue.contains(hl7Value))
	    vDataValue.remove(hl7Value);
    }

    /**
     * Create a DOM of this HL7 data element
     * @return a DOM node
     */
    public Node getDOM(Document doc) {
	Element ref = null;
	
	ref = (Element)doc.createElement("ReferenceData");
	// Name
	if(name != null)
	    ref.setAttribute("Name", name);
	// Qualifier
	if(qualifier != null) {
	    int idx = qualifier.indexOf("[");
	    if(idx != -1)
		ref.setAttribute("Qualifier", qualifier.substring(idx+1, qualifier.length()-1));
	    else
		ref.setAttribute("Qualifier", qualifier);
	}
	// Qualifier2
	if(qualifier != null) {
	    int idx = qualifier.indexOf("[");
	    if(idx != -1)
		ref.setAttribute("Qualifier2", qualifier.substring(0, idx));
	    else
		ref.setAttribute("Qualifier2", "");
	}
	// Segment
	if(segment != null)
	    ref.setAttribute("Segment", segment);
	// Datatype
	if(datatype != null)
	    ref.setAttribute("Datatype", datatype);
	// Length L
	if(length != 0)
	    ref.setAttribute("Length", String.valueOf(length));
	// Table ID L
	if(table_id != 0)
	    ref.setAttribute("Tableid", String.valueOf(table_id));
	// Reference
	if(reference != null)
	    ref.setAttribute("Reference", reference);
	// ConfigRef
	if(config_ref != null)
	    ref.setAttribute("Configref", config_ref);
	// Status
	if(status != null)
	    ref.setAttribute("Status", status);
	// Notes
	if(notes != null)
	    ref.setAttribute("Notes", notes);
	// Source
	if(source != null)
	    ref.setAttribute("Source", source);
	// The different values for this element
	for(int i=0;i<vDataValue.size();i++) {
	    HL7DataValue hl7dv = (HL7DataValue)vDataValue.get(i);
	    ref.appendChild(hl7dv.getDOM(doc));
	}
	return ref;
    }
}
