/**
 * This class stores information about the config data
 */

package gov.nist.hl7maker.control;

// JDK
import java.io.File;
import java.io.IOException;
import java.util.Vector;

// XML
import javax.xml.parsers.DocumentBuilder; 
import javax.xml.parsers.DocumentBuilderFactory;  
import javax.xml.parsers.ParserConfigurationException;

// SAX
import org.xml.sax.SAXException;  

// DOM
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Element;

// XSL
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

// MessageMaker
import gov.nist.hl7maker.util.Constants;
import gov.nist.hl7maker.view.config.ValueNode;

public class ConfigData {
    
    private Vector vConfig;

    /**
     * Constructor
     * @param xmlConfig The config data filename
     */
    public ConfigData(String xmlConfig) {
	try {
	    DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
	    DocumentBuilder builder = factory.newDocumentBuilder();
	    Document doc = builder.parse(new File(xmlConfig));
	    vConfig = new Vector();
	    browseXML(doc);
	}
	catch(ParserConfigurationException pce) {
	    System.err.println("ParserConfigurationException " + pce.getMessage());
	    pce.printStackTrace();
	}
	catch(IOException ioe) {
	    System.err.println("IOException " + ioe.getMessage());
	    ioe.printStackTrace();
	}
	catch(SAXException saxe) {
	    System.err.println("SAXException " + saxe.getMessage());
	    saxe.printStackTrace();
	}	
    }

    /**
     * Browse recursively the DOM
     * @param n The current node
     */
    private void browseXML(Node n) {
	String name = null, qualifier = null, qualifier2 = null, segment = null, value = null, usage = null;
	if(n.getNodeName().equals("config")) {
 	    Node attributes = n.getAttributes().getNamedItem("Name");
 	    if(attributes != null) {
		name = attributes.getNodeValue();
	    }
	    attributes = n.getAttributes().getNamedItem("Qualifier");
	    if(attributes != null) {
		qualifier = attributes.getNodeValue();
	    }
	    else {
		qualifier = "";
	    }
	    attributes = n.getAttributes().getNamedItem("Qualifier2");
	    if(attributes != null) {
		qualifier2 = attributes.getNodeValue();
	    }
	    else {
		qualifier2 = "";
	    }
	    attributes = n.getAttributes().getNamedItem("Segment");
	    if(attributes != null) {
		segment = attributes.getNodeValue();
	    }
	    attributes = n.getAttributes().getNamedItem("Value");
	    if(attributes != null) {
		value = attributes.getNodeValue();
	    }
	    attributes = n.getAttributes().getNamedItem("Usage");
	    if(attributes != null) {
		usage = attributes.getNodeValue();
	    }
	    ValueNode curValue = new ValueNode();
	    curValue.setName(name);
	    curValue.setSegment(segment);
	    curValue.setQualifier(qualifier);
	    curValue.setQualifier2(qualifier2);
	    curValue.setValue(value);
	    curValue.setCategory("Config");
	    curValue.setUsage(usage);
	    vConfig.add(curValue);
	}
 	// Apply method on children
 	NodeList nl = n.getChildNodes();
 	for(int i=0;i<nl.getLength();i++) {
 	    browseXML(nl.item(i));
	}
    }

    /** 
     * Get all the config data
     * @return The value
     */
    public Vector getConfigData() {
	return vConfig;
    }

    /**
     * Get the values for a name, qualifier and segment
     * @param name The name
     * @param segment The segment
     * @param qualifier The qualifier
     * @param qualifier2 The qualifier2
     * @return A vector with the values
     */
    public Vector getValues(String name, String segment, String qualifier, String qualifier2) {
	Vector v = new Vector();
	ValueNode tmp = new ValueNode();
	tmp.setName(name);
	tmp.setQualifier(qualifier);
	tmp.setQualifier2(qualifier2);
	tmp.setSegment(segment);

	for(int i=0;i<vConfig.size();i++) {
	    ValueNode vn = (ValueNode)vConfig.get(i);
	    if(tmp.equals(vn) && vn.getStatus() != ValueNode.REMOVED) {
		v.add(vConfig.get(i));
	    }
	}
	return v;
    }

    /**
     * Save the default value
     * @param xmlConfig The config data filename
     */
    public void save(String xmlConfig) {
	try {
	    DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
	    DocumentBuilder builder = factory.newDocumentBuilder();
	    Document doc = builder.newDocument();
	    
	    // Root element
	    Element root = (Element)doc.createElement("DefaultValue");
	    root.setAttribute("xmlns", "http://www.nist.gov/hl7/DefaultValue");
	    doc.appendChild(root);
	    
	    Element magic = (Element)doc.createElement("MessageMaker");
	    root.appendChild(magic);
	    
	    Element value = null;
	    Element item = null;
/*	    for(int i=0;i<vDV.size();i++) {
		DV dv = (DV)vDV.get(i);
		value = (Element)doc.createElement("value");
		// Datatype
		value.setAttribute("datatype", dv.getDatatype());
		Vector vDVItem = dv.getValue();
		for(int j=0;j<vDVItem.size();j++) {
		    DVItem dvItem = (DVItem)vDVItem.get(j);
		    item = (Element)doc.createElement("item");
		    // Value
		    item.setAttribute("value", dvItem.getValue());
		    // Category
		    item.setAttribute("category", dvItem.getCategory());
		    value.appendChild(item);
		}
		root.appendChild(value);
	    }
*/    
	    // Write the DOM document to the file
	    Transformer transformer = TransformerFactory.newInstance().newTransformer();
	    DOMSource source = new DOMSource(doc);
	    StreamResult result = new StreamResult(new File(xmlConfig));
	    transformer.transform(source, result);
	}
	catch(ParserConfigurationException pce) {
	    System.err.println("ParserConfigurationException " + pce.getMessage());
	    pce.printStackTrace();
	}
	catch(TransformerConfigurationException tce) {
	    System.err.println("TransformerConfigurationException " + tce.getMessage());
	    tce.printStackTrace();
	}
	catch(TransformerException te) {
	    System.err.println("TransformerException " + te.getMessage());
	    te.printStackTrace();
	}
    }
}
