/**
 * This class stores information about the default value
 */

package gov.nist.hl7maker.control;

// JDK
import java.io.File;
import java.io.IOException;
import java.util.Vector;

// XML
import javax.xml.parsers.DocumentBuilder; 
import javax.xml.parsers.DocumentBuilderFactory;  
import javax.xml.parsers.ParserConfigurationException;

// SAX
import org.xml.sax.SAXException;  

// DOM
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Element;

// XSL
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

// MessageMaker
import gov.nist.hl7maker.util.Constants;

public class DefaultValue {
    
    private final String DEFAULT_VALUE = "data" + Constants.FS + "DefaultValue.xml";
    private Vector vDV;
    private DV currentDV;

    /**
     * Constructor
     */
    public DefaultValue() {
	try {
	    DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
	    DocumentBuilder builder = factory.newDocumentBuilder();
	    Document doc = builder.parse(new File(DEFAULT_VALUE));
	    vDV = new Vector();
	    browseXML(doc);
	}
	catch(ParserConfigurationException pce) {
	    System.err.println("ParserConfigurationException " + pce.getMessage());
	    pce.printStackTrace();
	}
	catch(IOException ioe) {
	    System.err.println("IOException " + ioe.getMessage());
	    ioe.printStackTrace();
	}
	catch(SAXException saxe) {
	    System.err.println("SAXException " + saxe.getMessage());
	    saxe.printStackTrace();
	}	
    }

    /**
     * Browse recursively the DOM
     * @param n The current node
     */
    private void browseXML(Node n) {
	if(n.getNodeName().equals("value")) {
	    currentDV = new DV();
	    vDV.add(currentDV);
	    // Datatype
	    Node node = n.getAttributes().getNamedItem("datatype");
	    if(node != null) {
		currentDV.setDatatype(node.getNodeValue());
	    }
	}
	else if(n.getNodeName().equals("item")) {
	    // Value
	    String value = null;
	    Node node = n.getAttributes().getNamedItem("value");
	    if(node != null) {
		value = node.getNodeValue();
	    }
	    // Category
	    node = n.getAttributes().getNamedItem("category");
	    if(node != null) {
		currentDV.addValue(value, node.getNodeValue());
	    }
	}
 	// Apply method on children
 	NodeList nl = n.getChildNodes();
 	for(int i=0;i<nl.getLength();i++) {
 	    browseXML(nl.item(i));
	}
    }

    /** 
     * Get the value for all datatypes
     * @return The value
     */
    public Vector getDefaultValue() {
	return vDV;
    }

    /**
     * Get the value for a datatype
     * @param datatype The datatype
     * @return The value
     */
    public Vector getDefaultValue(String datatype) {
	DV tmp = new DV();
	tmp.setDatatype(datatype);
	int idx = vDV.indexOf(tmp);
	if(idx != -1)
	    return ((DV)vDV.get(idx)).getValue();
	return new Vector();
    }

    /**
     * Save the default value
     */
    public void save() {
	try {
	    DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
	    DocumentBuilder builder = factory.newDocumentBuilder();
	    Document doc = builder.newDocument();
	    
	    // Root element
	    Element root = (Element)doc.createElement("DefaultValue");
	    root.setAttribute("xmlns", "http://www.nist.gov/hl7/DefaultValue");
	    doc.appendChild(root);
	    
	    Element magic = (Element)doc.createElement("MessageMaker");
	    root.appendChild(magic);
	    
	    Element value = null;
	    Element item = null;
	    for(int i=0;i<vDV.size();i++) {
		DV dv = (DV)vDV.get(i);
		value = (Element)doc.createElement("value");
		// Datatype
		value.setAttribute("datatype", dv.getDatatype());
		Vector vDVItem = dv.getValue();
		for(int j=0;j<vDVItem.size();j++) {
		    DVItem dvItem = (DVItem)vDVItem.get(j);
		    item = (Element)doc.createElement("item");
		    // Value
		    item.setAttribute("value", dvItem.getValue());
		    // Category
		    item.setAttribute("category", dvItem.getCategory());
		    value.appendChild(item);
		}
		root.appendChild(value);
	    }
	    
	    // Write the DOM document to the file
	    Transformer transformer = TransformerFactory.newInstance().newTransformer();
	    DOMSource source = new DOMSource(doc);
	    StreamResult result = new StreamResult(new File(DEFAULT_VALUE).getPath());
	    transformer.transform(source, result);
	}
	catch(ParserConfigurationException pce) {
	    System.err.println("ParserConfigurationException " + pce.getMessage());
	    pce.printStackTrace();
	}
	catch(TransformerConfigurationException tce) {
	    System.err.println("TransformerConfigurationException " + tce.getMessage());
	    tce.printStackTrace();
	}
	catch(TransformerException te) {
	    System.err.println("TransformerException " + te.getMessage());
	    te.printStackTrace();
	}
    }

    // inner class to represent a default value datatype
    public class DV {

	private String datatype;
	private Vector vDVItem;

	public DV() {
	    vDVItem = new Vector();
	}

	public String getDatatype() {
	    return datatype;
	}

	public void setDatatype(String datatype) {
	    this.datatype = datatype;
	}

	public void addValue(String value, String category) {
	    vDVItem.add(new DVItem(value, category));
	}

	public void removeValue(String value, String category) {
	    DVItem item = new DVItem(value, category);
	    int idx = vDVItem.indexOf(item);
	    if(idx != -1)
		vDVItem.remove(vDVItem.get(idx));
	}

	public void editValue(String value, String category) {
	    DVItem item = new DVItem(value, category);
	    int idx = vDVItem.indexOf(item);
	    if(idx != -1)
		vDVItem.set(idx, item);
	}

	public Vector getValue() {
	    return vDVItem;
	}

	public boolean equals(Object o) {
	    if(o instanceof DV) {
		DV dv = (DV)o;
		return datatype.equals(dv.getDatatype());
	    }
	    return false;
	}

	public String toString() {
	    return datatype;
	}
    }

    // inner class to represent a default value item
    public class DVItem {

	private String value;
	private String category;

	public DVItem(String value, String category) {
	    setValue(value);
	    setCategory(category);
	}

	public String getValue() {
	    return value;
	}

	public String getCategory() {
	    return category;
	}

	public void setValue(String value) {
	    this.value = value;
	}

	public void setCategory(String category) {
	    this.category = category;
	}

	public boolean equals(Object o) {
	    if(o instanceof DVItem) {
		DVItem dvItem = (DVItem)o;
		return value.equals(dvItem.getValue()) && category.equals(dvItem.getCategory());
	    }
	    return false;
	}

	public String toString() {
	    return "[" + category + "] " + value;
	}
    }
}
