/**
 * This class represents a profile
 */

package gov.nist.hl7maker.control;

// JDK
import java.io.File;
import java.io.IOException;
import java.util.Vector;

// Swing
import javax.swing.tree.DefaultMutableTreeNode;

// XML
import javax.xml.parsers.DocumentBuilder; 
import javax.xml.parsers.DocumentBuilderFactory;  
import javax.xml.parsers.ParserConfigurationException;

// SAX
import org.xml.sax.SAXException;  

// DOM
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Element;

// XSL
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

// MessageMaker
import gov.nist.hl7maker.view.config.ProfileNode;
import gov.nist.hl7maker.view.config.SegmentGroupNode;
import gov.nist.hl7maker.view.config.SegmentNode;
import gov.nist.hl7maker.view.config.FieldNode;
import gov.nist.hl7maker.view.config.ComponentNode;
import gov.nist.hl7maker.view.config.SubComponentNode;
import gov.nist.hl7maker.view.config.ValueNode;
import gov.nist.hl7maker.view.config.RepositoryDataModel;

public class Profile {
    
    private String xmlProfile;

    private Vector vSegGroup;
    private Vector vSegment;
    private Vector vField;
    private Vector vComponent;
    private Vector vSubComponent;
    private Vector vConstant;
    private Vector vExample;
    private SegmentGroupNode curSegmentGroup;
    private SegmentNode curSegment;
    private FieldNode curField;
    private ComponentNode curComponent;
    private SubComponentNode curSubComponent;
    private ProfileNode curPN;
    private ProfileNode[] parent;
    private int level;

    private Vector vRoot;

    private DataElement dataElt;
    private int fieldSeqNumber, compSeqNumber, subCompSeqNumber;

    /** 
     * Constructor
     */
    public Profile() {
    }
    
    /**
     * Constructor 
     * @param xmlProfile The filename of the profile
     */
    public Profile(String xmlProfile) {
	this.xmlProfile = xmlProfile;
	loadProfile(xmlProfile);
    }
    
    /**
     * Load a profile
     * @param xmlProfile The filename of the profile
     */
    public void loadProfile(String xmlProfile) {
	try {
	    DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
	    DocumentBuilder builder = factory.newDocumentBuilder();
	    Document doc = builder.parse(new File(xmlProfile));
	    removeEmptyTextElement(doc);

	    vRoot = new Vector();
	    vSegGroup = new Vector();	
	    vSegment = new Vector();
	    vField = new Vector();
	    vComponent = new Vector();
	    vSubComponent = new Vector();
	    parent = new ProfileNode[1024];
	    level = 0;
	    dataElt = new DataElement();
	    vConstant = new Vector();
	    vExample = new Vector();
	    browseXML(doc);
	}
	catch(ParserConfigurationException pce) {
	    System.err.println("ParserConfigurationException " + pce.getMessage());
	    pce.printStackTrace();
	}
	catch(IOException ioe) {
	    System.err.println("IOException " + ioe.getMessage());
	    ioe.printStackTrace();
	}
	catch(SAXException saxe) {
	    System.err.println("SAXException " + saxe.getMessage());
	    saxe.printStackTrace();
	}
    }

    /**
     * This method removes all TEXT_NODE with empty text
     * @param node The current node
     */    
    private void removeEmptyTextElement(Node node) {
	if(node.hasChildNodes()) {
	    NodeList nl = node.getChildNodes();
	    for(int i=0;i<nl.getLength();i++) {
		Node n = nl.item(i);
		if(n.getNodeValue() != null && n.getNodeValue().trim().equals(""))
		    node.removeChild(n);
	    }
	}
	// Execute the method on the children
	if(node.hasChildNodes()) {
	    NodeList nl = node.getChildNodes();
	    for(int i=0;i<nl.getLength();i++) {
		removeEmptyTextElement(nl.item(i));
	    }
	}
    }

    /**
     * Browser recursively the DOM and creates the tree structure
     * @param n The current node
     */
    private void browseXML(Node n) {
	String name = null, table = null, datatype = null, usage = null, cardMin = null, cardMax = null, constant = null, example = null;
	Node nameNode, tableNode, datatypeNode, usageNode, cardMinNode, cardMaxNode, constantNode, exampleNode;
	// SegmentGroup
 	if(n.getNodeName().equals("SegGroup")) {
	    curSegmentGroup = new SegmentGroupNode();
 	    nameNode = n.getAttributes().getNamedItem("Name");
	    if(nameNode != null) {
		name = nameNode.getNodeValue();
	    }
	    tableNode = n.getAttributes().getNamedItem("Table");
	    if(tableNode != null) {
		table = tableNode.getNodeValue();
	    }
	    datatypeNode = n.getAttributes().getNamedItem("Datatype");
	    if(datatypeNode != null) {
		datatype = datatypeNode.getNodeValue();
	    }
	    usageNode = n.getAttributes().getNamedItem("Usage");
	    if(usageNode != null) {
		usage = usageNode.getNodeValue();
	    }
	    // Try Optionality
	    else {
		usageNode = n.getAttributes().getNamedItem("Optionality");
		if(usageNode != null) {
		    usage = usageNode.getNodeValue();
		}
	    }
	    cardMinNode = n.getAttributes().getNamedItem("Min");
	    if(cardMinNode != null) {
		cardMin = cardMinNode.getNodeValue();
	    }
	    cardMaxNode = n.getAttributes().getNamedItem("Max");
	    if(cardMaxNode != null) {
		cardMax = cardMaxNode.getNodeValue();
	    }
	    curSegmentGroup.setName(name);
	    curSegmentGroup.setQualifier("");
	    curSegmentGroup.setQualifier2("");
	    curSegmentGroup.setSegment("");
	    curSegmentGroup.setUsage(usage);
	    curSegmentGroup.setCardMin(cardMin);
	    curSegmentGroup.setCardMax(cardMax);
	    vSegGroup.add(curSegmentGroup);
	    curSegmentGroup.setParent(parent[level-1]);
	    if(!n.getParentNode().getNodeName().equals("SegGroup")) {
		vRoot.add(curSegmentGroup);
	    }
	    else {
		((SegmentGroupNode)curSegmentGroup.getParent()).addChild(curSegmentGroup);
	    }
	    parent[level] = curSegmentGroup;
 	}
	// Segment
 	if(n.getNodeName().equals("Segment")) {
	    curSegment = new SegmentNode();
 	    nameNode = n.getAttributes().getNamedItem("Name");
	    if(nameNode != null) {
		name = nameNode.getNodeValue();
	    }
	    tableNode = n.getAttributes().getNamedItem("Table");
	    if(tableNode != null) {
		table = tableNode.getNodeValue();
	    }
	    datatypeNode = n.getAttributes().getNamedItem("Datatype");
	    if(datatypeNode != null) {
		datatype = datatypeNode.getNodeValue();
	    }
	    usageNode = n.getAttributes().getNamedItem("Usage");
	    if(usageNode != null) {
		usage = usageNode.getNodeValue();
	    }
	    // Try Optionality
	    else {
		usageNode = n.getAttributes().getNamedItem("Optionality");
		if(usageNode != null) {
		    usage = usageNode.getNodeValue();
		}
	    }
	    cardMinNode = n.getAttributes().getNamedItem("Min");
	    if(cardMinNode != null) {
		cardMin = cardMinNode.getNodeValue();
	    }
	    cardMaxNode = n.getAttributes().getNamedItem("Max");
	    if(cardMaxNode != null) {
		cardMax = cardMaxNode.getNodeValue();
	    }
	    curSegment.setName(name);
	    curSegment.setQualifier("");
	    curSegment.setQualifier2("");
	    curSegment.setSegment(name);
	    curSegment.setTable(table);
	    curSegment.setDatatype(datatype);
	    curSegment.setUsage(usage);
	    curSegment.setCardMin(cardMin);
	    curSegment.setCardMax(cardMax);
	    curSegment.setParent(parent[level-1]);
	    if(n.getParentNode().getNodeName().equals("SegGroup")) {
		((SegmentGroupNode)curSegment.getParent()).addChild(curSegment);
	    }
	    else {
		vRoot.add(curSegment);	       
		curSegmentGroup = null;
	    }
	    vSegment.add(curSegment);

	    fieldSeqNumber = 0;
 	}
	// Field
 	else if(n.getNodeName().equals("Field")) {
	    curField = new FieldNode();
	    nameNode = n.getAttributes().getNamedItem("Name");
	    if(nameNode != null) {
		name = nameNode.getNodeValue();
	    }
	    tableNode = n.getAttributes().getNamedItem("Table");
	    if(tableNode != null) {
		table = tableNode.getNodeValue();
	    }
	    datatypeNode = n.getAttributes().getNamedItem("Datatype");
	    if(datatypeNode != null) {
		datatype = datatypeNode.getNodeValue();
	    }
	    usageNode = n.getAttributes().getNamedItem("Usage");
	    if(usageNode != null) {
		usage = usageNode.getNodeValue();
	    }
	    // Try Optionality
	    else {
		usageNode = n.getAttributes().getNamedItem("Optionality");
		if(usageNode != null) {
		    usage = usageNode.getNodeValue();
		}
	    }
	    cardMinNode = n.getAttributes().getNamedItem("Min");
	    if(cardMinNode != null) {
		cardMin = cardMinNode.getNodeValue();
	    }
	    cardMaxNode = n.getAttributes().getNamedItem("Max");
	    if(cardMaxNode != null) {
		cardMax = cardMaxNode.getNodeValue();
	    }
	    constantNode = n.getAttributes().getNamedItem("ConstantValue");
	    if(constantNode != null) {
		constant = constantNode.getNodeValue();
		ProfileConstantValue pcv = new ProfileConstantValue();
		pcv.setName(name);
		pcv.setQualifier("");
		pcv.setQualifier2("");
		pcv.setSegment(curSegment.getName());
		pcv.setValue(constant);
		vConstant.add(pcv);
	    }
	    curField.setName(name);
	    curField.setQualifier("");
	    curField.setQualifier2("");
	    curField.setSegment(curSegment.getName());
	    curField.setTable(table);
	    curField.setDatatype(datatype);
	    curField.setUsage(usage);
	    curField.setCardMin(cardMin);
	    curField.setCardMax(cardMax);
	    int tmpSeqNumber = dataElt.getFieldSequenceNum(curSegment.getName(), name);
	    if(tmpSeqNumber == -1) fieldSeqNumber++;
	    else fieldSeqNumber = tmpSeqNumber;
	    curField.setSequenceNum(String.valueOf(fieldSeqNumber));
	    //curField.setParent(curSegment.getParent());
	    curField.setParent(curSegment);
	    curSegment.addField(curField);
	    vField.add(curField);

	    compSeqNumber = 0;
	    subCompSeqNumber = 0;
	    curPN = curField;
	}
 	// Component
	else if(n.getNodeName().equals("Component")) {
	    curComponent = new ComponentNode();
	    nameNode = n.getAttributes().getNamedItem("Name");
	    if(nameNode != null) {
		name = nameNode.getNodeValue();
	    }
	    tableNode = n.getAttributes().getNamedItem("Table");
	    if(tableNode != null) {
		table = tableNode.getNodeValue();
	    }
	    datatypeNode = n.getAttributes().getNamedItem("Datatype");
	    if(datatypeNode != null) {
		datatype = datatypeNode.getNodeValue();
	    }
	    usageNode = n.getAttributes().getNamedItem("Usage");
	    if(usageNode != null) {
		usage = usageNode.getNodeValue();
	    }
	    // Try Optionality
	    else {
		usageNode = n.getAttributes().getNamedItem("Optionality");
		if(usageNode != null) {
		    usage = usageNode.getNodeValue();
		}
	    }
	    cardMinNode = n.getAttributes().getNamedItem("Min");
	    if(cardMinNode != null) {
		cardMin = cardMinNode.getNodeValue();
	    }
	    cardMaxNode = n.getAttributes().getNamedItem("Max");
	    if(cardMaxNode != null) {
		cardMax = cardMaxNode.getNodeValue();
	    }
	    constantNode = n.getAttributes().getNamedItem("ConstantValue");
	    if(constantNode != null) {
		constant = constantNode.getNodeValue();
		ProfileConstantValue pcv = new ProfileConstantValue();
		pcv.setName(name);
		pcv.setQualifier(curField.getName());
		pcv.setQualifier2("");
		pcv.setSegment(curSegment.getName());
		pcv.setValue(constant);
		vConstant.add(pcv);
	    }
	    compSeqNumber++;
	    curComponent.setName(name);
	    curComponent.setQualifier(curField.getName());
	    curComponent.setQualifier2("");
	    curComponent.setSegment(curSegment.getName());
	    curComponent.setTable(table);
	    curComponent.setDatatype(datatype);
	    curComponent.setUsage(usage);
	    curComponent.setCardMin(cardMin);
	    curComponent.setCardMax(cardMax);
	    curComponent.setSequenceNum(fieldSeqNumber + "." + compSeqNumber);
	    //curComponent.setParent(curField.getParent());
	    curComponent.setParent(curField);
	    curComponent.setMsgLocPrefix(curField.getDatatype());
	    curField.addComponent(curComponent);
	    vComponent.add(curComponent);

	    subCompSeqNumber = 0;
	    curPN = curComponent;
	}
	// SubComponent
 	else if(n.getNodeName().equals("SubComponent")) {
	    curSubComponent = new SubComponentNode();
	    nameNode = n.getAttributes().getNamedItem("Name");
	    if(nameNode != null) {
		name = nameNode.getNodeValue();
	    }
	    tableNode = n.getAttributes().getNamedItem("Table");
	    if(tableNode != null) {
		table = tableNode.getNodeValue();
	    }
	    datatypeNode = n.getAttributes().getNamedItem("Datatype");
	    if(datatypeNode != null) {
		datatype = datatypeNode.getNodeValue();
	    }
	    usageNode = n.getAttributes().getNamedItem("Usage");
	    if(usageNode != null) {
		usage = usageNode.getNodeValue();
	    }
	    // Try Optionality
	    else {
		usageNode = n.getAttributes().getNamedItem("Optionality");
		if(usageNode != null) {
		    usage = usageNode.getNodeValue();
		}
	    }
	    cardMinNode = n.getAttributes().getNamedItem("Min");
	    if(cardMinNode != null) {
		cardMin = cardMinNode.getNodeValue();
	    }
	    cardMaxNode = n.getAttributes().getNamedItem("Max");
	    if(cardMaxNode != null) {
		cardMax = cardMaxNode.getNodeValue();
	    }
	    constantNode = n.getAttributes().getNamedItem("ConstantValue");
	    if(constantNode != null) {
		constant = constantNode.getNodeValue();
		ProfileConstantValue pcv = new ProfileConstantValue();
		pcv.setName(name);
		pcv.setQualifier(curComponent.getName());
		pcv.setQualifier2(curComponent.getQualifier());
		pcv.setSegment(curSegment.getName());
		pcv.setValue(constant);
		vConstant.add(pcv);
	    }
	    subCompSeqNumber++;
	    curSubComponent.setName(name);
	    //curSubComponent.setQualifier(curComponent.getQualifier() + "{" + curComponent.getName() + "}");
	    curSubComponent.setQualifier(curComponent.getName());
	    curSubComponent.setQualifier2(curComponent.getQualifier());
	    curSubComponent.setSegment(curSegment.getName());
	    curSubComponent.setTable(table);
	    curSubComponent.setDatatype(datatype);
	    curSubComponent.setUsage(usage);
	    curSubComponent.setCardMin(cardMin);
	    curSubComponent.setCardMax(cardMax);
	    curSubComponent.setSequenceNum(fieldSeqNumber + "." + compSeqNumber + "." + subCompSeqNumber);
	    //curSubComponent.setParent(curComponent.getParent());
	    curSubComponent.setParent(curComponent);
	    curSubComponent.setMsgLocPrefix(curComponent.getDatatype());
	    curComponent.addSubComponent(curSubComponent);
	    vSubComponent.add(curSubComponent);

	    curPN = curSubComponent;
	}
	else if(n.getNodeName().equals("DataValues")) {
	    exampleNode = n.getAttributes().getNamedItem("ExValue");
	    if(exampleNode != null) {
		example = exampleNode.getNodeValue();
		ProfileExampleValue pev = new ProfileExampleValue();
		pev.setName(curPN.getName());
		pev.setQualifier(curPN.getQualifier());
		pev.setQualifier2(curPN.getQualifier2());
		pev.setSegment(curPN.getSegment());
		pev.addValue(example);
		int idx = vExample.indexOf(pev);
		if(idx == -1)
		    vExample.add(pev);
		else
		    ((ProfileExampleValue)vExample.get(idx)).addValue(example);		
	    }
	}

 	// Apply method on children
 	NodeList nl = n.getChildNodes();
 	for(int i=0;i<nl.getLength();i++) {
	    level++;
 	    browseXML(nl.item(i));
	    level--;
	}
    }

    /**
     * Return a tree structure of the current profile
     * @param rootText The text of the root node
     * @return The tree representation of the profile
     */
    public DefaultMutableTreeNode getTree(String rootText) { //throws ParserConfigurationException, SAXException, IOException {
/*	DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
	DocumentBuilder builder = factory.newDocumentBuilder();
	Document doc = builder.parse(new File(xmlProfile));

	// Browser the XML Profile	
	vSegment = new Vector();
        browseXML(doc);*/
	// Create the Tree Structure
	DefaultMutableTreeNode root = new DefaultMutableTreeNode(rootText);

	for(int i=0;i<vRoot.size();i++) {
	    ProfileNode pn = (ProfileNode)vRoot.get(i);
	    if(pn instanceof SegmentGroupNode) {
		root.add(processSegmentGroupNode((SegmentGroupNode)pn));
	    }
	    else if(pn instanceof SegmentNode) {
		root.add(processSegment((SegmentNode)pn));
	    }
	}
	return root;
    }

    /**
     * Process the segment groups
     * @param segGroupNode The segment group node
     * @return A part of the tree
     */
    public DefaultMutableTreeNode processSegmentGroupNode(SegmentGroupNode segGroupNode) {
	DefaultMutableTreeNode segGroup;
	segGroup = new DefaultMutableTreeNode(segGroupNode);
	Vector vChildren = segGroupNode.getChildren();
	for(int j=0;j<vChildren.size();j++) {
	    ProfileNode pn = (ProfileNode)vChildren.get(j);
	    if(pn instanceof SegmentNode)
		segGroup.add(processSegment((SegmentNode)pn));
	    else
		segGroup.add(processSegmentGroupNode((SegmentGroupNode)pn));
	}
	return segGroup;
    }

    /**
     * Process the segments
     * @param segNode The segment node
     * @return A part of the tree
     */
    public DefaultMutableTreeNode processSegment(SegmentNode segNode) {
	DefaultMutableTreeNode segment, field, component, subcomponent;

	segment = new DefaultMutableTreeNode(segNode);
	Vector vField = segNode.getFields();
	for(int j=0;j<vField.size();j++) {
	    FieldNode fieldNode = (FieldNode)vField.get(j);
	    field = new DefaultMutableTreeNode(fieldNode);
	    segment.add(field);
	    Vector vComponent = fieldNode.getComponents();
	    for(int k=0;k<vComponent.size();k++) {
		ComponentNode compNode = (ComponentNode)vComponent.get(k);
		component = new DefaultMutableTreeNode(compNode);
		field.add(component);
		Vector vSubComponent = compNode.getSubComponents();
		for(int l=0;l<vSubComponent.size();l++) {
		    SubComponentNode subcompNode = (SubComponentNode)vSubComponent.get(l);
		    subcomponent = new DefaultMutableTreeNode(subcompNode);
		    component.add(subcomponent);
		}
	    }
	}
	return segment;
    }

    /**
     * Return a tree structure of the current profile (only MSG segment)
     * @param rootText The text of the root node
     * @return The tree representation of the profile
     */
    public DefaultMutableTreeNode getMSHTree(String rootText) { 
	// Create the Tree Structure
	DefaultMutableTreeNode root = new DefaultMutableTreeNode(rootText);
	//	DefaultMutableTreeNode segGroup;

	for(int i=0;i<vRoot.size();i++) {
	    ProfileNode pn = (ProfileNode)vRoot.get(i);
/*	    if(pn instanceof SegmentGroupNode) {
		SegmentGroupNode segGroupNode = (SegmentGroupNode)vRoot.get(i);
		segGroup = new DefaultMutableTreeNode(segGroupNode);
		root.add(segGroup);
		Vector vSegmentOfGroup = segGroupNode.getSegments();
		for(int j=0;j<vSegmentOfGroup.size();j++) {
		    segGroup.add(processSegment((SegmentNode)vSegmentOfGroup.get(j)));
		}
		}*/
	    if(pn instanceof SegmentNode && pn.getName().equals("MSH")) {
		root.add(processSegment((SegmentNode)pn));
	    }
	}
	return root;
    }

    /**
     * Get the usage of a node
     * @param pn The ProfileNode
     * @return The usage
     */
    public String getUsage(ProfileNode pn) {
	int idx;
	// Segment Group
	if(pn instanceof SegmentGroupNode) {
	    idx = vSegGroup.indexOf(pn);
	    if(idx != -1) {
		return ((SegmentGroupNode)vSegGroup.get(idx)).getUsage();
	    }
	}
	// Segment
	else if(pn instanceof SegmentNode) {
	    idx = vSegment.indexOf(pn);
	    if(idx != -1) {
		return ((SegmentNode)vSegment.get(idx)).getUsage();
	    }
	}
	// Field
	else if(pn instanceof FieldNode) {
	    idx = vField.indexOf(pn);
	    if(idx != -1) {
		return ((FieldNode)vField.get(idx)).getUsage();
	    }
	}
	// Component
	else if(pn instanceof ComponentNode) {
	    idx = vComponent.indexOf(pn);
	    if(idx != -1) {
		return ((ComponentNode)vComponent.get(idx)).getUsage();
	    }
	}
	// SubComponent
	else if(pn instanceof SubComponentNode) {
	    idx = vSubComponent.indexOf(pn);
	    if(idx != -1) {
		return ((SubComponentNode)vSubComponent.get(idx)).getUsage();
	    }
	}
	return null;
    }

    /** 
     * Set the parent of a node from the profile
     * @param pn The ProfileNode
     * @param segGroup The segment group path (xxx/xxx/xxx)
     */
    public void setParent(ProfileNode pn, String segGroup) {
	// Segment Group
	if(pn instanceof SegmentGroupNode) {
	    for(int i=0;i<vSegGroup.size();i++) {
		SegmentGroupNode sgn = (SegmentGroupNode)vSegGroup.get(i);
		if(sgn.specialEquals(pn) && sgn.getParentPath().equals(segGroup)) {
		    pn.setParent(sgn.getParent());
		    i = vSegGroup.size() + 1;
		}
	    }
	}
	// Segment
	else if(pn instanceof SegmentNode) {
	    for(int i=0;i<vSegment.size();i++) {
		SegmentNode sn = (SegmentNode)vSegment.get(i);
		if(sn.specialEquals(pn) && sn.getParentPath().equals(segGroup)) {
		    pn.setParent(sn.getParent());
		    i = vSegment.size() + 1;
		}
	    }
	}
	// Field
	else if(pn instanceof FieldNode) {
	    for(int i=0;i<vField.size();i++) {
		FieldNode fn = (FieldNode)vField.get(i);
		if(fn.specialEquals(pn) && fn.getParentPath().equals(segGroup)) {
		    pn.setParent(fn.getParent());
		    i = vField.size() + 1;
		}
	    }
	}
	// Component
	else if(pn instanceof ComponentNode) {
	    for(int i=0;i<vComponent.size();i++) {
		ComponentNode cn = (ComponentNode)vComponent.get(i);
		if(cn.specialEquals(pn) && cn.getParentPath().equals(segGroup)) {
		    pn.setParent(cn.getParent());
		    i = vComponent.size() + 1;
		}
	    }
	}
	// SubComponent
	else if(pn instanceof SubComponentNode) {
	    for(int i=0;i<vSubComponent.size();i++) {
		SubComponentNode scn = (SubComponentNode)vSubComponent.get(i);
		if(scn.specialEquals(pn) && scn.getParentPath().equals(segGroup)) {
		    pn.setParent(scn.getParent());
		    i = vSubComponent.size() + 1;
		}
	    }
	}

    }

    /**
     * Set the cardinality of a node from the profile
     * @param pn The ProfileNode
     */
    public void setCardinality(ProfileNode pn) {
	int idx;
	// Segment Group
	if(pn instanceof SegmentGroupNode) {
	    idx = vSegGroup.indexOf(pn);
	    if(idx != -1) {
		pn.setCardMin(((SegmentGroupNode)vSegGroup.get(idx)).getCardMin());
		pn.setCardMax(((SegmentGroupNode)vSegGroup.get(idx)).getCardMax());
	    }
	}
	// Segment
	else if(pn instanceof SegmentNode) {
	    idx = vSegment.indexOf(pn);
	    if(idx != -1) {
		pn.setCardMin(((SegmentNode)vSegment.get(idx)).getCardMin());
		pn.setCardMax(((SegmentNode)vSegment.get(idx)).getCardMax());
	    }
	}
	// Field
	else if(pn instanceof FieldNode) {
	    idx = vField.indexOf(pn);
	    if(idx != -1) {
		pn.setCardMin(((FieldNode)vField.get(idx)).getCardMin());
		pn.setCardMax(((FieldNode)vField.get(idx)).getCardMax());
	    }
	}
	// Component
	else if(pn instanceof ComponentNode) {
	    idx = vComponent.indexOf(pn);
	    if(idx != -1) {
		pn.setCardMin(((ComponentNode)vComponent.get(idx)).getCardMin());
		pn.setCardMax(((ComponentNode)vComponent.get(idx)).getCardMax());
	    }
	}
	// SubComponent
	else if(pn instanceof SubComponentNode) {
	    idx = vSubComponent.indexOf(pn);
	    if(idx != -1) {
		pn.setCardMin(((SubComponentNode)vSubComponent.get(idx)).getCardMin());
		pn.setCardMax(((SubComponentNode)vSubComponent.get(idx)).getCardMax());
	    }
	}
    }

    /**
     * Get the message location
     * @param segGroupLoc The segment group location of the error
     * @param segLoc The segment location of the error
     * @param fieldLoc The field location of the error
     * @param compLoc The component location of the error
     * @param subCompLoc The subcomponent location of the error
     * @return Get the message location
     */
    public String getMsgLocation(String segGroupLoc, String segLoc, String fieldLoc, String compLoc, String subCompLoc) {
	// Find the type
	ProfileNode pn = null;
	if(subCompLoc != null && !subCompLoc.equals("")) {
	    pn = new SubComponentNode();
	    pn.setName(subCompLoc);
	    pn.setQualifier(compLoc);
	    pn.setQualifier2(fieldLoc);
	    pn.setSegment(segLoc);
	}
	else if(compLoc != null && !compLoc.equals("")) {
	    pn = new ComponentNode();
	    pn.setName(compLoc);
	    pn.setQualifier(fieldLoc);
	    pn.setQualifier2("");
	    pn.setSegment(segLoc);
	}
	else if(fieldLoc != null && !fieldLoc.equals("")) {
	    pn = new FieldNode();
	    pn.setName(fieldLoc);
	    pn.setQualifier("");
	    pn.setQualifier2("");
	    pn.setSegment(segLoc);
	}
	else if(segLoc != null && !segLoc.equals("")) {
	    return segLoc;
	}
	else if(segLoc == null && segLoc.equals("")) {
	    return segGroupLoc;
	}
	// Field
	if(pn instanceof FieldNode) {
	    for(int i=0;i<vField.size();i++) {
		FieldNode fn = (FieldNode)vField.get(i);
		if(fn.specialEquals(pn)) { 
		    return fn.getMsgLoc();
		}
	    }
	}
	// Component
	else if(pn instanceof ComponentNode) {
	    for(int i=0;i<vComponent.size();i++) {
		ComponentNode cn = (ComponentNode)vComponent.get(i);
		if(cn.specialEquals(pn)) {
		    return cn.getMsgLoc();
		}
	    }
	}
	// SubComponent
	else if(pn instanceof SubComponentNode) {
	    for(int i=0;i<vSubComponent.size();i++) {
		SubComponentNode scn = (SubComponentNode)vSubComponent.get(i);
		if(scn.specialEquals(pn)) {
		    return scn.getMsgLoc();
		}
	    }
	}	
	return "";
    }

    /**
     * Generate the specific data repository
     * @param refDBModel The reference data
     * @param configModel The config data
     */
    public void generateRepository(ReferenceData refData, ConfigData configData) {
	try {
	    // Create the document
	    DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
	    DocumentBuilder builder = factory.newDocumentBuilder();
	    Document doc = builder.newDocument();
	    
	    // Root element
	    Element root = (Element)doc.createElement("NIST_HL7_Repository"); 
	    root.setAttribute("xmlns", "http://www.nist.gov/hl7/ReferenceData");
	    doc.appendChild(root);
	    
	    Element magic = (Element)doc.createElement("MessageMaker");
	    root.appendChild(magic);
	    
	    appendElement(doc, root, vField, refData, configData);
	    appendElement(doc, root, vComponent, refData, configData);
	    appendElement(doc, root, vSubComponent, refData, configData);
	    
	    // Write the DOM document to the file
	    String filename = new File(xmlProfile).getParent() + File.separator + "Repository_" + new File(xmlProfile).getName();
	    Transformer transformer = TransformerFactory.newInstance().newTransformer();
	    DOMSource source = new DOMSource(doc);
	    StreamResult result = new StreamResult(new File(filename).getPath());
	    transformer.setOutputProperty(OutputKeys.METHOD, "xml");
	    transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "4");
	    transformer.setOutputProperty(OutputKeys.INDENT, "yes");
	    transformer.transform(source, result);
	}
	catch(ParserConfigurationException pce) {
	    System.err.println("ParserConfigurationException " + pce.getMessage());
	    pce.printStackTrace();
	}
	catch(TransformerConfigurationException tce) {
	    System.err.println("TransformerConfigurationException " + tce.getMessage());
	    tce.printStackTrace();
	}
	catch(TransformerException te) {
	    System.err.println("TransformerException " + te.getMessage());
	    te.printStackTrace();
	}
    }	

    /**
     * Browse the Vector and append the values to the DOM document
     * @param doc The document to create elements
     * @param root The root node
     * @param v A Vector of ProfileNode
     * @param refData The reference data
     * @param configData The config data
     */
    private void appendElement(Document doc, Element root, Vector v, ReferenceData refData, ConfigData configData) {
	Element ref = null;
	for(int i=0;i<v.size();i++) {
	    ProfileNode pn = (ProfileNode)v.get(i);
	    if(pn.isLeaf()) {
		ref = (Element)doc.createElement("ReferenceData");
		// Name
		if(pn.getName() != null)
		    ref.setAttribute("Name", pn.getName());
		// Qualifier
		if(pn.getQualifier() != null && !pn.getQualifier().equals(""))
		    ref.setAttribute("Qualifier", pn.getQualifier());
		// Qualifier2
		if(pn.getQualifier2() != null && !pn.getQualifier2().equals(""))
		    ref.setAttribute("Qualifier2", pn.getQualifier2());
		// Segment
		if(pn.getSegment() != null)
		    ref.setAttribute("Segment", pn.getSegment());
		// Datatype
		if(pn.getDatatype() != null)
		    ref.setAttribute("Datatype", pn.getDatatype());
		
		// The data value
		// 1 - Config
		boolean bConfig = false;
		Vector vConfig = configData.getValues(pn.getName(), pn.getSegment(), pn.getQualifier(), pn.getQualifier2());
		for(int j=0;j<vConfig.size();j++) {
		    if(((ValueNode)vConfig.get(j)).getCategory().equals("Config")) {
			bConfig = true;
			ref.appendChild(getDOMValue(doc, ((ValueNode)vConfig.get(j)).getValue(), "Config"));
		    }
		}
		// 2 - Constant
		if(!bConfig) {
		    ProfileConstantValue pcv = new ProfileConstantValue();
		    pcv.setName(pn.getName());
		    pcv.setQualifier(pn.getQualifier());
		    pcv.setQualifier2(pn.getQualifier2());
		    pcv.setSegment(pn.getSegment());
		    int idx = vConstant.indexOf(pcv);
		    if(idx != -1) {
			pcv = (ProfileConstantValue)vConstant.get(idx);
			ref.appendChild(getDOMValue(doc, pcv.getValue(), "Constant"));
		    }
		    // 3 4 - Example && RefDB 
		    else {
			// 3 - Example
			ProfileExampleValue pev = new ProfileExampleValue();
			pev.setName(pn.getName());
			pev.setQualifier(pn.getQualifier());
			pev.setQualifier2(pn.getQualifier2());
			pev.setSegment(pn.getSegment());
			idx = vExample.indexOf(pev);
			if(idx != -1) {
			    pev = (ProfileExampleValue)vExample.get(idx);
			    Vector vValue = pev.getValues();
			    for(int j=0;j<vValue.size();j++) {
				ref.appendChild(getDOMValue(doc, (String)vValue.get(j), "Example"));
			    }
			}
			// 4 - RefDB
			Vector vRefDB = refData.getValues(pn.getName(), pn.getSegment(), pn.getQualifier(), pn.getQualifier2());
			for(int j=0;j<vRefDB.size();j++) {
			    ref.appendChild(getDOMValue(doc, ((ValueNode)vRefDB.get(j)).getValue(), ((ValueNode)vRefDB.get(j)).getCategory()));
			}
		    }
		}
		root.appendChild(ref);
	    }
	}
    }

    /**
     * Get the DOM tree of a value
     * @param doc The document to create element
     * @param value The value
     * @param category The category of the value (valid, invalid, constant, example=valid)
     * @return a DOM tree
     */
    private Element getDOMValue(Document doc, String value, String category) {
	Element item = null;

	item = (Element)doc.createElement("Item");
	// Value
	if(value != null)
	    item.setAttribute("Value", value);
	// Category
	if(category != null)
	    item.setAttribute("Category", category);		    
	return item;   
    }

    /**
     * Get the ProfileNode that has no value in the repository
     * @param repositoryData The repository data
     * @return A vector of ProfileNode
     */
    public Vector getNoValueProfileNode(RepositoryDataModel repositoryData) {
	Vector vNoValue = new Vector();
	ProfileNode pn = null;
	// Field
	for(int i=0;i<vField.size();i++) {
	    pn = (ProfileNode)vField.get(i);
	    if(repositoryData.getValues(pn.getName(), pn.getSegment(), pn.getQualifier(), pn.getQualifier2()).size() == 0) {
		vNoValue.add(pn);
	    }
	}
	// Component
	for(int i=0;i<vComponent.size();i++) {
	    pn = (ProfileNode)vComponent.get(i);
	    if(repositoryData.getValues(pn.getName(), pn.getSegment(), pn.getQualifier(), pn.getQualifier2()).size() == 0) {
		vNoValue.add(pn);
	    }
	}
	// Subcomponent
	for(int i=0;i<vSubComponent.size();i++) {
	    pn = (ProfileNode)vSubComponent.get(i);
	    if(repositoryData.getValues(pn.getName(), pn.getSegment(), pn.getQualifier(), pn.getQualifier2()).size() == 0) {
		vNoValue.add(pn);
	    }
	}
	return vNoValue;
    }
}
