/**
 * This class contains all the configuration parameters for a particular Session
 */

package gov.nist.hl7maker.control;

// JDK
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Vector;
import java.util.StringTokenizer;
import java.util.Properties;
import java.util.Observable;
import java.util.Collections;
import java.util.Comparator;

// Swing
import javax.swing.tree.DefaultMutableTreeNode;

// XML
import javax.xml.parsers.DocumentBuilder; 
import javax.xml.parsers.DocumentBuilderFactory;  
import javax.xml.parsers.ParserConfigurationException;

// SAX
import org.xml.sax.SAXException;  

// DOM
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

// XSL
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;

// MessageMaker
import gov.nist.hl7maker.util.MVCArg;
import gov.nist.hl7maker.util.MMLog;
import gov.nist.hl7maker.view.config.TableDataModel;
import gov.nist.hl7maker.view.config.RepositoryDataModel;
import gov.nist.hl7maker.view.config.ProfileNode;
import gov.nist.hl7maker.view.config.SegmentGroupNode;
import gov.nist.hl7maker.view.config.SegmentNode;
import gov.nist.hl7maker.view.config.FieldNode;
import gov.nist.hl7maker.view.config.ComponentNode;
import gov.nist.hl7maker.view.config.SubComponentNode;
import gov.nist.hl7maker.view.config.TablePriority;
import gov.nist.hl7maker.view.init.ProfileTableModel;

public class Session extends Observable {

    private Properties props;
    private boolean modified;
    private Vector vLocalTable;
    private int nMsgVariant;

    private Profile ramProfile;
    private DefaultValue dv;
    private ReferenceData refData;
    private ConfigData configData;
    private TableDataModel tableModel;
    private RepositoryDataModel repositoryModel;

    /**
     * Constructor
     */
    public Session() {
	repositoryModel = new RepositoryDataModel();
	vLocalTable = new Vector();
	dv = new DefaultValue();
    }

    /**
     * Constructor
     * @param mmkFile A .mmk filename
     */
    public Session(String mmkFile) throws IOException {
	this();
	loadProperties(mmkFile);
    }

    /**
     * Constructor
     * @param mmkFile A .mmk File
     */
    public Session(File mmkFile) throws IOException {
	this();
	loadProperties(mmkFile);
    }

    /**
     * Load the properties from a mmk file
     * @param mmkFile The mmk file
     */
    public void loadProperties(String mmkFile) throws IOException {
	loadProperties(new File(mmkFile));
    }

    /**
     * Load the properties from a mmk file
     * @param mmkFile The mmk file
     */
    public void loadProperties(File mmkFile) throws IOException {
	vCardinalityTestValid = new Vector();
	vCardinalityTestInvalid = new Vector();
	vCardinalityTestBoth = new Vector();
	vUsageTestValid = new Vector();
	vUsageTestInvalid = new Vector();
	vUsageTestBoth = new Vector();
	vErrorCodesTestValid = new Vector();
	vErrorCodesTestInvalid = new Vector();
	vCardinalityTempTestValid = new Vector();
	vCardinalityTempTestInvalid = new Vector();
	vCardinalityTempTestBoth = new Vector();
	vUsageTempTestValid = new Vector();
	vUsageTempTestInvalid = new Vector();
	vUsageTempTestBoth = new Vector();
	vErrorCodesTempTestValid = new Vector();
	vErrorCodesTempTestInvalid = new Vector();
	props = new Properties();
	props.load(new FileInputStream(mmkFile));
	if(!getProfile().equals("")) {
	    ramProfile = new Profile(getProfile());
	    generateRepository(false);
	    /*	    setProfile(new File(getProfile()).getCanonicalPath());
	    setName(new File(getName()).getCanonicalPath());
	    setMsgPath(new File(getMsgPath()).getCanonicalPath());*/
	}
	else 
	    ramProfile = null;
	loadTables();

	gov.nist.hl7maker.util.MVCLog.info("[Session] Load the properties from a mmk file");
	setChanged();
	notifyObservers(this);
	modified = false;
    }

    /**
     * Load the options from a mmk file
     * @param mmkFile The mmk file
     */
    public void loadOptions(String mmkFile) throws IOException {
	loadOptions(new File(mmkFile));
    }

    /**
     * Load the options from a mmk file
     * @param mmkFile The mmk file
     */
    public void loadOptions(File mmkFile) throws IOException {
	// Save old values
	String oldProfile = getProfile();
	String oldName = getName();
	String oldMsgPath = getMsgPath();
	//String oldTypeMessageSet = getTypeMessageSet();
	String oldVolume = String.valueOf(getVolume());
	// Load the properties
	props = new Properties();
	props.load(new FileInputStream(mmkFile));
	// Restore old values
	props.setProperty("session.general.profile", oldProfile);
	props.setProperty("session.general.name", oldName);
	props.setProperty("session.general.msgPath", oldMsgPath);
	//props.setProperty("session.msg.typeMessageSet", oldTypeMessageSet);
	props.setProperty("session.msg.volume", oldVolume);
	// MVC
	gov.nist.hl7maker.util.MVCLog.info("[Session] Load the options from a mmk file");
	setChanged();
	notifyObservers(this);
	modified = true;
    }

    /**
     * Save the properties in a mmk file
     * @param mmkFile The mmk file
     */
    public void saveProperties(String mmkFile) throws IOException {
	props.store(new FileOutputStream(new File(mmkFile)), null);
	modified = false;
	gov.nist.hl7maker.util.MVCLog.info(mmkFile + " saved");
    }

    /**
     * Save the properties in a mmk file
     * @param mmkFile The mmk file
     */
    public void saveProperties(File mmkFile) throws IOException {
	props.store(new FileOutputStream(mmkFile), null);
	modified = false;
	gov.nist.hl7maker.util.MVCLog.info(mmkFile.getPath() + " saved");
    }

    // GeneralProperties

    /* Name identifying the Message Test Set (Add .mmk to get the filename) */
    public String getName() {
	String name = props.getProperty("session.general.name");
	try {
	    if(!name.equals(""))
		name = new File(name).getCanonicalPath();
	}
	catch(IOException ioe) {
	    System.err.println("IOException: " + ioe.getMessage());
	    ioe.printStackTrace();
	}
	return name;
    }

    public void setName(String name) {
	props.setProperty("session.general.name", name);
	modified = true;
	gov.nist.hl7maker.util.MVCLog.info("[Session] Set the session name");
	setChanged();
	notifyObservers(new MVCArg("name", name));
    }

    /* Name of the XML file containing the Profile used to created messages */
    public String getProfile() {
	String profile = props.getProperty("session.general.profile");
	try {
	    if(!profile.equals(""))
		profile = new File(profile).getCanonicalPath();
	}
	catch(IOException ioe) {
	    System.err.println("IOException: " + ioe.getMessage());
	    ioe.printStackTrace();
	}
	return profile;
    }

    public void setProfile(String profile) {
	if(!props.getProperty("session.general.profile").equals(profile)) {
	    props.setProperty("session.general.profile", profile);
	    if(!profile.equals("")) {
		ramProfile = new Profile(profile);
		setRepositoryData("");
		generateRepository(false);
	    }
	    clearTest();
	    clearTempTest();
	    saveTest();
	    modified = true;
	    gov.nist.hl7maker.util.MVCLog.info("[Session] Set the profile");
	}
	setChanged();
	notifyObservers(new MVCArg("profile", profile));	
    }

    /* Name of the path used to store messages */
    public String getMsgPath() {
	String msgPath = props.getProperty("session.general.msgPath");
	try {
	    if(!msgPath.equals(""))
		msgPath = new File(msgPath).getCanonicalPath();
	}
	catch(IOException ioe) {
	    System.err.println("IOException: " + ioe.getMessage());
	    ioe.printStackTrace();
	}
	return msgPath;
    }

    public void setMsgPath(String msgPath) {
	props.setProperty("session.general.msgPath", msgPath);
	modified = true;
	gov.nist.hl7maker.util.MVCLog.info("[Session] Set the messages path");
	setChanged();
	notifyObservers(new MVCArg("msgPath", msgPath));
    }

    /* Version */
    public String getVersion() {
	return props.getProperty("session.general.version");
    }

    public void setVersion(String version) {
	props.setProperty("session.general.version", version);
	modified = true;
	gov.nist.hl7maker.util.MVCLog.info("[Session] Set the HL7 version");
	setChanged();
	notifyObservers(new MVCArg("version", version));
    }

    /* Validate flag */
    public boolean getValidate() {
	if(props.getProperty("session.general.validate").equals("0")) {
	    return true;
	}
	return false;
    }

    public void setValidate(boolean validate) {
	String sValidate;
	if(validate) {
	    props.setProperty("session.general.validate", "0");
	    sValidate = "0";
	}
	else {
	    props.setProperty("session.general.validate", "1");
	    sValidate = "1";
	}
	modified = true;
	gov.nist.hl7maker.util.MVCLog.info("[Session] Set the validation on or off");
	setChanged();
	notifyObservers(new MVCArg("validate", sValidate));
    }

    /* HL7 Schema compliance flag */
/*    public boolean getHL7CheckCompliance() {
	if(props.getProperty("session.general.HL7Compliant").equals("0")) {
	    return true;
	}
	return false;
    }

    public void setHL7CheckCompliance(boolean HL7Compliant) {
	String sHL7Compliant;
	if(HL7Compliant) {
	    props.setProperty("session.general.HL7Compliant", "0");
	    sHL7Compliant = "0";
	}
	else {
	    props.setProperty("session.general.HL7Compliant", "1");
	    sHL7Compliant = "1";
	}
	modified = true;
	gov.nist.hl7maker.util.MVCLog.info("[Session] Set the HL7 Compliance test to on or off");
	setChanged();
	notifyObservers(new MVCArg("HL7Compliant", sHL7Compliant));
    }
*/
    
    /* Validate Profile flag */
    public String getValidateProfile() {
	return props.getProperty("session.general.validateprofile");
    }

    public void setValidateProfile(String validateProfile) {
	props.setProperty("session.general.validateprofile", validateProfile);
	modified = true;
    }

    /* Value of Star */

    public String getStar() {
	return props.getProperty("session.general.star");
    }

    public void setStar(String star) {
	props.setProperty("session.general.star", star);
    }

    // DataProperties

    /* Name of the XML file containing the data from reference database */
    public String getReferenceDB() {
	return props.getProperty("session.data.referenceDB");
    }

    public void setReferenceDB(String referenceDB) {
	props.setProperty("session.data.referenceDB", referenceDB);
	modified = true;
	gov.nist.hl7maker.util.MVCLog.info("[Session] Set the reference database file");
	setChanged();
	notifyObservers(new MVCArg("referenceDB", referenceDB));
    }

    /* Name of the XML file containing the data from the repository */
    public String getRepositoryData() {
	return props.getProperty("session.data.repositoryData");
    }

    public void setRepositoryData(String repositoryData) {
	props.setProperty("session.data.repositoryData", repositoryData);
	modified = true;
	gov.nist.hl7maker.util.MVCLog.info("[Session] Set the repository data file");
	setChanged();
	notifyObservers(new MVCArg("repositoryData", repositoryData));
    }

    /* Name of the XML file containing the HL7 tables data */
    public String getHL7Tables() {
	return props.getProperty("session.data.HL7Tables");
    }

    public void setHL7Tables(String HL7Tables) {
	if(!props.getProperty("session.data.HL7Tables").equals(HL7Tables)) {
	    props.setProperty("session.data.HL7Tables", HL7Tables);
	    modified = true;
	    gov.nist.hl7maker.util.MVCLog.info("[Session] Set the HL7 tables file");
	}
	setChanged();
	notifyObservers(new MVCArg("HL7Tables", HL7Tables));
    }

    /* Name of the XML file containing the User tables data */
    public String getUserTables() {
	return props.getProperty("session.data.userTables");
    }

    public void setUserTables(String userTables) {
	if(!props.getProperty("session.data.userTables").equals(userTables)) {
	    props.setProperty("session.data.userTables", userTables);
	    modified = true;
	    gov.nist.hl7maker.util.MVCLog.info("[Session] Set the user tables file");
	}
	setChanged();
	notifyObservers(new MVCArg("userTables", userTables));
    }

    /* Name of the XML files containing the external tables data */
    public String getExternalTables() {
	return props.getProperty("session.data.externalTables");
    }

    public void setExternalTables(String externalTables) {
	if(!props.getProperty("session.data.externalTables").equals(externalTables)) {
	    props.setProperty("session.data.externalTables", externalTables);
	    modified = true;
	    gov.nist.hl7maker.util.MVCLog.info("[Session] Set the external tables file");
	}
	setChanged();
	notifyObservers(new MVCArg("externalTables", externalTables));
    }

    /* Name of the XML file containing the Configuration data */
    public String getConfig() {
	return props.getProperty("session.data.configData");
    }

    public void setConfig(String configData) {
	if(!props.getProperty("session.data.configData").equals(configData)) {
	    props.setProperty("session.data.configData", configData);
	    modified = true;
	    gov.nist.hl7maker.util.MVCLog.info("[Session] Set the config data file");
	}
	setChanged();
	notifyObservers(new MVCArg("configData", configData));
    }

    // MessageProperties

    /* Type of message set (Valid, Invalid, Both, Sample, Custom) */
    public String getTypeMessageSet() {
	return props.getProperty("session.msg.typeMessageSet");
    }

    public void setTypeMessageSet(String typeMessageSet) {
	props.setProperty("session.msg.typeMessageSet", typeMessageSet);
	modified = true;
    }

    /* Volume option - How many messages should be created? */
    public int getVolume() {
	return Integer.parseInt(props.getProperty("session.msg.volume"));
    }

    public void setVolume(int volume, boolean mvcForward) {
	props.setProperty("session.msg.volume", String.valueOf(volume));
	modified = true;
	// MVC Forward to avoid recursive calls in MessageSetupPanel in insertString and removeString methods
	if(mvcForward) {
	    gov.nist.hl7maker.util.MVCLog.info("[Session] Set the volume of message");
	    setChanged();
	    notifyObservers(new MVCArg("volume", String.valueOf(volume)));
	}
    }

    /* Segment Cardinality Option - Create messages that vares segment 
       cardinality? If cardinality is "false", this option is meaningless. */
/*
    public boolean getSegmentCardinality() {
	boolean b = false;

	if(props.getProperty("session.msg.segmentCardinality").compareTo("1") == 0) {
	    b = true;
	}
	return b;
    }

    public void setSegmentCardinality(boolean b) {
	if(b) {
	    props.setProperty("session.msg.segmentCardinality", "1");
	}
	else {
	    props.setProperty("session.msg.segmentCardinality", "0");
	}
	modified = true;
    }
*/  
    /* Field Cardinality Option - Create messages that vares field 
       cardinality? If cardinality is "false", this option is meaningless. */
/*    public boolean getFieldCardinality() {
	boolean b = false;
	
	if(props.getProperty("session.msg.fieldCardinality").compareTo("1") == 0) {
	    b = true;
	}
	return b;
    }

    public void setFieldCardinality(boolean b) {
	if(b) {
	    props.setProperty("session.msg.fieldCardinality", "1");
	}
	else {
	    props.setProperty("session.msg.fieldCardinality", "0");
	}
	modified = true;
    }
*/
    /* Required Usage Option - Create messages that specifically examine the
       required usage option. */
 /*   public boolean getRequiredUsage() {
	boolean b = false;
	
	if(props.getProperty("session.msg.requiredUsage").compareTo("1") == 0) {
	    b = true;
	}
	return b;
    }

    public void setRequiredUsage(boolean b) {
	if(b) {
	    props.setProperty("session.msg.requiredUsage", "1");
	}
	else {
	    props.setProperty("session.msg.requiredUsage", "0");
	}    
	modified = true;
	}*/

    /* Required or Empty Usage Option - Create messages that specifically 
       examine the required or empty usage option. */
/*    public boolean getRequiredEmptyUsage() {
	boolean b = false;
	
	if(props.getProperty("session.msg.requiredEmptyUsage").compareTo("1") == 0) {
	    b = true;
	}
	return b;
    }

    public void setRequiredEmptyUsage(boolean b) {
	if(b) {
	    props.setProperty("session.msg.requiredEmptyUsage", "1");
	}
	else {
	    props.setProperty("session.msg.requiredEmptyUsage", "0");
	}    
	modified = true;
    }
*/
    
    /* Conditional Usage Option - Create messages that specifically examine 
       the conditional usage option. */
/*    public boolean getConditionalUsage() {
	boolean b = false;
	
	if(props.getProperty("session.msg.conditionalUsage").compareTo("1") == 0) {
	    b = true;
	}
	return b;
    }

    public void setConditionalUsage(boolean b) {
	if(b) {
	    props.setProperty("session.msg.conditionalUsage", "1");
	}
	else {
	    props.setProperty("session.msg.conditionalUsage", "0");
	}    
	modified = true;
    }
*/  
    /* Conditional or Empty Usage Option - Create messages that specifically 
       examine the conditional or Empty usage option. */
/*    public boolean getConditionalEmptyUsage() {
	boolean b = false;
	
	if(props.getProperty("session.msg.conditionalEmptyUsage").compareTo("1") == 0) {
	    b = true;
	}
	return b;
    }

    public void setConditionalEmptyUsage(boolean b) {
	if(b) {
	    props.setProperty("session.msg.conditionalEmptyUsage", "1");
	}
	else {
	    props.setProperty("session.msg.conditionalEmptyUsage", "0");
	}    
	modified = true;
    }
*/  
    /* Not Supported Usage Option - Create messages that specifically 
       examine the not supported usage option. */
/*    public boolean getNotSupportedUsage() {
	boolean b = false;
	
	if(props.getProperty("session.msg.notSupportedUsage").compareTo("1") == 0) {
	    b = true;
	}
	return b;
    }

    public void setNotSupportedUsage(boolean b) {
	if(b) {
	    props.setProperty("session.msg.notSupportedUsage", "1");
	}
	else {
	    props.setProperty("session.msg.notSupportedUsage", "0");
	}    
	modified = true;
    }
*/
    /* compositeLength - Create messages that specifically exmine the composite length
       of an element. */
/*    public boolean getCompositeLength() {
	boolean b = false;
	
	if(props.getProperty("session.msg.compositeLength").compareTo("1") == 0) {
	    b = true;
	}
	return b;
    }

    public void setCompositeLength(boolean b) {
	if(b) {
	    props.setProperty("session.msg.compositeLength", "1");
	}
	else {
	    props.setProperty("session.msg.compositeLength", "0");
	}    
	modified = true;
    }
*/  
    /* fieldLength - Create messages that specifically exmine the field length
       of an element. */
/*    public boolean getFieldLength() {
	boolean b = false;
	
	if(props.getProperty("session.msg.fieldLength").compareTo("1") == 0) {
	    b = true;
	}
	return b;
    }

    public void setFieldLength(boolean b) {
	if(b) {
	    props.setProperty("session.msg.fieldLength", "1");
	}
	else {
	    props.setProperty("session.msg.fieldLength", "0");
	}    
	modified = true;
    }
*/  
  /* componentLength - Create messages that specifically exmine the field 
     length of an element. */
/*    public boolean getComponentLength() {
	boolean b = false;
	
	if(props.getProperty("session.msg.componentLength").compareTo("1") == 0) {
	    b = true;
	}
	return b;
    }

    public void setComponentLength(boolean b) {
	if(b) {
	    props.setProperty("session.msg.componentLength", "1");
	}
	else {
	    props.setProperty("session.msg.componentLength", "0");
	}    
	modified = true;
    }
*/  
    /* subComponentLength - Create messages that specifically exmine the  
       subComponent length of an element. */
/*    public boolean getSubComponentLength() {
	boolean b = false;
	
	if(props.getProperty("session.msg.subComponentLength").compareTo("1") == 0) {
	    b = true;
	}
	return b;
    }

    public void setSubComponentLength(boolean b) {
	if(b) {
	    props.setProperty("session.msg.subComponentLength", "1");
	}
	else {
	    props.setProperty("session.msg.subComponentLength", "0");
	}    
	modified = true;
    }
*/
    /* Content Options - Create messages that contain certain data content. 
       The following options allow for the inclusion or exclusion of these
       content tests. */
    
    /* Data Value - Include content testing data values. */
/*    public boolean getDataValue() {
	boolean b = false;
	
	if(props.getProperty("session.msg.dataValue").compareTo("1") == 0) {
	    b = true;
	}
	return b;
    }
    
    public void setDataValue(boolean b) {
	if(b) {
	    props.setProperty("session.msg.dataValue", "1");
	}
	else {
	    props.setProperty("session.msg.dataValue", "0");
	}    
	modified = true;
	}*/
    
    /* Data Type - Include content testing the data type of elements. */
/*    public boolean getDataType() {
	boolean b = false;
	
	if(props.getProperty("session.msg.dataType").compareTo("1") == 0) {
	    b = true;
	}
	return b;
    }
    
    public void setDataType(boolean b) {
	if(b) {
	    props.setProperty("session.msg.dataType", "1");
	}
	else {
	    props.setProperty("session.msg.dataType", "0");
	}    
	modified = true;
    }
*/  
    /* Out of Range - Include content testing "out of range" data values. */
/*    public boolean getOutOfRange() {
	boolean b = false;
	
	if(props.getProperty("session.msg.outOfRange").compareTo("1") == 0) {
	    b = true;
	}
	return b;
    }

    public void setOutOfRange(boolean b) {
	if(b) {
	    props.setProperty("session.msg.outOfRange", "1");
	}
	else {
	    props.setProperty("session.msg.outOfRange", "0");
	}    
	modified = true;
    }
*/
    /* Format - Include content testing data format. */
/*    public boolean getFormat() {
	boolean b = false;
	
	if(props.getProperty("session.msg.format").compareTo("1") == 0) {
	    b = true;
	}
	return b;
    }

    public void setFormat(boolean b) {	
	if(b) {
	    props.setProperty("session.msg.format", "1");
	}
	else {
	    props.setProperty("session.msg.format", "0");
	}    
	modified = true;
    }
*/
    /* Usage Type */
    public String getUsageType() {
	return props.getProperty("session.msg.usageType");
    }

    public void setUsageType(String usageType) {
	props.setProperty("session.msg.usageType", usageType);
	modified = true;
    }

    /* Content Type */
/*    public String getContentType() {
	return props.getProperty("session.msg.contentType");
    }

    public void setContentType(String contentType) {
	props.setProperty("session.msg.contentType", contentType);
	modified = true;
	}*/

    /* Content Complexity */
/*    public String getContentComplexity() {
	return props.getProperty("session.msg.contentComplexity");
    }

    public void setContentComplexity(String contentComplexity) {
	props.setProperty("session.msg.contentComplexity", contentComplexity);
	modified = true;
    }
*/

    /* Global Length */
    public String getLengthGlobal() {
	return props.getProperty("session.msg.lengthGlobal");
    }

    public void setLengthGlobal(String lengthGlobal) {
	props.setProperty("session.msg.lengthGlobal", lengthGlobal);
	modified = true;
    }

    /* Coded Entry Type */
    public String getCodedEntryType() {
	return props.getProperty("session.msg.codedEntryType");
    }

    public void setCodedEntryType(String codedEntryType) {
	props.setProperty("session.msg.codedEntryType", codedEntryType);
	modified = true;
    }

    /* Length Type */
    public String getLengthType() {
	return props.getProperty("session.msg.lengthType");
    }

    public void setLengthType(String lengthType) {
	props.setProperty("session.msg.lengthType", lengthType);
	modified = true;
    }

    /* Cardinality Type */
    public String getCardinalityType() {
	return props.getProperty("session.msg.cardinalityType");
    }

    public void setCardinalityType(String cardinalityType) {
	props.setProperty("session.msg.cardinalityType", cardinalityType);
	modified = true;
    }

    /* minimum and maximum value for the volume field */
    public int getVolumeMin() {
	return Integer.parseInt(props.getProperty("session.msg.volumeMin"));
    }

    public int getVolumeMax() {
	return Integer.parseInt(props.getProperty("session.msg.volumeMax"));
    }

    public void setModified(boolean b) {
	modified = b;
    }

    /* Test */
    private Vector vCardinalityTestValid;
    private Vector vCardinalityTestInvalid;
    private Vector vCardinalityTestBoth;
    private Vector vUsageTestValid;
    private Vector vUsageTestInvalid;
    private Vector vUsageTestBoth;
    private Vector vErrorCodesTestValid;
    private Vector vErrorCodesTestInvalid;
    private Vector vMsgVariant;

    /* Temporary Test */
    private Vector vCardinalityTempTestValid;
    private Vector vCardinalityTempTestInvalid;
    private Vector vCardinalityTempTestBoth;
    private Vector vUsageTempTestValid;
    private Vector vUsageTempTestInvalid;
    private Vector vUsageTempTestBoth;
    private Vector vErrorCodesTempTestValid;
    private Vector vErrorCodesTempTestInvalid;

    /**
     * Add a Test
     * @param type The type of the test
     * @param pn The ProfileNode
     */
    public void addTest(String type, ProfileNode pn) {
	if(type.equals("cardinalityValid")) {
	    if(!vCardinalityTestValid.contains(pn))
		vCardinalityTestValid.add(pn);
	}
	else if(type.equals("cardinalityInvalid")) {
	    if(!vCardinalityTestInvalid.contains(pn))
		vCardinalityTestInvalid.add(pn);
	}
	else if(type.equals("cardinalityBoth")) {
	    if(!vCardinalityTestBoth.contains(pn))
		vCardinalityTestBoth.add(pn);
	}
	else if(type.equals("usageValid")) {
	    if(!vUsageTestValid.contains(pn))
		vUsageTestValid.add(pn);
	}
	else if(type.equals("usageInvalid")) {
	    if(!vUsageTestInvalid.contains(pn))
		vUsageTestInvalid.add(pn);
	}
	else if(type.equals("usageBoth")) {
	    if(!vUsageTestBoth.contains(pn))
		vUsageTestBoth.add(pn);
	}
	else if(type.equals("errorCodesValid")) {
	    if(!vErrorCodesTestValid.contains(pn))
		vErrorCodesTestValid.add(pn);
	}
	else if(type.equals("errorCodesInvalid")) {
	    if(!vErrorCodesTestInvalid.contains(pn))
		vErrorCodesTestInvalid.add(pn);
	}
    }

    /**
     * Add a Temporary (unsaved) Test
     * @param type The type of the test
     * @param pn The ProfileNode
     */
    public void addTempTest(String type, ProfileNode pn) {
	if(type.equals("cardinalityValid")) {
	    if(!vCardinalityTempTestValid.contains(pn))
		vCardinalityTempTestValid.add(pn);
	}
	else if(type.equals("cardinalityInvalid")) {
	    if(!vCardinalityTempTestInvalid.contains(pn))
		vCardinalityTempTestInvalid.add(pn);
	}
	else if(type.equals("cardinalityBoth")) {
	    if(!vCardinalityTempTestBoth.contains(pn))
		vCardinalityTempTestBoth.add(pn);
	}
	else if(type.equals("usageValid")) {
	    if(!vUsageTempTestValid.contains(pn))
		vUsageTempTestValid.add(pn);
	}
	else if(type.equals("usageInvalid")) {
	    if(!vUsageTempTestInvalid.contains(pn))
		vUsageTempTestInvalid.add(pn);
	}
	else if(type.equals("usageBoth")) {
	    if(!vUsageTempTestBoth.contains(pn))
		vUsageTempTestBoth.add(pn);
	}
	else if(type.equals("errorCodesValid")) {
	    if(!vErrorCodesTempTestValid.contains(pn))
		vErrorCodesTempTestValid.add(pn);
	}
	else if(type.equals("errorCodesInvalid")) {
	    if(!vErrorCodesTempTestInvalid.contains(pn))
		vErrorCodesTempTestInvalid.add(pn);
	}
    }

    /**
     * Remove a Test
     * @param type The type of the test
     * @param pn The ProfileNode
     */
    public void removeTest(String type, ProfileNode pn) {
	if(type.equals("cardinalityValid")) {
	    vCardinalityTestValid.remove(pn);
	}
	else if(type.equals("cardinalityInvalid")) {
	    vCardinalityTestInvalid.remove(pn);
	}
	else if(type.equals("cardinalityBoth")) {
	    vCardinalityTestBoth.remove(pn);
	}
	else if(type.equals("usageValid")) {
	    vUsageTestValid.remove(pn);
	}
	else if(type.equals("usageInvalid")) {
	    vUsageTestInvalid.remove(pn);
	}
	else if(type.equals("usageBoth")) {
	    vUsageTestBoth.remove(pn);
	}
	else if(type.equals("errorCodesValid")) {
	    vErrorCodesTestValid.remove(pn);
	}
	else if(type.equals("errorCodesInvalid")) {
	    vErrorCodesTestInvalid.remove(pn);
	}
    }

    /**
     * Remove a Temporary Test
     * @param type The type of the test
     * @param pn The ProfileNode
     */
    public void removeTempTest(String type, ProfileNode pn) {
	if(type.equals("cardinalityValid")) {
	    vCardinalityTempTestValid.remove(pn);
	}
	else if(type.equals("cardinalityInvalid")) {
	    vCardinalityTempTestInvalid.remove(pn);
	}
	else if(type.equals("cardinalityBoth")) {
	    vCardinalityTempTestBoth.remove(pn);
	}
	else if(type.equals("usageValid")) {
	    vUsageTempTestValid.remove(pn);
	}
	else if(type.equals("usageInvalid")) {
	    vUsageTempTestInvalid.remove(pn);
	}
	else if(type.equals("usageBoth")) {
	    vUsageTempTestBoth.remove(pn);
	}
	else if(type.equals("errorCodesValid")) {
	    vErrorCodesTempTestValid.remove(pn);
	}
	else if(type.equals("errorCodesInvalid")) {
	    vErrorCodesTempTestInvalid.remove(pn);
	}
    }

    /**
     * Save Test to the properties
     */
    public void saveTest() {
	// Cardinality
	String cardinality = "";
	for(int i=0;i<vCardinalityTestValid.size();i++) {
	    ProfileNode pn = (ProfileNode)vCardinalityTestValid.get(i);
	    cardinality += (pn.getParentPath() + "@" + pn.getSegment() + ":" + pn.getName() + "|" + pn.getQualifier() + "^" + pn.getQualifier2() + "|");
	    if(i != vCardinalityTestValid.size()-1)
		cardinality += "#";
	}
	//	System.out.println("Cardinality tests: " + cardinality);
	props.setProperty("session.msg.test.cardinalityValid", cardinality);
	cardinality = "";
	for(int i=0;i<vCardinalityTestInvalid.size();i++) {
	    ProfileNode pn = (ProfileNode)vCardinalityTestInvalid.get(i);
	    cardinality += (pn.getParentPath() + "@" + pn.getSegment() + ":" + pn.getName() + "|" + pn.getQualifier() + "^" + pn.getQualifier2() + "|");
	    if(i != vCardinalityTestInvalid.size()-1)
		cardinality += "#";
	}
	//	System.out.println("Cardinality tests: " + cardinality);
	props.setProperty("session.msg.test.cardinalityInvalid", cardinality);
/*	cardinality = "";
	for(int i=0;i<vCardinalityTestBoth.size();i++) {
	    ProfileNode pn = (ProfileNode)vCardinalityTestBoth.get(i);
	    cardinality += (pn.getParentPath() + "@" + pn.getSegment() + ":" + pn.getName() + "|" + pn.getQualifier() + "|");
	    if(i != vCardinalityTestBoth.size()-1)
		cardinality += "#";
		}
	//	System.out.println("Cardinality tests: " + cardinality);
	props.setProperty("session.msg.test.cardinalityBoth", cardinality);*/
	// Usage
	String usage = "";
	for(int i=0;i<vUsageTestValid.size();i++) {
	    ProfileNode pn = (ProfileNode)vUsageTestValid.get(i);
	    usage += (pn.getParentPath() + "@" + pn.getSegment() + ":" + pn.getName() + "|" + pn.getQualifier() + "^" + pn.getQualifier2() + "|");
	    if(i != vUsageTestValid.size()-1)
		usage += "#";
	}
	//	System.out.println("Usage tests: " + usage);
	props.setProperty("session.msg.test.usageValid", usage);
	usage = "";
	for(int i=0;i<vUsageTestInvalid.size();i++) {
	    ProfileNode pn = (ProfileNode)vUsageTestInvalid.get(i);
	    usage += (pn.getParentPath() + "@" + pn.getSegment() + ":" + pn.getName() + "|" + pn.getQualifier() + "^" + pn.getQualifier2() + "|");
	    if(i != vUsageTestInvalid.size()-1)
		usage += "#";
	}
	//	System.out.println("Usage tests: " + usage);
	props.setProperty("session.msg.test.usageInvalid", usage);
/*	usage = "";
	for(int i=0;i<vUsageTestBoth.size();i++) {
	    ProfileNode pn = (ProfileNode)vUsageTestBoth.get(i);
	    usage += (pn.getParentPath() + "@" + pn.getSegment() + ":" + pn.getName() + "|" + pn.getQualifier() + "|");
	    if(i != vUsageTestBoth.size()-1)
		usage += "#";
	}
	//	System.out.println("Usage tests: " + usage);
	props.setProperty("session.msg.test.usageBoth", usage);*/
	String errorCodes = "";
	for(int i=0;i<vErrorCodesTestValid.size();i++) {
	    ProfileNode pn = (ProfileNode)vErrorCodesTestValid.get(i);
	    errorCodes += (pn.getParentPath() + "@" + pn.getSegment() + ":" + pn.getName() + "|" + pn.getQualifier() + "^" + pn.getQualifier2() + "|");
	    if(i != vErrorCodesTestValid.size()-1)
		errorCodes += "#";
	}
	//	System.out.println("Usage tests: " + usage);
	props.setProperty("session.msg.test.errorCodesValid", errorCodes);
	errorCodes = "";
	for(int i=0;i<vErrorCodesTestInvalid.size();i++) {
	    ProfileNode pn = (ProfileNode)vErrorCodesTestInvalid.get(i);
	    errorCodes += (pn.getParentPath() + "@" + pn.getSegment() + ":" + pn.getName() + "|" + pn.getQualifier() + "^" + pn.getQualifier2() + "|");
	    if(i != vErrorCodesTestInvalid.size()-1)
		errorCodes += "#";
	}
	//	System.out.println("Usage tests: " + usage);
	props.setProperty("session.msg.test.errorCodesInvalid", errorCodes);
	modified = true;
    }

    /**
     * Load Test from the properties
     */
    public void loadTest() {
	clearTest();
	// Cardinality
	String tests = props.getProperty("session.msg.test.cardinalityValid");
	vCardinalityTestValid.addAll(parseTest(tests));
	tests = props.getProperty("session.msg.test.cardinalityInvalid");
	vCardinalityTestInvalid.addAll(parseTest(tests));
	// Usage
	tests = props.getProperty("session.msg.test.usageValid");
	vUsageTestValid.addAll(parseTest(tests));
	tests = props.getProperty("session.msg.test.usageInvalid");
	vUsageTestInvalid.addAll(parseTest(tests));
	// Error Codes
	tests = props.getProperty("session.msg.test.errorCodesValid");
	vErrorCodesTestValid.addAll(parseTest(tests));
	tests = props.getProperty("session.msg.test.errorCodesInvalid");
	vErrorCodesTestInvalid.addAll(parseTest(tests));
    }
	
    /**
     * Parse the test
     * @param tests The String test
     * @return A Vector with the information (parent / segment / name / qualifier)
     */
    private Vector parseTest(String tests) {
	StringTokenizer st = null;
	String token = null;
	int idx0 = -1, idx1 = -1, idx2 = -1, idx3 = -1;
	String segGroup = null, segment = null, name = null, qualifier = null, qualifier2 = null;
	Vector v = new Vector();
	if(tests != null) {
	    st = new StringTokenizer(tests,"#");
	    while(st.hasMoreTokens()) {
		// segGroup@:|| Segment Group where segGroup is like xxx/xxx/xxx
		// segGroup@MSH:MSH|| Segment
		// segGroup@MSH:xxx|| Field
		// segGroup@MSH:xxx|xxx| Component
		// segGroup@MSH:xxx|xxx^xxx| SubComponent
		token = st.nextToken();
		idx0 = token.indexOf("@");
		idx1 = token.indexOf(":");
		idx2 = token.indexOf("|");
		idx3 = token.indexOf("^");
		if(idx0 < idx1)
		    segGroup = token.substring(0, idx0);
		segment = token.substring(idx0+1, idx1);
		name = token.substring(idx1+1, idx2);
		if(idx3 == -1) {
		    qualifier = token.substring(idx2+1, token.length()-1);
		    qualifier2 = "";
		}
		else {
		    qualifier = token.substring(idx2+1, idx3);
		    qualifier2 = token.substring(idx3+1, token.length()-1);
		}

		ProfileNode pn = null;
		if(!segGroup.equals("") && segment.equals(""))
		    pn = new SegmentGroupNode();		
		else if(name.equals(segment))
		    pn = new SegmentNode();
		else if(qualifier.equals(""))
		    pn = new FieldNode();
		else if(qualifier2.equals(""))
		    pn = new ComponentNode();
		else 
		    pn = new SubComponentNode();

		pn.setSegment(segment);
		pn.setName(name);
		pn.setQualifier(qualifier);
		pn.setQualifier2(qualifier2);
		ramProfile.setParent(pn, segGroup);
		ramProfile.setCardinality(pn);
		v.add(pn);
	    }
	}
	return v;
    }

    /**
     * Load Temp Test
     */
    public void loadTempTest() {
	clearTempTest();
	vCardinalityTempTestValid = (Vector)vCardinalityTestValid.clone();
	vCardinalityTempTestInvalid = (Vector)vCardinalityTestInvalid.clone();
	vCardinalityTempTestBoth = (Vector)vCardinalityTestBoth.clone();
	vUsageTempTestValid = (Vector)vUsageTestValid.clone();
	vUsageTempTestInvalid = (Vector)vUsageTestInvalid.clone();
	vUsageTempTestBoth = (Vector)vUsageTestBoth.clone();
	vErrorCodesTempTestValid = (Vector)vErrorCodesTestValid.clone();
	vErrorCodesTempTestInvalid = (Vector)vErrorCodesTestInvalid.clone();
    }

    /**
     * Clear test Vectors
     */
    public void clearTest() {
	vCardinalityTestValid.clear();
	vCardinalityTestInvalid.clear();
	vCardinalityTestBoth.clear();
	vUsageTestValid.clear();
	vUsageTestInvalid.clear();
	vUsageTestBoth.clear();
	vErrorCodesTestValid.clear();
	vErrorCodesTestInvalid.clear();
    }

    /**
     * Clear temporary test Vectors
     */
    public void clearTempTest() {
	vCardinalityTempTestValid.clear();
	vCardinalityTempTestInvalid.clear();
	vCardinalityTempTestBoth.clear();
	vUsageTempTestValid.clear();
	vUsageTempTestInvalid.clear();
	vUsageTempTestBoth.clear();
	vErrorCodesTempTestValid.clear();
	vErrorCodesTempTestInvalid.clear();
    }

    /**
     * Return true or false if there is a cardinality test for this ProfileNode
     * @param pn The ProfileNode
     * @type Type of messages VALID INVALID BOTH
     * @return true or false
     */
    public boolean isCardinalityTest(ProfileNode pn, short type) {
	switch(type) {
	case ProfileTableModel.VALID:
	    return vCardinalityTestValid.contains(pn);
	case ProfileTableModel.INVALID:
	    return vCardinalityTestInvalid.contains(pn);
	case ProfileTableModel.BOTH:
	    return vCardinalityTestBoth.contains(pn);
	}
	return false;
    }

    /**
     * Return true or false if there is a usage test for this ProfileNode
     * @param pn The ProfileNode
     * @type Type of messages VALID INVALID BOTH
     * @return true or false
     */
    public boolean isUsageTest(ProfileNode pn, short type) {
	switch(type) {
	case ProfileTableModel.VALID:
	    return vUsageTestValid.contains(pn);
	case ProfileTableModel.INVALID:
	    return vUsageTestInvalid.contains(pn);
	case ProfileTableModel.BOTH:
	    return vUsageTestBoth.contains(pn);
	}
	return false;
    }

    /**
     * Return true or false if there is an error codes test for this ProfileNode
     * @param pn The ProfileNode
     * @type Type of messages VALID INVALID BOTH
     * @return true or false
     */
    public boolean isErrorCodesTest(ProfileNode pn, short type) {
	switch(type) {
	case ProfileTableModel.VALID:
	    return vErrorCodesTestValid.contains(pn);
	case ProfileTableModel.INVALID:
	    return vErrorCodesTestInvalid.contains(pn);
	}
	return false;
    }

    /**
     * Load local tables from Properties
     */
    public void loadLocalTable() {
	String local = props.getProperty("session.data.local");
	StringTokenizer st = null;
	String token = null;
	int idx = -1;
	vLocalTable.clear();
	if(local != null) {
	    st = new StringTokenizer(local, ",");
	    while(st.hasMoreTokens()) {
		// Priority|Filename
		token = st.nextToken();
		idx = token.indexOf("|");
		vLocalTable.add(new TablePriority(token.substring(idx+1), Integer.parseInt(token.substring(0, idx))));
	    }
	}
	setChanged();
	notifyObservers(vLocalTable);
    }

    /**
     * Save local tables to Properties
     */
    public void saveLocalTable() {
	String local = "";
	for(int i=0;i<vLocalTable.size();i++) {
	    TablePriority tp = (TablePriority)vLocalTable.get(i);
	    local += tp.getPriority() + "|" + tp.getFilename();
	    if(i+1 != vLocalTable.size())
		local += ",";
	}
	props.setProperty("session.data.local", local);
    }

    /**
     * Get the Vector of Local Tables
     * @return A vector
     */
    public Vector getLocalTable() {
	return vLocalTable;
    }

    /**
     * Set the Vector of Local Tables
     * @param vLocal The local tables vector
     * @param modified If the local table vector was really modified
     */
    public void setLocalTable(Vector vLocal, boolean modified) {
	vLocalTable = vLocal;
	saveLocalTable();
	setChanged();
	notifyObservers(vLocalTable);
	if(modified)
	    this.modified = modified;
    }

    /**
     * Return the list of all tables with their priorities
     */
    public Vector getTable() {
	if(props != null) {
	    Vector v = new Vector();
	    // HL7 / User / External
	    String priority = props.getProperty("session.data.tables.priority");
	    StringTokenizer st = new StringTokenizer(priority, ",");
	    String token = null;
	    int i=0;
	    while(st.hasMoreTokens()) {
		token = st.nextToken();
		try {
		    switch(i) {
		    case 0: // HL7
			v.add(new TablePriority(new File(getHL7Tables()).getCanonicalPath(), Integer.parseInt(token)));
			break;
		    case 1: // User
			v.add(new TablePriority(new File(getUserTables()).getCanonicalPath(), Integer.parseInt(token)));
			break;
		    case 2: // External
			v.add(new TablePriority(new File(getExternalTables()).getCanonicalPath(), Integer.parseInt(token)));
			break;
		    }
		}
		catch(IOException ioe) {
		    System.err.println("IOException: " + ioe.getMessage());
		    ioe.printStackTrace();
		}
		i++;
	    }
	    // Local
	    v.addAll(getLocalTable());
	    // Order by priority
	    Collections.sort(v, new Comparator() {
		    public int compare(Object o1, Object o2) {
			TablePriority tp1 = (TablePriority)o1;
			TablePriority tp2 = (TablePriority)o2;
			return tp1.getPriority() - tp2.getPriority();
		    }
		});
	    return v;
	}
	return null;
    }

    /**
     * Set the priority for the tables
     * @param vPriority The priorities
     */
    public void setPriority(Vector vPriority) {
	// Order by priority
	Collections.sort(vPriority, new Comparator() {
		public int compare(Object o1, Object o2) {
		    TablePriority tp1 = (TablePriority)o1;
		    TablePriority tp2 = (TablePriority)o2;
		    return tp1.getPriority() - tp2.getPriority();
		}
	    });
	// Reset the priority to fill in gaps 1 2 4 will become 1 2 3
	for(int i=0;i<vPriority.size();i++) {
	    TablePriority tp = (TablePriority)vPriority.get(i);
	    tp.setPriority(i);
	}
	// HL7 / User / External
	String priority = "";
	int idx = vPriority.indexOf(new TablePriority(getHL7Tables()));
	priority = "" + ((TablePriority)vPriority.get(idx)).getPriority();
	vPriority.remove(idx);
	priority += ",";
	idx = vPriority.indexOf(new TablePriority(getUserTables()));
	priority += ((TablePriority)vPriority.get(idx)).getPriority();
	vPriority.remove(idx);
	priority += ",";
	idx = vPriority.indexOf(new TablePriority(getExternalTables()));
	priority += ((TablePriority)vPriority.get(idx)).getPriority();
	vPriority.remove(idx);
	props.setProperty("session.data.tables.priority", priority);
	setLocalTable(vPriority, true);
    }

    /**
     * Return the status of the session, if it was modified or not since
     * the openning of the application
     * @return true if modified
     */
    public boolean isModified() {
	if(getName() == null || getName().compareTo("") == 0)
	    return false;
	return modified;
    }

    /**
     * Return the path of a file with a protocol
     * @param filename The filename
     * @return The canonical path
     */
    public static String getProtocolPath(String filename) {
	try {
	    File f = new File(filename);
	    if(!f.isAbsolute()) {
		return "file:///" + f.getCanonicalPath();
	    }
	    else {
		return "file:///" + f.getPath();
	    }
	}
	catch(IOException ioe) {
	    System.err.println("IOException " + ioe.getMessage());
	    ioe.printStackTrace();
	}
	return null;
    }

    /**
     * Fire an event
     * @param arg An MVCArg object
     */
    public void fireEvent(MVCArg arg) {
	gov.nist.hl7maker.util.MVCLog.info("[Session] Fire an event " + arg.getName() + "=" + arg.getValue());
	setChanged();
	notifyObservers(arg);
    }

    /**
     * Generate a XML file which contains all the options for the Message Generator
     * @return The name of the file generated
     */
    public String generateMGOptions() {
	try {
	    // XML properties
	    String validity = null, lengthConstraints = null, segmentCardinality = null, fieldCardinality = null, RE_Usage = null, minSegCardError = null, maxSegCardError = null, minFieldCardError = null, maxFieldCardError = null, RusageError = null, XusageError = null, fieldLengthError = null, componentLengthError = null, subComponentLengthError = null, buildErrorCodes = null, hl7ErrorCodes = null, userErrorCodes = null, localErrorCodes = null, externalErrorCodes = null;
	    if(vMsgVariant == null) {
		vMsgVariant = new Vector();
	    }
	    else {
		vMsgVariant.clear();
	    }
	    // Create a temporary file
	    File fOptions = File.createTempFile("MessageMakerConfig", ".xml");
	    BufferedWriter buf = new BufferedWriter(new FileWriter(fOptions));

	    // Global length (override in case of a custom length test
	    if(getLengthGlobal().equals("Ignore")) {
		lengthConstraints = "false";
	    }
	    else if(getLengthGlobal().equals("Enforce")) {
		lengthConstraints = "true";
	    }
	    int msgCount = getVolume();
	    String typeMessageSet = getTypeMessageSet();
	    // A valid set of messages
	    if(typeMessageSet.equals("Valid")) {
		validity = "VALID";
		segmentCardinality = "RANDOM";
		fieldCardinality = "RANDOM";
		RE_Usage = "SOME";

		minSegCardError = "false";
		maxSegCardError = "false";
		minFieldCardError = "false";
		maxFieldCardError = "false";
		RusageError = "false";
		XusageError = "false";
		fieldLengthError = "false";
		componentLengthError = "false";
		subComponentLengthError = "false";

		buildErrorCodes = "false";
		hl7ErrorCodes = "false";
		userErrorCodes = "false";
		localErrorCodes = "false";
		externalErrorCodes = "false";
	    }
	    // An invalid set of messages
	    else if(typeMessageSet.equals("Invalid")) {
		validity = "INVALID";
		segmentCardinality = "RANDOM";
		fieldCardinality = "RANDOM";
		RE_Usage = "SOME";

		minSegCardError = "true";
		maxSegCardError = "true";
		minFieldCardError = "true";
		maxFieldCardError = "true";
		RusageError = "true";
		XusageError = "true";
		if(getLengthGlobal().equals("Ignore")) {
		    fieldLengthError = "false";
		    componentLengthError = "false";
		    subComponentLengthError = "false";
		}
		else if(getLengthGlobal().equals("Enforce")) {
		    fieldLengthError = "true";
		    componentLengthError = "true";
		    subComponentLengthError = "true";
		}

		buildErrorCodes = "true";
		hl7ErrorCodes = "true";
		userErrorCodes = "true";
		localErrorCodes = "true";
		externalErrorCodes = "true";
	    }
	    // A set of both valid and invalid messages
	    else if(typeMessageSet.equals("Both")) {
		validity = "VALID_AND_INVALID";
		segmentCardinality = "RANDOM";
		fieldCardinality = "RANDOM";
		RE_Usage = "SOME";

		minSegCardError = "true";
		maxSegCardError = "true";
		minFieldCardError = "true";
		maxFieldCardError = "true";
		RusageError = "true";
		XusageError = "true";
		if(getLengthGlobal().equals("Ignore")) {
		    fieldLengthError = "false";
		    componentLengthError = "false";
		    subComponentLengthError = "false";
		}
		else if(getLengthGlobal().equals("Enforce")) {
		    fieldLengthError = "true";
		    componentLengthError = "true";
		    subComponentLengthError = "true";
		}

		buildErrorCodes = "true";
		hl7ErrorCodes = "true";
		userErrorCodes = "true";
		localErrorCodes = "true";
		externalErrorCodes = "true";
	    }
	    // A sample message (randomly populated)
	    else if(typeMessageSet.equals("SampleRandom")) {
		validity = "VALID";
		segmentCardinality = "RANDOM";
		fieldCardinality = "RANDOM";
		RE_Usage = "SOME";

		minSegCardError = "false";
		maxSegCardError = "false";
		minFieldCardError = "false";
		maxFieldCardError = "false";
		RusageError = "false";
		XusageError = "false";
		fieldLengthError = "false";
		componentLengthError = "false";
		subComponentLengthError = "false";

		buildErrorCodes = "false";
		hl7ErrorCodes = "false";
		userErrorCodes = "false";
		localErrorCodes = "false";
		externalErrorCodes = "false";
	    }
	    // A sample message (minimally populated)
	    else if(typeMessageSet.equals("SampleMin")) {
		validity = "VALID";
		segmentCardinality = "MIN";
		fieldCardinality = "MIN";
		RE_Usage = "NONE";

		minSegCardError = "false";
		maxSegCardError = "false";
		minFieldCardError = "false";
		maxFieldCardError = "false";
		RusageError = "false";
		XusageError = "false";
		fieldLengthError = "false";
		componentLengthError = "false";
		subComponentLengthError = "false";

		buildErrorCodes = "false";
		hl7ErrorCodes = "false";
		userErrorCodes = "false";
		localErrorCodes = "false";
		externalErrorCodes = "false";
	    }
	    // A sample message (maximally populated)
	    else if(typeMessageSet.equals("SampleMax")) {
		validity = "VALID";
		segmentCardinality = "MAX";
		fieldCardinality = "MAX";
		RE_Usage = "ALL";

		minSegCardError = "false";
		maxSegCardError = "false";
		minFieldCardError = "false";
		maxFieldCardError = "false";
		RusageError = "false";
		XusageError = "false";
		fieldLengthError = "false";
		componentLengthError = "false";
		subComponentLengthError = "false";

		buildErrorCodes = "false";
		hl7ErrorCodes = "false";
		userErrorCodes = "false";
		localErrorCodes = "false";
		externalErrorCodes = "false";
	    }
	    // A custom message
	    else if(typeMessageSet.equals("Custom")) {
		// Default
		validity = "VALID";
		segmentCardinality = "MIN";
		fieldCardinality = "MIN";
		RE_Usage = "NONE";

		minSegCardError = "false";
		maxSegCardError = "false";
		minFieldCardError = "false";
		maxFieldCardError = "false";
		RusageError = "false";
		XusageError = "false";
		fieldLengthError = "false";
		componentLengthError = "false";
		subComponentLengthError = "false";

		// Custom Cardinality
		if(getCardinalityType().equals("ValidRange")) {
		    segmentCardinality = "RANDOM";
		    fieldCardinality = "RANDOM";
		    RE_Usage = "SOME";
		}
		else if(getCardinalityType().equals("ValidMax")) {
		    segmentCardinality = "MAX";
		    fieldCardinality = "MAX";
		    RE_Usage = "ALL";
		}
		else if(getCardinalityType().equals("Invalid")) {
		    minSegCardError = "true";
		    maxSegCardError = "true";
		    minFieldCardError = "true";
		    maxFieldCardError = "true";
		    validity = "INVALID";
		}
		// Custom Usage
		if(getUsageType().equals("Invalid")) {
		    RusageError = "true";
		    XusageError = "true";
		    validity = "INVALID";
		}
		// Custom Length
		if(getLengthType().equals("Invalid")) {
		    fieldLengthError = "true";
		    componentLengthError = "true";
		    subComponentLengthError = "true";
		    validity = "INVALID";
		}
		if(getCodedEntryType().equals("Invalid")) {
		    buildErrorCodes = "true";
		    hl7ErrorCodes = "true";
		    userErrorCodes = "true";
		    localErrorCodes = "true";
		    externalErrorCodes = "true";
		}
	    }
	    else if(typeMessageSet.equals("Manual")) {
		msgCount = 0;
		// Default
		validity = "VALID";
		segmentCardinality = "MIN";
		fieldCardinality = "MIN";
		RE_Usage = "NONE";

		minSegCardError = "false";
		maxSegCardError = "false";
		minFieldCardError = "false";
		maxFieldCardError = "false";
		RusageError = "false";
		XusageError = "false";
		fieldLengthError = "false";
		componentLengthError = "false";
		subComponentLengthError = "false";

		buildErrorCodes = "false";
		hl7ErrorCodes = "false";
		userErrorCodes = "false";
		localErrorCodes = "false";
		externalErrorCodes = "false";

		loadTest();
		//loadProfile(getProfile());
		// MessageVariant Usage Valid
		for(int i=0;i<vUsageTestValid.size();i++) {
		    validity = "VALID";
		    ProfileNode pn = (ProfileNode)vUsageTestValid.get(i);
		    String usage = ramProfile.getUsage(pn);
		    if(usage.equals("RE")) {			
			addMsgVariant(pn);
			vMsgVariant.add("REinclude");
			vMsgVariant.add("true");
			vMsgVariant.add("segmentCardinality");
			vMsgVariant.add("MAX");
			vMsgVariant.add("fieldCardinality");
			vMsgVariant.add("MAX");
			vMsgVariant.add("validity");
			vMsgVariant.add(validity);
			if(pn.hasREUsage()) {
			    RE_Usage = "ALL";
			}
		    }
		}
		// MessageVariant Usage Invalid
		for(int i=0;i<vUsageTestInvalid.size();i++) {
		    validity = "INVALID";
		    ProfileNode pn = (ProfileNode)vUsageTestInvalid.get(i);
		    addMsgVariant(pn);
		    String usage = ramProfile.getUsage(pn);
		    if(usage.equals("R")) {
			vMsgVariant.add("RusageError");
			vMsgVariant.add("true");
		    }
		    else if(usage.equals("X")) {
			vMsgVariant.add("XusageError");
			vMsgVariant.add("true");
			/*vMsgVariant.add("segmentCardinality");
			vMsgVariant.add("MAX");
			vMsgVariant.add("fieldCardinality");
			vMsgVariant.add("MAX");*/
			vMsgVariant.add("cardinality");
			vMsgVariant.add("1");
		    }
		    vMsgVariant.add("validity");
		    vMsgVariant.add(validity);
		    if(pn.hasREUsage()) {
			RE_Usage = "ALL";
		    }
		}
		// MessageVariant Usage Both
		for(int i=0;i<vUsageTestBoth.size();i++) {
		}
		// MessageVariant Cardinality Valid
		for(int i=0;i<vCardinalityTestValid.size();i++) {
		    validity = "VALID";
		    ProfileNode pn = (ProfileNode)vCardinalityTestValid.get(i);
		    String usage = ramProfile.getUsage(pn);
		    //setCardinality(pn);
		    if(usage.equals("R") || usage.equals("RE")) {
			int min = Integer.parseInt(pn.getCardMin());
			int max = -1;
			if(!pn.getCardMax().equals("*"))
			    max = Integer.parseInt(pn.getCardMax());
			// MIN+1
			addMsgVariant(pn);
			if(min == 0) {
			    RE_Usage = "ALL";
			}
			if(pn.hasREUsage()) {
			    RE_Usage = "ALL";
			    vMsgVariant.add("segmentCardinality");
			    vMsgVariant.add("MAX");
			    vMsgVariant.add("fieldCardinality");
			    vMsgVariant.add("MAX");
			}
			vMsgVariant.add("cardinality");
			vMsgVariant.add(String.valueOf(min+1));
			vMsgVariant.add("validity");
			vMsgVariant.add(validity);
			// MAX or MIN + 3 if max=-1
			if(min + 1 != max) {
			    addMsgVariant(pn);
			    if(pn.hasREUsage()) {
				vMsgVariant.add("segmentCardinality");
				vMsgVariant.add("MAX");
				vMsgVariant.add("fieldCardinality");
				vMsgVariant.add("MAX");
			    }
			    vMsgVariant.add("cardinality");
			    if(max == -1)
				vMsgVariant.add(String.valueOf(min+3));
			    else
				vMsgVariant.add(String.valueOf(max));
			    vMsgVariant.add("validity");
			    vMsgVariant.add(validity);
			}
			// (MIN+MAX)/2 if max!=-1
			if(max != -1) {
			    if((min+max)/2 > min+1 && (min+max)/2 != max) {
				addMsgVariant(pn);
				if(pn.hasREUsage()) {
				    vMsgVariant.add("segmentCardinality");
				    vMsgVariant.add("MAX");
				    vMsgVariant.add("fieldCardinality");
				    vMsgVariant.add("MAX");
				}
				vMsgVariant.add("cardinality");
				vMsgVariant.add(String.valueOf((min+max)/2));
				vMsgVariant.add("validity");
				vMsgVariant.add(validity);
			    }
			}
		    }
		}
		// MessageVariant Cardinality Invalid
		for(int i=0;i<vCardinalityTestInvalid.size();i++) {
		    validity = "INVALID";
		    ProfileNode pn = (ProfileNode)vCardinalityTestInvalid.get(i);
		    String usage = ramProfile.getUsage(pn);
		    //setCardinality(pn);
		    int min = Integer.parseInt(pn.getCardMin());
		    int max = -1;
		    if(!pn.getCardMax().equals("*"))
			max = Integer.parseInt(pn.getCardMax());
		    //if(usage.equals("R")) {
		    if(min != 0) {
			addMsgVariant(pn);			
			if(pn instanceof SegmentNode) {
			    vMsgVariant.add("minSegCardError");
			}
			else if(pn instanceof FieldNode) {
			    vMsgVariant.add("minFieldCardError");
			}
			vMsgVariant.add("true");
			vMsgVariant.add("validity");
			vMsgVariant.add(validity);
			if(pn.hasREUsage()) {
			    vMsgVariant.add("segmentCardinality");
			    vMsgVariant.add("MAX");
			    vMsgVariant.add("fieldCardinality");
			    vMsgVariant.add("MAX");
			}
		    }
		    if(max != -1) {
			addMsgVariant(pn);			
			if(pn instanceof SegmentNode) {
			    vMsgVariant.add("maxSegCardError");
			}
			else if(pn instanceof FieldNode) {
			    vMsgVariant.add("maxFieldCardError");
			}
			vMsgVariant.add("true");
			vMsgVariant.add("validity");
			vMsgVariant.add(validity);
			if(pn.hasREUsage()) {
			    vMsgVariant.add("segmentCardinality");
			    vMsgVariant.add("MAX");
			    vMsgVariant.add("fieldCardinality");
			    vMsgVariant.add("MAX");
			}
		    }
		    if(usage.equals("RE")) {
			RE_Usage = "ALL";
		    }
		}
		// MessageVariant Cardinality Both
		for(int i=0;i<vCardinalityTestBoth.size();i++) {
		}
		// MessageVariant Error Codes Valid
		// MessageVariant Error Codes Invalid
		for(int i=0;i<vErrorCodesTestInvalid.size();i++) {
		    buildErrorCodes = "true";
		    hl7ErrorCodes = "true";
		    userErrorCodes = "true";
		    localErrorCodes = "true";
		    externalErrorCodes = "true";

		    validity = "INVALID";
		    ProfileNode pn = (ProfileNode)vErrorCodesTestInvalid.get(i);
		    addMsgVariant(pn);			
		    vMsgVariant.add("elementErrorCode");
		    vMsgVariant.add("true");
		    if(pn.hasREUsage()) {
			vMsgVariant.add("REinclude");
			vMsgVariant.add("true");
		    }
		    vMsgVariant.add("segmentCardinality");
		    vMsgVariant.add("MAX");
		    vMsgVariant.add("fieldCardinality");
		    vMsgVariant.add("MAX");
		    vMsgVariant.add("validity");
		    vMsgVariant.add(validity);
		}
	    }

	    // Put the content
	    buf.write("<?xml version=\"1.0\" ?>"); 
	    buf.write("<MMmessageProperties HL7Version=\"2.4\" Profile=\"" + getProfile() + "\">");
	    buf.write("<MetaData Name=\"Example\" OrgName=\"NIST\" Version=\"1.4\" />");
	    buf.write("<property name=\"messageControlId\" value=\"MSG\" />"); 
	    buf.write("<property name=\"msgCount\" value=\"" + msgCount + "\" />"); 
	    buf.write("<property name=\"offset\" value=\"-1\" />"); 
	    buf.write("<property name=\"validity\" value=\"" + validity + "\" />"); 
	    buf.write("<property name=\"lengthConstraints\" value=\"" + lengthConstraints + "\" />"); 
	    buf.write("<property name=\"compositeLength\" value=\"false\" />"); 
	    buf.write("<property name=\"segmentCardinality\" value=\"" + segmentCardinality + "\" />"); 
	    buf.write("<property name=\"fieldCardinality\" value=\"" + fieldCardinality + "\" />"); 
	    buf.write("<property name=\"RE_Usage\" value=\"" + RE_Usage + "\" />"); 
	    buf.write("<property name=\"minSegCardError\" value=\"" + minSegCardError + "\" />"); 
	    buf.write("<property name=\"maxSegCardError\" value=\"" + maxSegCardError + "\" />"); 
	    buf.write("<property name=\"minFieldCardError\" value=\"" + minFieldCardError + "\" />"); 
	    buf.write("<property name=\"maxFieldCardError\" value=\"" + maxFieldCardError + "\" />"); 
	    buf.write("<property name=\"RusageError\" value=\"" + RusageError + "\" />"); 
	    buf.write("<property name=\"XusageError\" value=\"" + XusageError + "\" />"); 
	    buf.write("<property name=\"fieldLengthError\" value=\"" + fieldLengthError + "\" />"); 
	    buf.write("<property name=\"componentLengthError\" value=\"" + componentLengthError + "\" />"); 
	    buf.write("<property name=\"subComponentLengthError\" value=\"" + subComponentLengthError + "\" />"); 
	    buf.write("<property name=\"starValue\" value=\"" + getStar() + "\" />");
	    buf.write("<property name=\"buildErrorCodes\" value=\"" + buildErrorCodes + "\" />");
	    buf.write("<property name=\"hl7ErrorCodes\" value=\"" + hl7ErrorCodes + "\" />");
	    buf.write("<property name=\"userErrorCodes\" value=\"" + userErrorCodes + "\" />");
	    buf.write("<property name=\"localErrorCodes\" value=\"" + localErrorCodes + "\" />");
	    buf.write("<property name=\"externalErrorCodes\" value=\"" + externalErrorCodes + "\" />");

	    nMsgVariant = 0;
	    for(int i=0;i<vMsgVariant.size();i++) {
		String s = (String)vMsgVariant.get(i);
		if(s.equals("Variant")) {
		    nMsgVariant++;
		    buf.write("<MessageVariant>");
		    // SegmentName
		    s = (String)vMsgVariant.get(++i);
		    if(!s.equals("")) {
			buf.write("<property name=\"segmentName\" value=\"" + s + "\"/>");
		    }
		    // FieldName
		    s = (String)vMsgVariant.get(++i);
		    if(!s.equals("")) {
			buf.write("<property name=\"fieldName\" value=\"" + s + "\"/>");
		    }
		    // ComponentName
		    s = (String)vMsgVariant.get(++i);
		    if(!s.equals("")) {
			buf.write("<property name=\"componentName\" value=\"" + s + "\"/>");
		    }
		    // SubComponentName
		    s = (String)vMsgVariant.get(++i);
		    if(!s.equals("")) {
			buf.write("<property name=\"subComponentName\" value=\"" + s + "\"/>");
		    }
		    // Options
		    boolean end = false;
		    while(!end && i+1 < vMsgVariant.size()) {
			s = (String)vMsgVariant.get(++i);
			if(s.equals("Variant")) {
			    end = true;
			    i--;
			}
			else {
			    buf.write("<property name=\"" + s + "\" value=\"" + vMsgVariant.get(++i) + "\"/>");
			}
		    }
		    buf.write("</MessageVariant>");
		}
	    }
	    buf.write("</MMmessageProperties>");
	    
	    buf.close();
/*	    if(typeMessageSet.equals("Manual")) {
		setVolume(nMsgVariant, true);
		}*/

	    // Debug
	    MMLog.log("<?xml version=\"1.0\" ?>"); 
	    MMLog.log("<MMmessageProperties HL7Version=\"2.4\" Profile=\"" + getProfile() + "\">");
	    MMLog.log("<MetaData Name=\"Example\" OrgName=\"NIST\" Version=\"1.4\" />");
	    MMLog.log("<property name=\"messageControlId\" value=\"MSG\" />"); 
	    MMLog.log("<property name=\"msgCount\" value=\"" + msgCount + "\" />"); 
	    MMLog.log("<property name=\"offset\" value=\"-1\" />"); 
	    MMLog.log("<property name=\"validity\" value=\"" + validity + "\" />"); 
	    MMLog.log("<property name=\"lengthConstraints\" value=\"" + lengthConstraints + "\" />"); 
	    MMLog.log("<property name=\"compositeLength\" value=\"false\" />"); 
	    MMLog.log("<property name=\"segmentCardinality\" value=\"" + segmentCardinality + "\" />"); 
	    MMLog.log("<property name=\"fieldCardinality\" value=\"" + fieldCardinality + "\" />"); 
	    MMLog.log("<property name=\"RE_Usage\" value=\"" + RE_Usage + "\" />"); 
	    MMLog.log("<property name=\"minSegCardError\" value=\"" + minSegCardError + "\" />"); 
	    MMLog.log("<property name=\"maxSegCardError\" value=\"" + maxSegCardError + "\" />"); 
	    MMLog.log("<property name=\"minFieldCardError\" value=\"" + minFieldCardError + "\" />"); 
	    MMLog.log("<property name=\"maxFieldCardError\" value=\"" + maxFieldCardError + "\" />"); 
	    MMLog.log("<property name=\"RusageError\" value=\"" + RusageError + "\" />"); 
	    MMLog.log("<property name=\"XusageError\" value=\"" + XusageError + "\" />"); 
	    MMLog.log("<property name=\"fieldLengthError\" value=\"" + fieldLengthError + "\" />"); 
	    MMLog.log("<property name=\"componentLengthError\" value=\"" + componentLengthError + "\" />"); 
	    MMLog.log("<property name=\"subComponentLengthError\" value=\"" + subComponentLengthError + "\" />"); 
	    MMLog.log("<property name=\"starValue\" value=\"" + getStar() + "\" />");
	    MMLog.log("<property name=\"buildErrorCodes\" value=\"" + buildErrorCodes + "\" />");
	    MMLog.log("<property name=\"hl7ErrorCodes\" value=\"" + hl7ErrorCodes + "\" />");
	    MMLog.log("<property name=\"userErrorCodes\" value=\"" + userErrorCodes + "\" />");
	    MMLog.log("<property name=\"localErrorCodes\" value=\"" + localErrorCodes + "\" />");
	    MMLog.log("<property name=\"externalErrorCodes\" value=\"" + externalErrorCodes + "\" />");
	    for(int i=0;i<vMsgVariant.size();i++) {
		String s = (String)vMsgVariant.get(i);
		if(s.equals("Variant")) {
		    MMLog.log("<MessageVariant>");
		    // SegmentName
		    s = (String)vMsgVariant.get(++i);
		    if(!s.equals("")) {
			MMLog.log("  <property name=\"segmentName\" value=\"" + s + "\"/>");
		    }
		    // FieldName
		    s = (String)vMsgVariant.get(++i);
		    if(!s.equals("")) {
			MMLog.log("  <property name=\"fieldName\" value=\"" + s + "\"/>");
		    }
		    // ComponentName
		    s = (String)vMsgVariant.get(++i);
		    if(!s.equals("")) {
			MMLog.log("  <property name=\"componentName\" value=\"" + s + "\"/>");
		    }
		    // SubComponentName
		    s = (String)vMsgVariant.get(++i);
		    if(!s.equals("")) {
			MMLog.log("  <property name=\"subComponentName\" value=\"" + s + "\"/>");
		    }
		    // Options
		    boolean end = false;
		    while(!end && i+1 < vMsgVariant.size()) {
			s = (String)vMsgVariant.get(++i);
			if(s.equals("Variant")) {
			    end = true;
			    i--;
			}
			else {
			    MMLog.log("  <property name=\"" + s + "\" value=\"" + vMsgVariant.get(++i) + "\"/>");
			}
		    }
		    MMLog.log("</MessageVariant>");
		}
	    }
	    MMLog.log("</MMmessageProperties>");

	    // Delete this file on exit
	    fOptions.deleteOnExit();

	    return fOptions.getPath();
	}
	catch(IOException ioe) {
	    System.err.println("IOException " + ioe.getMessage());
	}
	return null;
    }
    
    /**
     * Add a message variant
     * @param pn The Element on which the variation will be applied
     */
    private void addMsgVariant(ProfileNode pn) {
	vMsgVariant.add("Variant");
/*	if(pn instanceof SegmentGroupNode) {
	    //	    vMsgVariant.add(pn.getSegmentGroup());
vMsgVariant.add("");
vMsgVariant.add("");
vMsgVariant.add("");
vMsgVariant.add("");
	}*/
	if(pn instanceof SegmentNode) {
	    //	    vMsgVariant.add(pn.getSegmentGroup());
	    vMsgVariant.add(pn.getSegment());
	    vMsgVariant.add("");
	    vMsgVariant.add("");
	    vMsgVariant.add("");
	}
	else if(pn instanceof FieldNode) {
	    //	    vMsgVariant.add(pn.getSegmentGroup());
	    vMsgVariant.add(pn.getSegment());
	    vMsgVariant.add(pn.getName());
	    vMsgVariant.add("");
	    vMsgVariant.add("");
	}
	else if(pn instanceof ComponentNode) {
	    //	    vMsgVariant.add(pn.getSegmentGroup());
	    vMsgVariant.add(pn.getSegment());
	    vMsgVariant.add(pn.getQualifier());
	    vMsgVariant.add(pn.getName());
	    vMsgVariant.add("");
	}
	else if(pn instanceof SubComponentNode) {
	    //	    vMsgVariant.add(pn.getSegmentGroup());
	    vMsgVariant.add(pn.getSegment());
	    vMsgVariant.add(pn.getQualifier2());//.substring(0, idx));
	    vMsgVariant.add(pn.getQualifier());//.substring(idx+1, pn.getQualifier().length()-1));
	    vMsgVariant.add(pn.getName());
	}
    }

    /**
     * Count the number of message variants
     */
    public int countMsgVariants() {
	int nMsgVariant = 0;

	loadTest();
	//loadProfile(getProfile());
	// MessageVariant Usage Valid
	for(int i=0;i<vUsageTestValid.size();i++) {
	    ProfileNode pn = (ProfileNode)vUsageTestValid.get(i);
	    String usage = ramProfile.getUsage(pn);

	    if(usage.equals("RE")) {			
		nMsgVariant++;
	    }
	}
	// MessageVariant Usage Invalid
	for(int i=0;i<vUsageTestInvalid.size();i++) {
	    ProfileNode pn = (ProfileNode)vUsageTestInvalid.get(i);
	    String usage = ramProfile.getUsage(pn);
	    if(usage.equals("R") || usage.equals("X")) {
		nMsgVariant++;
	    }
	}
	// MessageVariant Cardinality Valid
	for(int i=0;i<vCardinalityTestValid.size();i++) {
	    ProfileNode pn = (ProfileNode)vCardinalityTestValid.get(i);
	    String usage = ramProfile.getUsage(pn);
	    //setCardinality(pn);
	    if(usage.equals("R") || usage.equals("RE")) {
		int min = Integer.parseInt(pn.getCardMin());
		int max = -1;
		if(!pn.getCardMax().equals("*"))
		    max = Integer.parseInt(pn.getCardMax());
		// MIN+1
		nMsgVariant++;
		// MAX or MIN + 3 if max=-1
		if(min + 1 != max)
		    nMsgVariant++;
		// (MIN+MAX)/2 if max!=-1
		if(max != -1) {
		    if((min+max)/2 > min+1 && (min+max)/2 != max) {
			nMsgVariant++;
		    }
		}
	    }
	}
	// MessageVariant Cardinality Invalid
	for(int i=0;i<vCardinalityTestInvalid.size();i++) {
	    ProfileNode pn = (ProfileNode)vCardinalityTestInvalid.get(i);
	    String usage = ramProfile.getUsage(pn);
	    //setCardinality(pn);
	    int min = Integer.parseInt(pn.getCardMin());
	    int max = -1;
	    if(!pn.getCardMax().equals("*"))
		max = Integer.parseInt(pn.getCardMax());
	    if(usage.equals("R")) {
		nMsgVariant++;
		if(max != -1) {
		    nMsgVariant++;
		}
	    }
	    else if(usage.equals("RE")) {
		if(max != -1) {
		    nMsgVariant++;
		}
	    }
	}
	// MessageVariant Error Codes Invalid
	for(int i=0;i<vErrorCodesTestInvalid.size();i++) {
	    nMsgVariant++;
	}
	return nMsgVariant;
    }

    /**
     * Count the number of temporary message variants
     */
    public int countTempMsgVariants() {
	int nMsgVariant = 0;

	loadTest();
	//loadProfile(getProfile());
	// MessageVariant Usage Valid
	for(int i=0;i<vUsageTempTestValid.size();i++) {
	    ProfileNode pn = (ProfileNode)vUsageTempTestValid.get(i);
	    String usage = ramProfile.getUsage(pn);
	    
	    if(usage.equals("RE")) {			
		nMsgVariant++;
	    }
	}
	// MessageVariant Usage Invalid
	for(int i=0;i<vUsageTempTestInvalid.size();i++) {
	    ProfileNode pn = (ProfileNode)vUsageTempTestInvalid.get(i);
	    String usage = ramProfile.getUsage(pn);
	    if(usage.equals("R") || usage.equals("X")) {
		nMsgVariant++;
	    }
	}
	// MessageVariant Cardinality Valid
	for(int i=0;i<vCardinalityTempTestValid.size();i++) {
	    ProfileNode pn = (ProfileNode)vCardinalityTempTestValid.get(i);
	    String usage = ramProfile.getUsage(pn);
	    //setCardinality(pn);
	    if(usage.equals("R") || usage.equals("RE")) {
		int min = Integer.parseInt(pn.getCardMin());
		int max = -1;
		if(!pn.getCardMax().equals("*"))
		    max = Integer.parseInt(pn.getCardMax());
		// MIN+1
		nMsgVariant++;
		// MAX or MIN + 3 if max=-1
		if(min + 1 != max)
		    nMsgVariant++;
		// (MIN+MAX)/2 if max!=-1
		if(max != -1) {
		    if((min+max)/2 > min+1 && (min+max)/2 != max) {
			nMsgVariant++;
		    }
		}
	    }
	}
	// MessageVariant Cardinality Invalid
	for(int i=0;i<vCardinalityTempTestInvalid.size();i++) {
	    ProfileNode pn = (ProfileNode)vCardinalityTempTestInvalid.get(i);
	    String usage = ramProfile.getUsage(pn);
	    //setCardinality(pn);
	    int min = Integer.parseInt(pn.getCardMin());
	    int max = -1;
	    if(!pn.getCardMax().equals("*"))
		max = Integer.parseInt(pn.getCardMax());
	    if(usage.equals("R")) {
		nMsgVariant++;
		if(max != -1) {
		    nMsgVariant++;
		}
	    }
	    else if(usage.equals("RE")) {
		if(max != -1) {
		    nMsgVariant++;
		}
	    }
	}
	// MessageVariant Error Codes Invalid
	for(int i=0;i<vErrorCodesTempTestInvalid.size();i++) {
	    nMsgVariant++;
	}
	return nMsgVariant;
    }

    /** 
     * Get the RAM Profile
     * @return A Profile object
     */
    public Profile getRAMProfile() {
	return ramProfile;
    }

    /**
     * Get the default value for missing data
     * @return The default values
     */
    public DefaultValue getDefaultValue() {
	return dv;
    }

    /**
     * Set the default value for missing data
     * @param dv The default value
     */
    public void setDefaultValue(DefaultValue dv) {
	this.dv = dv;
    }

    /**
     * Get the tableDataModel
     * @return The tableModelObject
     */
    public TableDataModel getTableDataModel() {
	if(tableModel == null)
	    tableModel = new TableDataModel(this);
	return tableModel;
    }

    /**
     * Get the ReferenceData
     * @return The refData object
     */
    private ReferenceData getReferenceData() {
	if(refData == null) {
	    refData = new ReferenceData();
	}
	return refData;
    }

    /**
     * Get the ConfigData
     * @return The configData object
     */
    private ConfigData getConfigData() {
	if(configData == null) {
	    configData = new ConfigData(getConfig());
	}
	return configData;
    }

    /**
     * Get the RepositoryDataModel
     * @return The repositoryModel object
     */
    public RepositoryDataModel getRepositoryDataModel() {
	return repositoryModel;
    }

    /**
     * Load a table
     * @param type The type
     * @param filename The filename of the table
     */
/*    public void loadTableData(String type, String filename) throws {
	tableModel.loadTableData(type, filename);
	}*/

    /**
     * Generate the repository
     * @param defaultValue Yes if include default value
     */
    public void generateRepository(boolean defaultValue) {
	try {
	    String repositoryFilename = new File(getProfile()).getParent() + File.separator + "Repository_" + new File(getProfile()).getName();
	    if(!getRepositoryData().equals(""))
		repositoryFilename = getRepositoryData();
	    else if(!defaultValue)
		ramProfile.generateRepository(getReferenceData(), getConfigData());
	    System.out.println("Load: " + repositoryFilename);
	    repositoryModel.loadRepository(repositoryFilename);
	    if(defaultValue) {
		Vector vNoValue = ramProfile.getNoValueProfileNode(repositoryModel);
		for(int i=0;i<vNoValue.size();i++) {
		    ProfileNode pn = (ProfileNode)vNoValue.get(i);
		    Vector vDV = dv.getDefaultValue(pn.getDatatype());
		    for(int j=0;j<vDV.size();j++) {
			DefaultValue.DVItem dvItem = (DefaultValue.DVItem)vDV.get(j);
			if(dvItem.getCategory().equals("DefaultValid"))
			    repositoryModel.addValue(pn, dvItem.getValue(), dvItem.getCategory()); 
		    }
		}
		//repositoryModel.clean();
		repositoryModel.saveToFile(new File(repositoryFilename));
	    }
	    else
		setRepositoryData(repositoryFilename);
	}
	catch(ParserConfigurationException pce) {
	    System.err.println("ParserConfigurationException: " + pce.getMessage());
	    pce.printStackTrace();
	}
	catch(SAXException saxe) {
	    System.err.println("SAXException: " + saxe.getMessage());
	    saxe.printStackTrace();
	}
	catch(IOException ioe) {
	    System.err.println("IOException: " + ioe.getMessage());
	    ioe.printStackTrace();
	}
	catch(TransformerConfigurationException tce) {
	    System.err.println("TransformerConfigurationException: " + tce.getMessage());
	    tce.printStackTrace();
	}
	catch(TransformerException te) {
	    System.err.println("TransformerException: " + te.getMessage());
	    te.printStackTrace();
	}
    }

    /**
     * Load the tables data
     */
    public void loadTables() {
	try {
	    // Table Data
	    tableModel = getTableDataModel();
	    tableModel.loadTableData("HL7", getHL7Tables());
	    tableModel.loadTableData("User", getUserTables());
	    tableModel.loadTableData("External", getExternalTables());
	    loadLocalTable();
	    for(int i=0;i<vLocalTable.size();i++) {
		TablePriority tp = (TablePriority)vLocalTable.get(i);
		tableModel.loadLocalTableData(tp.getFilename());
	    }
	}
	catch(ParserConfigurationException pce) {
	    System.err.println("ParserConfigurationException: " + pce.getMessage());
	    pce.printStackTrace();
	}
	catch(SAXException saxe) {
	    System.err.println("SAXException: " + saxe.getMessage());
	    saxe.printStackTrace();
	}
	catch(IOException ioe) {
	    System.err.println("IOException: " + ioe.getMessage());
	    ioe.printStackTrace();
	}
    }
}
