package gov.nist.hl7maker.message;

import java.io.*;
import java.util.*;

import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactoryConfigurationError;

import ca.uhn.hl7v2.HL7Exception;
import ca.uhn.hl7v2.model.Message;
import ca.uhn.hl7v2.conf.ProfileException;
import ca.uhn.hl7v2.conf.parser.ProfileParser;
import ca.uhn.hl7v2.conf.spec.message.StaticDef;
import ca.uhn.hl7v2.conf.spec.RuntimeProfile;
import ca.uhn.hl7v2.conf.check.DefaultValidator;

import org.w3c.dom.*;

import gov.nist.hl7maker.util.MMLog;
import gov.nist.hl7maker.util.Constants;
import gov.nist.hl7maker.util.ErrorCodes;

/**
  This class supports the automatic generation of HL7 messages.  The
  messages are generated based on a XML profile description of conformant 
  messages.  
  Methods are included for initiating and controlling the message
  generation process.  Messages are generated one at a time using
  the <i>gov.nist.hl7maker.message.XsltTransform</i> class. 
  <p>
  Two constructors provide a configurable means of initiating message
  generation.  One takes a String argument that names a file 
  containing a properties list as described in the <i>load</i> method 
  of the java.util.Properties class.  Using this constructor, the following
  code illustrates how message generation could be effected.
  <p>
  <i>
  MessageGenerator mg = new MessageGenerator("filenames.prop");
  <br>
  mg.generateMessages(false);
  <i>
  <p>
  The <i>false</i> argument to <i>generateMessages</i> turns
  off message validation.
  The String <i>filenames.prop</i> names a file that itself contains a 
  list of file names.  The file names are as described for  
  <i>MessageGenerator(Properties)</i>.  They identify the XSL transform
  files that are used to parse the profile and construct messages, the 
  XML input files that are used for configuration values 
  and message data values, the messages output directory and meta data file,
  and a file that holds the message generation options.  The options 
  file can take two different forms.  It can be formatted
  as a properties list as is <i>filenames.prop</i>.  Or, if the filename
  ends with ".xml", it is interpretted as an XML file.  The XML options
  file supports all options supported through the properties file, and, 
  in addition, supports the generation of what are designated
  "MessageVariants".  The message variants support control over specific
  elements of the output message.  A more detailed explanation is provided
  in the example file, msgGeneration.xml, located in the top level
  Message Maker directory.
  <p>
  The options supported through the use of a properties file
  are described in <i>setOptions(Properties)</i>.
  <p>
  A <i>MessageGeneration</i> object can also be instantiated using a
  java.util.Properties object.  The message generation process then
  can be effected in the manner shown in the sample code below.
  <p><i>
    Properties fileNames = new Properties();
    <br>
    fileNames.setProperty("preTransform",   "src\xslt\profilePreProcess.xsl");
    <br>
    fileNames.setProperty("preTransform",   "src\xslt\profile2msg.xsl");
    <br>
    fileNames.setProperty("xsltTransform", "profiles\NIST_EXAMPLE_PROFILE.xml");
    <br>
    fileNames.setProperty("xmlProfile",     "data\ConfigData.xml");
    <br>
    fileNames.setProperty("configTable",    "data<quote>\</quote>usertables.xml");
    <br>
    fileNames.setProperty("hl7Tables",      "data\externaltables.xml");
    <br>
    fileNames.setProperty("userTables",     "data\hl7tables.xml");
    <br>
    fileNames.setProperty("externalTables", "data\ReferenceData.xml");
    <br>
    fileNames.setProperty("referenceTable", "data\dataElements.xml");
    <br>
    fileNames.setProperty("dataElements",   "msgGeneration.properties");
    <br>
    fileNames.setProperty("optionsFile",    "MetaData.xml");
    <br>
    fileNames.setProperty("metaFile",       "messages\");
    <br>
    fileNames.setProperty("outputMsgDir",   "src\xslt\profilePreProcess.xsl");
    <br>
    MessageGenerator mg = new MessageGenerator(fileNames);
    <br>
    mg.generateMessages(false);
  </i>
  <p>
  The file names are shown in Windows format; they should 
  reflect the actual file names being used.
  <p><i>
  MessageGenerator mg = new MessageGenerator();
  <br>
  mg.generateMessages(false); 
  </i>
  
  @see #MessageGenerator(Properties) 
  @see  #setOptions(Properties)

  @author   Len Gebase, NIST
  */
 public class MessageGenerator {
/*  private static final String cName = "MessageGenerator";
  private final static boolean DEBUG = true;
*/  
  private static String mmVersion = "1.0";
  private static String hl7Version = "2.4";
  
  
  private final static String fileSeparator = Constants.FS;
  private final static String lineSeparator = Constants.LS;

  private static String optionsTransform = null;
  private static String preTransform = null;
  private static String xsltTransform = null;
  private static String xmlProfile = null;
  private static String configTable = null;
  private static String hl7Tables = null;
  private static String userTables = null;
  private static String localTables = null;
  private static String externalTables = null;    
  private static String errorCodesTables = null;    
  private static String referenceTable = null;
  private static String dataElements = null;
  private static String outputMsgDir = null;
  private static String outputMsgPath = null;

  private static String profile = null;   // content of the xml profile

  // profile properties
  private static int numberOfSegments = 0;
  private static int numberOfFields = 0;
  private static int numberOfComp = 0;
  private static int numberOfSubcomp = 0;
  private static Map segmentNames = null;
  private static Map fieldNames = null;
  private static Map compNames = null;
  private static Map subcompNames = null;
  private String sessionDescription = null;
  private static List segmentCardMinNums = null;
  private static List fieldCardMinNums = null;
  private static List rUsageSegments = null;
  private static List xUsageSegments = null;
  private static List rUsageFields = null;
  private static List xUsageFields = null;
  private static List rUsageComponents = null;
  private static List xUsageComponents = null;
  private static List rUsageSubcomps = null;
  private static List xUsageSubcomps = null;
  private static List primitiveFields = null;
  private static List primitiveComps = null;
  private static List primitiveSubcomps = null;

  private Properties errLocations = null;
  
  private String sessionName = null;
  private String orgName = null;
  private String userComment = "";

  private int totalMsgCount = 0;
  
  // available when generating msgs on a separate thread
  private boolean done = false;
  private static boolean stop = false;
  private int progressValue = 0; 
  private String statusMessage;

  private String optionsFile = null;
  private String metaFileName = null;

  private ProfileProperties profileProp = null;
  private ArrayList messageVariants = null;

  private GroupOptions options = null;
  private MessageOptions transformOptions = null;
  private MessageDescription messageDescription = null;
  private Properties fileNames = null;
  private Vector messageFileNames = null;

  private boolean errorCondition = false;

  /**
   * This class supports managing the parameter settings for each 
   * message generated. 
   */
  private class MessageGroupParams {
    final static String cName = "MessageParams";
    
    private GroupOptions defaultOptions = null;
    private GroupOptions groupOptions = null;
    private MessageOptions messageOptions = null;
    private int msgNumber = 0;

    // msg parameters
    private int messageControlId = 0;
    private int msgCount = 0;
    private int offset = 0;
    private String messageControlPrefix = ""; 
    private boolean minSegCardError = false;
    private boolean maxSegCardError = false;
    private boolean minFieldCardError = false;
    private boolean maxFieldCardError = false;
    private boolean RusageError = false;
    private boolean XusageError = false;
    private boolean fieldLengthError = false;
    private boolean componentLengthError = false;
    private boolean subComponentLengthError = false;

    //  count of how many of each type of err msg we'll generate
    private int minSegCardErrCount = 0;
    private int maxSegCardErrCount = 0;
    private int minFieldCardErrCount = 0;
    private int maxFieldCardErrCount = 0;
    private int RusageErrCount = 0;
    private int XusageErrCount = 0;
    private int fieldLengthErrCount = 0;
    private int compLengthErrCount = 0;
    private int subCompLengthErrCount = 0;
    private int errMsgTypes = 0;
    //  number of msgs to generate for each type of error
    private int errMsgsEachCount = 0; 
    
    /* 
     * Creates a new object and sets default and the initial group options.  
     * groupOptions is reset for each msg variant.  options not 
     * specified in the variant are taken from defaultOptions.
     */
    MessageGroupParams (GroupOptions opts) {
      /* set default options for initial group and all variants */
      defaultOptions = new GroupOptions(opts);  
      
      /* the initial set of group options are determined by the defaults */
      groupOptions = new GroupOptions(defaultOptions); 

      //initMessageOptions();
      initMessageControlId();
      //offset = Integer.parseInt(groupOptions.getOption("offset"));
      //initialOffset = offset;
      //setOffset();
      //setProfileProperties();
      
      /* set msg group params */
      setParams();
    }

    /*  
     * Sets group options for this set of Variant Messages.
     *  p should contain the Variant properties.
     *  all properties are optional in a MessageVariant, so any
     *  property not present is set to a default.
     */  
    void setVariantOptions (Properties p) {
      // initialize this group's options to defaultOptions  
      groupOptions = new GroupOptions(defaultOptions); 

      // variant only options, init to empty
      groupOptions.setOption("cardinality", "");
      groupOptions.setOption("REinclude", "");
      groupOptions.setOption("segmentName", "");
      groupOptions.setOption("fieldName", ""); 
      groupOptions.setOption("componentName", ""); 
      groupOptions.setOption("subComponentName", "");
        
      // the msgCount defaults to 1 
      groupOptions.setOption("msgCount", "1");
      
      // offset value defaults to -1
      groupOptions.setOption("offset", "-1");

      // validity defaults to VALID
      groupOptions.setOption("validity", "VALID");
      
      /* all error settings default to false */
      groupOptions.setOption("minSegCardError", "false");
      groupOptions.setOption("maxSegCardError", "false");
      groupOptions.setOption("minFieldCardError", "false");
      groupOptions.setOption("maxFieldCardError", "false");
      groupOptions.setOption("RusageError", "false");
      groupOptions.setOption("XusageError", "false");
      groupOptions.setOption("fieldLengthError", "false");
      groupOptions.setOption("componentLengthError", "false");
      groupOptions.setOption("subComponentLengthError", "false");
      
      // set the variant options: if the option is present in p, the
      //  default is overridden. 
      //  messageControlId and offset have to be checked before being reset.
      Enumeration e = p.propertyNames();
      while (e.hasMoreElements()) {
        String key = (String)e.nextElement();
        
        groupOptions.setOption(key, p.getProperty(key));

        /* msg cntl id & offset have to be checked before being set */
        if ("messageControlId".equalsIgnoreCase(key)) {
          initMessageControlId();
        }
        //if ("offset".equalsIgnoreCase(key)) {
          //setOffset();
        //}
      }

      /* init msg options for this group of variant msgs */
      //initMessageOptions();
      
      /* update params settings */
      setParams();
    }

    /*
     * set options for this message
     */
    void setMessageOptions (String valid, int mOffset) {
      initMessageOptions();
      
      messageOptions.setOption("valid", valid);
      messageOptions.setOption("offset", String.valueOf(mOffset));
      messageOptions.setOption("messageControlId", 
          messageControlPrefix +String.valueOf(messageControlId));

      /* init all err settings to false (only 1 error per msg) */
      messageOptions.setOption("minSegCardError", "false");
      messageOptions.setOption("maxSegCardError", "false");
      messageOptions.setOption("minFieldCardError", "false");
      messageOptions.setOption("maxFieldCardError", "false");
      messageOptions.setOption("RusageError", "false");
      messageOptions.setOption("XusageError", "false");
      messageOptions.setOption("fieldLengthError", "false");
      messageOptions.setOption("componentLengthError", "false");
      messageOptions.setOption("subComponentLengthError", "false");

      initErrorLocations();

      if ("false".equalsIgnoreCase(valid)) {
        /* set error option */ 
        updateErrorSetting();
        
        /* if an error element hasn't been selected, we select one randomly */
        if (!errorElementSet())
          setAnErrorElement();
      }
    }
    
    /* 
     * Set group params based on the group options settings 
     */
    void setParams () {
      offset = Integer.parseInt(groupOptions.getProperty("offset"));
      //  number of msgs to generate for this set of msgs
      msgCount = Integer.parseInt(
          groupOptions.getProperty("msgCount"));
      
      minSegCardError = new Boolean(
          groupOptions.getOption("minSegCardError")).booleanValue();
      maxSegCardError = new Boolean(
          groupOptions.getOption("maxSegCardError")).booleanValue();
      minFieldCardError = new Boolean(
          groupOptions.getOption("minFieldCardError")).booleanValue();
      maxFieldCardError = new Boolean(
          groupOptions.getOption("maxFieldCardError")).booleanValue();
      RusageError = new Boolean(
          groupOptions.getOption("RusageError")).booleanValue();
      XusageError = new Boolean(
          groupOptions.getOption("XusageError")).booleanValue();
      fieldLengthError = new Boolean(
          groupOptions.getOption("fieldLengthError")).booleanValue();
      componentLengthError = new Boolean(
          groupOptions.getOption("componentLengthError")).booleanValue();
      subComponentLengthError = new Boolean(
          groupOptions.getOption("subComponentLengthError")).booleanValue();

      /* when min usage is 0, min card errors aren't possible and we
       * reset to max error.
       */
      if (minSegCardError) {
        if (!existsMinCardSeg()) maxSegCardError = true;
      }
      if (minFieldCardError) {
        if (!existsMinCardField()) maxFieldCardError = true;
      }
      
      //  count of how many of each type of err msg we'll generate
      minSegCardErrCount = 0;
      maxSegCardErrCount = 0;
      minFieldCardErrCount = 0;
      maxFieldCardErrCount = 0;
      RusageErrCount = 0;
      XusageErrCount = 0;
      fieldLengthErrCount = 0;
      compLengthErrCount = 0;
      subCompLengthErrCount = 0;

      //  count how many types of error msgs to generate.
      //  if we're not generating an error type, set its counter
      //  above the number of msgs to be generated, so no msgs 
      //  of the type get generated.  
      errMsgTypes = 0;
      if (minSegCardError) errMsgTypes++;
      else minSegCardErrCount = msgCount + 1;
      if (maxSegCardError) errMsgTypes++;
      else maxSegCardErrCount = msgCount + 1;
      if (minFieldCardError) errMsgTypes++;
      else minFieldCardErrCount = msgCount + 1;
      if (maxFieldCardError) errMsgTypes++;
      else maxFieldCardErrCount = msgCount + 1;
      if (RusageError) errMsgTypes++;
      else RusageErrCount = msgCount + 1;
      if (XusageError) errMsgTypes++;
      else XusageErrCount = msgCount + 1;
      if (fieldLengthError) errMsgTypes++;
      else fieldLengthErrCount = msgCount + 1;
      if (componentLengthError) errMsgTypes++;
      else compLengthErrCount = msgCount + 1;
      if (subComponentLengthError) errMsgTypes++;
      else subCompLengthErrCount = msgCount + 1;
      
      String validity = groupOptions.getOption("validity");
      if (errMsgTypes > 0 && "VALID".equalsIgnoreCase(validity)) {
        MMLog.error( 
            "Invalid variantOptions settings: 'validity=VALID' and error option set!");
        errorCondition = true;
        return;
      }

      if (errMsgTypes > 0) {
        //  number of msgs to generate for each type of error.
        //  we round down -- we'll generate extra msgs if we need to
        errMsgsEachCount = msgCount/errMsgTypes;
        
        // reset count if only every 3rd msg is going to have an error
        if ("VALID_AND_INVALID".equalsIgnoreCase(validity)) {
          errMsgsEachCount = errMsgsEachCount/3;
        }
      }
    }
    
    void updateErrorSetting () {
      if (minSegCardErrCount++ < errMsgsEachCount) {
        messageOptions.setOption("minSegCardError", "true");
      }
      else if (maxSegCardErrCount++ < errMsgsEachCount) {
        messageOptions.setOption("maxSegCardError", "true");
      }
      else if (minFieldCardErrCount++ < errMsgsEachCount) {
        messageOptions.setOption("minFieldCardError", "true");
      }
      else if (maxFieldCardErrCount++ < errMsgsEachCount) {
        messageOptions.setOption("maxFieldCardError", "true");
      }
      else if (RusageErrCount++ < errMsgsEachCount) {
        messageOptions.setOption("RusageError", "true");
      }
      else if (XusageErrCount++ < errMsgsEachCount) {
        messageOptions.setOption("XusageError", "true");
      }
      else if (fieldLengthErrCount++ < errMsgsEachCount) {
        messageOptions.setOption("fieldLengthError", "true");
      }
      else if (compLengthErrCount++ < errMsgsEachCount) {
        messageOptions.setOption("componentLengthError", "true");
      }
      else if (subCompLengthErrCount++ < errMsgsEachCount) {
        messageOptions.setOption("subComponentLengthError", "true");
      }
      else {  /*  an extra msg may need to be generated because
                  errMsgsEachCount was rounded down */
        if (minSegCardError) {
          messageOptions.setOption("minSegCardError", "true");
          minSegCardError = false;    // done
        }
        else if (maxSegCardError) { 
          messageOptions.setOption("maxSegCardError", "true");
          maxSegCardError = false;
        }
        else if (minFieldCardError) { 
          messageOptions.setOption("minFieldCardError", "true");
          minFieldCardError = false;
        }
        else if (maxFieldCardError) { 
          messageOptions.setOption("maxFieldCardError", "true");
          maxFieldCardError = false;
        }
        else if (RusageError) { 
          messageOptions.setOption("RusageError", "true");
          RusageError = false;
        }
        else if (XusageError) { 
          messageOptions.setOption("XusageError", "true");
          XusageError = false;
        }
        else if (fieldLengthError) { 
          messageOptions.setOption("fieldLengthError", "true");
          fieldLengthError = false;
        }
        else if (componentLengthError) { 
          messageOptions.setOption("componentLengthError", "true");
          componentLengthError = false;
        }
        else if (subComponentLengthError) { 
          messageOptions.setOption("subComponentLengthError", "true");
          subComponentLengthError = false;
        }
      }
    }
    
    /* init message options based on group settings */
    void initMessageOptions () {
      Properties msgProperties = new Properties(groupOptions);
      msgProperties.setProperty("valid", "VALID");    // default
      messageOptions = new MessageOptions(msgProperties); 
    }

    void initMessageControlId () {
      //  messageControlId may not be a number but a string prefix
      //    to a number!
      messageControlPrefix = groupOptions.getOption("messageControlId"); 
      try {
        messageControlId = Integer.parseInt(messageControlPrefix);
        messageControlPrefix = "";
      }
      catch (NumberFormatException e) {
        messageControlPrefix += ".";
        messageControlId = 0;
      }
    }
    
    void setOffset () {
      // we recompute the offset unless it's set
      //  to a non-negative value (enables replicating a previous run)
      if (offset < 0) 
        offset = (int)(1000*Math.random() % 1000);
    }
  
    /*
    void setProfileProperties () {
      segmentNames = profileProp.getSegmentNames();
      fieldNames = profileProp.getFieldNames();
      compNames = profileProp.getComponentNames();
      subcompNames = profileProp.getSubcomponentNames();
      numberOfFields     = profileProp.getNumberOfFields();
      numberOfSegments   = profileProp.getNumberOfSegments();
      numberOfComp       = profileProp.getNumberOfComp();
      numberOfSubcomp    = profileProp.getNumberOfSubcomp();
      segmentCardMinNums = profileProp.getMinSegmentCardList();
      fieldCardMinNums   = profileProp.getMinFieldCardList();
      rUsageSegments     = profileProp.getRUsageSegments(); 
      xUsageSegments     = profileProp.getXUsageSegments();
      rUsageFields       = profileProp.getRUsageFields();
      xUsageFields       = profileProp.getXUsageFields();
      rUsageComponents   = profileProp.getRUsageComponents();
      xUsageComponents   = profileProp.getXUsageComponents();
      rUsageSubcomps     = profileProp.getRUsageSubcomps();
      xUsageSubcomps     = profileProp.getXUsageSubcomps();
      primitiveFields    = profileProp.getPrimitiveFields();
      primitiveComps     = profileProp.getPrimitiveComps();
      primitiveSubcomps  = profileProp.getPrimitiveSubcomps();
    }
*/
    void initErrorLocations () {
      errLocations = new Properties();
      
      errLocations.setProperty("rUSegment", "0");
      errLocations.setProperty("rUField", "0");
      errLocations.setProperty("rUComponent", "0");
      errLocations.setProperty("rUSubcomp", "0");
      errLocations.setProperty("xUField", "0");
      errLocations.setProperty("xUSegment", "0");
      errLocations.setProperty("xUComponent", "0");
      errLocations.setProperty("xUSubcomp", "0");
      errLocations.setProperty("cardErrorMinSeg", "0");
      errLocations.setProperty("cardErrorMinField", "0");
      errLocations.setProperty("cardErrorMaxSeg", "0");
      errLocations.setProperty("cardErrorMaxField", "0");
      errLocations.setProperty("fieldLengthError", "0");
      errLocations.setProperty("compLengthError", "0");
      errLocations.setProperty("subLengthError", "0");
    }

    boolean existsMinCardSeg () {
      if ((profileProp.getMinSegmentCardList()).isEmpty()) return false;
      else return true;
    }

    boolean existsMinCardField () {
      if ((profileProp.getMinFieldCardList()).isEmpty()) return false;
      else return true;
    }

    boolean errorElementSet () {
      if ("".equals(messageOptions.getOption("segmentName")))
        return false;
      else 
        return true;
    }
    
    void setAnErrorElement () {
      String fieldLengthErr = messageOptions.getOption("fieldLengthError");
      String compLengthErr = messageOptions.getOption("componentLengthError");
      String subCompLengthErr = 
          messageOptions.getOption("subComponentLengthError");
      String minSegCardErr = messageOptions.getOption("minSegCardError");
      String maxSegCardErr = messageOptions.getOption("maxSegCardError");
      String minFieldCardErr = messageOptions.getOption("minFieldCardError");
      String maxFieldCardErr = messageOptions.getOption("maxFieldCardError");
      String RusageErr = messageOptions.getOption("RusageError");
      String XusageErr = messageOptions.getOption("XusageError");
      
      String rUSegment = "0";
      String rUField   = "0";
      String rUComponent = "0";
      String rUSubcomp = "0";
/*      String reUSegment = "0";
      String reUField  = "0";
      String reUComponent = "0";   
      String reUSubcomp = "0";
*/      
      String xUField   = "0";
      String xUSegment = "0";
      String xUComponent = "0";
      String xUSubcomp = "0";
      String cardErrorMinSeg = "0"; 
      String cardErrorMinField = "0";
      String cardErrorMaxSeg = "0"; 
      String cardErrorMaxField = "0";
      String fieldLengthError = "0";
      String compLengthError = "0";
      String subLengthError = "0";
     
      if ("true".equalsIgnoreCase(minSegCardErr)) {
        cardErrorMinSeg = getaNodeNumber(segmentCardMinNums);
        setMsgOptionsFullSegmentName(cardErrorMinSeg);
      }
      if ("true".equalsIgnoreCase(minFieldCardErr)) {
        cardErrorMinField =
            getaNodeNumber(fieldCardMinNums);
        setMsgOptionsFullFieldName(cardErrorMinField);
      }
      // use required elements for max cardinality errs 
      if ("true".equalsIgnoreCase(maxSegCardErr)) {
        cardErrorMaxSeg = getaNodeNumber(rUsageSegments);
        setMsgOptionsFullSegmentName(cardErrorMaxSeg);
      }
      if ("true".equalsIgnoreCase(maxFieldCardErr)) {
        cardErrorMaxField = getaNodeNumber(rUsageFields);
        setMsgOptionsFullFieldName(cardErrorMaxField);
      }
      
      // when Usage errors are being generated and offset < 0, 
      //	we'll randomly select between segments, fields, comp, 
      //	and subcomp for inserting the error,
      //	otherwise we're (probably) trying to replicate a previous
      int selector = 0;
      if (offset < 0)
    	  selector = (int)(10*Math.random() % 6);
      else
    	  selector = offset % 6;
  
      // set R-Usage error. (we're favoring comp & subcomp)
      //  (there has to be at least 1 required segment in the profile, same 
      //  for Field, but no components or subcomponents is possible (in theory
      //  anyway), so we check what's returned in the latter 2 cases and look
      //  elsewhere if we don't locate an element)
      if ("true".equalsIgnoreCase(RusageErr)) {
        if (selector == 4 || selector == 5) {  // subcomponent
          rUSubcomp = getaNodeNumber(rUsageSubcomps);
          if ("0".equals(rUSubcomp))  
            selector = 2;   // try component
          else {
            setMsgOptionsFullSubName(rUSubcomp);
          }
        }
        if (selector == 2 || selector == 3) {  // component     
          rUComponent = getaNodeNumber(rUsageComponents);
          if ("0".equals(rUComponent))
            selector = 1;   // get a Field
          else {
            setMsgOptionsFullCompName(rUComponent);
          }
        }
        if (selector == 1) {  // Field
          rUField   = getaNodeNumber(rUsageFields);
          setMsgOptionsFullFieldName(rUField);
        }
        if (selector == 0) {  // Segment 
          rUSegment = getaNodeNumber(rUsageSegments);
          setMsgOptionsFullSegmentName(rUSegment);
        }
      }
      // set X-Usage error.
      //  (nothing has to have usage marked X, so if we don't find a 
      //  location we're looking for, we try everything before giving up)
      if ("true".equalsIgnoreCase(XusageErr)) {
        for (int i = 0; i < 4; i++) {
          if (selector == 0) {  // segment 
            xUSegment = getaNodeNumber(xUsageSegments);
            if ("0".equals(xUSegment))
              selector = 1;   // try field
            else {
              setMsgOptionsFullSegmentName(xUSegment);
              break;
            }
          }
          else if (selector == 1) { // field
            xUField   = getaNodeNumber(xUsageFields);
            if ("0".equals(xUField))  
              selector = 2;   // try component
            else {
              setMsgOptionsFullFieldName(xUField);
              break;
            }
          }
          else if (selector == 2 || selector == 3) { // comp
             xUComponent = getaNodeNumber(xUsageComponents);
            if ("0".equals(xUComponent))  
              selector = 4;   // try subcomp
            else {
              setMsgOptionsFullCompName(xUComponent);
              break;
            }
          }
          else if (selector == 4 || selector == 5) {   // subcomp
             xUSubcomp = getaNodeNumber(xUsageSubcomps);
            if ("0".equals(xUSegment))
              selector = 0;   // try segment
            else {
              setMsgOptionsFullSubName(xUSubcomp);
              break;
            }
          }
        }
      }
  
      // use required elements that are primitive for length errs
      if ("true".equalsIgnoreCase(fieldLengthErr)) {
        fieldLengthError = getaNodeNumber(primitiveFields);
        setMsgOptionsFullFieldName(fieldLengthError);
      }
      if ("true".equalsIgnoreCase(compLengthErr)) {
        compLengthError = getaNodeNumber(primitiveComps);
        setMsgOptionsFullCompName(compLengthError);
      }
      if ("true".equalsIgnoreCase(subCompLengthErr)) {
        subLengthError = 
            getaNodeNumber(primitiveSubcomps);
        setMsgOptionsFullSubName(subLengthError);
      }
    
      errLocations.setProperty("rUSegment", rUSegment);
      errLocations.setProperty("rUField", rUField);
      errLocations.setProperty("rUComponent", rUComponent);
      errLocations.setProperty("rUSubcomp", rUSubcomp);
      errLocations.setProperty("xUField", xUField);
      errLocations.setProperty("xUSegment", xUSegment);
      errLocations.setProperty("xUComponent", xUComponent);
      errLocations.setProperty("xUSubcomp", xUSubcomp);
      errLocations.setProperty("cardErrorMinSeg", cardErrorMinSeg);
      errLocations.setProperty("cardErrorMinField", cardErrorMinField);
      errLocations.setProperty("cardErrorMaxSeg", cardErrorMaxSeg);
      errLocations.setProperty("cardErrorMaxField", cardErrorMaxField);
      errLocations.setProperty("fieldLengthError", fieldLengthError);
      errLocations.setProperty("compLengthError", compLengthError);
      errLocations.setProperty("subLengthError", subLengthError);

      /*
System.out.println("rUSegment, name = " +rUSegment +profileProp.getSegmentName(rUSegment));
System.out.println("rUField, name = " +rUField +profileProp.getFieldName(rUField));
System.out.println("rUComponent, name = " +rUComponent +profileProp.getComponentName(rUComponent));
//System.out.println("rUSubcomp, name = " +rUSubcomp +profileProp.getSubName(rUSubcomp));
System.out.println("xUField, name = " +xUField +profileProp.getFieldName(xUField));
System.out.println("xUSegment, name = " +xUSegment +profileProp.getSegmentName(xUSegment));
System.out.println("xUComponent, name = " +xUComponent +profileProp.getComponentName(xUComponent));
//System.out.println("xUSubcomp, name = " +xUSubcomp +profileProp.getSubName(xUSubcomp));
System.out.println("cardErrorMinSeg, name = " +cardErrorMinSeg +profileProp.getSegmentName(cardErrorMinSeg));
System.out.println("cardErrorMinField, name = " +cardErrorMinField +profileProp.getFieldName(cardErrorMinField));
System.out.println("cardErrorMaxSeg, name = " +cardErrorMaxSeg +profileProp.getSegmentName(cardErrorMaxSeg));
System.out.println("cardErrorMaxField, name = " +cardErrorMaxField +profileProp.getFieldName(cardErrorMaxField));
*/
    }
    
    void setMsgOptionsFullSegmentName (String nodeNumber) {
      List qname = profileProp.getSegmentName(nodeNumber);
      String segName = "";
      try {
        segName = (String)qname.get(0);
        messageOptions.setOption("segmentName", segName);
      }
      catch (NullPointerException e) {
        e.printStackTrace();
        MMLog.error("Segment Name = " + segName +" Exception: " +e);
      }
    }
    
    void setMsgOptionsFullFieldName (String nodeNumber) {
      List qname = profileProp.getFieldName(nodeNumber);
      String segName = "";
      String fieldName = "";
      try {
        segName = (String)qname.get(0);
        fieldName = (String)qname.get(1);
        messageOptions.setOption("segmentName", segName);
        messageOptions.setOption("fieldName", fieldName);
      }
      catch (NullPointerException e) {
        e.printStackTrace();
        MMLog.error("Segment Name = " + segName 
            +"Field Name = " +fieldName
            +" Exception: " +e);
      }
    }
    
    void setMsgOptionsFullCompName (String nodeNumber) {
      List qname = profileProp.getComponentName(nodeNumber);
      String segName = "";
      String fieldName = "";
      String compName = "";
      try {
        segName = (String)qname.get(0);
        fieldName = (String)qname.get(1);
        compName = (String)qname.get(2);
        messageOptions.setOption("segmentName", segName);
        messageOptions.setOption("fieldName", fieldName);
        messageOptions.setOption("componentName", compName);
      }
      catch (NullPointerException e) {
        e.printStackTrace();
        MMLog.error("Segment Name = " + segName 
            +"Field Name = " +fieldName
            +"Component Name = " +compName
            +" Exception: " +e);
      }
    }
    
    void setMsgOptionsFullSubName (String nodeNumber) {
      List qname = profileProp.getSubComponentName(nodeNumber);
      String segName = "";
      String fieldName = "";
      String compName = "";
      String subName = "";
      try {
        segName = (String)qname.get(0);
        fieldName = (String)qname.get(1);
        compName = (String)qname.get(2);
        subName = (String)qname.get(3);
        messageOptions.setOption("segmentName", segName);
        messageOptions.setOption("fieldName", fieldName);
        messageOptions.setOption("componentName", compName);
        messageOptions.setOption("subComponentName", subName);
      }
      catch (NullPointerException e) {
        e.printStackTrace();
        MMLog.error("Segment Name = " + segName 
            +"Field Name = " +fieldName
            +"Component Name = " +compName
            +"SubComponent Name = " +subName
            +" Exception: " +e);
      }
    }
    
    //  return a random (if offset < 0) node number from the list of numbers,
    //	 or 0 if the list is empty or if optionSetting isn't negative.
    String getaNodeNumber (List nodeNumbers)
    {
      if (nodeNumbers.isEmpty())
        return String.valueOf(0);
              
      int index = 0;
      int size = nodeNumbers.size();
      // we don't want to include elements from the MSH segment 
      //  (number starts with '1' or '1.'). 
      String number = "";
      //  if the profile has only the MSH Segment, we'll loop forever
      //  without the counter
      int attempts = 0;
      for (;attempts < 1000; attempts++) {
    	  // random index or we want to enable replicating a previous run 
    	  if (offset < 0) {
    		  index = (int)(100*Math.random() % size);
    	  }
    	  else {
    		  index = (offset + attempts) % size;
    	  }
         //number = (String)nodeNumbers.get(index = (int)(100*Math.random() % size));
     	 number = (String)nodeNumbers.get(index);
        if (number.length() == 1) {
          if (number.charAt(0) != '1')
            break;
        }
        else if (number.charAt(0) != '1' || number.charAt(1) != '.')
            break;
      }                          

      if (attempts == 1000) {
        number = "0";
        MMLog.error("failed to get node number");
      }
      return number;
    }
  }
  
  /** 
    Creates a MessageGenerator object.
    
    @param files  The name of a file containing a list of properties as 
                  described in the <i>load</i> method of the 
                  java.util.Properties class. The property names and meanings
                  are exactly those described in 
                  <i>MessageGenerator(Properties)</i>.
    @see  #MessageGenerator(Properties)               
    */
  public MessageGenerator (String files) {
    /* get filenames from the named file */
    if ((fileNames  = getProperties(files)) != null) {
      init();      
    }
    else {
      MMLog.error("bad file name?");
      errorCondition = true;
    }
  }

  /**   
    Creates a MessageGenerator object.
    
    @param files  A Properties object with property names and their 
                  associated meanings as described below.
    <p>              
    <br>optionsTransform -     <i>name of xsl file for parsing XML options file</i>
    <br>preTransform     -     <i>name of xsl preProcessing transform file</i>
    <br>xsltTransform    -     <i>name of xsl message transform file        </i>
    <br>xmlProfile       -     <i>name of xml profile file          </i>
    <br>configTable      -     <i>name of xml config (MSH Segment) data file</i>
    <br>hl7Tables        -     <i>name of xml hl7Tables data file  </i>
    <br>localTables      -     <i>name of xml local data file  </i>
    <br>userTables       -     <i>name of xml userTables data file </i>
    <br>externalTables   -     <i>name of xml externalTables data file</i>
    <br>errorCodes       -     <i>name of xml error codes data file</i>
    <br>referenceTable   -     <i>name of xml referenceTable file</i>
    <br>dataElements     -     <i>name of xml dataElements file</i>
    <br>optionsFile      -     <i>name of MessageGenerator options file</i> 
    <br>metaFile         -     <i>name of xml output matadata file</i>
    <br>outputMsgDir     -     <i>name of directory where output messages and 
                                the metadata file are to be placed</i>
    <p>
    The options file contains a list of properties as 
    described in the <i>load</i> method of the 
    java.util.Properties class, or, if the file name ends with ".xml",
    it is interpreted as an XML file (see above). 
    The property names are exactly those
    listed in <i>setOptions(Properties)</i>.
      
    @see  #setOptions(Properties)
    */  
  public MessageGenerator (Properties files) {
    if (files != null) {
      // save file & dir names
      fileNames = files;

      init();
    }
    else {
      errorCondition = true;
    }
  }

  private void init () {
    // set each file & directory name
    setFileNames(fileNames);
    outputMsgDir = outputMsgDir + fileSeparator;
    File d = new File(outputMsgDir);
    outputMsgPath = d.getAbsolutePath();

    // read-in options from xml or properties file
    if (isXMLFile(optionsFile)) {
      XMLOptionsReader reader = null;

      // options & msg variants
      if ((reader = readOptionsXML(optionsFile)) == null) {
    	statusMessage += lineSeparator +"**ERROR** "
          	+"errors present, can't generate messages!";
        errorCondition = true;
        return;
      }

      // set options
      options = new GroupOptions(reader.getOptions());
      messageVariants = reader.getMessageVariants();
      
      totalMsgCount = reader.getMessageCount();
    }
    else { 
      options = readOptionsProp(optionsFile);
      totalMsgCount = Integer.parseInt(options.getOption("msgCount"));
      messageVariants = new ArrayList();    // empty list
    }
    
    /* update error codes if option set */
    if ("true".equalsIgnoreCase(options.getOption("buildErrorCodes"))) {
	    ErrorCodes tables = new ErrorCodes(errorCodesTables);

	    if ("true".equalsIgnoreCase(options.getOption("hl7ErrorCodes"))) 
	    	tables.addErrorCodes(hl7Tables);
	    if ("true".equalsIgnoreCase(options.getOption("userErrorCodes"))) 
	    	tables.addErrorCodes(userTables);
	    if ("true".equalsIgnoreCase(options.getOption("localErrorCodes"))) 
	    	tables.addErrorCodes(localTables);
	    if ("true".equalsIgnoreCase(options.getOption("externalErrorCodes"))) 
	    	tables.addErrorCodes(externalTables);
	   
	    /* codes aren't saved till this is called */
	    tables.saveErrorCodes();
    }

    /*	profile data has to be collected before msgs can be generated */
    try {
      profileProp = new ProfileProperties(preTransform, xmlProfile);
      setProfileProperties();
    } 
    catch (TransformerFactoryConfigurationError e) {
      e.printStackTrace();
      MMLog.error("Exception: " +e);
      errorCondition = true;
    } 
    catch (TransformerException e) {
      e.printStackTrace();
      MMLog.error("Exception: " +e);
      errorCondition = true;
    } 
    catch (IOException e) {
      e.printStackTrace();
      MMLog.error("Exception: " +e);
      errorCondition = true;
    } 
    catch (Exception e) {
      e.printStackTrace();
      MMLog.error("Exception: " +e);
      errorCondition = true;
    }
    
    sessionName = profileProp.getSessionName();
    orgName = profileProp.getOrgName();
  }
  
  /**
    A convenience method for generating messages without a Runnable object.
    
    @param validate  if true, generated messages are validated against 
            the xml profile, provided the messages weren't created with an
            error purposely included.
    @see  #generateMessages(boolean, Runnable)
    */
  public void generateMessages (boolean validate) {
    generateMessages(validate, null);
  }
  
  /**
    Generates a set of messages based on the XML profile and the option 
    settings.
    Messages are saved in the directory named by <i>outputMsgDir</i>.  
    The name of each saved message is determined by the String <i>xmlProfile</i>.

    @param validate  if true, generated messages are validated against 
            the xml profile, provided the messages weren't created with an
            error purposely included.
    @param  runner  A Runnable object that can be used to gain some control
              and monitoring of the message generation process, particularly
              from AWT or Swing GUI interface.
              
    @see  #MessageGenerator(Properties)              
    */
  public void generateMessages (boolean validate, Runnable runner) {
    if (errorCondition) {
        statusMessage += lineSeparator +"**ERROR** "
        +"errors present, can't generate messages!";
      MMLog.error("errors present, can't generate messages!");
      return;      
    }
    
    /* we save each msg file name */
    messageFileNames = new Vector();
    /* create a new object for managing each msgs's parameter settings */
    MessageGroupParams msgGroupParams = new MessageGroupParams(options);
    /* create a new object for describing each msg generated */
    messageDescription = new MessageDescription();
    messageDescription.startOfSession(options,
        outputMsgDir,
        metaFileName,
        mmVersion,
        hl7Version,
        sessionName,
        orgName,
        userComment,
        xmlProfile,
        totalMsgCount);

    /* our interface to the xsl transform */
    XsltTransform transformer = null; 
    try {
      transformer = new XsltTransform(xsltTransform, 
                            xmlProfile,
                            configTable,   
                            hl7Tables,
                            userTables,
                            localTables,
                            externalTables,
                            referenceTable,
                            dataElements,
                            errorCodesTables);
    }
    catch (TransformerFactoryConfigurationError e) {
      MMLog.error("Exception: " +e);
      e.printStackTrace();
      return;
    }
    catch (TransformerConfigurationException e) {
      MMLog.error("Exception: " +e);
      e.printStackTrace();
      return;
    }
    catch (IOException e) {
      MMLog.error("Exception: " +e);
      e.printStackTrace();
      return;
    }
    catch (Exception e) {
      MMLog.error("Exception: " +e);
      e.printStackTrace();
      return;
    }
    
    /*
     * generate groups of 1 or more messages.  after the initial group,
     * a group is generated for each MessageVariant in the list -- the
     * variant list may be empty
     */
    boolean statusOK = true;
    int msgOffset = 0;
    Iterator it = messageVariants.iterator();
    for (;;) {
      String validity = msgGroupParams.groupOptions.getOption("validity");
  
      /*
       *  generate messages for this group 
       */
      for (int i = 0; 
              i < msgGroupParams.msgCount; 
              i++, msgGroupParams.messageControlId++) {
        int msgNumber = i + 1;        //  this group.
        msgGroupParams.msgNumber++;   //  all messages.

        /* mark next msg valid or invalid */
        String valid = "true";
        if ("VALID_AND_INVALID".equalsIgnoreCase(validity)) {
          if (((i+1) % 3) == 0) valid = "false";
        }
        else if ("INVALID".equalsIgnoreCase(validity)) {
          valid = "false";
        }

        /* if the user set the offset to a negative value, we generate a random
         * value for each msg offset, otherwise the msg offset is determined
         * by the value set by the user (to enable replicating a previous run) 
         */ 
        if (msgGroupParams.offset < 0) msgOffset = (int)(1000*Math.random() % 1000);
        else msgOffset = msgGroupParams.offset + i;
        
        /* set options for next message */
        msgGroupParams.setMessageOptions(valid, msgOffset);
        
        /*  set msg description  */
        messageDescription.setMessageProperties(msgGroupParams.messageOptions,
            errLocations); 
        
        /* map the options to what the transform wants */
        transformOptions = 
            msgGroupParams.messageOptions.options2TransformOptions();
        
        /*  pass the transform options into the transformer */
        /*transformer.setOptions(messageDescription, transformOptions, errLocations);*/
        transformer.setOptions(transformOptions);
        
        /* generate the new message  */
        String xmlMessage = transformer.transform();
  
        /* see if we should continue */
        if (!continueMsgGeneration(runner,
            validate,
            msgNumber,
            valid,
            xmlMessage,
            transformer,
            msgGroupParams, 
            messageDescription,
            messageFileNames)) {
          statusOK = false;
          break; 
        }
      }
  
      /*  get next Variant and set options. */ 
      Properties msgVariant = null;
      if (it.hasNext()) { 
        msgVariant = (Properties)it.next();
        msgGroupParams.setVariantOptions(msgVariant);
      }
      else
        break;   // finished
    }
    
    /*  finish up  */
    messageDescription.endOfSession();
    
    if (statusOK) statusMessage = "Finished Sucessfully\n";
    done = true;
  }

  /**
    Sets the message generation options.
    
    @param p   <i>A Properties object with property names containing the
        following properties:
        <p><i>
          messageControlId        <br>
                      If the string can be parsed as an integer, 
                      then the message control id of first message created.
                      Otherwise, the prefix prepended to an integer value
                      to form the message control id of each message. 
                      <p>
          msgCount                <br>
                      The total number of messages to create.
                      <p>
          offset                  <br>
                      Determines the starting location used for searching 
                      the xml data files for the values that are used
                      when constructing message elements. Typically, 
                      this can be set to -1, and the offset will be 
                      automatically calculated.  It can also be used to set
                      the offset to a saved value from a previous run.  
                      <p>
          validity                <br>
                      Determines if the messages that are generated are to
                      be all valid, all invalid, or a combination of both
                      valid and invalid messages.  Allowed values are:
                      VALID, INVALID, and VALID_AND_INVALID.   
                      <p>                      
          segmentCardinality      <br>
                      Determines the cardinality of all Segments.
                      Allowed values are: MIN, MAX, and RANDOM.
                      <p>                      
          fieldCardinality        <br>
                      Determines the cardinality of all Fields.
                      Allowed values are: MIN, MAX, and RANDOM.
                      <p>                      
          RE_Usage                <br>
                      Determines which elements with usage equal to
                      RE are included in the output message.
                      Allowed values are: ALL, NONE, and SOME.
                      <p>                      
          lengthConstraints       <br>
                      For primitive elements:
                      true = enforce length constraints.
                      false = don't enforce length constraints.
                      <p>                      
          compositeLength         <br>
                      true = issue a warning for composite elements whose 
                      length constraint may be violated without viloating 
                      any of its component element's length 
                      constraints. 
                      false = no action.
                      <p>                      
          minSegCardError         <br> 
                      true = produce an element with a minimum Segment
                      cardinaity error in the output message. 
                      false = no action.
                      <p>                      
          maxSegCardError         <br> 
                      true = produce an element with a maximum Segment
                      cardinaity error in the output message. 
                      false = no action.
                      <p>                      
          minFieldCardError       <br>  
                      true = produce an element with a minimum Field
                      cardinaity error in the output message. 
                      false = no action.
                      <p>                      
          maxFieldCardError       <br> 
                      true = produce an element with a maximum Field
                      cardinaity error in the output message. 
                      false = no action.
                      <p>                      
          fieldLengthError        <br>
                      true = produce an element with a Field length
                      error in the output message. 
                      false = no action.
                      <p>                      
          componentLengthError    <br>
                      true = produce an element with a Component length
                      error in the output message. 
                      false = no action.
                      <p>                      
          subComponentLengthError <br>
                      true = produce an element with a SubComponent 
                      length error in the output message. 
                      false = no action.
                      <p>                      
        </i><p>
    */
  public void setOptions (Properties p) {
    if (p == null) {
      MMLog.error("Can't set options. Properties is null");
      errorCondition = true;
      return;
    }

    try {
      options = new GroupOptions(p);
    }
    catch (IllegalArgumentException e) {
      e.printStackTrace();
      MMLog.error(e.toString());
      errorCondition = true;
      return;
    }
  }
  
  /**
    Sets the name of the output message directory.
              
    @see  #MessageGenerator(Properties)              
    */
  public void setOutputDir (String name) {
    fileNames.setProperty("outputMsgDir", name);
    
    outputMsgDir = name + fileSeparator;
    File d = new File(outputMsgDir);
    outputMsgPath = d.getAbsolutePath();
  }
 
  /**
    Sets the name of the meta data file.
              
    @see  #MessageGenerator(Properties)              
    */
  public void setMetaFile (String metafile) {
    fileNames.setProperty("metaFile", metafile);
    messageDescription.setMetaFile(outputMsgDir, metafile);
  }

  /**
    @return   The current message generation default option settings.
    */
  public Properties getOptions () {
    return options;
  }
  
  /**
    @return   The input and output file and directories names 
              in current use.
              
    @see  #MessageGenerator(Properties)              
    */
  public Properties getFileNames () {
    return fileNames;
  }
  
  /**
    @return   All messages file names that have been created 
          for the current invocation of <i>generateMessages</i>.
          
    @see  #generateMessages(boolean, Runnable)          
    */
  public Vector getMessageFileNames () {
    return messageFileNames;
  }

  /**
    Validates a message against a XML profile.
    
    @param  msg       The message to validate.
    @param  profile   The profile to use for validating the message.
    
    @see  #generateMessages(boolean, Runnable)
    */
  public static void validate (Message msg, String profile) {
    
/* ProfileParser:    
import org.apache.xerces.parsers.DOMParser;
import org.apache.xerces.parsers.StandardParserConfiguration;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.EntityResolver;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
 */    
    
    ProfileParser pparser = new ProfileParser(false);
    RuntimeProfile rtProfile = null;
    try {
      rtProfile = pparser.parse(profile);
    }
    catch (ProfileException e) {
      MMLog.error("Exception: " +e);
    }
    catch (Exception e) {
      MMLog.error("Exception: " +e);
    }
    
    StaticDef def = rtProfile.getMessage();
    
    HL7Exception[] errs = null;
    try {
      DefaultValidator validator = new DefaultValidator();
      errs = validator.validate(msg, def);
    }
    catch (ProfileException e) {
      MMLog.error("Exception: " +e);
    }
    catch (HL7Exception e) {
      MMLog.error("Exception: " +e);
    }
    catch (Exception e) {
      MMLog.error("Exception: " +e);
    }
    
    for (int i = 0; i < errs.length; i++) {
      MMLog.warn(false, "", "", "VALIDATION ERROR " +errs[i].toString());
    }
  }
  
  public void setOrgName (String name) {
    orgName = name;
  }
  
  public void setSessionName (String name) {
    sessionName = name;
  }
  
  public void setUserComment (String comment) {
    userComment = comment;
  }
  
  /**
     @return  the count of the number of messages generated.

     @see  #generateMessages(boolean, Runnable)    
  */
  public int getProgressValue() {
    return progressValue;
  }
  
  /** 
   * This should terminate the message generation process when 
   * <i>generateMessages</i> was invoked with a Runnable object. 
   * All files (messages & metadata) are left in their current state,
   * except that the end tag is added to the metadata file before
   * it is closed.
    
    @see  #generateMessages (boolean, Runnable)
    */
  public void stop() {
    stop = true;
    done = true;
    messageDescription.endOfSession();
  }
  
  /**
     resets flags and counters and removes all generated messages
    @see  #generateMessages(boolean, Runnable)
    @see  #stop()    
    */
  public void reset() {
    stop = false;
    done = false;
    progressValue = 0;
  }
  
  /*
   * convenience method equivalent to stop() + reset()
   */
  public void abort() {
    progressValue = 0;
    removeMsgFiles(messageFileNames);
    messageDescription.resetMetaFile();
  }
  
  /*
   * equivalent to instantiating a new MessageGeneration object.
   * (call reset() first to get rid of message files already generated)
   */
  public void restart (Properties files) {
    if (files != null) {
      // save file & dir names
      fileNames = files;
  
      init();
    }
    else {
      errorCondition = true;
    }
  }
  
  /**
    Checks to see if generation of the current set of messages 
    has completed when <i>generateMessages</i>
    was invoked with a Runnable object.
    
    @see  #generateMessages(boolean, Runnable)
    */
  public boolean isDone() {
    return done;
  }

  /**
    @return   A message indicating the status of the current 
          set of messages being generated when <i>generateMessages</i>
          when invoked with a Runnable object.
    
    @see  #generateMessages(boolean, Runnable)
    */
  public String getStatusMessage() {
    return statusMessage;
  }

  public static void main(String[] args) {
    MessageGenerator mg = new MessageGenerator("filenames.properties");
    mg.generateMessages(false);
  }
  
  //------------ Private methods -------------------------------
  
  private static String  getProfile () {
    BufferedReader in = null;

    try {
      in = new BufferedReader(new FileReader(xmlProfile));
    }
    catch (FileNotFoundException e) {
      MMLog.error("Exception: " +e);
    }
    catch (Exception e) {
      MMLog.error("Exception: " +e);
    }
    
    StringBuffer buf = new StringBuffer();
    try {
      String line = null;
      for (line = in.readLine(); line != null; line = in.readLine()) {
        buf.append(line);
      }
    }
    catch (IOException e) {
      MMLog.error("Exception: " +e);
    }
    catch (Exception e) {
      MMLog.error("Exception: " +e);
    }

    try {
      in.close();
    }
    catch (IOException e) {
      MMLog.error("Exception: " +e);
    }
    catch (Exception e) {
      MMLog.error("Exception: " +e);
    }
    
    return buf.toString();
  }
  
  /* set up the names of our input files, the metadata file, and the 
   * msg output directory.
   */
  private void setFileNames (Properties config) {
    if (config == null) {
      errorCondition = true;
      return;
    }
    
    if ((optionsTransform = config.getProperty("optionsTransform")) == null) {
      MMLog.error("'optionsTransform' property missing");
      errorCondition = true;
    }
    if ((preTransform = config.getProperty("preTransform")) == null) {
      MMLog.error("'preTransform' property missing");
      errorCondition = true;
    }
    if ((xsltTransform = config.getProperty("xsltTransform")) == null) {
      MMLog.error("'xsltTransform' property missing");
      errorCondition = true;
    }      
    if ((xmlProfile = config.getProperty("xmlProfile")) == null) {
      MMLog.error("'xmlProfile' property missing");
      errorCondition = true;
    }      
    if ((configTable = config.getProperty("configTable")) == null) {
      MMLog.error("'configTable' property missing");
      errorCondition = true;
    }      
    if ((hl7Tables = config.getProperty("hl7Tables")) == null) {
      MMLog.error("'hl7Tables' property missing");
      errorCondition = true;
    }      
    if ((userTables = config.getProperty("userTables")) == null) {
      MMLog.error("'userTable' property missing");
      errorCondition = true;
    }
    if ((localTables = config.getProperty("localTables")) == null) {
      MMLog.error("'localTable' property missing");
      errorCondition = true;
    }
    if ((externalTables = config.getProperty("externalTables")) == null) {
      MMLog.error("'externalTable' property missing");
      errorCondition = true;
    }
    if ((errorCodesTables = config.getProperty("errorCodes")) == null) {
    	MMLog.error("'errorCodes' property missing");
    	errorCondition = true;
    }
    if ((referenceTable = config.getProperty("referenceTable")) == null) {
      MMLog.error("'referenceTable' property missing");
      errorCondition = true;
    }      
    if ((dataElements = config.getProperty("dataElements")) == null) {
      MMLog.error("'dataElements' property missing");
      errorCondition = true;
    }      
    if ((optionsFile = config.getProperty("optionsFile")) == null) {
      MMLog.error("'optionsFile' property missing");
      errorCondition = true;
    }      
    if ((metaFileName = config.getProperty("metaFile")) == null) {
      MMLog.error("'metaFile' property missing");
      errorCondition = true;
    }
    if ((outputMsgDir = config.getProperty("outputMsgDir")) == null) {
      MMLog.error("'outputMsgDir' property missing");
      errorCondition = true;
    }
  }   

  private void setProfileProperties () {
    segmentNames = profileProp.getSegmentNames();
    fieldNames = profileProp.getFieldNames();
    compNames = profileProp.getComponentNames();
    subcompNames = profileProp.getSubcomponentNames();
    numberOfFields     = profileProp.getNumberOfFields();
    numberOfSegments   = profileProp.getNumberOfSegments();
    numberOfComp       = profileProp.getNumberOfComp();
    numberOfSubcomp    = profileProp.getNumberOfSubcomp();
    segmentCardMinNums = profileProp.getMinSegmentCardList();
    fieldCardMinNums   = profileProp.getMinFieldCardList();
    rUsageSegments     = profileProp.getRUsageSegments(); 
    xUsageSegments     = profileProp.getXUsageSegments();
    rUsageFields       = profileProp.getRUsageFields();
    xUsageFields       = profileProp.getXUsageFields();
    rUsageComponents   = profileProp.getRUsageComponents();
    xUsageComponents   = profileProp.getXUsageComponents();
    rUsageSubcomps     = profileProp.getRUsageSubcomps();
    xUsageSubcomps     = profileProp.getXUsageSubcomps();
    primitiveFields    = profileProp.getPrimitiveFields();
    primitiveComps     = profileProp.getPrimitiveComps();
    primitiveSubcomps  = profileProp.getPrimitiveSubcomps();
  }

  /* get properties from the named file */
  private Properties getProperties (String filename) {
    FileInputStream in = null;

    Properties properties = new Properties();
    
    try {
      in = new FileInputStream(filename);
      properties.load(in);
    }
    catch (FileNotFoundException e) {
      MMLog.error("Exception: " +e);
      return null;
    }
    catch (IOException e) {
      MMLog.error("Exception: " +e);
      return null;
    }
    catch (Exception e) {
      MMLog.error("Exception: " +e);
      return null;
    }

    //System.out.println("PWD: " +System.getProperty("user.dir"));
    //System.out.println("Properties from: " +filename);
    //properties.list(System.out);
    
    return properties;
  }
  
  private void removeMsgFiles (Vector filenames) {
    if (filenames == null) return;
    
    Enumeration e = filenames.elements();
    while (e.hasMoreElements()) {
      String name = (String)e.nextElement();
      File file = new File (name);
      
      file.delete();
    }
  }

  /* read options from the named properties file */
  private GroupOptions readOptionsProp (String filename) {
    errorCondition = false;
      
    FileInputStream in = null;

    Properties properties = new Properties();
    
    try {
      in = new FileInputStream(filename);
      properties.load(in);
    }
    catch (FileNotFoundException e) {
      MMLog.error("Exception: " +e);
    }
    catch (IOException e) {
      MMLog.error("Exception: " +e);
    }
    catch (Exception e) {
      MMLog.error("Exception: " +e);
    }

    //System.out.println("PWD: " +System.getProperty("user.dir"));
    //System.out.println("Properties from: " +filename);
    //properties.list(System.out);
    
    GroupOptions opt = null;
    try {
      opt = new GroupOptions(properties);
    }
    catch (IllegalArgumentException e) {
      MMLog.error(e.toString());
      errorCondition = true;
      return null;
    }
    
    return opt;
  }
 
  private void saveMessage (String msg, String outputFile) {
    FileWriter outFile = null;

    // save the generated msg
    try {
      outFile = new FileWriter (outputFile);
      outFile.write(msg, 0, msg.length());
      outFile.flush();
      MMLog.log("Generated HL7 message saved to " +outputFile);
    }
    catch (IOException e) {
      MMLog.error("Message not saved. Exception: " +e);
      e.printStackTrace();
    }
    catch (Exception e) {
      e.printStackTrace();
      MMLog.error("Message not saved. Exception: " +e);
    }
  }
    
  private boolean isXMLFile (String file) {
    return file.endsWith(".xml");
  }
  
  private XMLOptionsReader readOptionsXML (String optionsFile) {
    XMLOptionsReader opts = null;
    try {
      opts = new XMLOptionsReader(optionsTransform, optionsFile);
    } 
    catch (TransformerFactoryConfigurationError e) {
      e.printStackTrace();
      MMLog.error("Exception: " +e);
      errorCondition = true;
      return null;
    } 
    catch (TransformerException e) {
      e.printStackTrace();
      MMLog.error("Exception: " +e);
      errorCondition = true;
      return null;
    } 
    catch (IOException e) {
      e.printStackTrace();
      MMLog.error("Exception: " +e);
      errorCondition = true;
      return null;
    }

    return opts;
  }
  
  private boolean continueMsgGeneration (Runnable runner,
        boolean validate,
        int msgNumber,
        String valid,
        String xmlMessage,
        XsltTransform transformer,
        MessageGroupParams msgGroupParams,
        MessageDescription msgDescription,
        Vector messageFileNames) {

    if (xmlMessage == null) {
      statusMessage += lineSeparator +"**ERROR** "
          +"null message.  Giving up message generation!";
      MMLog.error("null message.  Giving up message generation!");
      return false;
    }

    //  save the msg, possibly validate, see if we should continue, etc
    String msgName = new File(xmlProfile).getName();
  
    if (msgName.endsWith(".xml"))
      msgName = msgName.substring(0, msgName.length() - 4);
    
    // make the file number 5 digits
    String fileNumber = Integer.toString(100000 + msgGroupParams.msgNumber);
    fileNumber = fileNumber.substring(1);
    String localFileName = msgName +".msg" +fileNumber +".xml";
    String outputMsg = outputMsgDir +localFileName;
  
    saveMessage(xmlMessage, outputMsg);
    
    boolean er7Encoding = true;   // true unless encoding fails
    Message message = null;
    
    // try to validate the msg if the msg should be valid and 
    //  the validation flag is set.
    if (validate && "true".equalsIgnoreCase(valid)) {
      message = transformer.encodeMessage();
      
      if (message != null) {
        if (profile == null) {
          profile = getProfile();
        }
        
        try {
          validate(message, profile);
        }
        catch (Exception e) {
          MMLog.warn("Message " +msgGroupParams.msgNumber +". Validation failed.");
        }
      }
      else {
        er7Encoding = false;
        MMLog.error("Message " +msgGroupParams.msgNumber 
          + ". Could not validate, encoding failed.");
      }
  
      transformer.encodeHl7Message();
      String hl7Message = transformer.getHl7Message();
      System.out.println("msg: " +hl7Message);
      /*
      transformer.encodeDocument();
      Document doc = transformer.getDocument();
      */
    }
  
    messageFileNames.add(outputMsgPath+fileSeparator+localFileName);
    msgDescription.saveMessageDescription(localFileName, 
          msgGroupParams.msgNumber, valid, er7Encoding); 
  
    //  see if we should continue
    if (stop) {
      statusMessage = "\nProcess Terminated\n"
      + "Close this dialog and select Generate again to restart\n";
      stop = false;
      return false;
    }
    
    progressValue = msgGroupParams.msgNumber;
    statusMessage = msgGroupParams.msgNumber +" Messages generated of " 
      +" a total of "+totalMsgCount +" to be generated. "
      +"This set, completed " +msgNumber +" out of "
      +msgGroupParams.msgCount;
    
    /*
     * If msg generation is running on a separate thread and being 
     * monitored from another object, invoke its run method.
     */
    if (runner != null) {
      try {
        javax.swing.SwingUtilities.invokeLater(runner);
      } 
      catch (Exception e) {
        MMLog.error("Exception: " +e);
        e.printStackTrace();
        return false;
      }
    }
    
    return true;
  }
}
