package gov.nist.hl7maker.message;

import java.io.*;
import java.util.*;
import java.io.StringWriter;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.apache.xerces.xni.parser.XMLInputSource;

import gov.nist.hl7maker.util.MMLog;


/**
  ProfileProperties is designed to collect and make available through its
  'get' methods properties of an HL7 profile.
  
  @author   Len Gebase, NIST
  */
  
public class ProfileProperties
{ 
  private static final int SEGMENT = 1;
  private static final int FIELD = 2;
  private static final int COMPONENT = 3;
  private static final int SUBCOMPONENT = 4;
  
  private int numberOfSegments = 0;
  private int numberOfFields =   0;
  private int numberOfComp =     0;
  private int numberOfSubcomp =  0;   
  private static List minSegmentCardList = new ArrayList();
  private static List minFieldCardList = new ArrayList();
  private static List rUsageSegments = new ArrayList();
  private static List reUsageSegments = new ArrayList();
  private static List xUsageSegments = new ArrayList();
  private static List rUsageFields = new ArrayList();
  private static List reUsageFields = new ArrayList();
  private static List xUsageFields = new ArrayList();
  private static List rUsageComponents = new ArrayList();
  private static List reUsageComponents = new ArrayList();
  private static List xUsageComponents = new ArrayList();
  private static List rUsageSubcomps = new ArrayList();
  private static List reUsageSubcomps = new ArrayList();
  private static List xUsageSubcomps = new ArrayList();
  private static List primitiveFields = new ArrayList();
  private static List primitiveComps = new ArrayList();
  private static List primitiveSubcomps = new ArrayList();

  private static Map segmentNames = new HashMap();
  private static Map fieldNames   = new HashMap();
  private static Map compNames  =   new HashMap();
  private static Map subcompNames = new HashMap();  
  
  private static String orgName = "";
  private static String sessionName = "";

  /**
    The name of the HL7 profile along with the transform used to
    read it are required for creating new instances of this class
    */
  public ProfileProperties (String xslTransform, String xmlProfile)
        throws TransformerFactoryConfigurationError, 
        TransformerException,
        IOException {
    if (!new File(xslTransform).exists()) {
      MMLog.error("transform: " +xslTransform +" does not exist");
      throw new IOException("File not found: " +xslTransform);
    }
    if (!new File(xmlProfile).exists()) {
      MMLog.error("profile: " +xmlProfile +" does not exist");
      throw new IOException("File not found: " +xmlProfile);
    }
  
/*    InputStream xsltInStream = null;
    InputStream profileInStream = null;
    
    if ((xsltInStream = 
        this.getClass().getResourceAsStream(xslTransform)) == null) {
      MMLog.error(cName, mName, "transform: " +xslTransform +" does not exist");
      throw new IOException("File not found: " +xslTransform);
    }
    
    if ((profileInStream = 
        this.getClass().getResourceAsStream(xmlProfile)) == null) {
      MMLog.error(cName, mName, "profile: " +xmlProfile +" does not exist");
      throw new IOException("File not found: " +xmlProfile);
    }
*/      
    //  create a StreamSource for both files
    StreamSource xsltStream = new StreamSource(new File(xslTransform));
    StreamSource profileStream = new StreamSource(new File(xmlProfile));
    //StreamSource xsltStream = new StreamSource(xsltInStream);
    //StreamSource profileStream = new StreamSource(profileInStream);

    //  get a new Transformer object for the xslt stream
    TransformerFactory tfactory = TransformerFactory.newInstance();
    Transformer transformer = tfactory.newTransformer(xsltStream);
    
    StringWriter writer = new StringWriter();
    StreamResult result = new  StreamResult(writer);

    //  apply the transform to the profile stream
    transformer.transform(profileStream, result);
    processXsltOutput(writer);
  }
  
  /* The format of the transform output looks like:
       SessionName
       OrganizationName
       nodeNumber'name'-M{R/RP/RE/X} ... nodeNumber'name'-M{R/RP/RE/X}  
       numOfSegs
       nodeNumber'name'-M{R/RP/RE/X} ... nodeNumber'name'-M{R/RP/RE/X}  
       numOfFields
       nodeNumber'name'-M{R/RP/RE/X} ... nodeNumber'name'-R/RE/X  
       numOfComps
       nodeNumber'name'-M{R/RP/RE/X} ... nodeNumber'name'-R/RE/X  
       numOfSubComps
     The output represents segment, field, component, and subcomponents
     data, in that order.
     the quotes and dash appear literally, name may include white space,
     the M (used only with Segments & Fields) is included when Min > 0, 
     and exactly one of letter combination follows indicating Usage.  
     each substring (nodeNumber'name'-chars) in the list is separated 
     by exactly one white space char and the entire substring ends with 
     a white space char followed by a newline.
     name includes not only the element name but also the names of its
     ancestors and a string identifying each.
    */  
  private void processXsltOutput (StringWriter xsltOutput)
        throws IOException {
    String output = xsltOutput.toString();
    
    if (output == null || "".equals(output)) return;
    
    StringReader sreader = new StringReader(output);
    BufferedReader reader = new BufferedReader(sreader);

//System.out.println("PRE PROCESS DATA:..." +output +"...");
    
    sessionName = xmlEncode(reader.readLine());
    orgName = xmlEncode(reader.readLine());

    String segments = reader.readLine();
    processString(segments, SEGMENT);
    numberOfSegments = Integer.parseInt(reader.readLine());
//System.out.println("num of segs: " +numberOfSegments);
    
    String fields = reader.readLine();
    processString(fields, FIELD);
    numberOfFields = Integer.parseInt(reader.readLine());
//System.out.println("num of fields: " +numberOfFields);      
    
    String comp = reader.readLine();
    processString(comp, COMPONENT);
    numberOfComp = Integer.parseInt(reader.readLine());
    
    // the substring holding data for each element (segment, field ...)
    //  is passed to processString().
    String subcomp = reader.readLine();
    processString(subcomp, SUBCOMPONENT);
    numberOfSubcomp = Integer.parseInt(reader.readLine());
    
    //System.out.println("seg list:..." +minSegmentCardList +"...");
    //System.out.println("field list:..." +minFieldCardList +"...");
    //System.out.println("num of fields:..." +numberOfFields+"...");
    //System.out.println("num of segs:..." +numberOfSegments+"...");
  }

  //  elementType: SEGMENT, FIELD, COMPONENT, SUBCOMPONENT
  //
  //  string:
  //    nodeNumber'name'-M{R/RP/RE/X} ... nodeNumber'name'-M{R/RP/RE/X}
  //  as described above.
  //
  // the output of this processing is a number of lists we maintain for 
  //  holding profile info on Segments, Fields, Components, and 
  //  Subcomponents.  For each element we maintain Usage lists.   
  //  For Segments and Fields we also maintain Minimum cardinality lists 
  //  identifying elements with Min Card > 0.
  //  the lists are maintained in pairs, one holds a list of numbers 
  //  that identifies the element's position in the profile (from the
  //  XPath position function) and the other a corresponding list with
  //  the name of each identified element.
  private void processString (String string, int elementType) {
    int len = string.length();
    int indexStart = 0;
    int indexEnd = 0;
    boolean end = false;
    
    // go thru entire string, for each substring, save the element name and
    //  update the appro list
    for (int i = 0; indexStart < len; i++) {
      indexEnd = string.indexOf("'", indexStart);
      String nodeNumber = string.substring(indexStart, indexEnd);
      
      indexStart = indexEnd + 1;
      indexEnd = string.indexOf("'-", indexStart);
      String name = string.substring(indexStart, indexEnd);
      
      indexStart = indexEnd + 2;
      indexEnd = string.indexOf(" ", indexStart);
      String value = string.substring(indexStart, indexEnd);
      
      indexStart = indexEnd + 1;
      
      //System.out.println("number:..." +nodeNumber +"...");
      //System.out.println("name:..." +name +"...");
      //System.out.println("value:..." +value +"...");

      // add the nodeNumber to the appro list and save the node name
      //  in a Map using the number as the key. 
      // (it's not clear that there is any need for the RE-lists)
      switch (elementType)
      {
        case SEGMENT:
          segmentNames.put(nodeNumber, name);
        
          if (value.startsWith("M")) {
            minSegmentCardList.add(nodeNumber);
          }
          
          if (value.endsWith("RE")) {
            reUsageSegments.add(nodeNumber);
          }
          else if (value.endsWith("R")) {
            rUsageSegments.add(nodeNumber);
          }
          else if (value.endsWith("X")) {
            xUsageSegments.add(nodeNumber);
          }
          break;
          
        case FIELD:
          fieldNames.put(nodeNumber, name);
          
          if (value.startsWith("M")) {
            minFieldCardList.add(nodeNumber);
          }
          
          if (value.endsWith("RE")) {
            reUsageFields.add(nodeNumber);
          }
          else if (value.endsWith("R")) {
            rUsageFields.add(nodeNumber);
          }
          else if (value.endsWith("X")) {
            xUsageFields.add(nodeNumber);
          }
          else if (value.endsWith("RP")) {
            rUsageFields.add(nodeNumber);
            primitiveFields.add(nodeNumber);
          }
          break;
          
        case COMPONENT:
          //compNames.put(nodeNumber, name);
          compNames.put(nodeNumber, name);
          
          if (value.endsWith("RE")) {
            reUsageComponents.add(nodeNumber);
          }
          else if (value.endsWith("R")) {
            rUsageComponents.add(nodeNumber);
          }
          else if (value.endsWith("X")) {
            xUsageComponents.add(nodeNumber);
          }
          else if (value.endsWith("RP")) {
            rUsageComponents.add(nodeNumber);            
            primitiveComps.add(nodeNumber);
          }
          break;
          
        case SUBCOMPONENT:
          //subcompNames.put(nodeNumber, name);
          subcompNames.put(nodeNumber, name);
          
          if (value.endsWith("RE")) {
            reUsageSubcomps.add(nodeNumber);
          }
          else if (value.endsWith("R")) {
            rUsageSubcomps.add(nodeNumber);  
          }
          else if (value.endsWith("X")) {
            xUsageSubcomps.add(nodeNumber);
          }
          else if (value.endsWith("RP")) {
            rUsageSubcomps.add(nodeNumber);            
            primitiveSubcomps.add(nodeNumber);
          }
          break;
          
        default:
      }
    }
  }

  public String getSessionName ()
  {
    return sessionName;
  }
  
  public String getOrgName ()
  {
    return orgName;
  }
  
  public int getNumberOfSegments ()
  {
    return numberOfSegments;
  }
  
  public int getNumberOfFields ()
  {
    return numberOfFields;
  }
  
  public int getNumberOfComp ()
  {
    return numberOfComp;
  }
  
  public int getNumberOfSubcomp ()
  {
    return numberOfSubcomp;
  }
     
  /* each List contains a list of node numbers encoded as a list of Strings 
   * (a node's name can be retrieved using a node number -- see below) */ 
  public List getMinSegmentCardList ()
  {
    return minSegmentCardList;
  }
  
  public List getMinFieldCardList ()
  {
    return minFieldCardList;
  }
  
  public List getRUsageSegments()
  {
    return rUsageSegments;
  }
     
  public List getReUsageSegments()
  {
    return reUsageSegments;
  }
    
  public List getXUsageSegments()
  {
    return xUsageSegments;
  }
     
  public List getRUsageFields()
  {
    return rUsageFields;
  }
       
  public List getReUsageFields()
  {
    return reUsageFields;
  }
      
  public List getXUsageFields()
  {
    return xUsageFields;
  }
       
  public List getRUsageComponents()
  {
    return rUsageComponents;
  }
   
  public List getReUsageComponents()
  {
    return reUsageComponents;
  }
  
  public List getXUsageComponents()
  {
    return xUsageComponents;
  }
   
  public List getRUsageSubcomps()
  {
    return rUsageSubcomps;
  }
     
  public List getReUsageSubcomps()
  {
    return reUsageComponents;
  }
    
  public List getXUsageSubcomps()
  {
    return xUsageSubcomps;
  }
  
  public List getPrimitiveFields()
  {
    return primitiveFields;
  }
  
  public List getPrimitiveComps()
  {
    return primitiveComps;
  }
  
  public List getPrimitiveSubcomps()
  {
    return primitiveSubcomps;
  }
  
  public Map getSegmentNames()
  {
    return segmentNames;
  }
  
  public Map getFieldNames()
  {
    return fieldNames;
  }
  
  public Map getComponentNames()
  {
    return compNames;
  }
  
  public Map getSubcomponentNames()
  {
    return subcompNames;
  }
  
  /*
  public String getSegmentName (String nodeNumber)
  {
    return (String)segmentNames.get(nodeNumber);
  }
  
  
  public String getFieldName (String nodeNumber) {
    return (String)fieldNames.get(nodeNumber);
  }
  */
  
  /*
   * All profile element names are returned in a List of Strings
   * (including Segment for consistency).  The first list element 
   * contains the Segment name.  If there's a second element in the 
   * list, it contains the Field name, if a third, the Component name,
   * and if a forth, the SubComponent name.    
   */
  public List getSegmentName (String nodeNumber) {
    ArrayList qname = new ArrayList();
    
    String name = (String)segmentNames.get(nodeNumber);
    qname.add(0, name);
    
    return qname;
  }
  
  public List getFieldName (String nodeNumber) {
    ArrayList qname = new ArrayList();
    String name = (String)fieldNames.get(nodeNumber); 
    
    int segIndex = name.indexOf("#SEG#");
    if (segIndex < 0) return null;
    
    int fieldIndex = name.indexOf("#FIELD#");
    if (fieldIndex < 0) return null;
    
    //  "#SEG# ".length()=6
    String segName = name.substring(segIndex + 6, fieldIndex - 1);       
    String fieldName = name.substring(fieldIndex + 8); // "#FIELD# ".length()=8   
    
    qname.add(0, segName);
    qname.add(1, fieldName);
    
    return qname;
  }
   
  public List getComponentName (String nodeNumber) {
    ArrayList qname = new ArrayList();
    String name = (String)compNames.get(nodeNumber);
    
    int segIndex = name.indexOf("#SEG#");
    if (segIndex < 0) return null;
    
    int fieldIndex = name.indexOf("#FIELD#");
    if (fieldIndex < 0) return null;
    
    int compIndex = name.indexOf("#COMP#");
    if (compIndex < 0) return null;
    
    //  "#SEG# ".length()=6
    String segName = name.substring(segIndex + 6, fieldIndex - 1);    
    //  "#FIELD# ".length()=8
    String fieldName = name.substring(fieldIndex + 8, compIndex - 1);    
    String compName = name.substring(compIndex + 7); // "#COMP# ".length()=7   
    
    qname.add(0, segName);
    qname.add(1, fieldName);
    qname.add(2, compName);
    
    return qname;
  }
   
  public List getSubComponentName (String nodeNumber) {
    ArrayList qname = new ArrayList();
    String name = (String)subcompNames.get(nodeNumber);
    
    int segIndex = name.indexOf("#SEG#");
    if (segIndex < 0) return null;
    
    int fieldIndex = name.indexOf("#FIELD#");
    if (fieldIndex < 0) return null;
    
    int compIndex = name.indexOf("#COMP#");
    if (compIndex < 0) return null;
    
    int subIndex = name.indexOf("#SUBCOMP#");
    if (subIndex < 0) return null;
    
    //  "#SEG# ".length()=6
    String segName = name.substring(segIndex + 6, fieldIndex - 1);    
    //  "#FIELD# ".length()=8
    String fieldName = name.substring(fieldIndex + 8, compIndex - 1);
    //  "#COMP# ".length()=7
    String compName = name.substring(compIndex + 7, subIndex - 1);    
    String subName = name.substring(subIndex + 10); // "#SUBCOMP# ".length()=10   
    
    qname.add(0, segName);
    qname.add(1, fieldName);
    qname.add(2, compName);
    qname.add(3, subName);

    return qname;
  }
  
/*
  public String getComponentName (String nodeNumber)
  {
    return (String)compNames.get(nodeNumber);
  }
  
  public String getSubcomponentName (String nodeNumber)
  {
    return (String)subcompNames.get(nodeNumber);
  } */

  //  get segment number (we're using) based on segment name
  //  (inverse of getSegmentName)
  public String getSegmentNumber (String name)
  {
    if (name == null) return null;
    
    Iterator it = segmentNames.entrySet().iterator();
    while (it.hasNext()) {
      Map.Entry mapEntry = (Map.Entry)it.next();
      
      if (name.equals((String)mapEntry.getValue())) {
        return (String)mapEntry.getKey();
      }
    }
    
    return null;
  }
  
  //  get the number associated with the named Field, but it's not
  //  what we want, may be one of many (we're not maintaining
  //  any info on the profile's hierarchy -- we'll need to do 
  //  this is we have to have this functionality, but we'll
  //  probably take a different approach)
  public String getAFieldNumber (String name)
  {
    if (name == null) return null;
    
    Iterator it = fieldNames.entrySet().iterator();
    while (it.hasNext()) {
      Map.Entry mapEntry = (Map.Entry)it.next();
      
      if (name.equals((String)mapEntry.getValue())) {
        return (String)mapEntry.getKey();
      }
    }
    
    return null;
  }

  //  get the number associated with the named Comp, but it's not
  //  what we want -- see above.
  public String getAComponentNumber (String name)
  {
    if (name == null) return null;
    
    Iterator it = compNames.entrySet().iterator();
    while (it.hasNext()) {
      Map.Entry mapEntry = (Map.Entry)it.next();
      
      if (name.equals((String)mapEntry.getValue())) {
        return (String)mapEntry.getKey();
      }
    }
    
    return null;
  }
  
  //  get the number associated with the named SubComp, but it's not
  //  what we want -- see above.
  public String getASubcomponentNumber (String name)
  {
    if (name == null) return null;
    
    Iterator it = subcompNames.entrySet().iterator();
    while (it.hasNext()) {
      Map.Entry mapEntry = (Map.Entry)it.next();
      
      if (name.equals((String)mapEntry.getValue())) {
        return (String)mapEntry.getKey();
      }
    }
    
    return null;
  }

  /*	return string with xml special chars correctly encoded:  
		  & - &amp;
		  < - &lt;
		  > - &gt;
		  " - &quot;
		  ' - &#39;
		  (right now they're simply removed)  
  */
  private String xmlEncode (String instring) {
	  String newString = instring;
	  int len = newString.length();

	  for (int index = 0; index < len; index++) {
		  if (isXmlChar(newString.charAt(index))) {
			  newString = xmlEncodeChar(newString, newString.charAt(index), index);
			  len = newString.length();	// replaced char with xml string
			  
			  index = newString.indexOf(';', index) + 1;	// xml encoding always ends with ;
		  }
	  }
	  
	  return newString;
  }
  
  private boolean isXmlChar (char c) {
	  if (c == '&') return true;
	  if (c == '<') return true;
	  if (c == '>') return true;
	  if (c == '"') return true;
	  if (c == '\'') return true;
	     
	  return false;
  }
  
  private String xmlEncodeChar (String instring, char xmlChar, int index) {
	  String head = instring.substring(0, index);
	  String tail = instring.substring(index+1);

      String xml = "";
	  switch (xmlChar) {
	  case '&':
		  xml = "&amp;";
		  break;
	  case '<':
		  xml = "&lt;";
		  break;
	  case '>':
		  xml = "&gt;";
		  break;
	  case '"':
		  xml = "&quot;";
		  break;
	  case '\'':
		  xml = "&#39;";
		  break;
	  default:
		  xml = "";
	  }
	  
	  String newString = head.concat(xml);

	  return newString.concat(tail);
  }
}
