package gov.nist.hl7maker.message;

import java.io.*;
import java.io.File;
import java.util.*;
import java.io.StringWriter;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;
import org.w3c.dom.*; 
import ca.uhn.hl7v2.model.Message;
import ca.uhn.hl7v2.parser.*;
import ca.uhn.hl7v2.HL7Exception;
import gov.nist.hl7maker.util.MMLog;

/**
  This class serves as a Java interface to the message
  generation XSL transformation.
  The setOptions() method determines the paramter values that are
  passed to the transformation, and the transform() method effects
  the invocation of the transformation. The transformation results in 
  the generation of a single HL7 message.  The generated message 
  conforms to the profile specification, except that it may violate 
  specific conditions of the profile when the validity option is set to false.
  
  @author   Len Gebase, NIST
  */
  
public class XsltTransform {
  private static String errorStatus = null;
  
  private MessageDescription msgDescription = null;
  private StreamSource xsltStream = null;
  private StreamSource profileStream = null; 
  private Transformer transformer = null;
  
  // name of profile file
  private String xmlProfile = null;
  
  // data lookup files used by the xsl transformation
  private String configTable = null;   
  private String hl7Tables = null;
  private String userTables = null;
  private String localTables = null;
  private String externalTables = null;
  private String referenceTable = null;
  private String dataElements = null;
  private String errorCodes = null;
  
  // for now we only support the File protocol, may change later
  //  and make this more flexible
  private String protocol = "File";
  
  // various encodings of the same HL7 message:
  private String xmlMessage = null;   // xml encoded HL7 msg
  private String hl7Message = null;   // std er7 encoded HL7 msg
  private Document document = null;  
  private Message message = null;                                         
  
  private static final String cName = "XsltTransform";

  /* the xslt and profile file names, and the data input
      files needed by the transformation (except error codes),  
      are all required for creating new instances of this class.
   */
  public XsltTransform (String xsltTransform, 
                  String profile,
                  String cTable,   
                  String hTables,
                  String uTables,
                  String lTables,
                  String xTables,
                  String refTable,
                  String dElements,
                  String errCodes)
  throws TransformerFactoryConfigurationError,
  TransformerConfigurationException,
  IOException {
    String mName = "XsltTransform";
   
    xmlProfile = profile;
    // the file names passed to the transformation
    //  have to be absolute (the relative names will be wrong when the
    //  transformation sees them)
    configTable = new File(cTable).getAbsolutePath();
    hl7Tables =  new File(hTables).getAbsolutePath();
    userTables =  new File(uTables).getAbsolutePath();
    localTables =  new File(lTables).getAbsolutePath();
    externalTables =  new File(xTables).getAbsolutePath();
    referenceTable = new File(refTable).getAbsolutePath();
    dataElements = new File(dElements).getAbsolutePath();
    errorCodes = new File(errCodes).getAbsolutePath();

    /* error codes aren't required, but we don't want to set
     * a null transform parameter
     */
    if ((new File(errorCodes)).isFile() != true) {
    	errorCodes = "";
    }
    
    if (!filesExist()) {
      MMLog.error(cName, mName,
          "PWD: " +System.getProperty("user.dir") +"\n"
      +"One or more required files doesn't exist (or isn't normal). Check:\n"
          +configTable    +" "
          +hl7Tables      +" "
          +localTables    +" "
          +userTables     +" "
          +externalTables +" "
          +referenceTable +" "
          +dataElements);
      errorStatus = "1 or more files missing!";
    }
    
    if (errorStatus == null) {
      configTable =      protocol + ":///" +configTable;
      hl7Tables =        protocol +":///" +hl7Tables;
      userTables =       protocol +":///" +userTables; 
      localTables =      protocol +":///" +localTables; 
      externalTables =   protocol +":///" +externalTables;
      referenceTable =   protocol +":///" +referenceTable;
      dataElements =     protocol +":///" +dataElements;
    }

    xsltStream = new StreamSource(new File(xsltTransform));
    profileStream = new StreamSource(new File(xmlProfile));

    TransformerFactory tfactory = TransformerFactory.newInstance();
    transformer = tfactory.newTransformer(xsltStream);
    MMLog.log("New XSLT Transformation:\n   transform: " +xsltTransform
                +" profile: " +xmlProfile);
  }

	/*
	                          value range           
	                          -----------           
	valid                   true/false              
	segmentCardinality      0 <= number <= 1       
	fieldCardinality        0 <= number <= 1       
	RE-Usage                0 <= number <= 1       
	CE-Usage                0 <= number <= 1
	minSegCardError        -1 <= number <= -0.1
	maxSegCardError        -1 <= number <= -0.1
	minFieldCardError      -1 <= number <= -0.1
	maxFieldCardError      -1 <= number <= -0.1
	RusageError            -1 <= number <= -0.1
	XusageError            -1 <= number <= -0.1
	fieldLengthError       -1 <= number <= -0.1
	componentLengthError   -1 <= number <= -0.1
	subComponentLenError   -1 <= number <= -0.1
	lengthConstraints         0 or 1                
	compositeLength           0 or 1                
	*/
  public void setOptions (MessageOptions options) {
    String mName = "setOptions";

    MMLog.debug(cName, mName, options);

    String valid = options.getOption("valid");
    String offset = options.getOption("offset");
    String starValue = options.getOption("starValue");
    String msgControlID = options.getOption("messageControlId");
    String segmentCardinality = options.getOption("segmentCardinality");      
    String fieldCardinality = options.getOption("fieldCardinality");        
    String RE_Usage = options.getOption("RE_Usage");                

    String fieldLengthErr = options.getOption("fieldLengthError");
    String compLengthErr = options.getOption("componentLengthError");
    String subCompLengthErr = 
        options.getOption("subComponentLengthError");
    String minSegCardErr = options.getOption("minSegCardError");
    String maxSegCardErr = options.getOption("maxSegCardError");
    String minFieldCardErr = options.getOption("minFieldCardError");
    String maxFieldCardErr = options.getOption("maxFieldCardError");
    String RusageErr = options.getOption("RusageError");
    String XusageErr = options.getOption("XusageError");
    
    String lengthConstraints = options.getOption("lengthConstraints");
    String compositeLength = options.getOption("compositeLength");
    
    String segmentName = options.getOption("segmentName");
    String fieldName = options.getOption("fieldName");
    String componentName = options.getOption("componentName");
    String subComponentName = options.getOption("subComponentName");
    String elementCardinality = options.getOption("cardinality");
    String elementREinclude = options.getOption("REinclude");
    String elementErrorCode = options.getOption("elementErrorCode");
    
    //  no support yet:
    String C_Usage = "0";                 
    String CE_Usage = "0";             

    // transform input files
    transformer.setParameter("configTable", configTable);   
    transformer.setParameter("hl7Tables", hl7Tables);
    transformer.setParameter("userTables", userTables);
    transformer.setParameter("localTables", localTables);
    transformer.setParameter("externalTables", externalTables);
    transformer.setParameter("referenceTable", referenceTable);
    transformer.setParameter("dataElementsTable", dataElements);


    // transform options
    transformer.setParameter("valid", valid);
    transformer.setParameter("starValue", starValue);
    transformer.setParameter("offset", offset);    
    transformer.setParameter("messageControlID", msgControlID);
    transformer.setParameter("segmentCardinality", segmentCardinality);      
    transformer.setParameter("fieldCardinality", fieldCardinality);        
    transformer.setParameter("RE_Usage", RE_Usage);                
    transformer.setParameter("C_Usage", C_Usage);                 
    transformer.setParameter("CE_Usage", CE_Usage);                
    transformer.setParameter("lengthConstraints", lengthConstraints);       
    transformer.setParameter("fieldLength", fieldLengthErr); 
    transformer.setParameter("compLength", compLengthErr); 
    transformer.setParameter("subCompLength", subCompLengthErr);    
    transformer.setParameter("compositeLength", compositeLength);
    transformer.setParameter("minSegCardError", minSegCardErr);
    transformer.setParameter("maxSegCardError", maxSegCardErr);
    transformer.setParameter("minFieldCardError", minFieldCardErr);
    transformer.setParameter("maxFieldCardError", maxFieldCardErr);
    transformer.setParameter("RusageError", RusageErr);
    transformer.setParameter("XusageError", XusageErr);
    
    transformer.setParameter("namedSegment", segmentName);
    transformer.setParameter("namedField", fieldName);
    transformer.setParameter("namedComponent", componentName);
    transformer.setParameter("namedSubComponent", subComponentName);
    transformer.setParameter("elementCardinality", elementCardinality);
    transformer.setParameter("elementREinclude", elementREinclude);
    transformer.setParameter("elementErrorCode", elementErrorCode);
    
    if ("true".equalsIgnoreCase(elementErrorCode)) {
        if ((new File(errorCodes)).isFile() != true) {
        	errorStatus = "error codes file missing: " +errorCodes;
        	MMLog.error(errorStatus);
        }
    	errorCodes = protocol + ":///" + errorCodes;
    }
    transformer.setParameter("errorCodesTable", errorCodes);
    
    String dateOfMessage = getDate();
    transformer.setParameter("dateOfMessage", dateOfMessage);
    
    MMLog.log("configTable " +configTable);
    MMLog.log("hl7Tables " +hl7Tables);
    MMLog.log("userTables " +userTables);
    MMLog.log("localTables " +localTables);
    MMLog.log("externalTables " +externalTables);
    MMLog.log("referenceTable " +referenceTable);
    MMLog.log("dataElementsTable " +dataElements);
    MMLog.log("errorCodesTable " +errorCodes);
    /*MMLog.log("cardErrorMinSeg " +cardErrorMinSeg);
    MMLog.log("cardErrorMinField " +cardErrorMinField);
    MMLog.log("cardErrorMaxSeg " +cardErrorMaxSeg);
    MMLog.log("cardErrorMaxField " +cardErrorMaxField);
    MMLog.log("rUErrorSegment " +rUSegment);
    MMLog.log("rUErrorField " +rUField); 
    MMLog.log("rUErrorComponent " +rUComponent);
    MMLog.log("rUErrorSubcomp " +rUSubcomp);
    MMLog.log("xUErrorField " +xUField); 
    MMLog.log("xUErrorSegment " +xUSegment);
    MMLog.log("xUErrorComponent " +xUComponent);
    MMLog.log("xUErrorSubcomp " +xUSubcomp);*/
    MMLog.log("dateOfMessage " +dateOfMessage);
    MMLog.log("valid " +valid);
    MMLog.log("starValue " +starValue);
    MMLog.log("offset " +offset);
    MMLog.log("messageControlId " +msgControlID);
    MMLog.log("segmentCardinality " +segmentCardinality);
    MMLog.log("fieldCardinality " +fieldCardinality);
    MMLog.log("RE_Usage " +RE_Usage);
    MMLog.log("C_Usage " +C_Usage);
    MMLog.log("CE_Usage " +CE_Usage);
    MMLog.log("lengthConstraints " +lengthConstraints);
    MMLog.log("fieldLengthErr " +fieldLengthErr);
    MMLog.log("compLengthErr " +compLengthErr);
    MMLog.log("subCompLengthErr " +subCompLengthErr);
    MMLog.log("compositeLength " + compositeLength);
    MMLog.log("minSegCardError " +minSegCardErr);
    MMLog.log("maxSegCardError " +maxSegCardErr);
    MMLog.log("minFieldCardError " +minFieldCardErr);
    MMLog.log("maxFieldCardError " +maxFieldCardErr);
    MMLog.log("RusageError " +RusageErr);
    MMLog.log("XusageError " +XusageErr);
    MMLog.log("segmentName " + segmentName);
    MMLog.log("fieldName " + fieldName);
    MMLog.log("componentName " + componentName);
    MMLog.log("subComponentName " + subComponentName);
    MMLog.log("elementCardinality " + elementCardinality);
    MMLog.log("elementREinclude " + elementREinclude);
  }
  
  /* apply the xsl transformation and return the xml message generated */ 
  public String transform () {
    String mName = "transform";
    
    if (errorStatus != null) { 
      MMLog.error(cName, mName, "Cannot apply transform to profile. status: " 
                            +errorStatus);
      return null;
    }
      
    StringWriter writer = new StringWriter();

    try {
      StreamResult result = new  StreamResult(writer);
      transformer.transform(profileStream, result);
    }
    catch (TransformerException e) {
      MMLog.error(cName, mName, "Exception: " +e);
      return null;
    }
    catch (Exception e) {
      MMLog.error(cName, mName, "Exception: " +e);
      return null;
    }

    xmlMessage = writer.toString();
    
    return xmlMessage;
  }

  /* apply the xsl transformation and save the generated msg to a file
      before returning it. 
  */
  public String transform (String outputFile) {
    String mName = "transform(String)";
    
    // apply the transform
    String msg = transform();
    
    if (msg == null) return null;
    
    FileWriter outFile = null;
    
    // save the generated msg
    try {
      outFile = new FileWriter (outputFile);
      outFile.write(msg, 0, msg.length());
      outFile.flush();
    }
    catch (IOException e) {
      MMLog.error(cName, mName, "Exception: " +e);
      return null;
    }
    catch (Exception e) {
      MMLog.error(cName, mName, "Exception: " +e);
      return null;
    }

    MMLog.log("Generated HL7 message saved to " +outputFile);
    
    return msg;
  }

  /*  encode the xmlMessage as a Message object */  
  public Message encodeMessage ()
  {
    String mName = "encodeMessage";
    
    try {
        DefaultXMLParser parser = new DefaultXMLParser();
        message = parser.parse(xmlMessage);
    }
    catch (EncodingNotSupportedException e) {
      MMLog.error(cName, mName, "Exception: " +e);
      return null;
    }
    catch (HL7Exception e) {
      MMLog.error(cName, mName, "Exception: " +e);
      return null;
    }
    catch (Exception e) {
      MMLog.error(cName, mName, "Exception: " +e);
      return null;
    }

    MMLog.debug(cName, mName, 
        "Message.stringVal = " +checkMessage(message));
    
    return message;
  }

  // encode the std hl7 as a String  
  //  (if not already encoded, the Message encoding is generated)  
  public String encodeHl7Message ()
  {
    String mName = "encodeHl7Message";
    
    try {
      Parser parser = new PipeParser();
      
      if (message == null) encodeMessage();
      hl7Message = parser.encode(message);
    }
    catch (EncodingNotSupportedException e) {
      MMLog.error(cName, mName, "Exception: " +e);
    }
    catch (HL7Exception e) {
      MMLog.error(cName, mName, "Exception: " +e);
    }
    catch (Exception e) {
      MMLog.error(cName, mName, "Exception: " +e);
    }
    
    MMLog.log("Standard HL7 message encoding:\n" +hl7Message);
    
    return hl7Message;
  }
  
  // encode the Document representation of the message
  //  (if not already encoded, the Message encoding is generated)  
  public Document encodeDocument ()
  {
    String mName = "encodeDocument";
    
    try {
      DefaultXMLParser parser = new DefaultXMLParser();
      
      if (message == null) encodeMessage();        
      document = parser.encodeDocument(message);
    }
    catch (EncodingNotSupportedException e) {
      MMLog.error(cName, mName, "Exception: " +e);
    }
    catch (HL7Exception e) {
      MMLog.error(cName, mName, "Exception: " +e);
    }
    catch (Exception e) {
      MMLog.error(cName, mName, "Exception: " +e);
    }

    MMLog.debug(cName, mName, 
        "Document.stringVal = " +checkDocument(document));
    
    return document;
  }
  
  // return the message as a xml encoded string
  public String getXmlMessage ()
  {
    return xmlMessage;
  }

  // return the xml message encoded as a Message 
  public Message getMessage ()
  {
    return message;
  }

  // return a std HL7 encoding of the xml message
  public String getHl7Message ()
  {
    return hl7Message;
  }

  // return the xml message encoded as a Document 
  public Document getDocument ()
  {
    return document;
  }

  public MessageDescription getMessageDescription()
  {
    return msgDescription;
  }
   
  private String checkMessage (Message msg) {
    String mName = "checkMessage";
    
    if (msg == null) {
      MMLog.error(cName, mName, "null message!");
      return null;
    }
    
    String msgString = null;
    
    try {
      DefaultXMLParser parser = new DefaultXMLParser();
      
      msgString = parser.encode(msg);
    }
    catch (EncodingNotSupportedException e) {
      MMLog.error(cName, mName, "Exception: " +e);
    }
    catch (HL7Exception e) {
      MMLog.error(cName, mName, "Exception: " +e);
    }
    catch (Exception e) {
      MMLog.error(cName, mName, "Exception: " +e);
    }

    return msgString;
  }
  
  private String checkDocument (Document doc) {
    String mName = "checkDocument";
    
    if (doc == null) {
      MMLog.error(cName, mName, "null document!");
      return null;
    }
    
    Message msg = null;
    String version = "2.4";
    
    try {
      DefaultXMLParser parser = new DefaultXMLParser();
      
      msg = parser.parseDocument(doc, version);
    }
    catch (EncodingNotSupportedException e) {
      MMLog.error(cName, mName, "Exception: " +e);
    }
    catch (HL7Exception e) {
      MMLog.error(cName, mName, "Exception: " +e);
    }
    catch (Exception e) {
      MMLog.error(cName, mName, "Exception: " +e);
    }

    return checkMessage(msg);
  }

  // checks that all required files exist and are 'normal'  
  private boolean filesExist () {
    if ((new File(xmlProfile)).isFile() != true) {
      return false;
    }
    if ((new File(configTable)).isFile() != true) {
      return false;
    }
    if ((new File(hl7Tables)).isFile() != true) {
      return false;
    }
    if ((new File(userTables)).isFile() != true) {
      return false;
    }
    if ((new File(localTables)).isFile() != true) {
      return false;
    }
    if ((new File(externalTables)).isFile() != true) {
      return false;
    }
    if ((new File(referenceTable)).isFile() != true) {
      return false;
    }
    if ((new File(dataElements)).isFile() != true) {
      return false;
    }
    
    return true;
  }
  
/*     
 ------   -------                 ------------        -------
 G        era designator          (Text)              AD
 y        year                    (Number)            1996
 M        month in year           (Text & Number)     July & 07
 d        day in month            (Number)            10
 h        hour in am/pm (1~12)    (Number)            12
 H        hour in day (0~23)      (Number)            0
 m        minute in hour          (Number)            30
 s        second in minute        (Number)            55
 S        millisecond             (Number)            978
 E        day in week             (Text)              Tuesday
 D        day in year             (Number)            189
 F        day of week in month    (Number)            2 (2nd Wed in July)
 w        week in year            (Number)            27
 W        week in month           (Number)            2
 a        am/pm marker            (Text)              PM
 k        hour in day (1~24)      (Number)            24
 K        hour in am/pm (0~11)    (Number)            0
 z        time zone               (Text)              Pacific Standard Time
 '        escape for text         (Delimiter)
 ''       single quote            (Literal)           '
 */ 
  private String getDate () {
    java.text.SimpleDateFormat formatter = 
    	new java.text.SimpleDateFormat ("yyyyMMddHHmmss");
          //new java.text.SimpleDateFormat ("yyyyMMddHHmmss.SSS");
    Date currentTime = new Date();
    String date = formatter.format(currentTime);
    return date;
  }
}
