package gov.nist.hl7maker.util;

import java.io.*;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Properties;

import gov.nist.hl7maker.util.*;

public class ErrorCodes {

  public static final int HL7TABLE = 1;
  public static final int USERTABLE = 2;
  public static final int LOCALTABLE = 3;
  public static final int EXTERNALTABLE = 4;
  
  private String tablesFileIn = null;
  private StringBuffer tablesFileOut = null;
  private int tableIndex = 0;
  private boolean errorCondition = false;
  private PrintWriter outWriter = null;
  
  private int maxErrorNumber = 0;
  private String maxErrorChars = "";
	
  /**
  Generates a set of error codes.

  @param filename  name of the output file where the error codes will be stored. 
            
  @see  #addErrorCodes(String)  
  @see  #saveErrorCodes()            
  */
  public ErrorCodes (String filename) {
	  /* at least for now, it has to be a new file. */
	  if (new File(filename).exists()) {
		  MMLog.error("Error! File: " +filename +" already exists");
		  errorCondition = true;
		  return;
	  }
	  
	  /* set up to save output file */
	  try {
		outWriter = new PrintWriter(new BufferedWriter(new FileWriter(filename)));
	  } 
	  catch (IOException e) {
		e.printStackTrace();
		MMLog.error("Error! File: " +filename +" " +e);
		errorCondition = true;
		return;
	  }
	 
	  String header = getHeader();
	  tablesFileOut = new StringBuffer();
	  tablesFileOut.append(header);
  }
  
  /**
  Adds a set of error codes for the named file to the output.

  @param filename  name of the file of tables with valid codes that error
  						codes will be generated for.
            
  @see  #ErrorCodes(String)              
  @see  #saveErrorCodes()              
  */
  public void addErrorCodes (String filename) {
	if (errorCondition) {
		MMLog.error("Can't add error codes!");
		return;
	}
	
	tableIndex = 0;
	loadTables(filename);
	
	 /* for each input table, add a corresponding error table to output */ 
    for (String table = nextTable();
    		!"".equals(table);
    		table = nextTable()) {
    			
    	String newTable = getErrorTable(table);
    	tablesFileOut.append(newTable);
    	tablesFileOut.append("\n");
    }
  }
  
  /**
  Appends trailer and closes output file.  Must be called after all error 
  codes are added. 
            
  @see  #ErrorCodes(String)              
  @see  #saveErrorCodes()              
  */
  public void saveErrorCodes () {
	if (errorCondition) {
		MMLog.error("Can't save error codes!");
		return;
	}

	  tablesFileOut.append(getTrailer());
	  outWriter.println(tablesFileOut); 
	  outWriter.close();
  }

	/* load input tables */
  private void loadTables (String filename) {
	tablesFileIn = getFile(filename);
  }
  
  private String getFile (String filename) {
	  int read = 0;
	  int start = 0;
	  final int max = 512;
	  
	  char[] tmpBuf = new char[max];
	  StringBuffer fileBuf = new StringBuffer();
	  
	  BufferedReader reader = null;;
	  try {
	    reader = new BufferedReader(new FileReader(filename));
	  }
	  catch (FileNotFoundException e) {
	    MMLog.error("ERROR:  Exception = " +e);
	  }
		

	  while (read != -1) {
		  try {
			  read = reader.read(tmpBuf, start, max);
		  }
		  catch (IOException e) {
			  MMLog.error("ERROR:  Exception = " +e);
			  break;
		  }
		  
		  if (read == -1) 
			  break;
		  
		  fileBuf.append(tmpBuf, start, read);
	  }
	  
	  return (new String(fileBuf));
  }

  private String getHeader () {
	  return "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" +Constants.LS
	  	+"<errorCodes xmlns=\"http://www.nist.gov/hl7/errorCodes\">"
	  	+Constants.LS;
  }

  private String getTrailer () {
	  return "</errorCodes>" +Constants.LS;
  }

  private String nextTable () {
	  String startOfTable = "<Table";
	  String endOfTable = "</Table>";
	  int start = tablesFileIn.indexOf(startOfTable, tableIndex);
	  int end = tablesFileIn.indexOf(endOfTable, tableIndex);

	  if (end < 0) {
		  return "";
	  }
	  else {
          end += endOfTable.length();
		  String table = tablesFileIn.substring(start, end);
		  tableIndex =  end + 1;
		  return table;
	  }
  }
  
  private int itemIndex = 0;
  
  private String getErrorTable (String table) {
	  
	 StringBuffer errorTable = new StringBuffer();
	 
	 /* The existing table tag, with attrs, ids the new table.
	  * (we'll then add a single item with an invalid value below.)
	  * ex table:
	  * <Table Type="HL7User" Name="Marital status" ID="0002">
	  *  	<Item Code="A" Value="Separated"/>  */
	 String itemStart = "<Item";
	 int itemLoc = table.indexOf(itemStart);
	 String tableTag = table.substring(0, itemLoc - 1);
     errorTable.append(tableTag);
	 
     maxErrorNumber = 0;
     maxErrorChars = "";
     String errorCode = "";
	/* copy each item to output table and save item values */ 
    for (String item = nextItem(table);
    		!"".equals(item);
    		item = nextItem(table)) {
    			
    	String itemCode = getItemCode(item);
    	errorCode = getErrorCode(itemCode);
    }

    /* add item with error code */
    String item = getItem(errorCode, "The Code is Invalid so this Value is not significant");
    errorTable.append(item);
    /* mark end of table */
    String endOfTable = "</Table>";
    errorTable.append(endOfTable);
    
    return new String(errorTable);
  }
  
  private String nextItem (String table) {
	  String itemLabel = "<Item";
	  int start = table.indexOf(itemLabel, itemIndex);
	  if (start < 0) {
		  itemIndex = 0;
		  return "";
	  }
	  
	  int next = table.indexOf("</Item>", start + itemLabel.length());
	  if (next < 0)
		  next = table.indexOf("/>", start + itemLabel.length());

	  if (next < 0) {
		  itemIndex =  0;
		  return "";
	  }
	  else {
		  String item = table.substring(start, next - 1);
		  itemIndex =  next;
		  return item;
	  }
  }

  /* ex item: <Item Code="2028-9" Value="Asian"/> */
  private String getItemCode (String item) {
	  int index = item.indexOf("Code");
	  int start = item.indexOf("\"", index);
	  int end = item.indexOf("\"", start + 1);
	  String code = item.substring(start + 1, end);
	  
	  return code;  
  }
  
  /* get an error code based on the value of itemCode.
   * 
   * item codes within the same table can be both numbers and char strings,
   * so max code values are saved and should be reinitialized for each table.
   */
  private String getErrorCode (String itemCode) {
	  double itemValue = 0;
	  int len = 0;
	  boolean itemIsNumber = false;
  
	  try {
		  Double d = new Double(itemCode);
		  itemValue = d.doubleValue();
		  itemIsNumber = true;
	  }
	  catch (NumberFormatException e) {
		  len = itemCode.length();
	  }
	
 	  if (itemIsNumber) {
		  if ((int)itemValue >= maxErrorNumber) { 
			  maxErrorNumber = (int)itemValue + 1;
		  }
		  return String.valueOf((int)maxErrorNumber);
	  }
	  else {
		  /* we can get a new error code using an arbitrary starting 
		   * value -- any starting value should be ok. */
		  if ("".equals(maxErrorChars)) {
			  maxErrorChars = getNewCode(len, 'T');
		  }
		   
		  /* if the item code is lexographically .ge. max code value, 
		   * get a new errCode.
		   */
		  if (maxErrorChars.compareToIgnoreCase(itemCode) <= 0) {
	  		  char errChar = itemCode.charAt(0);
			  maxErrorChars = getNewCode(len, errChar);
		  }
		  
//System.out.println("return code = " +maxErrorChars);			  
		  return maxErrorChars;
	  }
  }
  
  /* return string of 'errChar's of length='len' or, if errChar=z, then of 
   * length='len'+1.
   */
  private String getNewCode (int len, char errChar) {
	StringBuffer b = new StringBuffer();

	int j = Character.digit(errChar, Character.MAX_RADIX);
	char newChar = Character.toUpperCase(
			Character.forDigit(j + 1, Character.MAX_RADIX));
	
	/* if newChar isn't a letter (errChar=z), then increase str length */
	if (!Character.isLetter(newChar)) { 
		newChar = errChar;
		len++;
	}

	for (int i = 0; i < len; i++) {
		b.append(newChar);
	}
	
	return new String(b);
  }
  
  private String getItem (String code, String value) {
	  return "<Item Code=\"" +code +"\" Value = \"" +value +"\"/>";  
  }
}
