/**
 * This class configures HAPI to take into account custom segments and groups.
 */

package gov.nist.hl7maker.util.hapi;

// JDK
import java.io.File;
import java.io.FileOutputStream;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.FileInputStream;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.IOException;
import java.util.Vector;
import java.util.Observable;
import java.util.jar.JarOutputStream;
import java.util.jar.JarEntry;
import java.util.jar.Manifest;

// Swing
import javax.swing.JLabel;

// XML
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.ParserConfigurationException;

// XSL
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamSource;
import javax.xml.transform.stream.StreamResult;

// DOM
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;
import org.w3c.dom.Node;
import org.w3c.dom.Element;

// SAX
import org.xml.sax.SAXException;

// MessageMaker
import gov.nist.hl7maker.util.MMLog;

public class CustomHAPI extends Observable {

    protected Vector vFileGroup;
    protected Vector vFileSegment;

    private final String SEGMENT_XSL_FILE = "xslt/hapi/javaZ.1.0.xsl";
    private final String GROUP_XSL_FILE = "xslt/hapi/javaGroup.1.0.xsl";

    public CustomHAPI() {
    }

    /**
     * Generates the custom segment (in files) based on the profile.
     * @param xmlProfile The profile
     * @param prefix The prefix of generated files
     */
    public void generateSegment(String xmlProfile, String prefix) throws ParserConfigurationException, SAXException, TransformerConfigurationException, TransformerException, IOException {
	setChanged();
	notifyObservers("Generating custom segments ...");

	// Generate all Groups in a single file using XSL Transformation
	DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
	DocumentBuilder builder = factory.newDocumentBuilder();
	Document doc = builder.parse(new File("file:///" + new File(xmlProfile).getCanonicalPath()).getPath());

	File f = new File("file:///" + new File(SEGMENT_XSL_FILE).getCanonicalPath()); 
	Transformer transformer = TransformerFactory.newInstance().newTransformer(new StreamSource(f.getPath()));
	transformer.setParameter("prefix", prefix);
	DOMSource source = new DOMSource(doc);
	StreamResult result = new StreamResult(new FileOutputStream("allZ.java"));
	transformer.transform(source, result);
	result.getOutputStream().close();

	// Split the single file into several .java file
	vFileSegment = new Vector();
	BufferedWriter bw = null;
	BufferedReader br = new BufferedReader(new FileReader("allZ.java"));
	String line = null;
	while((line = br.readLine()) != null) {
	    int idxSplitHere = line.indexOf("SplitHere");
	    int idxEndSplit = -1;
	    if(idxSplitHere != -1)
		idxEndSplit = line.indexOf("EndSplit");
	    if(idxSplitHere != -1 && idxEndSplit != -1) {
		if(bw != null)
		    bw.close();
		int idxSlash = line.lastIndexOf("/");
		// Create the destination directories if necessary
		new File(line.substring(idxSplitHere+9, idxSlash)).mkdirs();
		if(!vFileSegment.contains(line.substring(idxSplitHere+9, idxEndSplit))) {
		    bw = new BufferedWriter(new FileWriter(line.substring(idxSplitHere+9, idxEndSplit)));
		    vFileSegment.add(line.substring(idxSplitHere+9, idxEndSplit));
		}
		else {
		    bw = null;
		}
	    }
	    else if(bw != null) {
		bw.write(line, 0, line.length());
		bw.newLine();
	    }
	}
	if(bw != null)
	    bw.close();
	br.close();

	// Compile the generated java source file
	for(int i=0;i<vFileSegment.size();i++) {
	    //System.out.println("javac -classpath hapi.jar;commons-logging.jar " + vFileSegment.get(i));
	    //try {
	    setChanged();
	    notifyObservers("Generating custom segments " + " " + vFileSegment.get(i));
	    
	    com.sun.tools.javac.Main javac = new com.sun.tools.javac.Main();
	    String[] argsWin = new String[] {
		"-classpath", ".;lib/hapi/hapi.jar;lib/commons-logging.jar", (String)vFileSegment.get(i)
	    };
	    String[] argsIX = new String[] {
		"-classpath", ".:lib/hapi/hapi.jar:lib/commons-logging.jar", (String)vFileSegment.get(i)
	    };
	    int status = -1;
	    if(System.getProperty("os.name").indexOf("Win") != -1)
		status = javac.compile(argsWin);
	    else
		status = javac.compile(argsIX);
	    switch (status) {
	    case 0: MMLog.log("Compile status: OK " + (String)vFileSegment.get(i)); break;
	    case 1: MMLog.log("Compile status: ERROR"); break;
	    case 2: MMLog.log("Compile status: CMDERR"); break;
	    case 3: MMLog.log("Compile status: SYSERR"); break;
	    case 4: MMLog.log("Compile status: ABNORMAL"); break;
	    default:
		MMLog.log("Compile status: Unknown exit status");
	    }
		
		//Runtime.getRuntime().exec("javac -classpath lib/hapi/hapi.jar;lib/commons-logging.jar " + vFileSegment.get(i)).waitFor();
	    //}
	    /*catch(InterruptedException ie) {
		System.err.println("InterruptedException " + ie.getMessage());
		}*/
	}

	// Create a jar with the compiled source
	Vector v = new Vector();
	for(int i=0;i<vFileSegment.size();i++) {
	    v.add(((String)vFileSegment.get(i)).replaceAll(".java", ".class"));
	}
	//createJAR("lib/CustomZ.jar", v);
    }

    /**
     * Generates the custom group (in files) based on the profile.
     * @param xmlProfile The profile
     * @param prefix The prefix of generated files
     */
    public void generateGroup(String xmlProfile, String prefix) throws ParserConfigurationException, SAXException, TransformerConfigurationException, TransformerException, IOException {
	setChanged();
	notifyObservers("Generating custom groups ...");

	// Generate all Groups in a single file using XSL Transformation
	DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
	DocumentBuilder builder = factory.newDocumentBuilder();
	Document doc = builder.parse(new File("file:///" + new File(xmlProfile).getCanonicalPath()).getPath());
	
	File f = new File("file:///" + new File(GROUP_XSL_FILE).getCanonicalPath()); 
	Transformer transformer = TransformerFactory.newInstance().newTransformer(new StreamSource(f.getPath()));
	transformer.setParameter("prefix", prefix);
	DOMSource source = new DOMSource(doc);
	StreamResult result = new StreamResult(new FileOutputStream("allGroup.java"));
	transformer.transform(source, result);
	result.getOutputStream().close();

	// Split the single file into several .java file
	vFileGroup = new Vector();
	BufferedWriter bw = null;
	BufferedReader br = new BufferedReader(new FileReader("allGroup.java"));
	String line = null;
	while((line = br.readLine()) != null) {
	    int idxSplitHere = line.indexOf("SplitHere");
	    int idxEndSplit = -1;
	    if(idxSplitHere != -1)
		idxEndSplit = line.indexOf("EndSplit");
	    if(idxSplitHere != -1 && idxEndSplit != -1) {
		if(bw != null)
		    bw.close();
		int idxSlash = line.lastIndexOf("/");
		// Create the destination directories if necessary
		new File(line.substring(idxSplitHere+9, idxSlash)).mkdirs();
		bw = new BufferedWriter(new FileWriter(line.substring(idxSplitHere+9, idxEndSplit)));
		vFileGroup.insertElementAt(line.substring(idxSplitHere+9, idxEndSplit), 0);
	    }
	    else if(bw != null) {
		bw.write(line, 0, line.length());
		bw.newLine();
	    }
	}
	if(bw != null)
	    bw.close();
	br.close();

	// Compile the generated java source file
	for(int i=0;i<vFileGroup.size();i++) {
	    //System.out.println("javac -classpath .;hapi.jar;commons-logging.jar;CustomZ.jar " + vFileGroup.get(i));
	    //try {
		setChanged();
		notifyObservers("Generating custom segments " + " " + vFileGroup.get(i));

	    com.sun.tools.javac.Main javac = new com.sun.tools.javac.Main();
	    String[] argsWin = new String[] {
		"-classpath", ".;lib/hapi/hapi.jar;lib/commons-logging.jar", (String)vFileGroup.get(i)
	    };
	    String[] argsIX = new String[] {
		"-classpath", ".:lib/hapi/hapi.jar:lib/commons-logging.jar", (String)vFileGroup.get(i)
	    };
	    int status = -1;
	    if(System.getProperty("os.name").indexOf("Win") != -1)
		status = javac.compile(argsWin);
	    else
		status = javac.compile(argsIX);
	    switch (status) {
	    case 0: MMLog.log("Compile status: OK " + (String)vFileGroup.get(i)); break;
	    case 1: MMLog.log("Compile status: ERROR"); break;
	    case 2: MMLog.log("Compile status: CMDERR"); break;
	    case 3: MMLog.log("Compile status: SYSERR"); break;
	    case 4: MMLog.log("Compile status: ABNORMAL"); break;
	    default:
		MMLog.log("Compile status: Unknown exit status");
	    }


		//Runtime.getRuntime().exec("javac -classpath .;lib/hapi/hapi.jar;lib/commons-logging.jar;lib/CustomZ.jar " + vFileGroup.get(i)).waitFor();
	    /*}
	    catch(InterruptedException ie) {
		System.err.println("InterruptedException " + ie.getMessage());
		}*/
	}

	// Create a jar with the compiled source
	Vector v = new Vector();
	for(int i=0;i<vFileGroup.size();i++) {
	    v.add(((String)vFileGroup.get(i)).replaceAll(".java", ".class"));
	}
	//createJAR("lib/CustomGroup.jar", v);
	setChanged();
	notifyObservers(this);
    }

    /**
     * Generate custom segments and groups
     * @param xmlProfile The profile
     */
    public void generate(String xmlProfile) throws ParserConfigurationException, SAXException, TransformerConfigurationException, TransformerException, IOException {
	// Prefix
	String profileName = new File(xmlProfile).getName();

	generateSegment(xmlProfile, profileName.substring(0, profileName.length()-4).replaceAll("_", "XX").replaceAll(" ","XX"));
	generateGroup(xmlProfile, profileName.substring(0, profileName.length()-4).replaceAll("_", "XX").replaceAll(" ", "XX"));
    }

    /**
     * Clean all generated files
     */
    public void clean() {
	new File("allZ.java").delete();
	new File("allGroup.java").delete();
	File ca = new File("ca");
	clean(ca);
    }
    
    /**
     * Clean recursively the ca directory
     * @param The current file
     */
    private void clean(File node) {
	if(node.isDirectory()) {
	    File[] c = node.listFiles();
	    for(int i=0;i<c.length;i++)
		clean(c[i]);
	    node.delete();
	}
	// It's a file
	else {
	    node.delete();
	}
    }

    /**
     * Create a JAR file
     * @param jarName The name of the jar file
     * @param vFile The list of files in a Vector
     */
    public void createJAR(String jarName, Vector vFile) throws IOException {
	setChanged();
	notifyObservers("Creating " + jarName + " ...");
	JarOutputStream jar = new JarOutputStream(new FileOutputStream(jarName), new Manifest());
	
	// Allocate a buffer for reading the input files.
	byte[] buffer = new byte[1024];
	int bytesRead;
	
	    // Loop through the file names
	for(int i=0;i<vFile.size();i++) {
	    String fileName = (String)vFile.get(i);
	    
	    // Open the file.
	    FileInputStream file = new FileInputStream(fileName);
	    
	    // Create a jar entry and add it to the jar.
	    JarEntry entry = new JarEntry(fileName);
	    jar.putNextEntry(entry);

	    // Read the file the file and write it to the jar.
	    while((bytesRead = file.read(buffer)) != -1) {
		jar.write(buffer, 0, bytesRead);
	    }
	    file.close();
	}
	jar.close();
    }

    /**
     * Modify the XML Document to make it compliant with the customized HAPI
     * @param doc The XML document
     * @param prefix The prefix used to customized HAPI
     */
    public static Document modifyDocument(Document doc, String prefix) {
	// Modify the root
	Element newRoot = (Element)doc.createElement(prefix + "CUSTOM");
	NodeList nl = (doc.getDocumentElement().getChildNodes());
	for(int i=0;i<nl.getLength();i++) {
	    Node n = (Node)nl.item(i);
	    if(n.getNodeType() == Node.ELEMENT_NODE)
		newRoot.appendChild(doc.importNode(n, true));
	}
	doc.replaceChild(newRoot, doc.getDocumentElement());
	modifyDocument(doc, newRoot, prefix);

	return doc;
    }

    private static void modifyDocument(Document doc, Node n, String prefix) {
	// Apply to children
	NodeList nl = n.getChildNodes();
	for(int i=0;i<nl.getLength();i++) {
	    modifyDocument(doc, nl.item(i), prefix);
	}

	if(n.getNodeType() == Node.ELEMENT_NODE) {
	    String nodeName = n.getNodeName();
	    if(nodeName.startsWith("Z")) { 
		Element newZ = (Element)doc.createElement(prefix + nodeName);
		nl = n.getChildNodes();
		for(int i=0;i<nl.getLength();i++) {
		    newZ.appendChild(doc.importNode(nl.item(i), true));
		}
		n.getParentNode().replaceChild(newZ, n);
	    }
	    else {
		int idxDot = nodeName.indexOf(".");
		boolean segGroup = false;
		try {
		    Integer.parseInt(nodeName.substring(idxDot+1));
		}
		catch(NumberFormatException ne) {
		    segGroup = true;
		}
		if(idxDot != -1 && segGroup) {
		    Element newG = (Element)doc.createElement(prefix + nodeName.substring(idxDot+1));
		    nl = n.getChildNodes();
		    for(int i=0;i<nl.getLength();i++) {
			newG.appendChild(doc.importNode(nl.item(i), true));
		    }
		    n.getParentNode().replaceChild(newG, n);
		}
	    }
	}
    }

    /**
     * Clean the customized ER7 message
     * @param er7 The ER7 String
     * @param prefix The prefix used to customized HAPI
     * @return The cleaned ER7String
     */
    public static String cleanER7(String er7, String prefix) {
	Vector vLine = new Vector();
	StringWriter sw = new StringWriter();
	try {
	    BufferedReader br = new BufferedReader(new StringReader(er7));
	    BufferedWriter bw = new BufferedWriter(sw);
	    String line;
	    
	    while((line = br.readLine()) != null) {
		//		if(!vLine.contains(line)) {
		    if(line.startsWith(prefix)) bw.write(line.substring(prefix.length()));
		    else bw.write(line);
		    bw.write("\n");
		    vLine.add(line);
		    //		}
	    }
	    br.close();
	    bw.close();
	}
	catch(IOException ioe) {
	    System.err.println("IOException " + ioe.getMessage());
	    ioe.printStackTrace();
	}
	return sw.toString();
    }

    /**
     * Get the ER7 representation of a batch file
     * @param batchSeg The ER7 representation of batch segments (BHS, BTS, FHS, FTS)
     * @param msgSeg An array of ER7 representations of segments and messages
     * @return The ER7 representation
     */
    public String getBatchER7(String batchSeg, String[] msgSeg) {
	String er7 = "";
	try {
	    // Creat an array of String based on batchSeg
	    String[] bSeg = new String[50];
	    BufferedReader br = new BufferedReader(new StringReader(batchSeg));
	    int j = 0;
	    String line = "";
	    while((line = br.readLine()) != null) {
		if(line != null) {
		    bSeg[j] = line;
		    j++;
		}
	    }
	    
	    for(int i=0;i<msgSeg.length;i++) {
		if(msgSeg[i].length() == 3) {
		    for(int k=0;k<bSeg.length && bSeg[k] != null;k++) {
			if(bSeg[k].substring(0, 3).equals(msgSeg[i])) {
			    er7 += (bSeg[k] + "\n");
			    bSeg[k] = "XXX";
			    k = bSeg.length + 1;
			}
		    }
		}
		else {
		    er7 += (msgSeg[i] + "");
		}
	    }
	}
	catch(IOException ioe) {
	    System.err.println("IOException " + ioe.getMessage());
	    ioe.printStackTrace();
	}
	return er7;
    }

    /**
     * Does this message needs customization of HAPI?
     * @param profile The profile
     * @return true or false
     */
    public static boolean needCustom(String profile) {
	try {
	    DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
	    DocumentBuilder builder = factory.newDocumentBuilder();
	    Document doc = builder.parse(new File(profile));
	    NodeList nl = doc.getDocumentElement().getChildNodes();
	    for(int i=0;i<nl.getLength();i++) {
		Node n = nl.item(i);
		if(n.getNodeType() == Node.ELEMENT_NODE) {
		    if(n.getNodeName().equals("HL7v2xStaticDef") || n.getNodeName().equals("Message")) {
			NodeList nl1 = n.getChildNodes();
			for(int j=0;j<nl1.getLength();j++) {
			    Node n1 = nl1.item(j);
			    if(n1.getNodeType() == Node.ELEMENT_NODE) {
				if(n1.getNodeName().equals("Segment") && n1.getAttributes().getNamedItem("Name").getNodeValue().startsWith("Z")) {
				    return true;
				}
				else if(n1.getNodeName().equals("SegGroup")) {
				    return true;
				}
			    }
			}
		    }
		}
	    }
	}
	catch(ParserConfigurationException pce) {
	    System.err.println("ParserConfigurationException " + pce.getMessage());
	    pce.printStackTrace();
	}
	catch(SAXException saxe) {
	    System.err.println("SAXException " + saxe.getMessage());
	    saxe.printStackTrace();
	}
	catch(IOException ioe) {
	    System.err.println("IOException " + ioe.getMessage());
	    ioe.printStackTrace();
	}
	return false;
    }
}
