package gov.nist.hl7maker.verifier;

import gov.nist.hl7maker.verifier.hl7.HAPIValidator;
import gov.nist.hl7maker.verifier.html.ReportWriter;
import gov.nist.hl7maker.verifier.message.BasicMessageVerifier;
import gov.nist.hl7maker.verifier.profile.ProfileInspector;
import gov.nist.hl7maker.verifier.xmlverifier.DefaultXMLVerifier;
import gov.nist.hl7maker.verifier.xmlverifier.XMLVerifier;

import java.io.File;

import org.w3c.dom.Document;

/**
 * Common framework class for the creation of message verifiers. Provides
 * support for HAPI validation and message-to-profile verification.
 * 
 * @author Justin Manweiler
 */
abstract class AbstractVerifierFactory {

    // --- Private Fields ---
    
    private final File profileFile;
    
    private HAPIValidator hValidator = null;
    private XMLVerifier pVerifier = null;
    private ProfileInspector profileInspector = null;
    private ReportWriter rWriter = null;

    // --- Constructor ---
    
    public AbstractVerifierFactory( final File profile ) {
        profileFile = profile;
    }
    
    // --- Public Methods ---

    public void activateReporting() {
        if ( rWriter == null ) {
            rWriter = new ReportWriter();
        }
    }
    
    /**
     * Provides a XML verifier for a message profile.
     */
    public XMLVerifier getProfileVerifier() throws VerifierException {
        if ( pVerifier == null ) {
            pVerifier = new DefaultXMLVerifier( profileFile, "Profile" );
        }

        return pVerifier;
    }
    
    public boolean isReportingActive() {
        return rWriter != null;
    }
    
    public ReportWriter getReportWriter() {
        if ( rWriter == null ) {
            throw new IllegalStateException( "Reporting not enabled" );
        }
        
        return rWriter;
    }

    // --- Protected Methods ---

    /**
     * Provides a <code>HAPIValidator</code> to be shared between message
     * verifiers.
     */
    protected HAPIValidator getHAPIValidator() throws VerifierException {
        if ( hValidator == null ) {
            try {
                final String profXML = getProfileVerifier()
                        .getVerifiedContents();
                hValidator = new HAPIValidator( profXML );
            }
            catch ( final Exception e ) {
            }
        }

        return hValidator;
    }

    /**
     * Provides a <code>ProfileInspector</code> to be shared between message
     * verifiers.
     */
    protected ProfileInspector getProfileInspector() throws VerifierException {
        if ( profileInspector == null ) {
            final Document profDOM = getProfileVerifier().getDOM();
            profileInspector = new ProfileInspector( profDOM );
        }

        return profileInspector;
    }

    protected void setupMessageVerifier( final BasicMessageVerifier mV )
            throws VerifierException {
        mV.setHAPIValiator( getHAPIValidator() );
        mV.setProfileInspector( getProfileInspector() );
        
        if ( rWriter != null ) {
            mV.setupReporting( rWriter, getProfileVerifier().getDOMCopy() );
        }
    }
}
