package gov.nist.hl7maker.verifier;

import gov.nist.hl7maker.verifier.hl7.EncodingMapper;
import gov.nist.hl7maker.verifier.hl7.TableLookup;
import gov.nist.hl7maker.verifier.message.ExtendedMessageVerifier;
import gov.nist.hl7maker.verifier.metadata.GeneralInfo;
import gov.nist.hl7maker.verifier.metadata.MessageInfo;
import gov.nist.hl7maker.verifier.util.BadXPathException;
import gov.nist.hl7maker.verifier.xmlverifier.DefaultXMLVerifier;
import gov.nist.hl7maker.verifier.xmlverifier.XMLVerifier;

import java.io.File;

import javax.xml.transform.TransformerException;

import org.apache.xpath.XPathAPI;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Extends the <code>AbstractVerifierFactory</code> class to provide support
 * for the verification of implanted errors (possible due to the accessibity of
 * a metadata file).
 * 
 * @author Justin Manweiler
 */
public class FullVerifierFactory extends AbstractVerifierFactory {

    // --- Private Fields ---

    private final File metaFile;
    private final File msgFolderFile;

    private EncodingMapper encMapper = null;
    private XMLVerifier mVerifier = null;
    private XMLVerifier pVerifier = null;
    private GeneralInfo.Properties genProps = null;
    private NodeList msgNodeList = null;
    private TableLookup tableLookup = null;

    // --- Constructor ---

    public FullVerifierFactory( final File metadata, final File profile,
            final File msgFolder ) {
        super( profile );

        metaFile = metadata;
        msgFolderFile = msgFolder;
    }

    // --- Public Methods ---

    /**
     * Add a HL7 table file to the list available for value verification.
     */
    public void addTable( final File tableFile ) {
        getTableLookup()
                .addTable( new DefaultXMLVerifier( tableFile, "Table" ) );
    }

    public XMLVerifier getMetadataVerifier() {
        if ( mVerifier == null ) {
            mVerifier = new DefaultXMLVerifier( metaFile, "Metadata" );
        }

        return mVerifier;
    }

    /**
     * @return the number of message verifiers available.
     */
    public int getNumMessages() throws VerifierException {
        return getMessageNodeList().getLength();
    }

    /**
     * @return a verifier for the message associated with the given index.
     */
    public ExtendedMessageVerifier getMessageVerifier( final int msgIndex )
            throws VerifierException {
        final Node msgNode = getMessageNodeList().item( msgIndex );
        return newMessageVerifier( (Element)msgNode );
    }

    // --- Private Methods ---

    private EncodingMapper getEncodingMapper() throws VerifierException {
        if ( encMapper == null ) {
            encMapper = new EncodingMapper( getProfileVerifier().getDOM() );
        }

        return encMapper;
    }

    private GeneralInfo.Properties getGeneralProperties()
            throws VerifierException {
        if ( genProps == null ) {
            final Document metaDoc = getMetadataVerifier().getDOM();

            final GeneralInfo genInfo = new GeneralInfo();
            genInfo.parse( metaDoc, "/SESSION/" );

            genProps = genInfo.getProperties();
        }

        return genProps;
    }

    private NodeList getMessageNodeList() throws VerifierException {
        if ( msgNodeList == null ) {
            final String xPath = "//Message";
            final Document metaDoc = getMetadataVerifier().getDOM();

            try {
                msgNodeList = XPathAPI.selectNodeList( metaDoc, xPath );
            }
            catch ( TransformerException e ) {
                throw new BadXPathException( xPath, e.getMessage() );
            }
        }

        return msgNodeList;
    }

    private TableLookup getTableLookup() {
        if ( tableLookup == null ) {
            tableLookup = new TableLookup();
        }

        return tableLookup;
    }

    private ExtendedMessageVerifier newMessageVerifier( final Element msgElement )
            throws VerifierException {
        final MessageInfo.Properties mProps = newMsgProperties( msgElement );
        final GeneralInfo.Properties gProps = getGeneralProperties();

        final ExtendedMessageVerifier mV = new ExtendedMessageVerifier(
                msgFolderFile, mProps, gProps );

        setupMessageVerifier( mV );

        mV.setID( msgElement.getAttribute( "SequenceID" ) );
        mV.setTableLookup( getTableLookup() );

        return mV;
    }

    private MessageInfo.Properties newMsgProperties( final Element msgElement )
            throws VerifierException {
        final MessageInfo msgInfo = new MessageInfo( getEncodingMapper() );
        msgInfo.parse( msgElement );

        return msgInfo.getProperties();
    }
}
