package gov.nist.hl7maker.verifier.hl7;

import gov.nist.hl7maker.verifier.util.BadXPathException;

import javax.xml.transform.TransformerException;

import org.apache.xpath.XPathAPI;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * This class converts the description of a message segment, field, component,
 * or subcomponent and provides the XPath location of the element within a
 * message XML file. It allows the information stored within a metadata file
 * describing the location of a message element to be translated into a
 * navigable XPath.
 * 
 * @author Justin Manweiler
 */
public class EncodingMapper {

    // --- Inner Classes ---

    public static class Encoding {
        public final String sgmntEnc;
        public final String fieldEnc;
        public final String cmpntEnc;
        public final String scpntEnc;
        public final String xPath;
        
        private Encoding( final String sgmntEnc, final String fieldEnc,
                final String cmpntEnc, final String scpntEnc ) {
            this.sgmntEnc = sgmntEnc;
            this.fieldEnc = fieldEnc;
            this.cmpntEnc = cmpntEnc;
            this.scpntEnc = scpntEnc;

            xPath = getXPath( sgmntEnc, fieldEnc, cmpntEnc, scpntEnc );
        }
    }

    // --- Public Fields ---

    public static final String baseEnc = "/ADT_A01";

    // --- Private Fields ---

    private final Document profileDOM;

    // --- Constructor ---

    public EncodingMapper( final Document profile ) {
        profileDOM = profile;
    }

    // --- Public Methods ---

    public Encoding getEncoding( final String sgmnt, final String field,
            final String cmpnt, final String scpnt ) {
        final String fieldEnc = getFieldEnc( sgmnt, field );
        
        if ( fieldEnc == null ) {
            return new Encoding( sgmnt, null, null, null );
        }
        
        final String cmpntEnc = getComponentEnc( sgmnt, field, cmpnt );
        
        if ( cmpntEnc == null ) {
            return new Encoding( sgmnt, fieldEnc, null, null );
        }
        
        final String scpntEnc = getSubComponentEnc( sgmnt, field, cmpnt, scpnt );

        return new Encoding( sgmnt, fieldEnc, cmpntEnc, scpntEnc );
    }

    public static String getMessageXPath( final String sgmntEnc,
            final String fieldEnc, final String cmpntEnc, final String scpntEnc ) {
        String xPath = baseEnc;

        if ( sgmntEnc == null ) {
            return xPath;
        }

        xPath += "/" + sgmntEnc;

        if ( fieldEnc == null ) {
            return xPath;
        }

        xPath += "/" + fieldEnc;

        if ( cmpntEnc == null ) {
            return xPath;
        }

        xPath += "/" + cmpntEnc;

        if ( scpntEnc == null ) {
            return xPath;
        }

        return xPath + "/" + scpntEnc;
    }

    // --- Private Methods ---

    private String getSegmentEnc( final String sgmnt ) {
        return sgmnt;
    }

    private String getFieldEnc( final String sgmnt, final String field ) {
        if ( sgmnt == null || field == null ) {
            return null;
        }

        final String xPath = "//" + "Segment[@Name='" + sgmnt + "']";

        try {
            final NodeList nList = XPathAPI.selectNodeList( profileDOM, xPath );
            final int nListLen = nList.getLength();

            for ( int i = 0; i < nListLen; i++ ) {
                final Element element = (Element)nList.item( i );
                final NodeList fieldList = XPathAPI.selectNodeList( element,
                        "Field" );
                final int numFields = fieldList.getLength();

                for ( int j = 0; j < numFields; j++ ) {
                    final Element fieldElement = (Element)fieldList.item( j );

                    if ( fieldElement.getAttribute( "Name" ).equals( field ) ) {
                        return sgmnt + "." + (j + 1);
                    }
                }
            }
        }
        catch ( TransformerException e ) {
            throw new BadXPathException( xPath, e.getMessage() );
        }

        return null;
    }

    private String getFieldEncOLD( final String sgmnt, final String field ) {
        if ( sgmnt == null || field == null ) {
            return null;
        }

        final String xPath = "//" + "Segment[@Name='" + sgmnt + "']";

        try {
            final NodeList nList = XPathAPI.selectNodeList( profileDOM, xPath );
            final int nListLen = nList.getLength();

            for ( int i = 0; i < nListLen; i++ ) {
                final Element element = (Element)nList.item( i );
                final NodeList fieldList = XPathAPI.selectNodeList( element,
                        "Field" );
                final int numFields = fieldList.getLength();

                for ( int j = 0; j < numFields; j++ ) {
                    final Element fieldElement = (Element)fieldList.item( j );

                    if ( fieldElement.getAttribute( "Name" ).equals( field ) ) {
                        return sgmnt + "." + (j + 1);
                    }
                }
            }
        }
        catch ( TransformerException e ) {
            throw new BadXPathException( xPath, e.getMessage() );
        }

        return null;
    }

    private String getComponentEnc( final String sgmnt, final String field,
            final String cmpnt ) {
        if ( sgmnt == null || field == null || cmpnt == null ) {
            return null;
        }

        final String xPath = "//" + "Segment[@Name='" + sgmnt + "']/"
                + "Field[@Name='" + field + "']";

        try {
            final NodeList nList = XPathAPI.selectNodeList( profileDOM, xPath );
            final int nListLen = nList.getLength();

            for ( int i = 0; i < nListLen; i++ ) {
                final Element element = (Element)nList.item( i );
                final String dataType = element.getAttribute( "Datatype" );

                final NodeList componentList = XPathAPI.selectNodeList(
                        element, "Component" );

                final int numComps = componentList.getLength();
                for ( int j = 0; j < numComps; j++ ) {
                    final Element compElement = (Element)componentList.item( j );

                    if ( compElement.getAttribute( "Name" ).equals( cmpnt ) ) {
                        return dataType + "." + (j + 1);
                    }
                }
            }
        }
        catch ( TransformerException e ) {
            throw new BadXPathException( xPath, e.getMessage() );
        }

        return null;
    }

    private String getSubComponentEnc( final String sgmnt, final String field,
            final String cmpnt, final String scpnt ) {
        if ( sgmnt == null || field == null || cmpnt == null || scpnt == null ) {
            return null;
        }

        final String xPath = "//" + "Segment[@Name='" + sgmnt + "']/"
                + "Field[@Name='" + field + "']/" + "Component[@Name='" + cmpnt
                + "']";

        try {
            final NodeList nList = XPathAPI.selectNodeList( profileDOM, xPath );
            final int nListLen = nList.getLength();

            for ( int i = 0; i < nListLen; i++ ) {
                final Element element = (Element)nList.item( i );
                final String dataType = element.getAttribute( "Datatype" );

                final NodeList subCompList = XPathAPI.selectNodeList( element,
                        "SubComponent" );

                final int numSubComps = subCompList.getLength();
                for ( int j = 0; j < numSubComps; j++ ) {
                    final Element subCompElement = (Element)subCompList
                            .item( j );

                    if ( subCompElement.getAttribute( "Name" ).equals( scpnt ) ) {
                        return dataType + "." + (j + 1);
                    }
                }
            }
        }
        catch ( TransformerException e ) {
            throw new BadXPathException( xPath, e.getMessage() );
        }

        return null;
    }

    private static String getXPath( final String sgmntEnc,
            final String fieldEnc, final String cmpntEnc, final String scpntEnc ) {
        if ( sgmntEnc == null ) {
            return null;
        }

        String xPath = baseEnc + "/" + sgmntEnc;

        if ( fieldEnc == null ) {
            return xPath;
        }

        xPath += "/" + fieldEnc;

        if ( cmpntEnc == null ) {
            return xPath;
        }

        xPath += "/" + cmpntEnc;

        if ( scpntEnc == null ) {
            return xPath;
        }

        return xPath + "/" + scpntEnc;
    }
}
