package gov.nist.hl7maker.verifier.hl7;

import gov.nist.hl7maker.verifier.profile.ProfileInspector;
import gov.nist.hl7maker.verifier.util.RedirectOutput;

import org.w3c.dom.Document;

import ca.uhn.hl7v2.HL7Exception;
import ca.uhn.hl7v2.conf.check.DefaultValidator;
import ca.uhn.hl7v2.conf.parser.ProfileParser;
import ca.uhn.hl7v2.conf.spec.message.StaticDef;
import ca.uhn.hl7v2.model.Message;
import ca.uhn.hl7v2.parser.DefaultXMLParser;
import ca.uhn.hl7v2.parser.XMLParser;

/**
 * Wrapper class for using the HAPI validation API. Additionally, provides
 * support for the parsing of location information from a HAPI error message.
 * 
 * @author Justin Manweiler
 */
public class HAPIValidator {

    // --- Private Data Members ---

    private final StaticDef staticDef;
    private static final String HL7_VERSION = "2.4";
    private static final ProfileParser pparser = new ProfileParser( false );
    private static final DefaultValidator validator = new DefaultValidator();
    private static final XMLParser xmlParser = new DefaultXMLParser();

    // --- Constructor ---

    public HAPIValidator( final String xmlProfile ) throws Exception {

        /*
         * HAPI doesn't use its log4j logger correctly which produces the
         * following error messages:
         * 
         * 1) log4j:WARN No appenders could be found for logger
         * (ca.uhn.hl7v2.conf.parser.ProfileParser).
         * 
         * 2) log4j:WARN Please initialize the log4j system properly.
         * 
         * The I/O redirection code is in place strictly to hide these.
         */

        RedirectOutput.discardStdErr();

        try {
            staticDef = pparser.parse( xmlProfile ).getMessage();
        }
        finally {
            RedirectOutput.restoreStdErr();
        }
    }

    //  --- Public Methods ---

    public HL7Exception[] getErrors( final Document msgDOM ) throws Exception {
        final Message msg = HAPIValidator.convertDOMtoHAPI( msgDOM );
        return validator.validate( msg, staticDef );
    }

    public static String getMessageXPath( final String hapiError ) {
        final String[] substrings = hapiError.split( "\\s" );

        String xPath = EncodingMapper.baseEnc;

        final String sgmntName = getNextString( substrings, "Segment" );

        if ( sgmntName == null ) {
            return xPath;
        }

        xPath += "/" + sgmntName;

        final String fieldIndex = getNextString( substrings, "Field" );

        if ( fieldIndex == null ) {
            return xPath;
        }

        return xPath + "/" + sgmntName + "." + getNumString( fieldIndex );
    }
    
    public static String getProfileXPath( final String hapiError ) {
        final String[] substrings = hapiError.split( "\\s" );
        
        final String sgmntName = getNextString( substrings, "Segment" );
        final String fieldIndexString = getNextString( substrings, "Field" );
        
        int fieldIndexNum = -1;
        
        if ( fieldIndexString != null ) {
            try {
                fieldIndexNum = Integer.parseInt( getNumString( fieldIndexString ) );
            }
            catch ( final NumberFormatException e ) {
            }
        }
        
        return ProfileInspector.getProfileXPath( sgmntName, fieldIndexNum );
    }

    // --- Private Methods ---

    private static String getNextString( final String[] strings,
            final String key ) {
        for ( int i = 0; i < strings.length - 1; i++ ) {
            if ( strings[i].startsWith( key ) ) {
                return strings[i + 1];
            }
        }

        return null;
    }

    private static String getNumString( final String origString ) {
        final char[] origChars = origString.toCharArray();
        final char[] transChars = new char[origChars.length];

        int j = 0;

        for ( int i = 0; i < origChars.length; i++ ) {
            if ( Character.isDigit( origChars[i] ) ) {
                transChars[j++] = origChars[i];
            }
        }

        final char[] finalChars = new char[j];

        for ( int i = 0; i < j; i++ ) {
            finalChars[i] = transChars[i];
        }

        return new String( finalChars );
    }

    private static int getInt( final String string ) {
        return 0;
    }

    private static Message convertDOMtoHAPI( final Document doc )
            throws HL7Exception {
        return xmlParser.parseDocument( doc, HL7_VERSION );
    }
}
