package gov.nist.hl7maker.verifier.html;

import gov.nist.hl7maker.verifier.util.FileUtils;
import gov.nist.hl7maker.verifier.xmlverifier.XMLParser;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

public class ReportWriter {

    // --- Private Classes ---

    private static class ReportElement {
        public final String msgName;
        public final Document msgDOM;
        public final Document prfDOM;

        public ReportElement( final String msgName, final Document msgDOM,
                final Document prfDOM ) {
            this.msgName = msgName;
            this.msgDOM = msgDOM;
            this.prfDOM = prfDOM;
        }
    }

    // --- Private Fields ---
    
    private static final String boolT = "true";
    private static final String boolF = "";
    private static final String cssFileName = "style.css";
    private static final String xsltIndexFileName = "create_index.xml";
    private static final String xsltProfHTMLFileName = "profile_to_html.xml";
    private static final String xsltProfAbbrevFileName = "profile_to_html_abbrev.xml";
    private static final String xsltProfFullFileName = "profile_to_html_full.xml";

    private final Collection reportElements = new ArrayList();

    private boolean annXMLEnabled = true;
    private boolean abbrEnabled = true;
    private boolean fullEnabled = true;
    private boolean zipEnabled = true;
    private XSLTransformer indexXformer = null;
    private XSLTransformer profXformer = null;

    // --- Public Methods ---

    public void addMessage( final String msgName, final Document msgDOM,
            final Document profDOM ) {

        final ReportElement re = new ReportElement( msgName, msgDOM, profDOM );
        reportElements.add( re );
    }

    public void createReport( final File outputFolder ) {
        if ( outputFolder == null ) {
            throw new IllegalStateException( "Output folder may not be null." );
        }

        outputFolder.mkdirs();

        try {
            ZipOutputStream zipStream = null;

            if ( zipEnabled ) {
                final File zipFile = new File( outputFolder, "reports.zip" );
                zipStream = FileUtils.createZip( zipFile );
            }

            File msgFolder = null;
            File prfFolder = null;

            if ( abbrEnabled || fullEnabled ) {
                prfFolder = new File( outputFolder, "html" + File.separator );
                prfFolder.mkdirs();

                final File cssFile = new File( prfFolder, cssFileName );
                copyCSS( cssFile );

                if ( zipEnabled ) {
                    FileUtils.addToZip( zipStream, cssFile, new ZipEntry(
                            "html" + File.separator + cssFileName ) );
                }
                
                final XSLTransformer xformer = getProfileTransformer();
                xformer.setParam( "linkAnnXML", annXMLEnabled ? boolT : boolF );
            }

            if ( annXMLEnabled ) {
                msgFolder = new File( outputFolder, "messages" + File.separator );
                msgFolder.mkdirs();
            }

            final Iterator iter = reportElements.iterator();

            while ( iter.hasNext() ) {
                final ReportElement re = (ReportElement)iter.next();

                if ( annXMLEnabled ) {
                    final File msgFile = new File( msgFolder, re.msgName );
                    writeMessageReport( re.msgDOM, msgFile );

                    if ( zipEnabled ) {
                        FileUtils.addToZip( zipStream, msgFile, new ZipEntry(
                                "messages" + File.separator + msgFile.getName() ) );
                    }
                }

                if ( abbrEnabled ) {
                    final File prfAFile = new File( prfFolder, re.msgName
                            + ".a.html" );
                    
                    final XSLTransformer xformer = getProfileTransformer();
                    xformer.setParam( "isFull", boolF );
                    xformer.setParam( "linkOppDetail", fullEnabled ? boolT : boolF );
                    writeProfileReport( xformer, re.prfDOM, prfAFile );

                    if ( zipEnabled ) {
                        FileUtils.addToZip( zipStream, prfAFile, new ZipEntry(
                                "html" + File.separator + prfAFile.getName() ) );
                    }
                }

                if ( fullEnabled ) {
                    final File prfFFile = new File( prfFolder, re.msgName
                            + ".f.html" );
                    
                    final XSLTransformer xformer = getProfileTransformer();
                    xformer.setParam( "isFull", "true" );
                    xformer.setParam( "linkOppDetail", abbrEnabled ? boolT : boolF );
                    writeProfileReport( getProfileTransformer(), re.prfDOM,
                            prfFFile );

                    if ( zipEnabled ) {
                        FileUtils.addToZip( zipStream, prfFFile, new ZipEntry(
                                "html" + File.separator + prfFFile.getName() ) );
                    }
                }
            }

            final File indexFile = new File( outputFolder, "index.html" );
            writeIndex( indexFile );

            if ( zipEnabled ) {
                FileUtils.addToZip( zipStream, indexFile, new ZipEntry(
                        "index.html" ) );

                zipStream.close();
            }
        }
        catch ( final IOException e ) {
            e.printStackTrace();
        }
        catch ( final TransformerConfigurationException e ) {
            e.printStackTrace();
        }
    }

    public void setAnnXMLEnabled( final boolean enabled ) {
        annXMLEnabled = enabled;
    }

    public void setAbbrEnabled( final boolean enabled ) {
        abbrEnabled = enabled;
    }

    public void setFullEnabled( final boolean enabled ) {
        fullEnabled = enabled;
    }

    public void setZipEnabled( final boolean enabled ) {
        zipEnabled = enabled;
    }

    // --- Private Methods ---

    private void copyCSS( final File destFile ) {
        try {
            final InputStream is = getClass().getResourceAsStream( cssFileName );
            FileUtils.copyFile( is, destFile );
        }
        catch ( final Exception e ) {
        }
    }

    private XSLTransformer getProfileTransformer()
            throws TransformerConfigurationException {
        if ( profXformer == null ) {
            profXformer = new XSLTransformer();
            profXformer.setXSLT( getClass().getResourceAsStream(
                    xsltProfHTMLFileName ) );
        }

        return profXformer;
    }

    private XSLTransformer getIndexTransformer()
            throws TransformerConfigurationException {
        if ( indexXformer == null ) {
            indexXformer = new XSLTransformer();
            indexXformer.setXSLT( getClass().getResourceAsStream(
                    xsltIndexFileName ) );
        }

        return indexXformer;
    }

    private Document getIndexDOM() {
        Document indexDOM = null;

        try {
            final DocumentBuilderFactory factory = DocumentBuilderFactory
                    .newInstance();
            final DocumentBuilder builder = factory.newDocumentBuilder();

            indexDOM = builder.newDocument();

            final Element root = indexDOM.createElement( "IndexData" );
            indexDOM.appendChild( root );

            final Iterator iter = reportElements.iterator();

            while ( iter.hasNext() ) {
                final String msgName = ((ReportElement)iter.next()).msgName;
                final Element child = indexDOM.createElement( "MessageFile" );
                child.setAttribute( "name", msgName );

                root.appendChild( child );
            }

        }
        catch ( final ParserConfigurationException e ) {
        }

        return indexDOM;
    }

    private void writeIndex( final File outFile ) {
        try {
            final File inFile = File.createTempFile( "temp", ".xml" );
            XMLParser.writeXML( getIndexDOM(), inFile );

            final XSLTransformer xformer = getIndexTransformer();
            xformer.setParam( "genFull", fullEnabled ? boolT : boolF );
            xformer.setParam( "genAbbr", abbrEnabled ? boolT : boolF );
            xformer.setParam( "genAnnXML", annXMLEnabled ? boolT : boolF );
            xformer.transform( inFile, outFile );

            inFile.delete();
        }
        catch ( final Exception e ) {
            e.printStackTrace();
        }
    }

    private void writeMessageReport( final Document msgDOM, final File outFile ) {
        XMLParser.writeXML( msgDOM, outFile );
    }

    private void writeProfileReport( final XSLTransformer xformer,
            final Document prfDOM, final File outFile ) {
        try {
            final File inFile = File.createTempFile( "temp", ".xml" );
            XMLParser.writeXML( prfDOM, inFile );

            xformer.transform( inFile, outFile );

            inFile.delete();
        }
        catch ( final Exception e ) {
            e.printStackTrace();
        }
    }
}
