package gov.nist.hl7maker.verifier.message;

import gov.nist.hl7maker.verifier.VerifierException;
import gov.nist.hl7maker.verifier.hl7.HAPIValidator;
import gov.nist.hl7maker.verifier.html.ReportWriter;
import gov.nist.hl7maker.verifier.profile.ProfileElement;
import gov.nist.hl7maker.verifier.profile.ProfileInspector;
import gov.nist.hl7maker.verifier.util.ErrorSet;
import gov.nist.hl7maker.verifier.xmlverifier.AbstractXMLVerifier;

import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;

import javax.xml.transform.TransformerException;

import org.apache.xpath.XPathAPI;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * Verifies a single message file to information specified in a profile. This
 * class is used in the absence of metadata file information.
 * 
 * @author Justin Manweiler
 */
public class BasicMessageVerifier extends AbstractXMLVerifier {

    // --- Private Fields ---

    private final File file;

    private boolean hapiSet = false;
    private int nextErrID = 1;
    private Document profReportDOM = null;
    private Document reportDOM = null;
    private HAPIValidator hValidator = null;
    private MessageInspector msgInspector = null;
    private ProfileInspector profInspector = null;

    // --- Constructor ---

    public BasicMessageVerifier( final File msgFile ) {
        file = msgFile;
    }

    // --- Public Methods ---

    public File getFile() {
        return file;
    }

    public void setupReporting( final ReportWriter rw, final Document profileDOM )
            throws VerifierException {
        profReportDOM = profileDOM;

        final Element reportElement = profReportDOM
                .createElement( "ReportInfo" );
        reportElement.setAttribute( "msgFileName", file.getName() );
        reportElement.setAttribute( "time", "" + new Date() );

        profReportDOM.getDocumentElement().appendChild( reportElement );

        rw.addMessage( file.getName(), getReportDOM(), profileDOM );
    }

    public void setHAPIValiator( final HAPIValidator validator ) {
        hapiSet = true;
        hValidator = validator;
    }

    public void setProfileInspector( final ProfileInspector inspector ) {
        profInspector = inspector;
    }

    public VerifierException[] validateAgainstProfile()
            throws VerifierException {
        final ProfileElement[] elements = getProfileInspector()
                .getProfileElements();
        final ArrayList arrList = new ArrayList();

        for ( int i = 0; i < elements.length; i++ ) {
            final String msgXPath = elements[i].getMessageXPath();
            final int matchCnt = getMessageInspector()
                    .getCardinality( msgXPath );
            final ErrorSet errSet = MessageProfileValidator.validateErrorsOnly(
                    elements[i], matchCnt );

            if ( errSet.hasError() ) {
                final String[] errMsgs = errSet.getErrMessages();
                final String errString = "Error @ " + msgXPath;

                final VerifierException error = newPEVerifierException(
                        errString, elements[i], errMsgs, matchCnt );

                arrList.add( error );
                addErrorToReport( elements[i], error );
            }

            storeDataValuesInReport( msgXPath, elements[i].getProfileXPath() );
        }

        if ( arrList.isEmpty() ) {
            return null;
        }

        final VerifierException[] vEs = new VerifierException[arrList.size()];
        arrList.toArray( vEs );
        return vEs;
    }

    public void testHL7() throws VerifierException {
        final HAPIValidator hValidator = getHAPIValidator();

        final Exception[] errors;

        try {
            errors = hValidator.getErrors( getDOM() );
        }
        catch ( final Exception e ) {
            throw new VerifierException( "HAPI validator fatal error: "
                    + e.getClass().getName() );
        }

        if ( errors.length > 0 ) {
            final String[] errArr = new String[errors.length];

            for ( int i = 0; i < errors.length; i++ ) {
                errArr[i] = errors[i].getMessage();

                if ( profReportDOM != null ) {
                    final VerifierException ve = new VerifierException(
                            errArr[i] );
                    addErrorToMessageReport( HAPIValidator
                            .getMessageXPath( errArr[i] ), ve );
                    addErrorToProfileReport( HAPIValidator
                            .getProfileXPath( errArr[i] ), ve );
                }
            }

            throw new VerifierException( "Message failed HAPI validation.",
                    errArr );
        }
    }

    // --- Protected Methods ---

    protected void addErrorToReport( final ProfileElement location,
            final VerifierException error ) throws VerifierException {

        if ( profReportDOM != null ) {
            addErrorToMessageReport( location.getMessageXPath(), error );
            addErrorToProfileReport( location.getProfileXPath(), error );
        }
    }

    protected void addCommentToMessageReport( final String xPath,
            final String comment ) throws VerifierException {
        if ( profReportDOM != null ) {
            final Document reportDOM = getReportDOM();
            final Node child = reportDOM.createComment( comment );

            try {
                Node parent = null;
                String addXPath = xPath;

                do {
                    parent = XPathAPI.selectSingleNode( reportDOM, addXPath );

                    if ( parent != null ) {
                        break;
                    }

                    addXPath = getXPathParent( addXPath );
                }
                while ( addXPath != null );

                if ( parent == null ) {
                    reportDOM.appendChild( child );
                }
                else {
                    parent.appendChild( child );
                }
            }
            catch ( final TransformerException e ) {
                e.printStackTrace();
            }
        }
    }

    protected String getFileDescription() {
        return "Message";
    }

    protected HAPIValidator getHAPIValidator() throws VerifierException {
        if ( hValidator == null ) {
            if ( hapiSet ) {
                throw new VerifierException( "A HAPI validator could not be "
                        + "instantiated with the provided profile." );
            }
            else {
                throw new VerifierException( "HAPI validator not avaliable." );
            }
        }

        return hValidator;
    }

    protected MessageInspector getMessageInspector() throws VerifierException {
        if ( msgInspector == null ) {
            msgInspector = new MessageInspector( getDOM() );
        }

        return msgInspector;
    }

    protected ProfileInspector getProfileInspector() throws VerifierException {
        if ( profInspector == null ) {
            throw new VerifierException( "Profile inspector not available." );
        }

        return profInspector;
    }
    
    protected VerifierException newPEVerifierException( final String errMsg,
            final ProfileElement element, final String[] errMsgs,
            final int cardinality ) {
        return newPEVerifierException( errMsg, element, errMsgs, null, cardinality );
    }

    protected VerifierException newPEVerifierException( final String errMsg,
            final ProfileElement element, final String[] errMsgs, final String[] exDetails,
            final int cardinality ) {

        final ArrayList detailList = new ArrayList();
        
        if ( exDetails != null ) {
            detailList.addAll( Arrays.asList( exDetails ) );
        }

        detailList.add( "Profile location: " + element.getProfileLocDescription() );
        detailList.add( "Actual cardinality: " + cardinality );

        if ( !element.isUsageX() ) {
            if ( element.isMinCardinalitySet() ) {
                detailList.add( "Min cardinality: " + element.getMinCardinality() );
            }

            if ( element.isMaxCardinalitySet() ) {
                detailList.add( "Max cardinality: " + element.getMaxCardinality() );
            }
        }

        detailList.add( "Usage: " + element.getUsageDescription() );

        final String[] msgArr = new String[detailList.size()];
        detailList.toArray( msgArr );

        return new VerifierException( errMsg, msgArr, errMsgs );
    }
    
    protected void storeDataValuesInReport( final String msgXPath,
            final String profXPath ) throws VerifierException {
        if ( profReportDOM != null ) {
            final String[] dataValues = getMessageInspector().getDataValues(
                    msgXPath );

            if ( dataValues != null ) {
                for ( int i = 0; i < dataValues.length; i++ ) {
                    final Element dataElement = profReportDOM.createElement( "MsgDataValue" );
                    dataElement.setAttribute( "value", dataValues[i] );
                    
                    addNodeToProfileReport( dataElement, profXPath );
                }
            }
        }
    }

    // --- Private Methods ---

    private void addErrorToMessageReport( final String xPath,
            final VerifierException error ) throws VerifierException {
        if ( profReportDOM != null ) {
            String comment = error.getMessage();

            final String[] details = error.getDetails();

            if ( details != null ) {
                for ( int i = 0; i < details.length; i++ ) {
                    comment += "\n + " + details[i];
                }
            }

            final String[] faults = error.getFaults();

            if ( faults != null ) {
                for ( int i = 0; i < faults.length; i++ ) {
                    comment += "\n" + faults[i];
                }
            }

            addCommentToMessageReport( xPath, comment );
        }
    }

    private void addErrorToProfileReport( final String profXPath,
            final VerifierException error ) {
        if ( profReportDOM != null ) {
            final Element errElement = profReportDOM.createElement( "Error" );
            errElement.setAttribute( "title", error.getMessage() );
            errElement.setAttribute( "id", "" + nextErrID++ );

            final String[] details = error.getDetails();

            if ( details != null ) {
                for ( int i = 0; i < details.length; i++ ) {
                    final Element detailElement = profReportDOM
                            .createElement( "Detail" );
                    detailElement.setAttribute( "value", details[i] );

                    errElement.appendChild( detailElement );
                }
            }

            final String[] faults = error.getFaults();

            if ( faults != null ) {
                for ( int i = 0; i < faults.length; i++ ) {
                    final Element faultElement = profReportDOM
                            .createElement( "Fault" );
                    faultElement.setAttribute( "value", faults[i] );

                    errElement.appendChild( faultElement );
                }
            }
            
            addNodeToProfileReport( errElement, profXPath );
        }
    }

    private void addNodeToProfileReport( final Node toAdd,
            final String profXPath ) {
        try {
            final Node parent = XPathAPI.selectSingleNode( profReportDOM,
                    profXPath );

            if ( parent == null ) {
                profReportDOM.getDocumentElement().appendChild( toAdd );
            }
            else {
                parent.appendChild( toAdd );
            }
        }
        catch ( final TransformerException e ) {
        }
    }

    private String getXPathParent( final String origXPath ) {
        final int cutIndex = origXPath.lastIndexOf( "/" );

        return cutIndex < 0 ? null : origXPath.substring( 0, cutIndex );
    }

    private Document getReportDOM() throws VerifierException {
        if ( reportDOM == null ) {
            reportDOM = getDOMCopy();
        }

        return reportDOM;
    }
}
