package gov.nist.hl7maker.verifier.message;

import gov.nist.hl7maker.verifier.VerifierException;
import gov.nist.hl7maker.verifier.hl7.EncodingMapper;
import gov.nist.hl7maker.verifier.hl7.TableLookup;
import gov.nist.hl7maker.verifier.metadata.GeneralInfo;
import gov.nist.hl7maker.verifier.metadata.MessageInfo;
import gov.nist.hl7maker.verifier.profile.ExtendedProfileElement;
import gov.nist.hl7maker.verifier.profile.ProfileElement;
import gov.nist.hl7maker.verifier.util.BadXPathException;
import gov.nist.hl7maker.verifier.util.ErrorSet;

import java.io.File;
import java.util.ArrayList;

import javax.xml.transform.TransformerException;

import org.apache.xpath.XPathAPI;
import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Expands the verification abilities of the <code>BasicMessageVerifier</code>
 * with the information stored in a metadata file.
 * 
 * @author Justin Manweiler
 */
public class ExtendedMessageVerifier extends BasicMessageVerifier {

    // --- Private Fields ---

    private final MessageInfo.Properties msgProps;
    private final GeneralInfo.Properties genProps;

    private ExtendedProfileElement pErrorElement = null;
    private String id = null;
    private TableLookup tLookup = null;

    // --- Constructor ---

    public ExtendedMessageVerifier( final File folder,
            final MessageInfo.Properties mProps,
            final GeneralInfo.Properties gProps ) {
        super( getFile( folder, mProps.name ) );

        msgProps = mProps;
        genProps = gProps;
    }

    // --- Public Methods ---

    public String getErrorType() {
        String errDescrip = msgProps.getErrorDescription();
        
        if ( msgProps.encoding.xPath != null ) {
            errDescrip += " @ " + msgProps.encoding.xPath;
        }
        
        return errDescrip;
    }

    public String getID() {
        return id;
    }

    public void setTableLookup( final TableLookup lookup ) {
        tLookup = lookup;
    }

    public void setID( final String value ) {
        id = value;
    }

    public VerifierException[] validateAgainstProfile()
            throws VerifierException {
        final ProfileElement[] elements = getProfileInspector()
                .getProfileElements();
        final String ignoreXPath = msgProps.encoding.xPath;
        final ArrayList arrList = new ArrayList();
        final MessageProfileValidator validator = new MessageProfileValidator(
                genProps );

        for ( int i = 0; i < elements.length; i++ ) {
            final String msgXPath = elements[i].getMessageXPath();

            if ( ignoreXPath == null || !msgXPath.startsWith( ignoreXPath ) ) {
                final int matchCnt = getMessageInspector().getCardinality(
                        msgXPath );
                final ErrorSet errSet = validator.validate( elements[i],
                        matchCnt );

                if ( errSet.hasError() ) {
                    final String[] errMsgs = errSet.getErrMessages();
                    final String errString = "Error @ " + msgXPath;

                    final VerifierException error = newPEVerifierException(
                            errString, elements[i], errMsgs, matchCnt );

                    arrList.add( error );
                    addErrorToReport( elements[i], error );
                }

                storeDataValuesInReport( msgXPath, elements[i]
                        .getProfileXPath() );
            }
        }

        if ( arrList.isEmpty() ) {
            return null;
        }

        final VerifierException[] vEs = new VerifierException[arrList.size()];
        arrList.toArray( vEs );
        return vEs;
    }

    public void testImplantedError() throws VerifierException {
        final String xPath = msgProps.encoding.xPath;

        if ( xPath == null ) {
            return;
        }

        final int actlCard = getMessageInspector().getCardinality( xPath );
        final ExtendedProfileElement pElement = getErrorProfileElement();
        final ErrorSet pErrSet = MessageProfileValidator.validateErrorsOnly(
                pElement, actlCard );
        final ErrorSet iErrSet = new ImplantedErrorSet();

        /* RusageError */

        if ( pErrSet.hasError( ProfileErrorSet.ET_RusageError ) ) {
            if ( !msgProps.hasRusageError ) {
                iErrSet.addError( ImplantedErrorSet.UE_RusageError );
            }
        }
        else if ( msgProps.hasRusageError ) {
            iErrSet.addError( ImplantedErrorSet.ME_RusageError );
        }

        /* XusageError */

        if ( pErrSet.hasError( ProfileErrorSet.ET_XusageError ) ) {
            if ( !msgProps.hasXusageError ) {
                iErrSet.addError( ImplantedErrorSet.UE_XusageError );
            }
        }
        else if ( msgProps.hasXusageError ) {
            iErrSet.addError( ImplantedErrorSet.ME_XusageError );
        }

        /* MinSegCardError */

        if ( pErrSet.hasError( ProfileErrorSet.ET_MinSegCardError ) ) {
            if ( !msgProps.hasMinSegCardError ) {
                iErrSet.addError( ImplantedErrorSet.UE_MinSegCardError );
            }
        }
        else if ( msgProps.hasMinSegCardError ) {
            iErrSet.addError( ImplantedErrorSet.ME_MinSegCardError );
        }

        /* MaxSegCardError */

        if ( pErrSet.hasError( ProfileErrorSet.ET_MaxSegCardError ) ) {
            if ( !msgProps.hasMaxSegCardError ) {
                iErrSet.addError( ImplantedErrorSet.UE_MaxSegCardError );
            }
        }
        else if ( msgProps.hasMaxSegCardError ) {
            iErrSet.addError( ImplantedErrorSet.ME_MaxSegCardError );
        }

        /* MinFieldCardError */

        if ( pErrSet.hasError( ProfileErrorSet.ET_MinFieldCardError ) ) {
            if ( !msgProps.hasMinFieldCardError ) {
                iErrSet.addError( ImplantedErrorSet.UE_MinFieldCardError );
            }
        }
        else if ( msgProps.hasMinFieldCardError ) {
            iErrSet.addError( ImplantedErrorSet.ME_MinFieldCardError );
        }

        /* MaxFieldCardError */

        if ( pErrSet.hasError( ProfileErrorSet.ET_MaxFieldCardError ) ) {
            if ( !msgProps.hasMaxFieldCardError ) {
                iErrSet.addError( ImplantedErrorSet.UE_MaxFieldCardError );
            }
        }
        else if ( msgProps.hasMaxFieldCardError ) {
            iErrSet.addError( ImplantedErrorSet.ME_MaxFieldCardError );
        }

        /* FieldLengthError */

        if ( pElement.isField() ) {
            final int actualLen = getMaxElementLength();
            final int maxLen = pElement.getMaxLength();

            if ( actualLen > maxLen ) {
                if ( !msgProps.hasFieldLengthError ) {
                    iErrSet.addError( ImplantedErrorSet.ME_FieldLengthError );
                }
            }
            else if ( msgProps.hasComponentLengthError ) {
                iErrSet.addError( ImplantedErrorSet.UE_FieldLengthError );
            }
        }

        /* ComponentLengthError */

        if ( pElement.isComponent() ) {
            final int actualLen = getMaxElementLength();
            final int maxLen = pElement.getMaxLength();

            if ( actualLen > maxLen ) {
                if ( !msgProps.hasComponentLengthError ) {
                    iErrSet
                            .addError( ImplantedErrorSet.ME_ComponentLengthError );
                }
            }
            else if ( msgProps.hasComponentLengthError ) {
                iErrSet.addError( ImplantedErrorSet.UE_ComponentLengthError );
            }
        }

        /* SubComponentLengthError */

        if ( pElement.isSubComponent() ) {
            final int actualLen = getMaxElementLength();
            final int maxLen = pElement.getMaxLength();

            if ( actualLen > maxLen ) {
                if ( !msgProps.hasSubComponentLengthError ) {
                    iErrSet
                            .addError( ImplantedErrorSet.UE_SubComponentLengthError );
                }
            }
            else if ( msgProps.hasSubComponentLengthError ) {
                iErrSet.addError( ImplantedErrorSet.ME_SubComponentLengthError );
            }
        }

        /* Set REinclude */

        if ( msgProps.elmntREInclude != null ) {
            if ( msgProps.elmntREInclude.equals( "true" ) ) {
                if ( actlCard < 1 || !pElement.isUsageRE() ) {
                    iErrSet.addError( ImplantedErrorSet.BS_REincludeTrue );
                }
            }
            else if ( msgProps.elmntREInclude.equals( "false" ) ) {
                if ( actlCard > 0 || !pElement.isUsageRE() ) {
                    iErrSet.addError( ImplantedErrorSet.BS_REincludeFalse );
                }
            }
        }

        /* Set Cardinality */

        if ( msgProps.elmntCardinality != null ) {
            try {
                final int intVal = Integer.parseInt( msgProps.elmntCardinality );

                if ( actlCard != intVal ) {
                    iErrSet.addError( ImplantedErrorSet.BS_Cardinality );
                }
            }
            catch ( final NumberFormatException e ) {
            }
        }

        /* Test invalid table value */

        if ( tLookup != null ) {
            final String table = pElement.getTableID();

            if ( table != null ) {
                final NodeList nList = getElementNodes();
                final int nListLen = nList.getLength();

                for ( int i = 0; i < nListLen; i++ ) {
                    final String value = getTextValue( nList.item( i ) );

                    if ( !tLookup.isInTable( table, value ) ) {
                        iErrSet.addError( ImplantedErrorSet.UE_ValueNotInTable );
                    }
                }
            }
        }

        if ( iErrSet.hasError() ) {
            final String details[] = { "Expected element modifier: " + msgProps.getErrorDescription() };
            
            final VerifierException error = newPEVerifierException(
                    "Problem with implanted error, expected @ " + xPath,
                    pElement, iErrSet.getErrMessages(), details, actlCard );

            addErrorToReport( pElement, error );

            throw error;
        }
    }

    public void removeImplantedError() throws VerifierException {
        if ( msgProps.isValid ) {
            return;
        }

        final String xPath = msgProps.encoding.xPath;

        if ( xPath == null ) {
            return;
        }

        final int actualCard = getMessageInspector().getCardinality( xPath );

        if ( msgProps.hasRusageError ) {
            repairAdd( 1 );
        }

        if ( msgProps.hasXusageError ) {
            repairRemove();
        }

        if ( msgProps.hasMinSegCardError || msgProps.hasMinFieldCardError ) {
            final ProfileElement pElement = getErrorProfileElement();
            final int minCard = pElement.getMinCardinality();

            repairAdd( minCard - actualCard );
        }

        if ( msgProps.hasMaxSegCardError || msgProps.hasMaxFieldCardError ) {
            final ProfileElement pElement = getErrorProfileElement();
            final int maxCard = pElement.getMaxCardinality();

            repairRemove( actualCard - maxCard );
        }

        if ( msgProps.hasFieldLengthError || msgProps.hasComponentLengthError
                || msgProps.hasSubComponentLengthError ) {
            final ProfileElement pElement = getErrorProfileElement();
            final int minCard = pElement.getMinCardinality();

            repairRemove();
            repairAdd( minCard );
        }
    }

    // --- Private Methods ---

    private NodeList getElementNodes() throws VerifierException {
        try {
            return XPathAPI.selectNodeList( getDOM(), msgProps.encoding.xPath );
        }
        catch ( final TransformerException e ) {
            throw new BadXPathException( msgProps.encoding.xPath, e
                    .getMessage() );
        }
    }

    private ExtendedProfileElement getErrorProfileElement()
            throws VerifierException {
        if ( pErrorElement == null ) {
            pErrorElement = getProfileInspector().getProfileElement(
                    msgProps.elmntSgmnt, msgProps.elmntField,
                    msgProps.elmntCmpnt, msgProps.elmntScpnt );

            if ( pErrorElement == null ) {
                throw new VerifierException(
                        "Invalid error element specified (not found in profile) - sgmnt=\""
                                + msgProps.elmntSgmnt + "\", field=\""
                                + msgProps.elmntField + "\", cmpnt=\""
                                + msgProps.elmntCmpnt + "\", scpnt=\""
                                + msgProps.elmntScpnt + "\"" );
            }
        }

        return pErrorElement;
    }

    private static File getFile( final File folder, final String name ) {
        if ( name == null ) {
            throw new IllegalStateException(
                    "Message file must have a specified filename" );
        }

        return new File( folder, name );
    }

    private int getMaxElementLength() throws VerifierException {
        final NodeList nList = getElementNodes();
        final int numNodes = nList.getLength();
        int maxLen = 0;

        for ( int i = 0; i < numNodes; i++ ) {
            final int len = getLength( nList.item( i ) );

            if ( len > maxLen ) {
                maxLen = len;
            }
        }

        return maxLen;
    }

    private static int getLength( final Node node ) {
        final NodeList childNodes = node.getChildNodes();
        final int numChildren = childNodes.getLength();

        int length = 0;

        for ( int i = 0; i < numChildren; i++ ) {
            final Node child = childNodes.item( i );
            final short type = child.getNodeType();

            if ( type == Node.TEXT_NODE ) {
                length += child.getNodeValue().trim().length();
            }
            else if ( type == Node.ELEMENT_NODE ) {
                length += getLength( child );
            }
        }

        return length;
    }

    private static String getTextValue( final Node node ) {
        final NodeList childNodes = node.getChildNodes();
        final int numChildren = childNodes.getLength();

        String text = "";

        for ( int i = 0; i < numChildren; i++ ) {
            final Node child = childNodes.item( i );
            final short type = child.getNodeType();

            if ( type == Node.TEXT_NODE ) {
                text += child.getNodeValue().trim();
            }
        }

        return text;
    }

    private void repairAdd( final int copies ) throws VerifierException {
        final String newElementEnc;

        boolean incSgmnt = false;
        boolean incField = false;
        boolean incCmpnt = false;

        if ( msgProps.encoding.scpntEnc == null ) {
            if ( msgProps.encoding.cmpntEnc == null ) {
                if ( msgProps.encoding.fieldEnc == null ) {
                    newElementEnc = msgProps.encoding.sgmntEnc;
                }
                else {
                    newElementEnc = msgProps.encoding.fieldEnc;
                    incSgmnt = true;
                }
            }
            else {
                newElementEnc = msgProps.elmntCmpnt;
                incSgmnt = incField = true;
            }
        }
        else {
            newElementEnc = msgProps.elmntScpnt;
            incSgmnt = incField = incCmpnt = true;
        }

        String parentXPath = EncodingMapper.baseEnc;

        if ( incSgmnt ) {
            parentXPath += "/" + msgProps.encoding.sgmntEnc;

            if ( incField ) {
                parentXPath += "/" + msgProps.encoding.fieldEnc;

                if ( incCmpnt ) {
                    parentXPath += "/" + msgProps.encoding.cmpntEnc;
                }
            }
        }

        try {
            final Document doc = getDOM();
            final Node parent = XPathAPI.selectSingleNode( doc, parentXPath );

            for ( int i = 0; i < copies; i++ ) {
                parent.appendChild( doc.createElement( newElementEnc ) );
            }
        }
        catch ( final TransformerException e ) {
        }
        catch ( final DOMException e ) {
        }
    }

    private void repairRemove() throws VerifierException {
        final NodeList nList = getElementNodes();
        final int nodeCnt = nList.getLength();

        for ( int i = 0; i < nodeCnt; i++ ) {
            repairRemove( nList.item( i ) );
        }
    }

    private void repairRemove( final int maxRemove ) throws VerifierException {
        final NodeList nList = getElementNodes();
        final int nodeCnt = nList.getLength();

        for ( int i = 0; i < nodeCnt && i < maxRemove; i++ ) {
            repairRemove( nList.item( i ) );
        }
    }

    private void repairRemove( final Node node ) {
        node.getParentNode().removeChild( node );
    }
}
