package gov.nist.hl7maker.verifier.message;

import gov.nist.hl7maker.verifier.metadata.GeneralInfo;
import gov.nist.hl7maker.verifier.profile.ProfileElement;
import gov.nist.hl7maker.verifier.util.ErrorSet;

/**
 * Compares the cardinality of a message element to a
 * <code>ProfileElement</code> and metadata message settings.
 * 
 * @author Justin Manweiler
 */
class MessageProfileValidator {

    // --- Private Fields ---

    final GeneralInfo.Properties genProps;

    // --- Constructor ---

    public MessageProfileValidator( final GeneralInfo.Properties genProps ) {
        this.genProps = genProps;
    }

    // --- Public Methods ---

    public static ErrorSet validateErrorsOnly( final ProfileElement pElement,
            final int matchCnt ) {
        final ErrorSet errSet = new ProfileErrorSet();
        checkErrors( errSet, pElement, matchCnt );
        return errSet;
    }

    public ErrorSet validate( final ProfileElement pElement, final int matchCnt ) {
        final ErrorSet errSet = new ProfileErrorSet();
        checkErrors( errSet, pElement, matchCnt );
        checkViolations( errSet, pElement, matchCnt, genProps );
        return errSet;
    }

    // --- Private Methods ---

    private static void checkErrors( final ErrorSet errSet,
            final ProfileElement pElement, final int matchCnt ) {
        if ( pElement.isUsageX() ) {

            /* X-usage Errors */

            if ( matchCnt > 0 ) {
                errSet.addError( ProfileErrorSet.ET_XusageError );
            }
        }
        else {
            if ( matchCnt < 1 ) {
                /* R-usage Errors */
                
                if ( pElement.isUsageR() ) {
                    errSet.addError( ProfileErrorSet.ET_RusageError );
                }
            }
            else {
                /* Cardinality Errors */

                final int minCard = pElement.getMinCardinality();
                final int maxCard = pElement.getMaxCardinality();

                if ( pElement.isSegment() ) {
                    if ( matchCnt < minCard ) {
                        errSet.addError( ProfileErrorSet.ET_MinSegCardError );
                    }

                    if ( matchCnt > maxCard ) {
                        errSet.addError( ProfileErrorSet.ET_MaxSegCardError );
                    }
                }

                if ( pElement.isField() ) {
                    if ( matchCnt < minCard ) {
                        errSet.addError( ProfileErrorSet.ET_MinFieldCardError );
                    }

                    if ( matchCnt > maxCard ) {
                        errSet.addError( ProfileErrorSet.ET_MaxFieldCardError );
                    }
                }
            }
        }
    }

    private static void checkViolations( final ErrorSet errSet,
            final ProfileElement pElement, final int matchCnt,
            final GeneralInfo.Properties genProps ) {
        
        if ( !pElement.isUsageX() ) {
            /* Usage Violations */
            
            if ( pElement.isUsageRE() ) {
                if ( genProps.hasREAll && matchCnt < 1 ) {
                    errSet.addError( ProfileErrorSet.VT_REElementNotFound );
                }

                if ( genProps.hasRENone && matchCnt > 0 ) {
                    errSet.addError( ProfileErrorSet.VT_REElementFound );
                }
            }

            /* Required Cardinality Violations */

            final int minCard = pElement.getMinCardinality();
            final int maxCard = pElement.getMaxCardinality();

            if ( pElement.isSegment() ) {
                if ( genProps.isSgmntCardMin && matchCnt != minCard ) {
                    errSet.addError( ProfileErrorSet.VT_SegCardNEMin );
                }

                if ( genProps.isSgmntCardMax && matchCnt != maxCard ) {
                    errSet.addError( ProfileErrorSet.VT_SegCardNEMax );
                }
            }

            if ( pElement.isField() ) {
                if ( genProps.isFieldCardMin && matchCnt != minCard ) {
                    errSet.addError( ProfileErrorSet.VT_FieldCardNEMin );
                }

                if ( genProps.isFieldCardMax && matchCnt != maxCard ) {
                    errSet.addError( ProfileErrorSet.VT_FieldCardNEMax );
                }
            }
        }
    }
}
