package gov.nist.hl7maker.verifier.metadata;

import gov.nist.hl7maker.verifier.hl7.EncodingMapper;

import org.w3c.dom.Node;

/**
 * Parses, stores, and provides the properties of a single generated message.
 * 
 * @author Justin Manweiler
 */
public class MessageInfo extends SharedInfo {

    // --- Inner Classes ---

    public class Properties extends SharedInfo.Properties {
        public final boolean isValid = isValid();
        public final EncodingMapper.Encoding encoding;
        public final String name = getName();
        public final String elmntSgmnt = getElementSegment();
        public final String elmntField = getElementField();
        public final String elmntCmpnt = getElementComponent();
        public final String elmntScpnt = getElementSubComponent();
        public final String elmntREInclude = getElementREInclude();
        public final String elmntCardinality = getElementCardinality();

        private Properties() {
            encoding = encMapper.getEncoding( elmntSgmnt, elmntField,
                    elmntCmpnt, elmntScpnt );
        }
        
        public String getErrorDescription() {
            if ( elmntREInclude != null ) {
                if ( elmntREInclude.equals( "true" ) ) {
                    return "REInclude=true";
                }

                if (elmntREInclude.equals( "false" ) ) {
                    return "REInclude=false";
                }
            }

            if ( elmntCardinality != null ) {
                return "Cardinality set to " + elmntCardinality;
            }
            
            return super.getErrorDescription();
        }
    }

    // --- Private Fields ---

    private static final String valueValid = "VALID";
    private static final String mp_Name = "Name";
    private static final String mp_Type = "Type";
    private static final String mp_ElementSegment = "SEGMENT";
    private static final String mp_ElementField = "FIELD";
    private static final String mp_ElementComponent = "COMPONENT";
    private static final String mp_ElementSubComponent = "SUBCOMPONENT";
    private static final String mp_ElementREInclude = "REInclude";
    private static final String mp_ElementCardinality = "cardinality";

    private static final String[] mRootProperties = {
            mp_Name, mp_Type
    };

    private static final String[] mElementProperties = {
            mp_ElementSegment, mp_ElementField, mp_ElementComponent,
            mp_ElementSubComponent, mp_ElementREInclude, mp_ElementCardinality
    };

    private final EncodingMapper encMapper;

    // --- Constructor ----

    public MessageInfo( final EncodingMapper mapper ) {
        encMapper = mapper;
    }

    // --- Public Methods ---

    public Properties getProperties() {
        return new Properties();
    }

    public void parse( final Node parseRoot ) {
        super.parse( parseRoot, "ELEMENT/" );
        propertyParser.parse( parseRoot, mRootProperties );
        propertyParser.parse( parseRoot, "ELEMENT/", mElementProperties );
    }

    // --- Private Methods ---

    private String getElementSegment() {
        return propertyParser.getValue( mp_ElementSegment );
    }

    private String getElementField() {
        return propertyParser.getValue( mp_ElementField );
    }

    private String getElementComponent() {
        return propertyParser.getValue( mp_ElementComponent );
    }

    private String getElementSubComponent() {
        return propertyParser.getValue( mp_ElementSubComponent );
    }

    private String getElementREInclude() {
        return propertyParser.getValue( mp_ElementREInclude );
    }

    private String getElementCardinality() {
        return propertyParser.getValue( mp_ElementCardinality );
    }

    private String getName() {
        return propertyParser.getValue( mp_Name );
    }

    private boolean isValid() {
        final String type = propertyParser.getValue( mp_Type );
        return type != null && type.equals( valueValid );
    }
}
