package gov.nist.hl7maker.verifier.metadata;

import org.w3c.dom.Node;

/**
 * Parses, stores, and provides properties defined as general defaults in a
 * metadata file, but may be overridden for individual messages.
 * 
 * @author Justin Manweiler
 */
class SharedInfo {

    // --- Inner Classes ---

    public class Properties {
        public final boolean hasRusageError = hasRusageError();
        public final boolean hasXusageError = hasXusageError();
        public final boolean hasMinSegCardError = hasMinSegCardError();
        public final boolean hasMaxSegCardError = hasMaxSegCardError();
        public final boolean hasMinFieldCardError = hasMinFieldCardError();
        public final boolean hasMaxFieldCardError = hasMaxFieldCardError();
        public final boolean hasFieldLengthError = hasFieldLengthError();
        public final boolean hasComponentLengthError = hasComponentLengthError();
        public final boolean hasSubComponentLengthError = hasSubComponentLengthError();
        
        public String getErrorDescription() {
            if ( hasRusageError ) {
                return sp_RusageError;
            }

            if ( hasXusageError ) {
                return sp_XusageError;
            }

            if ( hasMinSegCardError ) {
                return sp_MinSegCardError;
            }

            if ( hasMaxSegCardError ) {
                return sp_MaxSegCardError;
            }

            if ( hasMinFieldCardError ) {
                return sp_MinFieldCardError;
            }

            if ( hasMaxFieldCardError ) {
                return sp_MaxFieldCardError;
            }

            if ( hasFieldLengthError ) {
                return sp_FieldLengthError;
            }

            if ( hasComponentLengthError ) {
                return sp_ComponentLengthError;
            }

            if ( hasSubComponentLengthError ) {
                return sp_SubComponentLengthError;
            }
            
            return "no implanted error expected";
        }
    }

    // --- Protected Fields ---

    protected final PropertyParser propertyParser = new PropertyParser();

    // --- Private Fields ---

    private static final String sp_MinSegCardError = "MinSegCardError";
    private static final String sp_MaxSegCardError = "MaxSegCardError";
    private static final String sp_MinFieldCardError = "MinFieldCardError";
    private static final String sp_MaxFieldCardError = "MaxFieldCardError";
    private static final String sp_RusageError = "RusageError";
    private static final String sp_XusageError = "XusageError";
    private static final String sp_FieldLengthError = "FieldLengthError";
    private static final String sp_ComponentLengthError = "ComponentLengthError";
    private static final String sp_SubComponentLengthError = "SubComponentLengthError";

    private static final String[] sProperties = {
            sp_MinSegCardError, sp_MaxSegCardError, sp_MinFieldCardError,
            sp_MaxFieldCardError, sp_RusageError, sp_XusageError,
            sp_FieldLengthError, sp_ComponentLengthError,
            sp_SubComponentLengthError
    };

    // --- Public Methods ---

    public void parse( final Node parseRoot ) {
        propertyParser.parse( parseRoot, sProperties );
    }

    public void parse( final Node parseRoot, final String baseXPath ) {
        propertyParser.parse( parseRoot, baseXPath, sProperties );
    }

    public void setDefaultInfo( final SharedInfo defaultInfo ) {
        propertyParser.setDefaults( defaultInfo.propertyParser );
    }

    // --- Private Methods ---

    private boolean hasMinSegCardError() {
        return propertyParser.isTrue( sp_MinSegCardError );
    }

    private boolean hasMaxSegCardError() {
        return propertyParser.isTrue( sp_MaxSegCardError );
    }

    private boolean hasMinFieldCardError() {
        return propertyParser.isTrue( sp_MinFieldCardError );
    }

    private boolean hasMaxFieldCardError() {
        return propertyParser.isTrue( sp_MaxFieldCardError );
    }

    private boolean hasRusageError() {
        return propertyParser.isTrue( sp_RusageError );
    }

    private boolean hasXusageError() {
        return propertyParser.isTrue( sp_XusageError );
    }

    private boolean hasFieldLengthError() {
        return propertyParser.isTrue( sp_FieldLengthError );
    }

    private boolean hasComponentLengthError() {
        return propertyParser.isTrue( sp_ComponentLengthError );
    }

    private boolean hasSubComponentLengthError() {
        return propertyParser.isTrue( sp_SubComponentLengthError );
    }
}
