package gov.nist.hl7maker.verifier.profile;

import gov.nist.hl7maker.verifier.hl7.EncodingMapper;

/**
 * Stores information from a profile about a single segment, field, component,
 * or subcomponent.
 * 
 * @author Justin Manweiler
 */
public class ProfileElement {

    // --- Private Fields ---

    private static final int TYPE_SEGMENT = 1;
    private static final int TYPE_FIELD = 2;
    private static final int TYPE_COMPONENT = 3;
    private static final int TYPE_SUBCOMPONENT = 4;
    private static final int USAGE_R = 1;
    private static final int USAGE_RE = 2;
    private static final int USAGE_X = 3;
    private static final String prop_UsageR = "R";
    private static final String prop_UsageRE = "RE";
    private static final String prop_UsageX = "X";

    private int type = 0;
    private int usage = 0;
    private int minCardinality = Integer.MIN_VALUE;
    private int maxCardinality = Integer.MAX_VALUE;
    private String segment;
    private String segmentEnc;
    private String field;
    private String fieldEnc;
    private String component;
    private String componentEnc;
    private String subcomponent;
    private String subcomponentEnc;

    // --- Public Methods ---

    public void copyLocation( final ProfileElement base ) {
        type = base.type;
        segment = base.segment;
        segmentEnc = base.segmentEnc;
        field = base.field;
        fieldEnc = base.fieldEnc;
        component = base.component;
        componentEnc = base.componentEnc;
        subcomponent = base.subcomponent;
        subcomponentEnc = base.subcomponentEnc;
    }

    public String getMessageXPath() {
        return EncodingMapper.getMessageXPath( segmentEnc, fieldEnc,
                componentEnc, subcomponentEnc );
    }

    public String getProfileXPath() {
        return ProfileInspector.getProfileXPath( segment, field, component,
                subcomponent );
    }

    public String getProfileLocDescription() {
        if ( segmentEnc == null ) {
            return "NO PROFILE LOCATION SPECIFIED";
        }

        String loc = "segment=\"" + segment + "\"";

        if ( field == null ) {
            return loc;
        }

        loc += ", field=\"" + field + "\"";

        if ( component == null ) {
            return loc;
        }

        loc += ", component=\"" + component + "\"";

        if ( subcomponent == null ) {
            return loc;
        }

        return loc + ", subcomponent=\"" + subcomponent + "\"";
    }
    
    public String getUsageDescription() {
        switch ( usage ) {
            case USAGE_R: return "required";
            case USAGE_RE: return "optional";
            case USAGE_X: return "not allowed"; 
            default: return "usage not specified";
        }
    }

    public int getMinCardinality() {
        return minCardinality;
    }

    public int getMaxCardinality() {
        return maxCardinality;
    }
    
    public boolean isMinCardinalitySet() {
        return minCardinality != Integer.MIN_VALUE;
    }
    
    public boolean isMaxCardinalitySet() {
        return maxCardinality != Integer.MAX_VALUE;
    }

    public boolean isUsageR() {
        return usage == USAGE_R;
    }

    public boolean isUsageRE() {
        return usage == USAGE_RE;
    }

    public boolean isUsageX() {
        return usage == USAGE_X;
    }

    public boolean isSegment() { 
        return type == TYPE_SEGMENT;
    }

    public boolean isField() {
        return type == TYPE_FIELD;
    }

    public boolean isComponent() {
        return type == TYPE_COMPONENT;
    }

    public boolean isSubComponent() {
        return type == TYPE_SUBCOMPONENT;
    }

    public void setMinCardinality( final String value ) {
        if ( value != null ) {
            try {
                minCardinality = Integer.parseInt( value );
            }
            catch ( final NumberFormatException e ) {
            }
        }
    }

    public void setMaxCardinality( final String value ) {
        if ( value != null ) {
            try {
                maxCardinality = Integer.parseInt( value );
            }
            catch ( final NumberFormatException e ) {
            }
        }
    }

    public void setUsage( final String value ) {
        if ( value != null ) {
            if ( value.equals( prop_UsageR ) ) {
                usage = USAGE_R;
            }
            else if ( value.equals( prop_UsageRE ) ) {
                usage = USAGE_RE;
            }
            else if ( value.equals( prop_UsageX ) ) {
                usage = USAGE_X;
            }
        }
    }

    public void setSegment( final String name, final String encoding ) {
        if ( type < TYPE_SEGMENT ) {
            type = TYPE_SEGMENT;
        }

        segment = name;
        segmentEnc = encoding;
    }

    public void setField( final String name, final String encoding ) {
        if ( type < TYPE_FIELD ) {
            type = TYPE_FIELD;
        }

        field = name;
        fieldEnc = encoding;
    }

    public void setComponent( final String name, final String encoding ) {
        if ( type < TYPE_COMPONENT ) {
            type = TYPE_COMPONENT;
        }

        component = name;
        componentEnc = encoding;
    }

    public void setSubComponent( final String name, final String encoding ) {
        if ( type < TYPE_SUBCOMPONENT ) {
            type = TYPE_SUBCOMPONENT;
        }

        subcomponent = name;
        subcomponentEnc = encoding;
    }
}