package gov.nist.hl7maker.verifier.profile;

import java.util.Collection;
import java.util.LinkedList;

import javax.xml.transform.TransformerException;

import org.apache.xpath.XPathAPI;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * Provides <code>ProfileElement</code> s for every segment, field, component,
 * and subcomponent described in a profile file.
 * 
 * @author Justin Manweiler
 */
public class ProfileInspector {

    // --- Private Fields ---

    public static final String attName = "Name";
    private static final String attMin = "Min";
    private static final String attMax = "Max";
    private static final String attDatatype = "Datatype";
    private static final String attLength = "Length";
    private static final String attTable = "Table";
    private static final String attUsage = "Usage";
    private static final String baseXPath = "//";
    private static final String elementSegment = "Segment";
    private static final String elementField = "Field";
    private static final String elementComponent = "Component";
    private static final String elementSubComponent = "SubComponent";
    private static final String encUnkown = "ENCODING_UNKOWN";

    private final Document profileDOM;

    private Collection elements = null;
    private ProfileElement[] retArray = null;

    // --- Constructor ---

    public ProfileInspector( final Document profileDoc ) {
        profileDOM = profileDoc;
    }

    // --- Public Methods ---

    public ExtendedProfileElement getProfileElement( final String segment,
            final String field, final String component,
            final String subcomponent ) {

        final String profileXPath = getProfileXPath( segment, field, component,
                subcomponent );

        ExtendedProfileElement pElement = null;

        try {
            final Element element = (Element)XPathAPI.selectSingleNode(
                    profileDOM, profileXPath );

            if ( element != null ) {
                pElement = new ExtendedProfileElement();

                pElement.setSegment( segment, encUnkown );

                if ( field != null ) {
                    pElement.setField( field, encUnkown );

                    if ( component != null ) {
                        pElement.setComponent( component, encUnkown );

                        if ( subcomponent != null ) {
                            pElement.setSubComponent( subcomponent, encUnkown );
                        }
                    }
                }

                if ( pElement.isSegment() || pElement.isField() ) {
                    pElement.setMinCardinality( element.getAttribute( attMin ) );
                    pElement.setMaxCardinality( element.getAttribute( attMax ) );
                }

                if ( element.hasAttribute( attUsage ) ) {
                    pElement.setUsage( element.getAttribute( attUsage ) );
                }

                if ( element.hasAttribute( attLength ) ) {
                    pElement.setMaxLength( element.getAttribute( attLength ) );
                }

                if ( element.hasAttribute( attTable ) ) {
                    pElement.setTableID( element.getAttribute( attTable ) );
                }
            }
        }
        catch ( final TransformerException e ) {
        }

        return pElement;
    }

    public ProfileElement[] getProfileElements() {
        if ( elements == null ) {
            inspectProfile();
            retArray = new ProfileElement[elements.size()];
            elements.toArray( retArray );
        }

        return retArray;
    }

    public static String getProfileXPath( final String segment,
            final String field, final String component,
            final String subcomponent ) {

        final String bgnAtt = "[@" + attName + "='";
        final String endAtt = "']";

        String xPath = baseXPath + elementSegment + bgnAtt + segment + endAtt;

        if ( field == null ) {
            return xPath;
        }

        xPath += "/" + elementField + bgnAtt + field + endAtt;

        if ( component == null ) {
            return xPath;
        }

        xPath += "/" + elementComponent + bgnAtt + component + endAtt;

        if ( subcomponent == null ) {
            return xPath;
        }

        return xPath + "/" + elementSubComponent + bgnAtt + subcomponent
                + endAtt;
    }

    public static String getProfileXPath( final String segment, final int fieldIndex ) {
        final String bgnAtt = "[@" + attName + "='";
        final String endAtt = "']";
        
        String xPath = baseXPath + elementSegment + bgnAtt + segment + endAtt;

        if ( fieldIndex < 1 ) {
            return xPath;
        }

        return xPath + "/*[" + fieldIndex + "]";
    }

    //  --- Private Methods ---

    private void inspectProfile() {
        try {
            elements = new LinkedList();

            final NodeList sgmntList = XPathAPI.selectNodeList( profileDOM,
                    baseXPath + elementSegment );

            final int sgmntCnt = sgmntList.getLength();

            for ( int i = 0; i < sgmntCnt; i++ ) {
                inspectSegment( (Element)sgmntList.item( i ) );
            }
        }
        catch ( final TransformerException e ) {
        }
    }

    private void inspectSegment( final Element sgmnt ) {
        final String sgmntName = sgmnt.getAttribute( attName );
        final String sgmntUsage = sgmnt.getAttribute( attUsage );

        final ProfileElement sgmntEntry = new ProfileElement();
        sgmntEntry.setSegment( sgmntName, sgmntName );
        sgmntEntry.setUsage( sgmntUsage );
        sgmntEntry.setMinCardinality( sgmnt.getAttribute( attMin ) );
        sgmntEntry.setMaxCardinality( sgmnt.getAttribute( attMax ) );
        elements.add( sgmntEntry );

        if ( !sgmntEntry.isUsageX() ) {
            try {
                final NodeList fieldList = XPathAPI.selectNodeList( sgmnt,
                        elementField );

                final int fieldCnt = fieldList.getLength();

                for ( int i = 0; i < fieldCnt; i++ ) {
                    final String fieldEnc = sgmntName + "." + (i + 1);
                    inspectField( (Element)fieldList.item( i ), sgmntEntry,
                            fieldEnc );
                }
            }
            catch ( final TransformerException e ) {
            }
        }
    }

    private void inspectField( final Element field, final ProfileElement base,
            final String fieldEnc ) {
        final String fieldDt = field.getAttribute( attDatatype );
        final String fieldUsage = field.getAttribute( attUsage );

        final ProfileElement fieldEntry = new ProfileElement();
        fieldEntry.copyLocation( base );
        fieldEntry.setField( field.getAttribute( attName ), fieldEnc );
        fieldEntry.setUsage( fieldUsage );
        fieldEntry.setMinCardinality( field.getAttribute( attMin ) );
        fieldEntry.setMaxCardinality( field.getAttribute( attMax ) );
        elements.add( fieldEntry );

        if ( !fieldEntry.isUsageX() ) {
            try {
                final NodeList cmptList = XPathAPI.selectNodeList( field,
                        elementComponent );

                final int cmptCnt = cmptList.getLength();

                for ( int i = 0; i < cmptCnt; i++ ) {
                    final String cmptEnc = fieldDt + "." + (i + 1);
                    inspectComponent( (Element)cmptList.item( i ), fieldEntry,
                            cmptEnc );
                }
            }
            catch ( final TransformerException e ) {
            }
        }
    }

    private void inspectComponent( final Element cmpt,
            final ProfileElement base, final String cmptEnc ) {
        final String cmptUsage = cmpt.getAttribute( attUsage );
        final String cmptDt = cmpt.getAttribute( attDatatype );

        final ProfileElement cmptEntry = new ProfileElement();
        cmptEntry.copyLocation( base );
        cmptEntry.setComponent( cmpt.getAttribute( attName ), cmptEnc );
        cmptEntry.setUsage( cmptUsage );
        elements.add( cmptEntry );

        if ( !cmptEntry.isUsageX() ) {
            try {
                final NodeList scmptList = XPathAPI.selectNodeList( cmpt,
                        elementSubComponent );

                final int scmptCnt = scmptList.getLength();

                for ( int i = 0; i < scmptCnt; i++ ) {
                    final String scmptEnc = cmptDt + "." + (i + 1);
                    inspectSubComponent( (Element)scmptList.item( i ),
                            cmptEntry, scmptEnc );
                }
            }
            catch ( final TransformerException e ) {
            }
        }
    }

    private void inspectSubComponent( final Element scmpt,
            final ProfileElement base, final String scmptName ) {
        final String scmptUsage = scmpt.getAttribute( attUsage );

        final ProfileElement scmptEntry = new ProfileElement();
        scmptEntry.copyLocation( base );
        scmptEntry.setSubComponent( scmpt.getAttribute( attName ), scmptName );
        scmptEntry.setUsage( scmptUsage );
        elements.add( scmptEntry );
    }
}
