package gov.nist.hl7maker.verifier.util;

/**
 * Framework for the creation of <code>Thread</code> s which may be safely
 * terminated. Derived classes should implement <code>killableRun</code>,
 * instead of overriding <code>run()</code> and should call checkKilled() at
 * points when they may be safely terminated.
 * 
 * @author Justin Manweiler
 */
public abstract class KillableThread extends Thread {

    // --- Inner Classes ---

    public static class KilledException extends Exception {
    };

    // --- Private Fields ---

    private volatile boolean isKilled = false;

    // --- Public Methods ---

    public final void run() {
        try {
            killableRun();
        }
        catch ( final KilledException e ) {
        }
    }

    public void kill( final boolean join ) {
        isKilled = true;

        if ( join ) {
            try {
                join();
            }
            catch ( final InterruptedException e ) {
            }
        }
    }

    // --- Protected Methods ---

    protected void checkKilled() throws KilledException {
        if ( isKilled ) {
            throw new KilledException();
        }
    }

    protected abstract void killableRun() throws KilledException;
}
