package gov.nist.hl7maker.verifier.util;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;

/**
 * Allows for the redirection and storage of output intended for System.out and
 * System.err.
 * 
 * @author Justin Manweiler
 */
public class RedirectOutput {

    // --- Private Data Members ---

    private static final ByteArrayOutputStream errOS = new ByteArrayOutputStream();
    private static final ByteArrayOutputStream outOS = new ByteArrayOutputStream();
    private static final PrintStream stdErr = System.err;
    private static final PrintStream stdOut = System.out;
    private static final PrintStream storedErr;
    private static final PrintStream storedOut;
    private static final PrintStream nullStream = new PrintStream(
            NullOutputStream.SINGLETON );

    static {
        storedErr = new PrintStream( errOS );
        storedOut = new PrintStream( outOS );
    }

    // --- Inner Classes ---

    private static class NullOutputStream extends OutputStream {

        // --- Public Fields ---

        public static final NullOutputStream SINGLETON = new NullOutputStream();

        // --- Constructor ---

        private NullOutputStream() {
        }

        // --- Public Methods ---

        public void write( final int b ) {
        }
    }

    // --- Constructor ---

    private RedirectOutput() {
    }

    // --- Public Methods ---

    /**
     * Permanently discards output sent to System.err.
     */
    public static void discardStdErr() {
        System.setErr( nullStream );
    }

    /**
     * Permanently discards output sent to System.out.
     */
    public static void discardStdOut() {
        System.setOut( nullStream );
    }

    /**
     * Blocks and saves output sent to System.err.
     */
    public static void storeStdErr() {
        System.setErr( storedErr );
    }

    /**
     * Blocks and saves output sent to System.out.
     */
    public static void storeStdOut() {
        System.setOut( storedOut );
    }

    /**
     * Restores the previous destination of output sent to System.err.
     */
    public static void restoreStdErr() {
        System.setErr( stdErr );
    }

    /**
     * Restores the previous destination of output sent to System.out.
     */
    public static void restoreStdOut() {
        System.setOut( stdOut );
    }

    /**
     * Prints stored output blocked from System.err back to System.err.
     */
    public static void retrieveStdErr() {
        try {
            errOS.writeTo( stdErr );
            errOS.reset();
        }
        catch ( IOException e ) {
        }
    }

    /**
     * Prints stored output blocked from System.out back to System.out.
     */
    public static void retrieveStdOut() {
        try {
            outOS.writeTo( stdOut );
            outOS.reset();
        }
        catch ( IOException e ) {
        }
    }
}
