package gov.nist.hl7maker.verifier.xmlverifier;

import gov.nist.hl7maker.verifier.VerifierException;
import gov.nist.hl7maker.verifier.util.TestResult;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;

import org.w3c.dom.Document;

/**
 * Provides infrastructure supporting the validation and verification of an XML
 * file.
 * 
 * @author Justin Manweiler
 */
public abstract class AbstractXMLVerifier implements XMLVerifier {

    // --- Private Fields ---

    private final TestResult testFileResult = new TestResult();

    private Document document = null;

    // --- Public Methods ---

    public Document getDOM() throws VerifierException {
        if ( document == null ) {
            testFile();

            try {
                document = XMLParser.parseXMLDOM( getFile() );
            }
            catch ( final Exception e ) {
                throw new VerifierException( getFileDescription() + " file \""
                        + getFile().getName()
                        + "\" could not be parsed into DOM." );
            }
        }

        return document;
    }
    
    public Document getDOMCopy() throws VerifierException {
        return (Document)getDOM().cloneNode( true );
    }

    public abstract File getFile();

    /**
     * If file passes verification, provides its contents.
     * 
     * @return the contents of the file in stringized form
     * @throws VerifierException
     *             on verification failure
     */
    public String getVerifiedContents() throws VerifierException {
        testFile();

        try {
            return readFileToString( getFile() );
        }
        catch ( final IOException e ) {
            throw new VerifierException( getFileDescription() + " file \""
                    + getFile().getName() + "\" could not be read." );
        }
    }

    public void resetFileTest() {
        testFileResult.reset();
    }

    public void testFile() throws VerifierException {
        final File file = getFile();

        if ( testFileResult.isUntested() ) {
            final String path = file.getName();
            final String type = getFileDescription();
            
            if ( !file.exists() ) {
                testFileResult.failure( type + " file \"" + path
                        + "\" could not be found." );
            }
            else if ( file.isDirectory() ) {
                testFileResult.failure( type + " file \"" + path
                        + "\" is a directory." );
            }
            else if ( !file.isFile() ) {
                testFileResult.failure( type + " file \"" + path
                        + "\" is not a normal file." );
            }
            else if ( !file.canRead() ) {
                testFileResult.failure( type + " file \"" + path
                        + "\" can not be read." );
            }
            else if ( !XMLParser.isValidXML( file ) ) {
                testFileResult.failure( type + " file \"" + path
                        + "\" is not valid XML." );
            }
            else {
                testFileResult.passed();
            }
        }

        if ( testFileResult.hasFailed() ) {
            throw new VerifierException( testFileResult.getFailCause() );
        }
    }

    // --- Protected Methods ---    
    
    protected String getFileDescription() {
        return "Generic";
    }

    // --- Private Methods ---

    private static String readFileToString( final File file )
            throws IOException {
        final FileInputStream fis = new FileInputStream( file );

        final byte b[] = new byte[fis.available()];
        fis.read( b );

        return new String( b );
    }
}
