/**
 * This class is a dialog that shows the progress of the generation process
 */

package gov.nist.hl7maker.view;

// JDK
import java.io.File;
import java.util.Vector;

// AWT
import java.awt.Dimension;
import java.awt.Color;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;

// Swing
import javax.swing.JDialog;
import javax.swing.JPanel;
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.BorderFactory;
import javax.swing.JProgressBar;
import javax.swing.BoxLayout;
import javax.swing.Box;
import javax.swing.Timer;

// MessageMaker
import gov.nist.hl7maker.control.Session;
import gov.nist.hl7maker.control.User;
import gov.nist.hl7maker.control.MessageGenerationProperties;
import gov.nist.hl7maker.control.MakerMessage;
import gov.nist.hl7maker.util.Constants;
import gov.nist.hl7maker.util.MVCArg;
import gov.nist.hl7maker.message.MessageGenerator;

public class MonitorGenerationDialog extends JDialog implements ActionListener {

    private Session session;
    private User user;
    private MessageGenerator generator;

    private JButton btnState;
    private JProgressBar pb;
    private JLabel lblInfo;

    private Timer timer;
    private boolean abort;

    /**
     * Constructor
     * @param frame A JFrame
     * @param ses The session
     * @param user The user
     */
    public MonitorGenerationDialog(JFrame frame, Session ses, User user) {
	super(frame, "Generate Messages", true);
	session = ses;
	this.user = user;

	BoxLayout layout = new BoxLayout(getContentPane(), BoxLayout.PAGE_AXIS);
	getContentPane().setLayout(layout);

	JPanel p0 = new JPanel();
	p0.setLayout(new BoxLayout(p0, BoxLayout.LINE_AXIS));
	btnState = new JButton("Start");
	btnState.addActionListener(this);
	p0.add(btnState);
	pb = new JProgressBar();
	pb.setStringPainted(true);
	p0.add(Box.createRigidArea(new Dimension(5, 0)));
	p0.add(pb);

	JPanel p1 = new JPanel();
	p1.setLayout(new BoxLayout(p1, BoxLayout.LINE_AXIS));
	lblInfo = new JLabel(" ");
	p1.add(lblInfo);

	JPanel p2 = new JPanel();
	p2.setLayout(new BoxLayout(p2, BoxLayout.PAGE_AXIS));
	p2.add(p0);
	p2.add(Box.createRigidArea(new Dimension(0, 5)));
	p2.add(p1);
	p2.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
	getContentPane().add(p2);
    }

    /**
     * Init the dialog
     */
    public void init() {
	/* 
	 * Create a timer. Report progress periodically. Upon termination or
	 * sucessful completion, reset the generator settings.
	 */
	btnState.setText("Start");
	pb.setValue(0);
	lblInfo.setText("Click on start to generate messages");
	
	timer = new Timer(500, new ActionListener() {
		public void actionPerformed(ActionEvent evt) {
		    pb.setValue(generator.getProgressValue());
		    if(generator.getProgressValue() == 0)
			lblInfo.setText("Generating Messages");
		    else
			lblInfo.setText(generator.getProgressValue() + " messages generated out of " + pb.getMaximum());
		    if(generator.isDone()) {
			//Toolkit.getDefaultToolkit().beep();
			pb.setValue(generator.getProgressValue());
			timer.stop();
			btnState.setText("Close");
			lblInfo.setText("Generation finished");
			//generatePanel.setCursor(null); //turn off the wait cursor
			//progressBar.setValue(0);
			//generator.reset();
		    }
		}
	    });
	session.generateRepository(true);
    }

    /**
     * Handles a click on a button
     * @param e The Event
     */
    public void actionPerformed(ActionEvent e) {
	if(e.getSource() == btnState) {
	    String state = btnState.getText();
	    if(state.equals("Start")) {
		// Create the messages directory if it doesn't exist
		File f = new File(session.getMsgPath());
		if(!f.exists()) {
		    f.mkdir();
		}
	
		/* Create a properties object from the current user settings */
		MessageGenerationProperties prop = new MessageGenerationProperties(session);
		// Set the options
		prop.setOptionsFileProperty(session.generateMGOptions());
		
		/* Obtain an instance and initialize the Message Generator with default settings. */
		if (generator == null) {
		    generator = new MessageGenerator(prop.getProperties());
		}
		else {
		    generator.restart(prop.getProperties());
		}
		abort = false;
/*		else {
		    // Set the output dir
		    generator.setOutputDir(session.getMsgPath());
		    // Set the meta file
		    generator.setMetaFile("MD_" + new File(session.getProfile()).getName());
		}
		
		generator.setOptions(prop.getProperties());*/

		// Set the volume of messages that will be generated
		pb.setMaximum(session.getVolume());
		
		/* Generate the messages and monitor and display the progress */
		GenerateMessagesThread gmt = new GenerateMessagesThread();
		gmt.start();
		timer.start();

		btnState.setText("Abort");
	    }
	    else if(btnState.getText().equals("Abort")) {
		generator.stop();
		abort = true;
		btnState.setText("Close");
		// Remove everything which has been generated
		//taskOutput.append(generator.getStatusMessage());
	    }
	    else if(btnState.getText().equals("Close")) {
		if(abort) generator.abort();
		generator.reset();
		setVisible(false);
	    }
	}
    }

    /*
     * Launches the the message generation process while providing the input
     * to a JProgressBar. This is a separate thread. The progress value is
     * obtained from the Message Generation object (generator).
     *
     */
    class GenerateMessagesThread extends Thread {
	boolean stop = false;

	public GenerateMessagesThread() {
	}

	/* 
	 * This is called from another object, the MessageGeneration object
	 * in this case. It gets called after a message is created. The 
	 * progress of the message creation process is captured and updated
	 * with the setValue() method.
	 */

	public void run() {
	    Runnable runner = new Runnable() {
		    public void run() {
			//			MMLog.log("MainMenu:" + "GenerateMessagesThread:" +  "run# " + generator.getProgressValue());
			pb.setValue(generator.getProgressValue());
		    }
		};

	    /* Generate the messages */
	    generator.generateMessages(session.getValidate(), runner);
	    if(!abort) {
		session.fireEvent(new MVCArg("mountDir", session.getMsgPath()));
		// Update the list of messages in the user object
		Vector vMsg = generator.getMessageFileNames();
		user.addMD(session.getMsgPath() + Constants.FS + "TD_" + new File(session.getProfile()).getName());
		for(int i=0;i<vMsg.size();i++) {
		    String sMsg = (String)vMsg.get(i);
		    if(!sMsg.startsWith("TD")) {
			user.addMsg(sMsg, session.getProfile());
			user.setStatus(sMsg, MakerMessage.GENERATED);
		    }
		}	
	    }
	}
    }

}
