/**
 * This class is a TreeModel for an XML file
 */

package gov.nist.hl7maker.view;

// JDK
import java.util.Vector;

// Swing
import javax.swing.JTree;
import javax.swing.tree.TreeModel;
import javax.swing.tree.TreePath;
import javax.swing.event.TreeModelListener;
import javax.swing.event.TreeModelEvent;

// DOM
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

public class ProfileXMLTreeModel implements TreeModel {

    private Document doc;
    private Vector listeners;

    /**
     * Constructor
     * @param doc A Document object
     */
    public ProfileXMLTreeModel(Document doc) {
	this.doc = doc;
	removeEmptyTextElement(doc);
	listeners = new Vector();
    }

    public Object getRoot() {
	return new MyProfileNode(doc.getDocumentElement());
    }

    public int getChildCount(Object parent) {
	return ((MyProfileNode)parent).getNode().getChildNodes().getLength();
    }

    public Object getChild(Object parent, int index) {
    	if(((MyProfileNode)parent).getNode().hasChildNodes())
	    return new MyProfileNode(((MyProfileNode)parent).getNode().getChildNodes().item(index));
	return null;
    }

    public int getIndexOfChild(Object parent, Object child) {
	if(((MyProfileNode)parent).getNode().hasChildNodes()) {
	    NodeList nl = ((MyProfileNode)parent).getNode().getChildNodes();
	    for(int i=0;i<nl.getLength();i++) {
		if(nl.item(i).equals(child)) {
		    return i;
		}
	    }
	}
	return -1;
    }

    /**
     * Return true if the node is a leaf
     * @param node The Node
     * @return is a leaf?
     */
    public boolean isLeaf(Object node) {
	boolean isLeaf = false;
	/*	Node n = ((MyProfileNode)node).getNode();
		if(n.hasChildNodes()) {
		NodeList nl = n.getChildNodes();
		for(int i=0;i<nl.getLength() && !isLeaf;i++)
		if(nl.item(i).getNodeType() == Node.TEXT_NODE)
		isLeaf = true;
		}
		return isLeaf;*/
	return !(((MyProfileNode)node).getNode().hasChildNodes());
    }

    public void addTreeModelListener(TreeModelListener listener) {
	if (listener != null && !listeners.contains(listener))
	    listeners.add(listener);
    }

    public void removeTreeModelListener(TreeModelListener listener) {
	if (listener != null) 
	    listeners.remove(listener);
    }
    
   public void fireTreeNodesChanged(TreeModelEvent event) {
       for(int i=0;i<listeners.size();i++) {
	   TreeModelListener listener = (TreeModelListener)listeners.get(i);
	   listener.treeNodesChanged(event);
       }
   }    
    
    public void valueForPathChanged(TreePath path, Object newValue) {
	MyProfileNode node = (MyProfileNode)path.getLastPathComponent();
	Node n = node.getNode();

	if(n.hasChildNodes()) {
	    Node updatedNode = n.getChildNodes().item(0);
	    if(updatedNode.getNodeType() == Node.TEXT_NODE) {
		updatedNode.setNodeValue((String)newValue);
		TreeModelEvent event = new TreeModelEvent(this, path);
		fireTreeNodesChanged(event);
	    }
	}
    }

    /**
     * This method removes all TEXT_NODE with empty text
     * @param node The current node
     */    
    private void removeEmptyTextElement(Node node) {
	if(node.hasChildNodes()) {
	    NodeList nl = node.getChildNodes();
	    for(int i=0;i<nl.getLength();i++) {
		Node n = nl.item(i);
		if(n.getNodeValue() != null && n.getNodeValue().trim().equals(""))
		    node.removeChild(n);
	    }
	}
	// Execute the method on the children
	if(node.hasChildNodes()) {
	    NodeList nl = node.getChildNodes();
	    for(int i=0;i<nl.getLength();i++) {
		removeEmptyTextElement(nl.item(i));
	    }
	}
    }

    public class MyProfileNode {

	private Node domNode;

    /**
     * Constructor
     * @param node a Node
     */
    public MyProfileNode(Node node) {
	domNode = node;
    }

    public String toString() {
	StringBuffer buffy = new StringBuffer();

	if(domNode.getNodeType() == Node.ELEMENT_NODE) {
	    buffy.append(domNode.getNodeName());
	    if(domNode.hasAttributes()) {
		for(int i=0;i<domNode.getAttributes().getLength();i++) {
		    Node attr = domNode.getAttributes().item(i);
		    buffy.append(" ").append(attr.getNodeName()).append("=\"").append(attr.getNodeValue()).append("\"");
		}
	    }    
	}
	else if(domNode.getNodeType() == Node.TEXT_NODE) {
	    buffy.append(domNode.getNodeValue());
	}
	else {
	    System.err.println("This type of node is not processed: " + domNode.getNodeType() + " " + domNode.getNodeValue());
	}
	return new String(buffy);	
    }

    /**
     * Return true if the node is a leaf
     * In this node, we consider that a leaf node is a node with one child
     * which is a TEXT_NODE
     * @return is a leaf?
     */
    public boolean isLeaf() {
	if(domNode.getNodeType() == Node.ELEMENT_NODE) {
	    if(domNode.hasChildNodes()) {
		NodeList nl = domNode.getChildNodes();
		if(nl.getLength() == 1 && nl.item(0).getNodeType() == Node.TEXT_NODE) {
		    return true;
		}
	    }
	}
	return false;
    }

    /**
     * Return the first text node value if it is a leaf
     * @return The value of the node
     */
    public String getFirstNodeValue() {
	if(isLeaf()) {
	    return domNode.getChildNodes().item(0).getNodeValue().trim();
	}
	return null;
    }

    /**
     * Get the Node object
     * @return The Node object
     */
    public Node getNode() {
	return domNode;
    }
}

}
