/**
 * This class represents the profile as a JTree
 */

package gov.nist.hl7maker.view.config;

// JDK
import java.io.IOException;
import java.util.Vector;

// AWT
import java.awt.CardLayout;
import java.awt.Component;
import java.awt.Color;

// Swing
import javax.swing.JTree;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreeCellRenderer;
import javax.swing.event.TreeSelectionListener;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.TreePath;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.Icon;
import javax.swing.JPanel;
import javax.swing.JLabel;
import javax.swing.BorderFactory;

// XML
import javax.xml.parsers.ParserConfigurationException;

// SAX
import org.xml.sax.SAXException;

// MessageMaker
import gov.nist.hl7maker.control.Session;
import gov.nist.hl7maker.control.Profile;
import gov.nist.hl7maker.view.config.SegmentGroupNode;
import gov.nist.hl7maker.util.MVCArg;

public class ProfileTree extends JTree implements TreeSelectionListener, TreeCellRenderer {

    private Session session;
    private Vector vMissing;

    /**
     * Constructor
     */
    public ProfileTree() {
	setModel(null);
	addTreeSelectionListener(this);
	this.setCellRenderer(this);
	this.putClientProperty("JTree.lineStyle", "None");
    }

    /**
     * Set the tree model
     * @param ses A session object
     */
    public void setTreeModel(Session ses) {
	session = ses;
	Profile p = session.getRAMProfile();
	if(p != null) {
	    DefaultTreeModel model = new DefaultTreeModel(p.getTree("Profile"));
	    vMissing = new Vector();
	    setModel(model);
	}
	else {
	    setModel(null);
	}
    }

    /**
     * Update views when the user selects a file
     * @param e The Event of selecting a file
     */
    public void valueChanged(TreeSelectionEvent e) {
	DefaultMutableTreeNode curNode = (DefaultMutableTreeNode)getLastSelectedPathComponent();
	if(curNode != null && curNode.getUserObject() instanceof ProfileNode && curNode.isLeaf()) {
	    ProfileNode pNode = (ProfileNode)curNode.getUserObject();
	    if(session.getRepositoryDataModel().isRepository(pNode)) {
		session.getRepositoryDataModel().setProfileNode(pNode);
		if(session.getRepositoryDataModel().hasConfig(pNode))
		    session.getRepositoryDataModel().fireEvent(new MVCArg("config", ""));
	    }
	    else if(session.getTableDataModel().isTable(pNode)) {
		session.getTableDataModel().setProfileNode(pNode);
	    }
	}
	else {
	    session.getRepositoryDataModel().setInitState();
	    //configModel.setInitState();
	    session.getTableDataModel().setInitState();
	    if(curNode != null && curNode.getUserObject() instanceof ProfileNode) {
		ProfileNode pNode = (ProfileNode)curNode.getUserObject();
		if(!pNode.isLeaf())
		    if(!(pNode instanceof SegmentNode) && !(pNode instanceof SegmentGroupNode))
			session.getRepositoryDataModel().setInitState((ProfileNode)curNode.getUserObject());
	    }
	}
	TreePath tp = e.getPath();
	if(this.isExpanded(tp))
	    this.collapsePath(tp);
	else
	    this.expandPath(tp);
    }

    /**
     * Show/Hide missing values in red
     * @param showhide true to show, false to hide
     */
    public void showMissing(boolean showhide) {
	Profile p = session.getRAMProfile();
	if(p != null) {
	    DefaultTreeModel model = new DefaultTreeModel(p.getTree("Profile"));
	    vMissing = new Vector();
	    if(showhide) {
		missingValues(model);
	    }
	    setModel(model);
	}
    }

    private void missingValues(DefaultTreeModel model) {
	for(int i=0;i<model.getChildCount(model.getRoot());i++) {
	    missingValues((DefaultMutableTreeNode)model.getChild(model.getRoot(), i));
	}
	for(int i=0;i<vMissing.size();i++) {
	    boolean end = false;
	    ProfileNode pn = (ProfileNode)vMissing.get(i);
	    while(!end && pn.getParent() != null) {
		if(!vMissing.contains(pn.getParent()))
		    vMissing.add(pn.getParent());
		else
		    end = true;
		pn = pn.getParent();
	    }
	}
	//System.out.println("Missing values: " + vMissing.size());
    }

    private void missingValues(DefaultMutableTreeNode node) {
	if(node != null && node.getUserObject() instanceof ProfileNode) {
	    if(node.isLeaf()) {
		ProfileNode pNode = (ProfileNode)node.getUserObject();
		if(!pNode.getUsage().equals("X")) {
		    if(session.getRepositoryDataModel().isRepository(pNode)) {
			if(session.getRepositoryDataModel().getValues(pNode.getName(), pNode.getSegment(), pNode.getQualifier(), pNode.getQualifier2()).size() == 0)
			    vMissing.add(pNode);
		    }
		    else if(session.getTableDataModel().isTable(pNode)) {
			if(session.getTableDataModel().getValues(pNode.getTable()).size() == 0)
			    vMissing.add(pNode);
		    }
		}
	    }
	    else {
		for(int i=0;i<node.getChildCount();i++) {
		    ProfileNode pNode = (ProfileNode)node.getUserObject();
		    if(!pNode.getUsage().equals("X"))
			missingValues((DefaultMutableTreeNode)node.getChildAt(i));
		}
	    }
	}
    }

    /**
     * Sets the value of the current tree cell to value. FROM TreeCellRenderer interface
     */
   public Component getTreeCellRendererComponent(JTree tree, Object value, boolean selected, boolean expanded, boolean leaf, int row, boolean hasFocus) {
       JLabel lblRenderer = new JLabel();
       lblRenderer.setText(value.toString());
       lblRenderer.setFont(lblRenderer.getFont().deriveFont(0));
       lblRenderer.setBorder(BorderFactory.createLineBorder(Color.WHITE));
       if(selected) {
	   lblRenderer.setOpaque(true);
	   lblRenderer.setBackground(new Color(204, 204, 255));
	   lblRenderer.setBorder(BorderFactory.createLineBorder(new Color(153, 153, 204)));
       }
       if(value instanceof DefaultMutableTreeNode && ((DefaultMutableTreeNode)value).getUserObject() instanceof ProfileNode) {
	   ProfileNode pNode = (ProfileNode)((DefaultMutableTreeNode)value).getUserObject();
	   if(vMissing.contains(pNode))
	       lblRenderer.setForeground(Color.RED);
       }
       if(value instanceof DefaultMutableTreeNode && ((DefaultMutableTreeNode)value).getUserObject() instanceof SegmentGroupNode) {
	   lblRenderer.setFont(lblRenderer.getFont().deriveFont(1));
       }
       return lblRenderer;
   }
}
