/**
 * This class is a model for the repository data
 */

package gov.nist.hl7maker.view.config;

// JDK
import java.io.File;
import java.io.IOException;
import java.util.Vector;
import java.util.Observable;
import java.util.StringTokenizer;

// XML
import javax.xml.parsers.DocumentBuilder; 
import javax.xml.parsers.DocumentBuilderFactory;  
import javax.xml.parsers.ParserConfigurationException;

// DOM
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Element;

// SAX
import org.xml.sax.SAXException;  

// XSL
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

// MessageMaker
import gov.nist.hl7maker.control.Session;
import gov.nist.hl7maker.util.MVCArg;

public class RepositoryDataModel extends Observable {

    //    private Session session;
    private Vector vData;
    private String name, segment, qualifier, qualifier2, datatype;

    private ProfileNode curPN;
    private ValueNode curVN;
    private Object[] curVVN;

    private Node refNode, itemNode;
    private Element configNode;
    private String xmlRepository;

    /**
     * Constructor
     */
    public RepositoryDataModel() {
	vData = new Vector();
    }

    /**
     * Read the XML repository and collects all values
     * @param xmlRepository The repository filename
     */
    public void loadRepository(String xmlRepository) throws ParserConfigurationException, SAXException, IOException {
	DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
	DocumentBuilder builder = factory.newDocumentBuilder();
	Document doc = builder.parse(xmlRepository);//session.getProtocolPath(session.getReferenceDB()));
	// Browser the XML file
	vData.clear();
        browseXML(doc);
	//gov.nist.hl7maker.util.MVCLog.info("Reference Data XML => " + session.getReferenceDB() + " loaded");
	this.xmlRepository = xmlRepository;
    }
    
    /**
     * Browse recursively the DOM and collects the values
     * @param n The current node
     */
    private void browseXML(Node n) {
	// Reference Data
	if(n.getNodeName().equals("ReferenceData")) {
 	    Node attributes = n.getAttributes().getNamedItem("Name");
 	    if(attributes != null) {
		name = attributes.getNodeValue();
	    }
	    attributes = n.getAttributes().getNamedItem("Qualifier");
	    if(attributes != null) {
		qualifier = attributes.getNodeValue();
	    }
	    else {
		qualifier = "";
	    }
	    attributes = n.getAttributes().getNamedItem("Qualifier2");
	    if(attributes != null) {
		qualifier2 = attributes.getNodeValue();
	    }
	    else {
		qualifier2 = "";
	    }
	    attributes = n.getAttributes().getNamedItem("Segment");
	    if(attributes != null) {
		segment = attributes.getNodeValue();
	    }
	    attributes = n.getAttributes().getNamedItem("Datatype");
	    if(attributes != null) {
		datatype = attributes.getNodeValue();
	    }
	}
	// Item
	else if(n.getNodeName().equals("Item")) {
	    ValueNode curValue = new ValueNode();
	    curValue.setName(name);
	    curValue.setQualifier(qualifier);
	    curValue.setQualifier2(qualifier2);
	    curValue.setSegment(segment);
	    curValue.setDatatype(datatype);
	    vData.add(curValue);
	    Node attributes = n.getAttributes().getNamedItem("Value");
	    if(attributes != null) {
		curValue.setValue(attributes.getNodeValue());
	    }
/*	    attributes = n.getAttributes().getNamedItem("Complexity");
	    if(attributes != null) {
		curValue.setComplexity(attributes.getNodeValue());
	    }
	    attributes = n.getAttributes().getNamedItem("Priority");
	    if(attributes != null) {
		curValue.setPriority(attributes.getNodeValue());
		}*/
	    attributes = n.getAttributes().getNamedItem("Category");
	    if(attributes != null) {
		curValue.setCategory(attributes.getNodeValue());
	    }
	    attributes = n.getAttributes().getNamedItem("Description");
	    if(attributes != null) {
		curValue.setDescription(attributes.getNodeValue());
	    }
	}
 	// Apply method on children
 	NodeList nl = n.getChildNodes();
 	for(int i=0;i<nl.getLength();i++) {
 	    browseXML(nl.item(i));
	}
    }

    /**
     * Return true if the ProfileNode has  data
     * @param pNode The ProfileNode
     * @return true or false
     */
    public boolean isRepository(ProfileNode pNode) {
	return getValues(pNode.getName(), pNode.getSegment(), pNode.getQualifier(), pNode.getQualifier2()).size() != 0 || pNode.getTable() == null;	
    }

    /**
     * Return true if the ProfileNode has  a config data
     * @param pNode The ProfileNode
     * @return true or false
     */
    public boolean hasConfig(ProfileNode pNode) {
	Vector v = getValues(pNode.getName(), pNode.getSegment(), pNode.getQualifier(), pNode.getQualifier2());
	if(v.size() != 0) {
	    if(((ValueNode)v.get(0)).getCategory().equals("Config"))
		return true;
	}
	return false;
    }

    /**
     * Set the current ProfileNode
     * @param pNode The current ProfileNode
     */
    public void setProfileNode(ProfileNode pNode) {
	curPN = pNode;
	gov.nist.hl7maker.util.MVCLog.info("[ReferenceDBModel] Set the current profile node");
	setChanged();
	notifyObservers(pNode);
    }

    /**
     * Get the current ProfileNode
     * @return The current ProfileNode
     */
    public ProfileNode getCurrentProfileNode() {
	return curPN;
    }

    /**
     * Set the current ValueNode
     * @param vNode The current ValueNode
     */
    public void setValueNode(ValueNode vNode) {
	curVN = vNode;
	gov.nist.hl7maker.util.MVCLog.info("[ReferenceDBModel] Set the current value node");
	setChanged();
	notifyObservers(vNode);
    }

    /**
     * Set the ValueNode selected
     * @param vVN An array of ValueNode
     */
    public void setValueNodes(Object[] vVN) {
	curVVN = vVN;
    }

    /**
     * Get the current ValueNode
     * @return The current ValueNode
     */
    public ValueNode getCurrentValueNode() {
	return curVN;
    }

    /**
     * Get the ValueNode selected
     * @return The array of ValueNode
     */
    public Object[] getCurrentValueNodes() {
	return curVVN;
    }

    /**
     * Prepare for a new value
     */
    public void newValue() {
	gov.nist.hl7maker.util.MVCLog.info("[ReferenceDBModel] Prepare GUI for a new value");
	setChanged();
	notifyObservers("new");
    }

    /**
     * Add a new value
     * @param vn The new value
     */
    public void addValue(ValueNode vn) {
	curVN = vn;
	vn.setUsage(curPN.getUsage());
	if(isUnique(vn))
	    vData.add(vn);
	gov.nist.hl7maker.util.MVCLog.info("[ReferenceDBModel] Add a new value");
	setChanged();
	notifyObservers(curPN);
    }

    /**
     * Add a new value
     * @param vn The new value
     * @param forward Forward the addition to MVC listeners
     */
    public void addValue(ValueNode vn, boolean forward) {
	curVN = vn;
	if(isUnique(vn))
	    vData.add(vn);
	if(forward) {
	    gov.nist.hl7maker.util.MVCLog.info("[ReferenceDBModel] Add a new value");
	    setChanged();
	    notifyObservers(curPN);
	}
    }

    /**
     * Modify a value
     * @param vn The modified value
     */
    public void modifyValue(ValueNode vn) {
	curVN = vn;
	if(vn.getStatus() == ValueNode.REMOVED)
	    curVN = null;
	gov.nist.hl7maker.util.MVCLog.info("[ReferenceDBModel] Modify or remove an existing value");
	setChanged();
	notifyObservers(curPN);
    }

    /**
     * Modify a value
     * @param vn The modified value
     * @param forward Forward the addition to MVC listeners
     */
    public void modifyValue(ValueNode vn, boolean forward) {
	curVN = vn;
	if(vn.getStatus() == ValueNode.REMOVED)
	    curVN = null;
	if(forward) {
	    gov.nist.hl7maker.util.MVCLog.info("[ReferenceDBModel] Modify or remove an existing value");
	    setChanged();
	    notifyObservers(curPN);
	}
    }

    /**
     * Init
     */
    public void setInitState() {
	curVN = null;
	gov.nist.hl7maker.util.MVCLog.info("[ReferenceDBModel] Init the GUI");
	setChanged();
	notifyObservers("init");
    }

    /**
     * Init
     * @param pn A ProfileNode
     */
    public void setInitState(ProfileNode pn) {
	setChanged();
	notifyObservers(new MVCArg("init", pn));
    }

    /**
     * Get the values for a name, segment and qualifier
     * @param name The name
     * @param segment The segment
     * @param qualifier The qualifier
     * @param qualifier2 The 2nd qualifier
     * @return A vector with the values
     */
    public Vector getValues(String name, String segment, String qualifier, String qualifier2) {
	Vector v = new Vector();
	ValueNode tmp = new ValueNode();
	tmp.setName(name);
	tmp.setQualifier(qualifier);
	tmp.setQualifier2(qualifier2);
	tmp.setSegment(segment);

	for(int i=0;i<vData.size();i++) {
	    ValueNode vn = (ValueNode)vData.get(i);
	    if(tmp.equals(vn) && vn.getStatus() != ValueNode.REMOVED && !vn.getCategory().equals("DefaultValid") && !vn.getCategory().equals("DefaultInvalid")) {
		v.add(vn);
		if(vn.getCategory().equals("Config")) { 
		    v.clear();
		    v.add(vn);
		    return v;
		}
	    }
	}
	return v;
    }

    /**
     * Get the values
     * @return A vector with the values
     */
    public Vector getValues() {
	return vData;
    }

    /**
     * Save the reference database
     * @param f The destination file
     */
    public void saveToFile(File f) throws ParserConfigurationException, SAXException, IOException, TransformerConfigurationException, TransformerException {
	DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
	DocumentBuilder builder = factory.newDocumentBuilder();
	Document doc = builder.parse(xmlRepository);//new File(""));//session.getReferenceDB()));
	Vector v = getValues();
	long start = System.currentTimeMillis();
	refNode = null;
	for(int i=0;i<v.size();i++) {
	    //	    System.out.println(i + " " + v.size());
	    ValueNode vn = (ValueNode)v.get(i);
	    //	    Node n = null;
	    if(vn.getStatus() == ValueNode.NEW) {
		/*refNode = null;
		  getRefDataNode(doc, vn.getName(), vn.getQualifier(), vn.getQualifier2(), vn.getSegment());*/
		Node tmpNode = refNode;
		refNode = null;
		if(tmpNode != null)
		    getRefDataNode(tmpNode, vn.getName(), vn.getQualifier(), vn.getQualifier2(), vn.getSegment());
		if(refNode == null) {
		    getRefDataNode(doc, vn.getName(), vn.getQualifier(), vn.getQualifier2(), vn.getSegment());
		}
		// Create the reference node if not exists
		if(refNode == null) {
		    Element newRefNode = (Element)doc.createElement("ReferenceData");
		    newRefNode.setAttribute("Name", vn.getName());
		    if(!vn.getQualifier().equals(""))
		    newRefNode.setAttribute("Qualifier", vn.getQualifier());
		    if(!vn.getQualifier2().equals(""))
		    newRefNode.setAttribute("Qualifier2", vn.getQualifier2());
		    newRefNode.setAttribute("Segment", vn.getSegment());
		    newRefNode.setAttribute("Datatype", vn.getDatatype());
		    doc.getDocumentElement().appendChild(newRefNode);
		    refNode = (Node)newRefNode;
		}
		// Create the item node
		Element item = null;
		item = (Element)doc.createElement("Item");
		// Value
		item.setAttribute("Value", vn.getValue());
		// Description
		item.setAttribute("Description", vn.getDescription());
		// Complexity
		//item.setAttribute("Complexity", vn.getComplexity());
		// Priority
		//item.setAttribute("Priority", vn.getPriority());
		// Category
		item.setAttribute("Category", vn.getCategory());
		refNode.appendChild(item);
	    }
	    else if(vn.getStatus() == ValueNode.EDITED) {
/*		refNode = null;
		getRefDataNode(doc, vn.getName(), vn.getQualifier(), vn.getQualifier2(), vn.getSegment());*/
		Node tmpNode = refNode;
		refNode = null;
		if(tmpNode != null)
		    getRefDataNode(tmpNode, vn.getName(), vn.getQualifier(), vn.getQualifier2(), vn.getSegment());
		if(refNode == null) {
		    getRefDataNode(doc, vn.getName(), vn.getQualifier(), vn.getQualifier2(), vn.getSegment());
		}
		itemNode = null;
		getItemNode(refNode, vn.getOriginalValue());
		// Modify the attributes
		NamedNodeMap attrs = itemNode.getAttributes();
		attrs.getNamedItem("Value").setNodeValue(vn.getValue());
		//attrs.getNamedItem("Complexity").setNodeValue(vn.getComplexity());
		//attrs.getNamedItem("Priority").setNodeValue(vn.getPriority());
		attrs.getNamedItem("Category").setNodeValue(vn.getCategory());
		if(attrs.getNamedItem("Description") == null) {
		    ((Element)itemNode).setAttribute("Description", vn.getDescription());
		}
		else {
		    attrs.getNamedItem("Description").setNodeValue(vn.getDescription());
		}
	    }
	    else if(vn.getStatus() == ValueNode.REMOVED) {
/*		refNode = null;
		getRefDataNode(doc, vn.getName(), vn.getQualifier(), vn.getQualifier2(), vn.getSegment());*/
		Node tmpNode = refNode;
		refNode = null;
		if(tmpNode != null)
		    getRefDataNode(tmpNode, vn.getName(), vn.getQualifier(), vn.getQualifier2(), vn.getSegment());
		if(refNode == null) {
		    getRefDataNode(doc, vn.getName(), vn.getQualifier(), vn.getQualifier2(), vn.getSegment());
		}
		if(refNode != null) {
		    itemNode = null;
		    getItemNode(refNode, vn.getOriginalValue());
		    if(itemNode != null)
			refNode.removeChild(itemNode);
		}
	    }
	}
	// Write the DOM document to the file
	Transformer transformer = TransformerFactory.newInstance().newTransformer();
	DOMSource source = new DOMSource(doc);
	StreamResult result = new StreamResult(f.getPath());
	transformer.setOutputProperty(OutputKeys.METHOD, "xml");
	transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "4");
	transformer.setOutputProperty(OutputKeys.INDENT, "yes");
	transformer.transform(source, result);
	loadRepository(f.getPath());
	//session.setReferenceDB(f.getPath());
    }

    /**
     * Get a reference data node
     * @param n The current node
     * @param name The name
     * @param qualifier The qualifier
     * @param qualifier2 The 2nd qualifier
     * @param segment The segment
     * @return The node
     */
    public void getRefDataNode(Node n, String name, String qualifier, String qualifier2, String segment) {
	if(refNode == null) {
	    if(n.getNodeName().compareTo("ReferenceData") == 0) {
		NamedNodeMap attrs = n.getAttributes();
		
		if(attrs.getNamedItem("Name").getNodeValue().compareTo(name) == 0 && (attrs.getNamedItem("Qualifier") != null && attrs.getNamedItem("Qualifier").getNodeValue().compareTo(qualifier) == 0  || attrs.getNamedItem("Qualifier") == null) && (attrs.getNamedItem("Qualifier2") != null && attrs.getNamedItem("Qualifier2").getNodeValue().compareTo(qualifier2) == 0  || attrs.getNamedItem("Qualifier2") == null) && attrs.getNamedItem("Segment").getNodeValue().compareTo(segment) == 0) {
		    refNode = n;
		}
	    }
	    // Apply method on children
	    if(refNode == null) {
		NodeList nl = n.getChildNodes();
		for(int i=0;i<nl.getLength();i++) {
		    getRefDataNode(nl.item(i), name, qualifier, qualifier2, segment);
		}
	    }
	}
    }

    /**
     * Get a item node
     * @param n The current node
     * @param value The value
     * @return The node
     */
    public void getItemNode(Node n, String value) {
	if(itemNode == null) {
	    if(n.getNodeName().compareTo("Item") == 0) {
		NamedNodeMap attrs = n.getAttributes();
		if(attrs.getNamedItem("Value").getNodeValue().compareTo(value) == 0) {
		    itemNode = n;
		}
	    }
	}
	if(itemNode == null) {
	    // Apply method on children
	    NodeList nl = n.getChildNodes();
	    for(int i=0;i<nl.getLength();i++) {
		getItemNode(nl.item(i), value);
	    }
	}
    }

    /**
     * Save the ConfigData file
     * @param f The destination file
     */
    public void saveConfigToFile(File f) throws ParserConfigurationException, SAXException, IOException, TransformerConfigurationException, TransformerException {
	DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
	DocumentBuilder builder = factory.newDocumentBuilder();
	//	Document doc = builder.parse(new File(session.getConfig()));
	Document doc = builder.parse(f);

	//	Name/Segment/Qualifier/Value/Type(ConfigData)
	Vector v = getValues();
	for(int i=0;i<v.size();i++) {
	    ValueNode vn = (ValueNode)v.get(i);
	    //	    Node n = null;
	    if(vn.getStatus() == ValueNode.NEW && vn.getCategory().equals("Config")) {
		//		configNode = null;
		//getConfigDataNode(doc, vn.getName(), vn.getQualifier());
		//if(configNode == null) {
		configNode = (Element)doc.createElement("config");
		// Name
		configNode.setAttribute("Name", vn.getName());
		// Segment
		configNode.setAttribute("Segment", vn.getSegment());
		// Qualifier
		if(!vn.getQualifier().equals(""))
		configNode.setAttribute("Qualifier", vn.getQualifier());
		// Qualifier
		if(!vn.getQualifier2().equals(""))
		configNode.setAttribute("Qualifier2", vn.getQualifier2());
		// Type
		configNode.setAttribute("Type", "ConfigData");
		// Value
		configNode.setAttribute("Value", vn.getValue());
		// Usage
		configNode.setAttribute("Usage", vn.getUsage());
		doc.getDocumentElement().appendChild(configNode);
/*
		// Create the item node
		Element item = null;
		item = (Element)doc.createElement("Item");
		// Value
		item.setAttribute("value", vn.getValue());
		// Description
		item.setAttribute("description", vn.getDescription());
		// Complexity
		item.setAttribute("complexity", vn.getComplexity());
		// Priority
		item.setAttribute("priority", vn.getPriority());
		// Category
		item.setAttribute("category", vn.getCategory());
		refNode.appendChild(item);
*/
	    }
	    else if(vn.getStatus() == ValueNode.EDITED && vn.getCategory().equals("Config")) {
		configNode = null;
		getConfigDataNode(doc, vn.getName(), vn.getSegment(), vn.getQualifier(), vn.getQualifier2());
		// Modify the attributes
		NamedNodeMap attrs = configNode.getAttributes();
		attrs.getNamedItem("Value").setNodeValue(vn.getValue());

/*		refNode = null;
		getRefDataNode(doc, vn.getName(), vn.getQualifier(), vn.getSegment());
		itemNode = null;
		getItemNode(refNode, vn.getOriginalValue());
		// Modify the attributes
		NamedNodeMap attrs = itemNode.getAttributes();
		attrs.getNamedItem("value").setNodeValue(vn.getValue());
		attrs.getNamedItem("complexity").setNodeValue(vn.getComplexity());
		attrs.getNamedItem("priority").setNodeValue(vn.getPriority());
		attrs.getNamedItem("category").setNodeValue(vn.getCategory());
		attrs.getNamedItem("description").setNodeValue(vn.getDescription());
*/
	    }
	    else if(vn.getStatus() == ValueNode.REMOVED && vn.getCategory().equals("Config")) {
		configNode = null;
		getConfigDataNode(doc, vn.getName(), vn.getSegment(), vn.getQualifier(), vn.getQualifier2());
		doc.getDocumentElement().removeChild(configNode);
/*		refNode = null;
		getRefDataNode(doc, vn.getName(), vn.getQualifier(), vn.getSegment());
		itemNode = null;
		getItemNode(refNode, vn.getOriginalValue());
		refNode.removeChild(itemNode);*/
	    }
	}
	// Write the DOM document to the file
	Transformer transformer = TransformerFactory.newInstance().newTransformer();
	DOMSource source = new DOMSource(doc);
/*	if(f == null) {
	    f = new File(session.getConfigData());
	    }*/
	StreamResult result = new StreamResult(f.getPath());
	transformer.transform(source, result);
	//session.setConfigData(f.getPath());
    }

    /**
     * Get a config data node
     * @param n The current node
     * @param name The name
     * @param qualifier The qualifier
     * @param qualifier2 The qualifier2
     * @return The node
     */
    private void getConfigDataNode(Node n, String name, String segment, String qualifier, String qualifier2) {
	if(configNode == null) {
	    if(n.getNodeName().compareTo("config") == 0) {
		NamedNodeMap attrs = n.getAttributes();

		if(attrs.getNamedItem("Name").getNodeValue().compareTo(name) == 0 && (attrs.getNamedItem("Qualifier") != null && attrs.getNamedItem("Qualifier").getNodeValue().compareTo(qualifier) == 0 || attrs.getNamedItem("Qualifier") == null) && (attrs.getNamedItem("Qualifier2") != null && attrs.getNamedItem("Qualifier2").getNodeValue().compareTo(qualifier2) == 0  || attrs.getNamedItem("Qualifier2") == null) && attrs.getNamedItem("Segment").getNodeValue().compareTo(segment) == 0) {
		    configNode = (Element)n;
		}
	    }
	    // Apply method on children
	    if(configNode == null) {
		NodeList nl = n.getChildNodes();
		for(int i=0;i<nl.getLength();i++) {
		    getConfigDataNode(nl.item(i), name, segment, qualifier, qualifier2);
		}
	    }
	}
    }
















    /**
     * Get all the ProfileNode close to the ProfileNode in parameter
     * @param pn a ProfileNode
     * @return A Vector of ProfileNode
     */
    public Vector getSimilarProfileNode(ProfileNode pn) {
	/** An array containing some common English words that are not usually useful
	    for searching. */
	String[] ENGLISH_STOP_WORDS = {
	    "a", "an", "and", "are", "as", "at", "be", "but", "by",
	    "for", "if", "in", "into", "is", "it",
	    "no", "not", "of", "on", "or", "s", "such",
	    "t", "that", "the", "their", "then", "there", "these",
	    "they", "this", "to", "was", "will", "with"
	};

	Vector vTmp = new Vector();
	// Get the datatype
	String datatype = pn.getDatatype();
	// Create a StringTokenizer (delimiter is the space character) with the name of the ProfileNode
	StringTokenizer tokens = new StringTokenizer(pn.getName(), " ");
	// For all ProfileNode with the same datatype, check the name
	boolean stop;
	int xWord = 0, nWord = 0;
	float maxScore = 0;
	for(int i=0;i<vData.size();i++) {
	    ValueNode vn = (ValueNode)vData.get(i);
	    // Same datatype ?
	    //if(datatype.equals(vn.getDatatype())) {
		tokens = new StringTokenizer(pn.getName(), " ");
		// Loop on each word of the name
		xWord = 0;
		nWord = 0;
		while(tokens.hasMoreTokens()) {
		    String token = tokens.nextToken();
		    // Check if the token is a stop word
		    stop = false;
		    for(int s=0;s<ENGLISH_STOP_WORDS.length && !stop;s++) {
			if(token.toLowerCase().equals(ENGLISH_STOP_WORDS[s])) {
			    stop = true;
			}
		    }
		    if(!stop) {
			nWord++;
		    }
		    if(!stop && token.length() > 1 && vn.getName().toLowerCase().indexOf(token.toLowerCase()) != -1) {
			xWord++;
		    }
		}
		if(xWord > 0) {
		    if((float)xWord/nWord > maxScore) {
			maxScore = (float)xWord/nWord;		    
			if(!vTmp.contains(vn.getName())) {
			    vTmp.add(0, vn.getName());
			}
		    }
		    else {
			if(!vTmp.contains(vn.getName())) {
			    vTmp.add(vn.getName());
			}
		    }
		    //			System.out.println(token + "***" + vn.getName());
		}
		//}
	}
	Vector v = new Vector();
	for(int i=0;i<vTmp.size();i++) {
	    //System.out.println(vTmp.get(i));
	    v.add((String)vTmp.get(i));
	    v.addAll(getValues((String)vTmp.get(i)));
	}
	return v;
    }

    /**
     * Get the values for a name
     * @param name The name
     * @return A vector with the values
     */
    private Vector getValues(String name) {
	Vector v = new Vector();
	Vector vTmp = new Vector();
	for(int i=0;i<vData.size();i++) {
	    ValueNode vn = (ValueNode)vData.get(i);
	    if(vn.getName().equals(name)) {
		if(!vTmp.contains(vn.getValue())) {
		    vTmp.add(vn.getValue());
		    v.add(vn);
		}
	    }
	}
	return v;
    }

    /**
     * Is a value unique?
     * @param vn The ValueNode
     * @return Yes or No
     */
    public boolean isUnique(ValueNode vn) {
	for(int i=0;i<vData.size();i++) {
	    ValueNode vnTmp = (ValueNode)vData.get(i);
	    if(vn.equals(vnTmp) && vnTmp.getStatus() != ValueNode.REMOVED) {
		if(vn.getValue().equals(vnTmp.getValue())) {
		    return false;
		}
	    } 
	}
	return true;
    }

    /**
     * Add a value (used for default value)
     * @param pn The ProfileNode
     * @param value The value
     * @param category The category
     */
    public void addValue(ProfileNode pn, String value, String category) {
	ValueNode vn = new ValueNode();

	vn.setStatus(ValueNode.NEW);
	vn.setName(pn.getName());
	vn.setSegment(pn.getSegment());
	vn.setQualifier(pn.getQualifier());
	vn.setQualifier2(pn.getQualifier2());
	vn.setValue(value);
	vn.setCategory(category);
    
	vData.add(vn);
    }

    /**
     * Merge 2 repositories
     * @param rep The other repository
     */
    public void merge(RepositoryDataModel rep) {
	Vector vOtherData = rep.getValues();

	boolean add;
	for(int i=0;i<vOtherData.size();i++) {
	    ValueNode vn = (ValueNode)vOtherData.get(i);
	    if(!vn.getCategory().equals("Constant") && !vn.getCategory().equals("Example")) {
		add = true;
		for(int j=0;j<vData.size();j++) {
		    ValueNode vn1 = (ValueNode)vData.get(j);
		    if(vn.fullEquals(vn1)) {
			add = false;
			j = vData.size() + 1;
		    }
		}
		if(add)
		    vData.add(vn);
	    }
	}
    }

    /**
     * Clean the repository
     */
    public void clean() {
	for(int i=0;i<vData.size();i++) {
	    ValueNode vn = (ValueNode)vData.get(i);
	    if(vn.getCategory().equals("Config") || vn.getCategory().equals("DefaultValid") || vn.getCategory().equals("DefaultInvalid"))
		vn.setStatus(ValueNode.REMOVED);
	}
    }

    /**
     * Fire an event
     * @param arg An MVCArg object
     */
    public void fireEvent(MVCArg arg) {
	setChanged();
	notifyObservers(arg);
    }
}
