/**
 * This class is a model for the table data
 */

package gov.nist.hl7maker.view.config;

// JDK
import java.io.File;
import java.io.IOException;
import java.util.Vector;
import java.util.Observable;
import java.util.Collections;
import java.util.Comparator;
import java.util.StringTokenizer;

// XML
import javax.xml.parsers.DocumentBuilder; 
import javax.xml.parsers.DocumentBuilderFactory;  
import javax.xml.parsers.ParserConfigurationException;

// DOM
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.w3c.dom.NamedNodeMap;

// SAX
import org.xml.sax.SAXException;  

// XSL
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

// MessageMaker
import gov.nist.hl7maker.control.Session;
import gov.nist.hl7maker.util.MVCArg;
import gov.nist.hl7maker.util.MMLog;

public class TableDataModel extends Observable {

    private Session session;
    private Vector vTable, vTableType;
    private String id, type, name;
    private ProfileNode curPN;
    private TableValueNode curTVN;
    private boolean isAllowedToModify;
    private Object[] curVTVN;

    private Node tableNode, itemNode;

    /**
     * Constructor
     * @param session A Session object
     */
    public TableDataModel(Session session) {
	this.session = session;
	vTable = new Vector();
	vTableType = new Vector();
    }

    /**
     * Constructor
     */
    public TableDataModel() {
	vTable = new Vector();
	vTableType = new Vector();
    }

    /**
     * Read the XML table files and collects all values
     */
    public void loadTableData() throws ParserConfigurationException, SAXException, IOException {
	DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
	DocumentBuilder builder = factory.newDocumentBuilder();
	Document doc = null;
	vTable.clear();
	vTableType.clear();

	// Browser the XML file
	loadTableData("HL7", session.getHL7Tables());
	loadTableData("User", session.getUserTables());
	loadTableData("External", session.getExternalTables());
	// Load local tables
	Vector vLocalTable = session.getLocalTable();
	for(int i=0;i<vLocalTable.size();i++) {
	    TablePriority tp = (TablePriority)vLocalTable.get(i);
	    loadLocalTableData(tp.getFilename());
	}
    }

    /**
     * Load a table files and collects all values
     * @param type The type
     *@param filename The filename
     */
    public void loadTableData(String type, String filename) throws ParserConfigurationException, SAXException, IOException {
	// Remove all old data of this type
	for(int i=0;i<vTable.size();i++) {
	    TableValueNode tvn = (TableValueNode)vTable.get(i);
	    // Special Case HL7
	    if(type.equals("HL7")) {
		if(tvn.getType().compareTo("HL7") == 0 || tvn.getType().compareTo("HL7User") == 0) {
		    vTable.remove(tvn);
		    i--;
		}
	    }
	    else {
		if(tvn.getType().compareTo(type) == 0) {
		    vTable.remove(tvn);
		    i--;
		}
	    }
	}
	for(int i=0;i<vTableType.size();i++) {
	    TableNode tn = (TableNode)vTableType.get(i);
	    // Special Case HL7
	    if(type.equals("HL7")) {
		if(tn.getType().compareTo("HL7") == 0 || tn.getType().compareTo("HL7User") == 0) {
		    vTableType.remove(tn);
		    i--;
		}
	    }
	    else {
		if(tn.getType().compareTo(type) == 0) {
		    vTableType.remove(tn);
		    i--;
		}
	    }
	}
	// Check
	for(int i=0;i<vTable.size();i++) {
	    TableValueNode tvn = (TableValueNode)vTable.get(i);
	    if(type.equals("HL7")) {
		if(tvn.getType().compareTo("HL7") == 0 || tvn.getType().compareTo("HL7User") == 0) {
		    System.out.println("TVN Problem in removing old HL7 data");
		    System.exit(-1);
		}
	    }
	    else {
		if(tvn.getType().compareTo(type) == 0) {
		    System.out.println("TVN Problem in removing old data of type [" + type + "]");
		    System.exit(-1);
		}
	    }
	}
	for(int i=0;i<vTableType.size();i++) {
	    TableNode tn = (TableNode)vTableType.get(i);
	    // Special Case HL7
	    if(type.equals("HL7")) {
		if(tn.getType().compareTo("HL7") == 0 || tn.getType().compareTo("HL7User") == 0) {
		    System.out.println("TN Problem in removing old HL7 data");
		    System.exit(-1);
		}
	    }
	    else {
		if(tn.getType().compareTo(type) == 0) {
		    System.out.println("TN Problem in removing old data of type [" + type + "]");
		    System.exit(-1);
		}
	    }
	}

	DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
	DocumentBuilder builder = factory.newDocumentBuilder();
	Document doc = null;
	// Browser the XML file
	doc = builder.parse(filename);//session.getProtocolPath(filename));
        browseXML(doc, null);
	gov.nist.hl7maker.util.MVCLog.info("Tables XML => " + filename + " loaded");
    }
    
    /**
     * Browse recursively the DOM and collects the values
     * @param n The current node
     * @param localType Type is specified for local tables, null otherwise
     */
    private void browseXML(Node n, String localType) {
	// Table
	if(n.getNodeName().equals("Table")) {
	    Node attributes = n.getAttributes().getNamedItem("ID");
	    if(attributes != null) {
		id = attributes.getNodeValue();
	    }
	    attributes = n.getAttributes().getNamedItem("Type");
	    if(attributes != null) {
		type = attributes.getNodeValue();
	    }
	    attributes = n.getAttributes().getNamedItem("Name");
	    if(attributes != null) {
		name = attributes.getNodeValue();
	    }
	    TableNode tn = new TableNode();
	    tn.setId(id);
	    tn.setName(name);
	    if(localType == null)
		tn.setType(type);
	    else 
		tn.setType(localType);
	    if(!vTableType.contains(tn))
		vTableType.add(tn);
	    //else {
		//MMLog.log("Type: " + tn.getType() + " " + localType);
		//MMLog.log("Redondant Table: " + tn.getName());
		//System.err.println("Type: " + tn.getType() + " " + localType);
		//System.err.println("Redondant Table: " + tn.getName());
		//new Exception().printStackTrace();
		//System.exit(0);
	    //}    
	}
	// Item
	else if(n.getNodeName().equals("Item")) {
	    TableValueNode tvn = new TableValueNode();
	    tvn.setId(id);
	    if(localType == null)
		tvn.setType(type);
	    else
		tvn.setType(localType);
	    tvn.setName(name);
	    Node attributes = n.getAttributes().getNamedItem("Code");
	    if(attributes != null) {
		tvn.setCode(attributes.getNodeValue());
	    }
	    else {
		tvn.setCode("");
	    }
	    attributes = n.getAttributes().getNamedItem("Value");
	    if(attributes != null) {
		tvn.setItem(attributes.getNodeValue());
	    }
	    else {
		tvn.setItem("");
	    }
	    if(!vTable.contains(tvn))
		vTable.add(tvn);
	    //else {
		//MMLog.log("Type: " + tvn.getType() + " " + localType);
		//MMLog.log("Redondant Item: " + tvn.getName() + " " + tvn.getCode() + " " + tvn.getItem());
		//		System.err.println("Type: " + tvn.getType() + " " + localType);
		//System.err.println("Redondant Item: " + tvn.getName() + " " + tvn.getCode() + " " + tvn.getItem());
	    //}
	}
 	// Apply method on children
 	NodeList nl = n.getChildNodes();
 	for(int i=0;i<nl.getLength();i++) {
 	    browseXML(nl.item(i), localType);
	}
    }

    /**
     * Load a local table
     * @param filename
     */
    public void loadLocalTableData(String filename) throws ParserConfigurationException, SAXException, IOException {
	// Generate the type
	String type = "Local[" + new File(filename).getName() + "]";
	cleanType(type);

	DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
	DocumentBuilder builder = factory.newDocumentBuilder();
	Document doc = null;
	// Browser the XML file
	doc = builder.parse(session.getProtocolPath(filename));
        browseXML(doc, type);
	gov.nist.hl7maker.util.MVCLog.info("Tables XML => " + filename + " loaded");
    }

    /**
     * Remove old value for a certain type
     * @param type The type
     */
    public void cleanType(String type) {
	// Remove all old data of this type
	for(int i=0;i<vTable.size();i++) {
	    TableValueNode tvn = (TableValueNode)vTable.get(i);
	    if(tvn.getType().compareTo(type) == 0) {
		vTable.remove(tvn);
		i--;
	    }
	}
	for(int i=0;i<vTableType.size();i++) {
	    TableNode tn = (TableNode)vTableType.get(i);
	    if(tn.getType().compareTo(type) == 0) {
		vTableType.remove(tn);
		i--;
	    }
	}
	// Check
	for(int i=0;i<vTable.size();i++) {
	    TableValueNode tvn = (TableValueNode)vTable.get(i);
	    if(tvn.getType().compareTo(type) == 0) {
		System.out.println("TVN Problem in removing old data of type [" + type + "]");
		System.exit(-1);
	    }
	}
	for(int i=0;i<vTableType.size();i++) {
	    TableNode tn = (TableNode)vTableType.get(i);
	    if(tn.getType().compareTo(type) == 0) {
		System.out.println("TN Problem in removing old data of type [" + type + "]");
		System.exit(-1);
	    }
	}
    }

    /**
     * Return true if the ProfileNode has table data
     * @param pNode The ProfileNode
     * @return true or false
     */
    public boolean isTable(ProfileNode pNode) {
	return pNode.getTable() != null;
    }

     /**
     * Set the current ProfileNode
     * @param pNode The current ProfileNode
     */
    public void setProfileNode(ProfileNode pNode) {
	curPN = pNode;
	gov.nist.hl7maker.util.MVCLog.info("[TableDataModel] Set the current profile node");
	setChanged();
	notifyObservers(pNode);
    }

    /**
     * Get the current ProfileNode
     * @return The current ProfileNode
     */
    public ProfileNode getCurrentProfileNode() {
	return curPN;
    }

    /**
     * Set the current TableValueNode
     * @param tvNode The current TableValueNode
     */
    public void setTableValueNode(TableValueNode tvNode) {
	curTVN = tvNode;
	gov.nist.hl7maker.util.MVCLog.info("[TableDataModel] Set the current table value node");
	setChanged();
	notifyObservers(tvNode);
    }

    /**
     * Set the TableValueNode selected
     * @param vTVN An array of TableValueNode
     */
    public void setTableValueNodes(Object[] vTVN) {
	curVTVN = vTVN;
    }

    /**
     * Get the current TableValueNode
     * @return The current TableValueNode
     */
    public TableValueNode getCurrentTableValueNode() {
	return curTVN;
    }

    /**
     * Get the TableValueNode selected
     * @return The array of TableValueNode
     */
    public Object[] getCurrentTableValueNodes() {
	return curVTVN;
    }

    /**
     * Prepare for a new value
     */
    public void newValue() {
	gov.nist.hl7maker.util.MVCLog.info("[TableDataModel] Prepare GUI for a new value");
	setChanged();
	notifyObservers("new");
    }

    /**
     * Add a new value
     * @param tvn The new value
     */
    public void addValue(TableValueNode tvn) {
	if(isUnique(tvn)) {
	    tvn.setId(curPN.getTable());
	    vTable.add(tvn);
	}
	gov.nist.hl7maker.util.MVCLog.info("[TableDataModel] Add a new value");
	setChanged();
	notifyObservers(curPN);
    }

    /**
     * Add a new value (for TableViewer)
     * @param tvn The new value
     */
    public void addValueTableViewer(TableValueNode tvn) {
	vTable.add(tvn);
	gov.nist.hl7maker.util.MVCLog.info("[TableDataModel] Add a new value (for TableViewer)");
	setChanged();
	notifyObservers(new MVCArg("updateLists", tvn.getId()));
    }

    /**
     * Modify a value
     * @param tvn The modified value
     */
    public void modifyValue(TableValueNode tvn) {
	gov.nist.hl7maker.util.MVCLog.info("[TableValueModel] Modify or remove an existing value");
	setChanged();
	notifyObservers(curPN);
    }

    /**
     * Modify a value (for TableViewer)
     * @param tvn The modified value
     */
    public void modifyValueTableViewer(TableValueNode tvn) {
	gov.nist.hl7maker.util.MVCLog.info("[TableValueModel] Modify or remove an existing value (for TableViewer)");
	setChanged();
	notifyObservers(new MVCArg("updateLists", tvn.getId()));
    }

    /**
     * Init
     */
    public void setInitState() {
	curTVN = null;
	gov.nist.hl7maker.util.MVCLog.info("[TableValueModel] Init the GUI");
	setChanged();
	notifyObservers("init");
    }

    /**
     * Switch from Table to Config data
     */
    public void switchConfig() {
	gov.nist.hl7maker.util.MVCLog.info("[TableValueModel] Switch to the config panel");
	setChanged();
	notifyObservers("switch");
    }

    /**
     * Get the values for a table id
     * @param tableId The id of the table
     * @return A vector with the values
     */
    public Vector getValues(String tableId) {
	Vector v = new Vector();
	//TableValueNode tmp = new TableValueNode();
	//tmp.setId(tableId);
	for(int i=0;i<vTable.size();i++) {
	    TableValueNode tvn = (TableValueNode)vTable.get(i);
	    if(tvn.getId().equals(tableId) && tvn.getStatus() != TableValueNode.REMOVED) {
		v.add(vTable.get(i));
	    }
	}
	return v;
    }

    /**
     * Get the values
     * @return A vactor with the values
     */
    public Vector getValues() {
	return vTable;
    }

    /**
     * Return true if the values are allowed to be modified
     * @return true if you are allowed to modify
     */
    public boolean isAllowedToModify() {
	//TableValueNode tmp = new TableValueNode();
	if(curPN == null)
	    return true;
	//	tmp.setId(curPN.getTable());
	
	isAllowedToModify = true;
	for(int i=0;i<vTable.size();i++) {
	    TableValueNode tvn = (TableValueNode)vTable.get(i);
	    if(tvn.getId().equals(curPN.getTable()) && tvn.getStatus() != TableValueNode.REMOVED) {
		if(tvn != null && tvn.getType().compareTo("HL7") == 0) {
		    isAllowedToModify = false;
		}
		i = vTable.size() + 1;
	    }
	} 
	return isAllowedToModify;
    }

    /**
     * Return true if the values are allowed to be modified
     * @param tableId The Table Id
     * @return true if you are allowed to modify
     */
    public boolean isAllowedToModify(String tableId) {
	TableValueNode tmp = new TableValueNode();
	tmp.setId(tableId);
	
	isAllowedToModify = true;
	for(int i=0;i<vTable.size();i++) {
	    TableValueNode tvn = (TableValueNode)vTable.get(i);
	    if(tvn.getId().equals(tableId) && tvn.getStatus() != TableValueNode.REMOVED) {
		if(tvn != null && tvn.getType().compareTo("HL7") == 0) {
		    isAllowedToModify = false;
		}
		i = vTable.size() + 1;
	    }
	} 
	return isAllowedToModify;
    }

    /**
     * Get the type of the table
     * @return The type of the table
     */
    public String getTableType() {
	String type = "";
	for(int i=0;i<vTableType.size();i++) {
	    TableNode tn = (TableNode)vTableType.get(i);
	    if(tn.getId().compareTo(curPN.getTable()) == 0) {
		type = tn.getType();
		i = vTableType.size() + 1;
	    }
	}
	return type;
    }

    /**
     * Get the type of the table
     * @param id The table id
     * @return The type of the table
     */
    public String getTableType(String id) {
	String type = "";
	for(int i=0;i<vTableType.size();i++) {
	    TableNode tn = (TableNode)vTableType.get(i);
	    if(tn.getId().compareTo(id) == 0) {
		type = tn.getType();
		i = vTableType.size() + 1;
	    }
	}
	return type;
    }

    /**
     * Get the name of the table
     * @param id The table id
     * @return The name of the table
     */
    public String getTableName(String id) {
	String name = "";
	for(int i=0;i<vTableType.size();i++) {
	    TableNode tn = (TableNode)vTableType.get(i);
	    if(tn.getId().compareTo(id) == 0) {
		name = tn.getName();
		i = vTableType.size() + 1;
	    }
	}
	return name;
    }

    /**
     * Get information about all the table 
     * @return A vector of TableNode
     */
    public Vector getTablesInformation() {
	Collections.sort(vTableType, new Comparator() {
		public int compare(Object o1, Object o2) {
		    TableNode tn1 = (TableNode)o1;
		    TableNode tn2 = (TableNode)o2;
		    return tn1.getId().compareTo(tn2.getId());
		}
	    });			
	return vTableType;
    }

    /**
     * is id unique?
     * @param id The Id
     * @return Return true if the id is unique
     */
    public boolean isUnique(String id) {
	// Format the id with 4 digits
	id = formatId(id);
	for(int i=0;i<vTableType.size();i++) {
	    TableNode tn = (TableNode)vTableType.get(i);
	    if(tn.getId().compareTo(id) == 0) {
		return false;
	    }
	}
	return true;
    }

    /** 
     * Format the id with this format xxxx
     * @param id The id
     * @return The formatted id
     */
    private String formatId(String id) {
	if(id.length() < 4) {
	    int n = 4 - id.length();
	    for(int i=0;i<n;i++) {
		id = "0" + id;
	    }
	}
	return id;
    }
    /**
     * Add a Table
     * @param id The id
     * @param name The name
     * @param type The Type
     */
    public void addTable(String id, String name, String type) {
	TableNode tn = new TableNode();
	tn.setId(formatId(id));
	tn.setName(name);
	tn.setType(type);
	vTableType.add(tn);
	gov.nist.hl7maker.util.MVCLog.info("[TableDataModel] Add a table");
	setChanged();
	notifyObservers("addTable");
    }

    /**
     * Modify a Table
     * @param id The id
     * @param name The name
     * @param type The Type
     */
    public void modifyTable(String id, String name, String type) {
	TableNode tmp = new TableNode();
	tmp.setId(formatId(id));
	TableNode tn = (TableNode)vTableType.get(vTableType.indexOf(tmp));
	tn.setName(name);
	tn.setType(type);
	// Modify the TableValueNode
	for(int i=0;i<vTable.size();i++) {
	    TableValueNode tvn = (TableValueNode)vTable.get(i);
	    if(tvn.getId().compareTo(id) == 0) {
		tvn.setStatus(TableValueNode.EDITED);
		tvn.setName(name);
	    }
	}
	gov.nist.hl7maker.util.MVCLog.info("[TableDataModel] Modify an existing table");
	setChanged();
	notifyObservers("modifyTable");
    }

    /**
     * Remove a Table
     * @param id The id
     */
    public void removeTable(String id) {
	TableNode tmp = new TableNode();
	tmp.setId(formatId(id));
	int idx = vTableType.indexOf(tmp);
	if(idx != -1) {
	    vTableType.removeElementAt(idx);
	    // Modify the TableValueNode
	    for(int i=0;i<vTable.size();i++) {
		TableValueNode tvn = (TableValueNode)vTable.get(i);
		if(tvn.getId().compareTo(id) == 0) {
		    tvn.setStatus(TableValueNode.REMOVED);
		}
	    }
	    gov.nist.hl7maker.util.MVCLog.info("[TableDataModel] Remove an existing table");
	    setChanged();
	    notifyObservers("removeTable");
	}
    }

    /**
     * New Table Click
     */
    public void newTable() {
	gov.nist.hl7maker.util.MVCLog.info("[TableDataModel] Prepare GUI for a new table");
	setChanged();
	notifyObservers("newTable");
    }

    /**
     * Edit Table Click
     */
    public void editTable() {
	gov.nist.hl7maker.util.MVCLog.info("[TableDataModel] Prepare GUI for editing table");
	setChanged();
	notifyObservers("editTable");
    }

    /**
     * Cancel Click
     */
    public void cancel() {
	gov.nist.hl7maker.util.MVCLog.info("[TableDataModel] GUI Cancel button");
	setChanged();
	notifyObservers("cancel");
    }

    /**
     * Init the TableViewer Dialog
     * @param param A parameter
     */
    public void initTableViewer(Object param) {
	//setChanged();
	if(param != null) {
	    gov.nist.hl7maker.util.MVCLog.info("[TableDataModel] Init the GUI of the Table Viewer");
	    setChanged();
	    notifyObservers(param);
	}
	else {
	    System.err.println("initTableViewer: The parameter is null, check the context");
	    System.exit(0);
	    //notifyObservers("initTable");
	}
    }

    /**
     * Update the values of code and item JList
     * @param param A parameter
     */
    public void updateLists(Object param) {
	if(param != null) {
	    gov.nist.hl7maker.util.MVCLog.info("[TableDataModel] Update the data lists in the Table Viewer");
	    setChanged();
	    notifyObservers(param);
	}
	else {
	    System.err.println("updateLists: The parameter is null, check the context");
	    System.exit(0);
	    //notifyObservers("updateLists");
	}
    }

    /**
     * Return true if a value has been modified
     * @return true or false
     */
    public boolean needToSave() {
	Vector v = getValues();

	for(int i=0;i<v.size();i++) {
	    TableValueNode tvn = (TableValueNode)v.get(i);
	    if(tvn.getStatus() == TableValueNode.NEW || tvn.getStatus() == TableValueNode.EDITED || tvn.getStatus() == TableValueNode.REMOVED) {
		return true;
	    }
	}
	return false;
    }

    /**
     * Return true if a value has been modified
     * @param type The type
     * @return true or false
     */
    public boolean needToSave(String type) {
	Vector v = getValues();

	for(int i=0;i<v.size();i++) {
	    TableValueNode tvn = (TableValueNode)v.get(i);
	    if(tvn.getType().equals(type)) {
		if(tvn.getStatus() == TableValueNode.NEW || tvn.getStatus() == TableValueNode.EDITED || tvn.getStatus() == TableValueNode.REMOVED) {
		    return true;
		}
	    }
	}
	return false;
    }

    /**
     * Save the Tables
     * @param type The type
     * @param srcName The original filename
     * @param dst The destination file
     */
    public void saveTableFile(String type, String srcName, File dst) throws ParserConfigurationException, SAXException, IOException, TransformerConfigurationException, TransformerException {
	DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
	DocumentBuilder builder = factory.newDocumentBuilder();
	Document doc = builder.parse(new File(srcName));
	Vector vRemove = new Vector();

	Vector v = getValues();
	for(int i=0;i<v.size();i++) {
	    TableValueNode tvn = (TableValueNode)v.get(i);
	    Node n = null;
	    if(tvn.getStatus() == TableValueNode.NEW) {
		tableNode = null;
		if(tvn.getType().compareTo(type) == 0) {
		    getTableNode(doc, tvn.getId());
		    // Create the table Node if needed
		    if(tableNode == null) {
			Element newTableNode = (Element)doc.createElement("Table");
			if(type.startsWith("Local"))
			    newTableNode.setAttribute("Type", "Local");
			else
			    newTableNode.setAttribute("Type", type);
			newTableNode.setAttribute("Name", tvn.getName());
			newTableNode.setAttribute("ID", tvn.getId());
			doc.getDocumentElement().appendChild(newTableNode);
			tableNode = newTableNode;
		    }
		    // Create the item node
		    Element item = null;
		    if(tvn.getType().compareTo(type) == 0) {
			item = (Element)doc.createElement("Item");
		    }
		    // Code
		    item.setAttribute("Code", tvn.getCode());
		    // Value
		    item.setAttribute("Value", tvn.getItem());
		    tableNode.appendChild(item);
		}
	    }
	    else if(tvn.getStatus() == TableValueNode.EDITED) {
		tableNode = null;
		if(tvn.getType().compareTo(type) == 0) {
		    getTableNode(doc, tvn.getId());
		    // Set the name
		    NamedNodeMap attrs = tableNode.getAttributes();
		    attrs.getNamedItem("Name").setNodeValue(tvn.getName());
		    itemNode = null;
		    getItemNode(tableNode, tvn.getOriginalCode(), tvn.getOriginalItem());
		    // Modify the attributes
		    attrs = itemNode.getAttributes();
		    attrs.getNamedItem("Code").setNodeValue(tvn.getCode());
		    attrs.getNamedItem("Value").setNodeValue(tvn.getItem());
		}
	    }
	    else if(tvn.getStatus() == TableValueNode.REMOVED) {
		tableNode = null;
		if(tvn.getType().compareTo(type) == 0) {
		    getTableNode(doc, tvn.getId());		
		    itemNode = null;
		    if(!vRemove.contains(tvn.getId())) {
			vRemove.add(tvn.getId());
		    }
		    getItemNode(tableNode, tvn.getOriginalCode(), tvn.getOriginalItem());
		    if(itemNode != null) {
			tableNode.removeChild(itemNode);
		    }
		}
	    }
	}

	// Check if there is still elements for this table
	for(int i=0;i<vRemove.size();i++) {
	    //	    if(getValues((String)vRemove.get(i)).size() == 0) {
	    if(getTableType((String)vRemove.get(i)).compareTo("") == 0) {
		tableNode = null;
		getTableNode(doc, (String)vRemove.get(i));				
		doc.getDocumentElement().removeChild(tableNode);
	    }
	}
	// Write the DOM document to the file
	Transformer transformer = TransformerFactory.newInstance().newTransformer();
	DOMSource source = new DOMSource(doc);
	StreamResult result = new StreamResult(dst.getPath());
	transformer.transform(source, result);
	if(type.equals("HL7User")) {
	    session.setHL7Tables(dst.getPath());
	    loadTableData("HL7", session.getHL7Tables());
	}
	else if(type.equals("User")) {
	    session.setUserTables(dst.getPath());
	    loadTableData("User", session.getUserTables());
	}
	else if(type.equals("External")) {
	    session.setExternalTables(dst.getPath());
	    loadTableData("External", session.getExternalTables());
	}
	else if(type.startsWith("Local")) {
	    Vector vLocalTable = session.getLocalTable();
	    for(int i=0;i<vLocalTable.size();i++) {
		TablePriority tp = (TablePriority)vLocalTable.get(i);
		if(srcName.equals(tp.getFilename())) {
		    cleanType(type);
		    tp.setFilename(dst.getPath());
		    vLocalTable.setElementAt(tp, i);
		}
	    }
	    session.setLocalTable(vLocalTable, !srcName.equals(dst.getPath()));
	    for(int i=0;i<vLocalTable.size();i++) {
		TablePriority tp = (TablePriority)vLocalTable.get(i);
		loadLocalTableData(tp.getFilename());
	    }
	}
    }

    /**
     * Cancel the modifications made in TableViewer
     */
    public void cancelModification() {
	try {
	    loadTableData();
	}
	catch(ParserConfigurationException pce) {
	    System.err.println("ParserConfigurationException " + pce.getMessage());
	    pce.printStackTrace();
	}
	catch(SAXException saxe) {
	    System.err.println("SAXException " + saxe.getMessage());
	    saxe.printStackTrace();
	}
	catch(IOException ioe) {
	    System.err.println("IOException " + ioe.getMessage());
	    ioe.printStackTrace();
	}
    }

    /**
     * Get a table node
     * @param n The current node
     * @param id The id
     * @return The node
     */
    public void getTableNode(Node n, String id) {
	if(tableNode == null) {
	    if(n.getNodeName().compareTo("Table") == 0) {
		NamedNodeMap attrs = n.getAttributes();

		if(attrs.getNamedItem("ID").getNodeValue().compareTo(id) == 0) {
		    tableNode = n;
		}
	    }
	    // Apply method on children
	    if(tableNode == null) {
		NodeList nl = n.getChildNodes();
		for(int i=0;i<nl.getLength();i++) {
		    getTableNode(nl.item(i), id);
		}
	    }
	}
    }

    /**
     * Get a item node
     * @param n The current node
     * @param code The code
     * @param value The value
     * @return The node
     */
    public void getItemNode(Node n, String code, String value) {
	if(itemNode == null) {
	    if(n.getNodeName().compareTo("Item") == 0) {
		NamedNodeMap attrs = n.getAttributes();
		if(attrs.getNamedItem("Code").getNodeValue().compareTo(code) == 0 && attrs.getNamedItem("Value").getNodeValue().compareTo(value) == 0) {
		    itemNode = n;
		}
	    }
	}
	if(itemNode == null) {
	    // Apply method on children
	    NodeList nl = n.getChildNodes();
	    for(int i=0;i<nl.getLength();i++) {
		getItemNode(nl.item(i), code, value);
	    }
	}
    }

    /**
     * Get all the ProfileNode close to the ProfileNode in parameter
     * @param pn a ProfileNode
     * @return A Vector of ProfileNode
     */
    public Vector getSimilarProfileNode(ProfileNode pn) {
	/** An array containing some common English words that are not usually useful
	    for searching. */
	String[] ENGLISH_STOP_WORDS = {
	    "a", "an", "and", "are", "as", "at", "be", "but", "by",
	    "for", "if", "in", "into", "is", "it",
	    "no", "not", "of", "on", "or", "s", "such",
	    "t", "that", "the", "their", "then", "there", "these",
	    "they", "this", "to", "was", "will", "with"
	};

	Vector vTmp = new Vector();
/*	// Get the tableID
	String datatype = pn.getDatatype();*/
	// Create a StringTokenizer (delimiter is the space character) with the name of the ProfileNode
	StringTokenizer tokens = new StringTokenizer(pn.getName(), " ");
	// For all ProfileNode with the same datatype, check the name
	boolean stop;
	int xWord = 0, nWord = 0;
	float maxScore = 0;
	for(int i=0;i<vTable.size();i++) {
	    TableValueNode tvn = (TableValueNode)vTable.get(i);
	    /*	    // Same datatype ?
		    if(datatype.equals(vn.getDatatype())) {*/
	    tokens = new StringTokenizer(pn.getName(), " ");
	    // Loop on each word of the name
	    xWord = 0;
	    nWord = 0;
	    while(tokens.hasMoreTokens()) {
		String token = tokens.nextToken();
		// Check if the token is a stop word
		stop = false;
		for(int s=0;s<ENGLISH_STOP_WORDS.length && !stop;s++) {
		    if(token.toLowerCase().equals(ENGLISH_STOP_WORDS[s])) {
			stop = true;
		    }
		}
		if(!stop) {
		    nWord++;
		}
		if(!stop && token.length() > 1 && tvn.getName().toLowerCase().indexOf(token.toLowerCase()) != -1) {
		    xWord++;
		}
	    }
	    if(xWord > 0) {
		if((float)xWord/nWord > maxScore) {
		    maxScore = (float)xWord/nWord;		    
			if(!vTmp.contains(tvn.getName())) {
			    vTmp.add(0, tvn.getName());
			}
		    }
		else {
		    if(!vTmp.contains(tvn.getName())) {
			vTmp.add(tvn.getName());
		    }
		}
		//System.out.println(token + "***" + tvn.getName());
		//}
	    }
	}
	Vector v = new Vector();
	for(int i=0;i<vTmp.size();i++) {
	    //System.out.println(vTmp.get(i));
	    v.add((String)vTmp.get(i));
	    v.addAll(getValuesByName((String)vTmp.get(i)));
	}
	return v;
    }

    /**
     * Get the values for a name
     * @param name The name
     * @return A vector with the values
     */
    private Vector getValuesByName(String name) {
	Vector v = new Vector();
	Vector vTmp = new Vector();
	for(int i=0;i<vTable.size();i++) {
	    TableValueNode tvn = (TableValueNode)vTable.get(i);
	    if(tvn.getName().equals(name)) {
		if(!vTmp.contains(tvn.getItem())) {
		    vTmp.add(tvn.getItem());
		    //		    tvn.setLstType(TableValueNode.ITEM);
		    TableValueNode newTVN = new TableValueNode();
		    newTVN.setId(curPN.getTable());
		    newTVN.setType("User");
		    newTVN.setName(tvn.getName());
		    newTVN.setCode(tvn.getOriginalCode());
		    newTVN.setItem(tvn.getOriginalItem());
		    newTVN.modifyCode(tvn.getCode());
		    newTVN.modifyItem(tvn.getItem());
		    newTVN.setLstType(TableValueNode.ITEM);
		    newTVN.setStatus(TableValueNode.NEW);
		    v.add(newTVN);
		}
	    }
	}
	return v;
    }

    /**
     * Is a value unique?
     * @param tvn The TableValueNode
     * @return Yes or No
     */
    public boolean isUnique(TableValueNode tvn) {
	for(int i=0;i<vTable.size();i++) {
	    TableValueNode tvnTmp = (TableValueNode)vTable.get(i);
	    if(tvn.equals(tvnTmp) && tvnTmp.getStatus() != TableValueNode.REMOVED) {
		if(tvn.getValue().equals(tvnTmp.getValue())) {
		    return false;
		}
	    } 
	}
	return true;
    }
}
