/**
 * This class loads a meta data file and provide information about it
 */

package gov.nist.hl7maker.view.display;

// JDK
import java.io.File;
import java.io.IOException;
import java.util.Vector;

// XML
import javax.xml.parsers.DocumentBuilder; 
import javax.xml.parsers.DocumentBuilderFactory;  
import javax.xml.parsers.ParserConfigurationException;

// DOM
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

// SAX
import org.xml.sax.SAXException;  

// MessageMaker
import gov.nist.hl7maker.util.Constants;
import gov.nist.hl7maker.util.SAXErrorHandler;

public class MetaData {

    private String xmlMD;

    private String HL7Version;
    private String session;
    private String profile;
    private String msgPath;
    private String date;
    private String organization;
    private String description;
    private String volume;
    private String cardinality;
    private String usage;
    private String type;

    private boolean bHL7Version;
    private boolean bSession;
    private boolean bProfile;
    private boolean bMsgPath;
    private boolean bDate;
    private boolean bOrganization;
    private boolean bDescription;
    private boolean bVolume;
    private boolean bCardinality;
    private boolean bUsage;
    private boolean bType;

    private boolean bMsg;
    private boolean bMsgId;
    private boolean bMsgName;
    private boolean bMsgType;
    private boolean bMsgDescription;
    private boolean bMsgSegment;
    private boolean bMsgField;
    private boolean bMsgComponent;
    private boolean bMsgSubComponent;

    private Vector vMsg;
    private MDMsg currentMsg;

    /**
     * Constructor
     *@param xmlMD The metadata filename
     */
    public MetaData(String xmlMD) throws ParserConfigurationException, SAXException, IOException {
	loadMetaData(xmlMD);
    }

    // Get Methods
    public String getName() {
	return xmlMD;
    }

    public String getHL7Version() {
	return HL7Version;
    }

    public String getSession() {
	return session;
    }

    public String getProfile() {
	return profile;
    }

    public String getMsgPath() {
	return msgPath;
    }

    public String getDate() {
	return date;
    }

    public String getOrganization() {
	return organization;
    }

    public String getDescription() {
	return description;
    }

    public String getVolume() {
	return volume;
    }
	
    public String getCardinality() {
	return cardinality;
    }
	
    public String getUsage() {
	return usage;
    }

    public String getType() {
	return type;
    }

    // Methods

    /**
     * Get the list of messages
     * @return The list of messages
     */
    public Vector getMessageList() {
	Vector vRes = new Vector();

	for(int i=0;i<vMsg.size();i++) {
	    if(new File(msgPath + ((MDMsg)vMsg.get(i)).name).exists()) {
		vRes.add(((MDMsg)vMsg.get(i)).name);
	    }
	}
	return vRes;
    }

    /**
     * Get the type of a message
     * @param msgFile
     * @return The type
     */
    public String getMsgType(String msgFile) {
	int idx = vMsg.indexOf(new MDMsg(msgFile));
	if(idx != -1) {
	    return ((MDMsg)vMsg.get(idx)).type;
	}
	return null;
    }

    /**
     * Get the description of a message
     * @param msgFile
     * @return The description
     */
    public String getMsgDescription(String msgFile) {
	int idx = vMsg.indexOf(new MDMsg(msgFile));
	if(idx != -1) {
	    return ((MDMsg)vMsg.get(idx)).description;
	}
	return null;
    }

    /**
     * Get the message variant description
     * @param msgFile
     * @return The message variant description
     */
    public String getVarDescription(String msgFile) {
	int idx = vMsg.indexOf(new MDMsg(msgFile));
	if(idx != -1) {
	    return ((MDMsg)vMsg.get(idx)).variantDescription;
	}
	return null;
    }

    /**
     * Get the specific location of the variant of error
     * @param msgFile
     * @return The specific location
     */
    public String getLocation(String msgFile) {
	int idx = vMsg.indexOf(new MDMsg(msgFile));
	if(idx != -1) {
	    return ((MDMsg)vMsg.get(idx)).getLocation();
	}
	return null;
    }

    /**
     * Get the MDMsg object
     * @param msgFile
     * @return The MDMsg object
     */
    public MDMsg getMDMsg(String msgFile) {
	int idx = vMsg.indexOf(new MDMsg(msgFile));
	if(idx != -1) {
	    return (MDMsg)vMsg.get(idx);
	}
	return null;
    }

    /**
     * Return true if the message is part of the generation process associated with the metadata
     * @param xmlFile The Message
     * @return true or false
     */
    public boolean hasMessage(String xmlFile) {
	int idx = xmlFile.lastIndexOf("\\");
	if(idx == -1)
	    idx = xmlFile.lastIndexOf("/");
	return vMsg.indexOf(new MDMsg(xmlFile.substring(idx+1))) != -1;
    }

    /**
     * Get the highest msg number
     * @return The Highest message number
     */
/*    public int getMaxMsgNumber() {
	int max = -1;
	for(int i=0;i<vMsg.size();i++) {
	    MDMsg msg = (MDMsg)vMsg.get(i);
	    max = Math.max(max, Integer.parseInt(msg.name.substring(msg.name.length()-9, msg.name.length()-4)));
	}
	return max;
	}*/

    /**
     * Load the metadata file
     * @param xmlMD The Metadata filename
     */
    public void loadMetaData(String xmlMD) throws ParserConfigurationException, SAXException, IOException {
	DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
	DocumentBuilder builder = factory.newDocumentBuilder();
	builder.setErrorHandler(new SAXErrorHandler());
	Document doc = builder.parse(new File(xmlMD));

	vMsg = new Vector();
	browseXML(doc);
	this.xmlMD = xmlMD;
    }

    /**
     * Browser recursively the DOM and creates the tree structure
     * @param n The current node
     */
    private void browseXML(Node n) {
	if(n.getNodeName().equals("HL7version")) {
	    bHL7Version = true;
	}
	else if(n.getNodeType() == Node.TEXT_NODE && bHL7Version) {
	    HL7Version = n.getNodeValue();
	    bHL7Version = false;
	}
	else if(n.getNodeName().equals("Session")) {
	    bSession = true;
	}
	else if(n.getNodeType() == Node.TEXT_NODE && bSession) {
	    session = n.getNodeValue();
	    bSession = false;
	}
	else if(n.getNodeName().equals("Profile")) {
	    bProfile = true;
	}
	else if(n.getNodeType() == Node.TEXT_NODE && bProfile) {
	    profile = n.getNodeValue();
	    bProfile = false;
	}
	else if(n.getNodeName().equals("MessagePath")) {
	    bMsgPath = true;
	}
	else if(n.getNodeType() == Node.TEXT_NODE && bMsgPath) {
	    msgPath = n.getNodeValue();
	    bMsgPath = false;
	}
	else if(n.getNodeName().equals("Date")) {
	    bDate = true;
	}
	else if(n.getNodeType() == Node.TEXT_NODE && bDate) {
	    date = n.getNodeValue();
	    bDate = false;
	}
	else if(n.getNodeName().equals("Organization")) {
	    bOrganization = true;
	}
	else if(n.getNodeType() == Node.TEXT_NODE && bOrganization) {
	    organization = n.getNodeValue();
	    bOrganization = false;
	}
	else if(!bMsg && n.getNodeName().equals("Description")) {
	    bDescription = true;
	}
	else if(n.getNodeType() == Node.TEXT_NODE && bDescription) {
	    description = n.getNodeValue();
	    bDescription = false;
	}
	else if(n.getNodeName().equals("MessageCountTotal")) {
	    bVolume = true;
	}
	else if(n.getNodeType() == Node.TEXT_NODE && bVolume) {
	    volume = n.getNodeValue();
	    bVolume = false;
	}
	else if(n.getNodeName().equals("Cardinality")) {
	    bCardinality = true;
	}
	else if(n.getNodeType() == Node.TEXT_NODE && bCardinality) {
	    cardinality = n.getNodeValue();
	    bCardinality = false;
	}
	else if(n.getNodeName().equals("Usage")) {
	    bUsage = true;
	}
	else if(n.getNodeType() == Node.TEXT_NODE && bUsage) {
	    usage = n.getNodeValue();
	    bUsage = false;
	}
	else if(!bMsg && n.getNodeName().equals("MessageTypes")) {
	    bType = true;
	}
	else if(n.getNodeType() == Node.TEXT_NODE && bType) {
	    type = n.getNodeValue();
	    bType = false;
	}
	else if(n.getNodeName().equals("Message")) {
	    bMsg = true;
	    currentMsg = new MDMsg();
	    Node idNode = n.getAttributes().getNamedItem("SequenceID");
	    if(idNode != null) {
		currentMsg.id = idNode.getNodeValue();
	    }
	}
	else if(n.getNodeName().equals("Name")) {
	    bMsgName = true;
	}
	else if(n.getNodeType() == Node.TEXT_NODE && bMsgName) {
	    currentMsg.name = n.getNodeValue();
	    bMsgName = false;
	}
	else if(bMsg && n.getNodeName().equals("Type")) {
	    bMsgType = true;
	}
	else if(n.getNodeType() == Node.TEXT_NODE && bMsgType) {
	    currentMsg.type = n.getNodeValue();
	    bMsgType = false;
	}
	else if(bMsg && n.getNodeName().equals("Description")) {
	    bMsgDescription = true;
	}
	else if(n.getNodeType() == Node.TEXT_NODE && bMsgDescription) {
	    currentMsg.description = n.getNodeValue();
	    bMsgDescription = false;
	    vMsg.add(currentMsg);
	}
	else if(n.getNodeName().equals("ELEMENT")) {
	    Node varDescNode = n.getAttributes().getNamedItem("description");
	    if(varDescNode != null) {
		currentMsg.variantDescription = varDescNode.getNodeValue();
	    }
	}
	else if(n.getNodeName().equals("SEGMENT")) {
	    bMsgSegment = true;
	}
	else if(n.getNodeType() == Node.TEXT_NODE && bMsgSegment) {
	    currentMsg.segment = n.getNodeValue();
	    bMsgSegment = false;
	}
	else if(n.getNodeName().equals("FIELD")) {
	    bMsgField = true;
	}
	else if(n.getNodeType() == Node.TEXT_NODE && bMsgField) {
	    currentMsg.field = n.getNodeValue();
	    bMsgField = false;
	}
	else if(n.getNodeName().equals("COMPONENT")) {
	    bMsgComponent = true;
	}
	else if(n.getNodeType() == Node.TEXT_NODE && bMsgComponent) {
	    currentMsg.component = n.getNodeValue();
	    bMsgComponent = false;
	}
	else if(n.getNodeName().equals("SUBCOMPONENT")) {
	    bMsgSubComponent = true;
	}
	else if(n.getNodeType() == Node.TEXT_NODE && bMsgSubComponent) {
	    currentMsg.subcomponent = n.getNodeValue();
	    bMsgSubComponent = false;
	}
	// Apply to children
 	NodeList nl = n.getChildNodes();
 	for(int i=0;i<nl.getLength();i++) {
 	    browseXML(nl.item(i));
	}
    }

    /**
     * This is a private class to store information on message files
     */
    public class MDMsg {
	public String id;
	public String name;
	public String type;
	public String description;
	public String segmentGroup;
	public String segment;
	public String field;
	public String component;
	public String subcomponent;
	public String variantDescription;

	public MDMsg() {
	}

	public MDMsg(String name) {
	    this.name = name;
	}

	public String getLocation() {
	    StringBuffer sb = new StringBuffer();
	    if(segment != null && !segment.equals("")) {
		sb.append("Segment: " + segment + "\n");
	    }
	    if(field  != null && !field.equals("")) {
		sb.append("Field: " + field + "\n");
	    }
	    if(component != null && !component.equals("")) {
		sb.append("Component: " + component + "\n");
	    }
	    if(subcomponent != null && !subcomponent.equals("")) {
		sb.append("Subcomponent: " + subcomponent + "\n");
	    }
	    return sb.toString();
	}

	public boolean equals(Object o) {
	    if(o instanceof MDMsg) {
		MDMsg x = (MDMsg)o;
		return name.compareTo(x.name) == 0;
	    }
	    return false;
	}
    }

}
