/**
 * This class handles the XML file of a directory
 */

package gov.nist.hl7maker.view.display;

// JDK
import java.io.File;
import java.io.FileFilter;
import java.io.IOException;
import java.util.Vector;
import java.util.Collections;
import java.util.Comparator;

// XML
import javax.xml.parsers.DocumentBuilder; 
import javax.xml.parsers.DocumentBuilderFactory;  
import javax.xml.parsers.ParserConfigurationException;

// SAX
import org.xml.sax.SAXException;  

// DOM
import org.w3c.dom.Document;

// Swing
import javax.swing.tree.DefaultMutableTreeNode;

// Message Maker
import gov.nist.hl7maker.util.Constants;

public class XMLDirectory {

    private Vector vMsg;
    private String dirName;

    /**
     * Constructor
     * @param dirName The directory
     */
    public XMLDirectory(String dirName) {
	this.dirName = dirName;
	vMsg = new Vector();
    }

    /**
     * Get the current directory
     * @return The current directory
     */
    public String getDirectory() {
	return dirName;
    }

    /**
     * Get the list of all .xml file
     * @return The list of all .xml file
     */
    public Vector getMessages() throws ParserConfigurationException, SAXException, IOException {
	Vector vMD = getMDlist();

	vMsg.clear();
	for(int i=0;i<vMD.size();i++) {
	    MetaData md = new MetaData((String)vMD.get(i));
	    vMsg.add("TD_" + md.getName());
	    vMsg.addAll(md.getMessageList());
	}
	return vMsg;
    }

    /**
     * Get the list of all MetaData files
     * @return The list
     */
    private Vector getMDlist() {
	Vector vMD = new Vector();
	File f = new File(dirName);
	File[] files = f.listFiles(new FileFilter() {
		public boolean accept(File f) {
		    // Check MD_.xml
		    String filename = f.getName().toLowerCase();
		    if(filename.startsWith("td_") && filename.endsWith(".xml")) {
			return true;
		    }
		    return false;
		}
	    });
	if(files != null) {
	    for(int i=0;i<files.length;i++) {
		vMD.add(files[i].getAbsolutePath());
	    }
	}
	return vMD;
    }

    /**
     * Get a tree representation of the XML directory
     * @return The tree representation
     */
    public DefaultMutableTreeNode getTree() throws ParserConfigurationException, SAXException, IOException {
	Vector vMsg = getMessages();

	if(dirName.compareTo("") != 0) {
	    DefaultMutableTreeNode top = new DefaultMutableTreeNode(dirName);
	    DefaultMutableTreeNode metadata = null;
	    DefaultMutableTreeNode msg = null;

	    if(vMsg.size() > 0) {
		for(int i=0;i<vMsg.size();i++) {
		    String elt = (String)vMsg.get(i);
		    if(elt.startsWith("TD_")) {
			metadata = new DefaultMutableTreeNode(elt.substring(3));
			top.add(metadata);
		    }
		    else {	
			msg = new DefaultMutableTreeNode(elt);
			metadata.add(msg);
		    }
		}
	    }
	    return top;
	}
	return null;
    }

    /**
     * Return true if both objects are equal
     * @return Equals?
     */
    public boolean equals(Object o) {
	if(o instanceof XMLDirectory) {
	    XMLDirectory x = (XMLDirectory)o;
	    if(x.getDirectory().equals(getDirectory())) {
		return true;
	    }
	}
	return false;
    }

    /**
     * Load a xml file to get a DOM object
     * @param xmlFile The XML File
     * @return The DOM representation
     */
    public static Document getMessageDocument(File xmlFile) throws ParserConfigurationException, IOException, SAXException {
	DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
	DocumentBuilder builder = factory.newDocumentBuilder();
	return builder.parse(xmlFile);
    }

    /**
     * Load a xml file to get a DOM object
     * @param idxMsg The index of the XML file in the XMLDirectory
     * @return The DOM representation
     */
    public Document getMessageDocument(int idxMsg) throws ParserConfigurationException, IOException, SAXException {
	return getMessageDocument(new File(dirName + Constants.FS + (String)vMsg.get(idxMsg)));
    }

    /**
     * Load a xml file to get a DOM object
     * @param xmlFile The xml fileName
     * @return The DOM representation
     */
    public static Document getMessageDocument(String xmlFile) throws ParserConfigurationException, IOException, SAXException {
	return getMessageDocument(new File(xmlFile));
    }

    /**
     * Get the profile associated to the message
     * @param xmlFile The filename of the message
     * @return The filename of the profile
     */
    public String getProfileFromMessage(String xmlFile) throws ParserConfigurationException, SAXException, IOException {
	Vector vMD = getMDlist();

	for(int i=0;i<vMD.size();i++) {
	    MetaData md = new MetaData((String)vMD.get(i));
	    if(md.hasMessage(xmlFile)) {
		return md.getProfile();
	    }
	}
	return "";
    }
}
