/**
 * This class is a TreeModel for an XML file
 */

package gov.nist.hl7maker.view.display;

// JDK
import java.util.Vector;

// Swing
import javax.swing.JTree;
import javax.swing.tree.TreeModel;
import javax.swing.tree.TreePath;
import javax.swing.event.TreeModelListener;
import javax.swing.event.TreeModelEvent;

// DOM
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

public class XMLTreeModel implements TreeModel {

    private Document doc;
    private Vector listeners;

    /**
     * Constructor
     * @param doc A Document object
     */
    public XMLTreeModel(Document doc) {
	this.doc = doc;
	removeEmptyTextElement(doc);
	listeners = new Vector();
    }

    public Object getRoot() {
	return new MyNode(doc.getDocumentElement());
    }

    public int getChildCount(Object parent) {
	return ((MyNode)parent).getNode().getChildNodes().getLength();
    }

    public Object getChild(Object parent, int index) {
    	if(((MyNode)parent).getNode().hasChildNodes())
	    return new MyNode(((MyNode)parent).getNode().getChildNodes().item(index));
	return null;
    }

    public int getIndexOfChild(Object parent, Object child) {
	if(((MyNode)parent).getNode().hasChildNodes()) {
	    NodeList nl = ((MyNode)parent).getNode().getChildNodes();
	    for(int i=0;i<nl.getLength();i++) {
		if(nl.item(i).equals(child)) {
		    return i;
		}
	    }
	}
	return -1;
    }

    /**
     * Return true if the node is a leaf
     * In this model, we consider that a leaf node is a node with one child
     * which is a TEXT_NODE
     * @param node The Node
     * @return is a leaf?
     */
    public boolean isLeaf(Object node) {
	boolean isLeaf = false;
	Node n = ((MyNode)node).getNode();
	if(n.hasChildNodes()) {
	    NodeList nl = n.getChildNodes();
	    for(int i=0;i<nl.getLength() && !isLeaf;i++)
		if(nl.item(i).getNodeType() == Node.TEXT_NODE)
		    isLeaf = true;
	}
	return isLeaf;
	//	return !(((MyNode)node).getNode().hasChildNodes());
    }

    public void addTreeModelListener(TreeModelListener listener) {
	if (listener != null && !listeners.contains(listener))
	    listeners.add(listener);
    }

    public void removeTreeModelListener(TreeModelListener listener) {
	if (listener != null) 
	    listeners.remove(listener);
    }
    
   public void fireTreeNodesChanged(TreeModelEvent event) {
       for(int i=0;i<listeners.size();i++) {
	   TreeModelListener listener = (TreeModelListener)listeners.get(i);
	   listener.treeNodesChanged(event);
       }
   }    
    
    public void valueForPathChanged(TreePath path, Object newValue) {
	MyNode node = (MyNode)path.getLastPathComponent();
	Node n = node.getNode();

	if(n.hasChildNodes()) {
	    Node updatedNode = n.getChildNodes().item(0);
	    if(updatedNode.getNodeType() == Node.TEXT_NODE) {
		updatedNode.setNodeValue((String)newValue);
		TreeModelEvent event = new TreeModelEvent(this, path);
		fireTreeNodesChanged(event);
	    }
	}
    }

    /**
     * This method removes all TEXT_NODE with empty text
     * @param node The current node
     */    
    private void removeEmptyTextElement(Node node) {
	if(node.hasChildNodes()) {
	    NodeList nl = node.getChildNodes();
	    for(int i=0;i<nl.getLength();i++) {
		Node n = nl.item(i);
		if(n.getNodeValue() != null && n.getNodeValue().trim().equals(""))
		    node.removeChild(n);
	    }
	}
	// Execute the method on the children
	if(node.hasChildNodes()) {
	    NodeList nl = node.getChildNodes();
	    for(int i=0;i<nl.getLength();i++) {
		removeEmptyTextElement(nl.item(i));
	    }
	}
    }
}
