/**
 * This class represents a message like a Tree
 */

package gov.nist.hl7maker.view.display;

// JDK
import java.io.IOException;
import java.util.Vector;

// Swing
import javax.swing.JTree;
import javax.swing.Icon;
import javax.swing.event.TreeSelectionListener;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeModelListener;
import javax.swing.event.TreeModelEvent;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

// SAX
import org.xml.sax.SAXException;

// DOM
import org.w3c.dom.Document;
import org.w3c.dom.Element;

// XML
import javax.xml.parsers.DocumentBuilder; 
import javax.xml.parsers.DocumentBuilderFactory;  
import javax.xml.parsers.ParserConfigurationException;

// XSL
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;

// HAPI
//import ca.uhn.hl7v2.HL7Exception;

public class XMLTreeView extends JTree implements TreeSelectionListener, TreeModelListener {

    private XMLView xmlView;
    private ER7View er7View;
    private EnhancedView enhancedView;

    /**
     * Constructor
     */
    public XMLTreeView() {
	this.setEditable(true);
	DefaultTreeModel treeModel = (DefaultTreeModel)JTree.createTreeModel(new Vector());
	this.setModel(treeModel);

        MyTreeCellRenderer renderer = new MyTreeCellRenderer();
        Icon personIcon = null;
        renderer.setLeafIcon(personIcon);
        renderer.setClosedIcon(personIcon);
        renderer.setOpenIcon(personIcon);
        this.setCellRenderer(renderer);
	this.putClientProperty("JTree.lineStyle", "None");

	this.getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
	this.addTreeSelectionListener(this);
	//	this.getModel().addTreeModelListener(this);

	MyTreeCellEditor tce = new MyTreeCellEditor((JTree)this, (DefaultTreeCellRenderer)this.getCellRenderer());
	this.setCellEditor(tce);
    }

    /**
     * Set the Other View (XML and ER7)
     * @param xmlView The XMLView
     * @param er7View The ER7View
     * @param enhancedView The EnhancedView
     */
    public void setOtherView(XMLView xmlView, ER7View er7View, EnhancedView enhancedView) {
	this.xmlView = xmlView;
	this.er7View = er7View;
	this.enhancedView = enhancedView;
    }

    /**
     * This method is called when a value is changed in the tree
     * It collapses and expands a tree node 
     * @param e TreeSelectionEvent The event
     */
    public void valueChanged(TreeSelectionEvent e) {
	TreePath tp = e.getPath();

	if(this.isExpanded(tp))
	    this.collapsePath(tp);
	else
	    this.expandPath(tp);
    }

    /**
     * Get the DOM object of the message from the current TreeModel
     * @return The DOM representation of the message
     */
    private Document getDocumentFromModel() throws TransformerConfigurationException, TransformerException, ParserConfigurationException {
	DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
	DocumentBuilder builder = factory.newDocumentBuilder();
	Document doc = builder.newDocument();

	Element root = (Element)((MyNode)getModel().getRoot()).getNode();
	root = (Element)doc.importNode(root, true);
	doc.appendChild(root);

	return doc;
    }

    // TreeModelListener

    /**
     * Invoked after a node (or a set of siblings) has changed in some way.
     * @param e The Event
     */
    public void treeNodesChanged(TreeModelEvent e) {
	// Get the Document
	try { 
	    Document doc = getDocumentFromModel();
	    xmlView.displayMessage(doc);
	    er7View.displayMessage(doc);
	    enhancedView.setHasBeenModified(true);
	}
/*	catch(SAXException saxe) {
	    System.err.println("SAXException " + saxe.getMessage());
	    saxe.printStackTrace();
	    }*/
	catch(TransformerConfigurationException tce) {
	    System.err.println("TransformerConfigurationException " + tce.getMessage());
	    tce.printStackTrace();
	}
	catch(TransformerException te) {
	    System.err.println("TransformerException " + te.getMessage());
	    te.printStackTrace();
	}
	catch(ParserConfigurationException pce) {
	    System.err.println("ParserConfigurationException " + pce.getMessage());
	    pce.printStackTrace();
	}
/*	catch(HL7Exception hl7e) {
	    System.err.println("HL7Exception " + hl7e.getMessage());
	    hl7e.printStackTrace();
	    }*/
/*	catch(IOException ioe) {
	    System.err.println("IOException " + ioe.getMessage());
	    ioe.printStackTrace();
	    }*/
    }

    /**
     * Invoked after nodes have been inserted into the tree.
     * @param e The Event
     */
    public void treeNodesInserted(TreeModelEvent e) {
    }
          
    /**
     * Invoked after nodes have been removed from the tree.
     * @param e The Event
     */
    public void treeNodesRemoved(TreeModelEvent e) {
    }

    /**
     *  Invoked after the tree has drastically changed structure from a given node down.
     * @param e The Event
     */
    public void treeStructureChanged(TreeModelEvent e) {
    }
}
