/**
 * This is the MVC model for the batch GUI
 */

package gov.nist.hl7maker.view.management;

// JDK
import java.io.File;
import java.io.IOException;
import java.util.Calendar;
import java.util.Observable;
import java.util.Vector;
import java.util.StringTokenizer;
import java.util.Enumeration;

// Swing
import javax.swing.tree.DefaultMutableTreeNode;

// XML
import javax.xml.parsers.DocumentBuilder; 
import javax.xml.parsers.DocumentBuilderFactory;  
import javax.xml.parsers.ParserConfigurationException;

// DOM
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

// SAX
import org.xml.sax.SAXException;  

// MessageMaker
import gov.nist.hl7maker.util.MVCArg;
import gov.nist.hl7maker.view.config.SegmentNode;
import gov.nist.hl7maker.view.config.FieldNode;
import gov.nist.hl7maker.view.config.ComponentNode;
import gov.nist.hl7maker.view.display.XMLDirectory;

public class BatchModel extends Observable {

    private final String XML_BATCH_DEFINITION = "data/BatchSegments.xml";
    private String structure;
    private DefaultMutableTreeNode root;

    private Vector vSegment;
    private SegmentNode curSegment;
    private FieldNode curField;
    private ComponentNode curComponent;

    private Vector vDir;
    private MessageNode currentMessageNode;

    /**
     * Constructor
     */
    public BatchModel() {
	init();
    }

    /**
     * Init the model
     */
    public void init() {
	vDir = new Vector();
    }

    /**
     * Fire an event
     * @param arg An MVCArg object
     */
    public void fireEvent(MVCArg arg) {
	setChanged();
	notifyObservers(arg);
    }

    /**
     * Set the structure
     * @param struct The Structure
     */
    public void setStructure(String struct) {
	structure = struct;
	setChanged();
	notifyObservers(new MVCArg("structure", structure));
    }

    /**
     * Get the structure
     * @return The structure
     */
    public String getStructure() {
	return structure;
    }

    /**
     * Get the directories
     * @return The directories
     */
    public Vector getDirectory() {
	return vDir;
    }

    /**
     * Add a directory
     * @param dir The directory
     */
    public void addDirectory(XMLDirectory xmlDir) {
	if(!vDir.contains(xmlDir))
	    vDir.add(xmlDir);
    }

    /**
     * Remove a directory
     * @param dir The directory
     */
    public void removeDirectory(XMLDirectory xmlDir) {
	vDir.remove(xmlDir);
    }

    /**
     * Remove all directories
     */
    public void removeAllDirectory() {	
	if(vDir != null)
	    vDir.clear();
    }

    /**
     * Set the current MessageNode
     * @param mn The current MessageNode
     */
    public void setMessageNode(MessageNode mn) {
	currentMessageNode = mn;
    }
    
    /**
     * Get the messages
     * @return The messages
     */
    public Vector getMessage() {
	return currentMessageNode.getMessage();
    }
    
    /**
     * Add a message
     * @param msg The message filename
     */
    public void addMessage(String msg) {
	currentMessageNode.addMessage(msg);
    }

    /**
     * Remove a message
     * @param msg The message filename
     */
    public void removeMessage(String msg) {
	currentMessageNode.removeMessage(msg);
    }

    /**
     * Read the XML Definition of a batch and return a tree structure
     * @param structure The structure of the batch
     * @return The tree representation of the batch
     */
    public DefaultMutableTreeNode getTree(String structure) throws ParserConfigurationException, SAXException, IOException {
	DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
	DocumentBuilder builder = factory.newDocumentBuilder();
	Document doc = builder.parse(new File(XML_BATCH_DEFINITION));

	// Parse the structure
	Vector vStruct = new Vector();
	StringTokenizer tokens = new StringTokenizer(structure, ",");
	while(tokens.hasMoreTokens()) {
	    vStruct.add(tokens.nextToken().trim());
	}

	// Browser the XML Profile	
	vSegment = new Vector();
        browseXML(doc);
	// Create the Tree Structure
	root = new DefaultMutableTreeNode("Batch");
	DefaultMutableTreeNode segment, field, component;//, subcomponent;

	for(int i=0;i<vStruct.size();i++) {
	    String segName = (String)vStruct.get(i);
	    if(segName.equals("MSG")) {
		root.add(new DefaultMutableTreeNode(new MessageNode()));
	    }
	    else {
		SegmentNode segTmp = new SegmentNode();
		segTmp.setName(segName);
		segTmp.setQualifier("");
		segTmp.setQualifier2("");
		segTmp.setSegment(segName);
		int idx = vSegment.indexOf(segTmp);
		if(idx != -1) {
		    SegmentNode segNode = (SegmentNode)vSegment.get(idx);
		    segment = new DefaultMutableTreeNode(new SegmentNode(segNode));
		    root.add(segment);
		    Vector vField = segNode.getFields();
		    for(int j=0;j<vField.size();j++) {
			FieldNode fieldNode = (FieldNode)vField.get(j);
			field = new DefaultMutableTreeNode(new FieldNode(fieldNode));
			segment.add(field);
			Vector vComponent = fieldNode.getComponents();
			for(int k=0;k<vComponent.size();k++) {
			    ComponentNode compNode = (ComponentNode)vComponent.get(k);
			    component = new DefaultMutableTreeNode(new ComponentNode(compNode));
			    field.add(component);
			}
		    }
		}
	    }
	}
	return root;
    }
    
    /**
     * Browser recursively the DOM and creates the tree structure
     * @param n The current node
     */
    private void browseXML(Node n) {
	String name = null, datatype = null, usage = null, min = null, max = null, length = null, itemNo = null;
	Node nameNode, datatypeNode, usageNode, minNode, maxNode, lengthNode, itemNoNode;
	// Segment
 	if(n.getNodeName().equals("Segment")) {
	    curSegment = new SegmentNode();
 	    nameNode = n.getAttributes().getNamedItem("Name");
	    if(nameNode != null) {
		name = nameNode.getNodeValue();
	    }
	    curSegment.setName(name);
	    curSegment.setQualifier("");
	    curSegment.setQualifier2("");
	    curSegment.setSegment(name);
	    vSegment.add(curSegment);
 	}
	// Field
 	else if(n.getNodeName().equals("Field")) {
	    curField = new FieldNode();
	    nameNode = n.getAttributes().getNamedItem("Name");
	    if(nameNode != null) {
		name = nameNode.getNodeValue();
		if(name.equals("Batch Field Separator") || name.equals("File Field Separator"))
		    curField.setBatchValue("|");
		else if(name.equals("Batch Encoding Characters") || name.equals("File Encoding Characters"))
		    curField.setBatchValue("^~\\&");
	    }
	    usageNode = n.getAttributes().getNamedItem("Usage");
	    if(usageNode != null) {
		usage = usageNode.getNodeValue();
	    }
	    minNode = n.getAttributes().getNamedItem("Min");
	    if(minNode != null) {
		min = minNode.getNodeValue();
	    }
	    maxNode = n.getAttributes().getNamedItem("Max");
	    if(maxNode != null) {
		max = maxNode.getNodeValue();
	    }
	    datatypeNode = n.getAttributes().getNamedItem("Datatype");
	    if(datatypeNode != null) {
		datatype = datatypeNode.getNodeValue();
	    }
	    lengthNode = n.getAttributes().getNamedItem("Length");
	    if(lengthNode != null) {
		length = lengthNode.getNodeValue();
	    }
	    itemNoNode = n.getAttributes().getNamedItem("ItemNo");
	    if(itemNoNode != null) {
		itemNo = itemNoNode.getNodeValue();
	    }
	    curField.setName(name);
	    curField.setQualifier("");
	    curField.setQualifier2("");
	    curField.setSegment(curSegment.getName());
	    curField.setUsage(usage);
	    curField.setCardMin(min);
	    curField.setCardMax(max);
	    curField.setDatatype(datatype);
	    curField.setLength(length);
	    curField.setItemNo(itemNo);
	    curSegment.addField(curField);
	}
	else if(n.getNodeName().equals("Component")) {
	    curComponent = new ComponentNode();
	    nameNode = n.getAttributes().getNamedItem("Name");
	    if(nameNode != null) {
		name = nameNode.getNodeValue();
		// Set the current date
		if(name.equals("Date/Time")) {
		    Calendar cal = Calendar.getInstance();
		    int y = cal.get(Calendar.YEAR);
		    int m = cal.get(Calendar.MONTH)+1;
		    int d = cal.get(Calendar.DAY_OF_MONTH);
		    String date = String.valueOf(y);
		    if(m < 10) date += "0" + m;
		    else date += m;
		    if(d < 10) date += "0" + d;
		    else date += d;
		    curComponent.setBatchValue(date);
		}
	    }
	    usageNode = n.getAttributes().getNamedItem("Usage");
	    if(usageNode != null) {
		usage = usageNode.getNodeValue();
	    }
	    datatypeNode = n.getAttributes().getNamedItem("Datatype");
	    if(datatypeNode != null) {
		datatype = datatypeNode.getNodeValue();
	    }
	    curComponent.setName(name);
	    curComponent.setQualifier(curField.getName());
	    curComponent.setQualifier2("");
	    curComponent.setSegment(curSegment.getName());
	    curComponent.setUsage(usage);
	    curComponent.setCardMin(curField.getCardMin());
	    curComponent.setCardMax(curField.getCardMax());
	    curComponent.setDatatype(datatype);
	    curComponent.setLength(curField.getLength());
	    curComponent.setItemNo(curField.getItemNo());
	    curField.addComponent(curComponent);
	}
	// SubComponent
/* 	else if(n.getNodeName().equals("SubComponent")) {
	    curSubComponent = new SubComponentNode();
	    nameNode = n.getAttributes().getNamedItem("Name");
	    if(nameNode != null) {
		name = nameNode.getNodeValue();
	    }
	    tableNode = n.getAttributes().getNamedItem("Table");
	    if(tableNode != null) {
		table = tableNode.getNodeValue();
	    }
	    datatypeNode = n.getAttributes().getNamedItem("Datatype");
	    if(datatypeNode != null) {
		datatype = datatypeNode.getNodeValue();
	    }
	    usageNode = n.getAttributes().getNamedItem("Usage");
	    if(usageNode != null) {
		usage = usageNode.getNodeValue();
	    }
	    curSubComponent.setName(name);
	    curSubComponent.setQualifier(curComponent.getQualifier() + "{" + curComponent.getName() + "}");
	    curSubComponent.setQualifier(curComponent.getQualifier());
	    curSubComponent.setQualifier2(curComponent.getName());
	    curSubComponent.setSegment(curSegment.getName());
	    curSubComponent.setTable(table);
	    curSubComponent.setDatatype(datatype);
	    curSubComponent.setUsage(usage);
	    curComponent.addSubComponent(curSubComponent);
	    }*/
 	// Apply method on children
 	NodeList nl = n.getChildNodes();
 	for(int i=0;i<nl.getLength();i++) {
 	    browseXML(nl.item(i));
	}
    }

    /**
     * Get the tree representation of the batch
     * @return The tree representation of the batch
     */
    public DefaultMutableTreeNode getTree() {
	return root;
    }

    /**
     * Add a segment
     * @param pos The position
     * @param type Type of segment
     * @param node Add the segment
     */
    public void addSegment(String pos, String type, DefaultMutableTreeNode node) {
	DefaultMutableTreeNode addedNode = null;
	if(type.equals("FHS")) addedNode = getFHSNode();
	else if(type.equals("BHS")) addedNode = getBHSNode();
	else if(type.equals("MSG")) addedNode = getMSGNode();
	else if(type.equals("BTS")) addedNode = getBTSNode();
	else if(type.equals("FTS")) addedNode = getFTSNode();
	if(pos.equals("Before"))
	    root.insert(addedNode, root.getIndex(node));
	else if(pos.equals("After"))
	    root.insert(addedNode, root.getIndex(node)+1);
	updateStructure();
    }

    /**
     * Remove a segment
     * @param node Remove this node
     */
    public void removeSegment(DefaultMutableTreeNode node) {
	root.remove(node);
	updateStructure();
    }

    /**
     * Update the structure
     */
    private void updateStructure() {
	String newStructure = "";
	for(Enumeration e=root.children();e.hasMoreElements();) {
	    newStructure += (getTypeNode((DefaultMutableTreeNode)e.nextElement()) + ",");
	}	
	if(newStructure.length() > 0)
	    structure = newStructure.substring(0, newStructure.length()-1);
    }

    /**
     * Get a new FHS node
     * @return a FHS node
     */    
    private DefaultMutableTreeNode getFHSNode() {
	DefaultMutableTreeNode fhs = null, field = null, component = null;
	for(int i=0;i<vSegment.size();i++) {
	    SegmentNode segNode = (SegmentNode)vSegment.get(i);
	    if(segNode.getName().equals("FHS")) {
		fhs = new DefaultMutableTreeNode(new SegmentNode(segNode));
		Vector vField = segNode.getFields();
		for(int j=0;j<vField.size();j++) {
		    FieldNode fieldNode = (FieldNode)vField.get(j);
		    field = new DefaultMutableTreeNode(new FieldNode(fieldNode));
		    fhs.add(field);
		    Vector vComponent = fieldNode.getComponents();
		    for(int k=0;k<vComponent.size();k++) {
			ComponentNode compNode = (ComponentNode)vComponent.get(k);
			component = new DefaultMutableTreeNode(new ComponentNode(compNode));
			field.add(component);
		    }
		}
		i = vSegment.size() + 1;
	    }
	}
	return fhs;
    }

    /**
     * Get a new BHS node
     * @return a BHS node
     */
    private DefaultMutableTreeNode getBHSNode() {
	DefaultMutableTreeNode bhs = null, field = null, component = null;
	for(int i=0;i<vSegment.size();i++) {
	    SegmentNode segNode = (SegmentNode)vSegment.get(i);
	    if(segNode.getName().equals("BHS")) {
		bhs = new DefaultMutableTreeNode(new SegmentNode(segNode));
		Vector vField = segNode.getFields();
		for(int j=0;j<vField.size();j++) {
		    FieldNode fieldNode = (FieldNode)vField.get(j);
		    field = new DefaultMutableTreeNode(new FieldNode(fieldNode));
		    bhs.add(field);
		    Vector vComponent = fieldNode.getComponents();
		    for(int k=0;k<vComponent.size();k++) {
			ComponentNode compNode = (ComponentNode)vComponent.get(k);
			component = new DefaultMutableTreeNode(new ComponentNode(compNode));
			field.add(component);
		    }
		}
		i = vSegment.size() + 1;
	    }
	}
	return bhs;
    }

    /**
     * Get a new MSG node
     * @return a MSG node
     */
    private DefaultMutableTreeNode getMSGNode() {
	return new DefaultMutableTreeNode(new MessageNode());
    }

    /**
     * Get a new BTS node
     * @return a BTS node
     */
    private DefaultMutableTreeNode getBTSNode() {
	DefaultMutableTreeNode bts = null, field = null, component = null;
	for(int i=0;i<vSegment.size();i++) {
	    SegmentNode segNode = (SegmentNode)vSegment.get(i);
	    if(segNode.getName().equals("BTS")) {
		// Clone segNode
		bts = new DefaultMutableTreeNode(new SegmentNode(segNode));
		Vector vField = segNode.getFields();
		for(int j=0;j<vField.size();j++) {
		    FieldNode fieldNode = (FieldNode)vField.get(j);
		    field = new DefaultMutableTreeNode(new FieldNode(fieldNode));
		    bts.add(field);
		    Vector vComponent = fieldNode.getComponents();
		    for(int k=0;k<vComponent.size();k++) {
			ComponentNode compNode = (ComponentNode)vComponent.get(k);
			component = new DefaultMutableTreeNode(new ComponentNode(compNode));
			field.add(component);
		    }
		}
		i = vSegment.size() + 1;
	    }
	}
	return bts;
    }

    /**
     * Get a new FTS node
     * @return a FTS node
     */    
    private DefaultMutableTreeNode getFTSNode() {
	DefaultMutableTreeNode fts = null, field = null, component = null;
	for(int i=0;i<vSegment.size();i++) {
	    SegmentNode segNode = (SegmentNode)vSegment.get(i);
	    if(segNode.getName().equals("FTS")) {
		fts = new DefaultMutableTreeNode(new SegmentNode(segNode));
		Vector vField = segNode.getFields();
		for(int j=0;j<vField.size();j++) {
		    FieldNode fieldNode = (FieldNode)vField.get(j);
		    field = new DefaultMutableTreeNode(new FieldNode(fieldNode));
		    fts.add(field);
		    Vector vComponent = fieldNode.getComponents();
		    for(int k=0;k<vComponent.size();k++) {
			ComponentNode compNode = (ComponentNode)vComponent.get(k);
			component = new DefaultMutableTreeNode(new ComponentNode(compNode));
			field.add(component);
		    }
		}
		i = vSegment.size() + 1;
	    }
	}
	return fts;
    }

    /**
     * Find the last sibling node to the first BHS after the node passed in parameter
     * @param node The current node
     * @return The last sibling node
     */
    private DefaultMutableTreeNode getLastSiblingNode(DefaultMutableTreeNode node) {
	boolean end = false;
	DefaultMutableTreeNode lastSiblingNode = null;
	while(!end) {
	    DefaultMutableTreeNode tmpNode = node.getNextSibling();
	    if(tmpNode == null)
		end = true;
	    else {
		if(isBHS(tmpNode) || isFTS(tmpNode)) {
		    end = true;
		    lastSiblingNode = node;
		}
		else if(isBTS(tmpNode)) {
		    end = true;
		    lastSiblingNode = tmpNode;
		}    
		node = tmpNode;
	    }
	}
	if(lastSiblingNode == null)
	    lastSiblingNode = node;
	return lastSiblingNode;
    }

    /**
     * Is is a FHS node?
     * @param node The node
     * @return True of false
     */
    public static boolean isFHS(DefaultMutableTreeNode node) {
	if(node == null) return false;
	Object obj = node.getUserObject();
	return obj instanceof SegmentNode && ((SegmentNode)obj).getName().equals("FHS");
    }
    
    /**
     * Is is a BHS node?
     * @param node The node
     * @return True of false
     */
    public static boolean isBHS(DefaultMutableTreeNode node) {
	if(node == null) return false;
	Object obj = node.getUserObject();
	return obj instanceof SegmentNode && ((SegmentNode)obj).getName().equals("BHS");
    }
    
    /**
     * Is is a MSG node?
     * @param node The node
     * @return True of false
     */
    public static boolean isMSG(DefaultMutableTreeNode node) {
	if(node == null) return false;
	Object obj = node.getUserObject();
	return obj instanceof MessageNode;
    }
    
    /**
     * Is is a BTS node?
     * @param node The node
     * @return True of false
     */
    public static boolean isBTS(DefaultMutableTreeNode node) {
	if(node == null) return false;
	Object obj = node.getUserObject();
	return obj instanceof SegmentNode && ((SegmentNode)obj).getName().equals("BTS");
    }
    
    /**
     * Is is a FTS node?
     * @param node The node
     * @return True of false
     */
    public static boolean isFTS(DefaultMutableTreeNode node) {
	if(node == null) return false;
	Object obj = node.getUserObject();
	return obj instanceof SegmentNode && ((SegmentNode)obj).getName().equals("FTS");
    }

    /**
     * Get the type of the node
     * @param node The node
     * @return The type
     */
    public static String getTypeNode(DefaultMutableTreeNode node) {
	if(isFHS(node)) return "FHS";
	if(isBHS(node)) return "BHS";
	if(isMSG(node)) return "MSG";
	if(isBTS(node)) return "BTS";
	if(isFTS(node)) return "FTS";
	return null;
    }
}
