/**
 * This panel displays the tree structure of the batch message
 */

package gov.nist.hl7maker.view.management;

// JDK
import java.io.IOException;
import java.util.Observer;
import java.util.Observable;
import java.util.Vector;

// AWT
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

// Swing
import javax.swing.JPanel;
import javax.swing.JTree;
import javax.swing.JScrollPane;
import javax.swing.JButton;
import javax.swing.BoxLayout;
import javax.swing.Box;
import javax.swing.BorderFactory;
import javax.swing.event.TreeSelectionListener;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.DefaultMutableTreeNode;

// XML
import javax.xml.parsers.ParserConfigurationException;

// SAX
import org.xml.sax.SAXException;  

// MessageMaker
import gov.nist.hl7maker.util.MVCArg;
import gov.nist.hl7maker.view.config.SegmentNode;

public class BatchTreePanel extends JPanel implements Observer, ActionListener, TreeSelectionListener {

    private JTree batchTree;
    private JScrollPane btScrollPane;
    private JButton btnAddSegment;
    private JButton btnRemoveSegment;

    private BatchModel bModel;
    private AddSegmentDialog dlgAddSegment;

    /**
     * Constructor
     */
    public BatchTreePanel() {
	setLayout(new BoxLayout(this, BoxLayout.PAGE_AXIS));

	JPanel p0 = new JPanel();
	p0.setLayout(new BoxLayout(p0, BoxLayout.LINE_AXIS));
	batchTree = new JTree(new Vector());
	batchTree.addTreeSelectionListener(this);
	btScrollPane = new JScrollPane(batchTree);
	btScrollPane.setPreferredSize(new Dimension(100, 100));
	p0.add(btScrollPane);

	JPanel p1 = new JPanel();
	p1.setLayout(new BoxLayout(p1, BoxLayout.LINE_AXIS));
	btnAddSegment = new JButton("Add Segment");
        btnAddSegment.setBackground(new Color(204, 204, 255));
	btnAddSegment.addActionListener(this);
	p1.add(btnAddSegment);
	p1.add(Box.createRigidArea(new Dimension(10, 0)));
	btnRemoveSegment = new JButton("Remove Segment");
	btnRemoveSegment.setBackground(new Color(204, 204, 255));
	btnRemoveSegment.addActionListener(this);
	p1.add(btnRemoveSegment);

	add(p0);
	add(Box.createRigidArea(new Dimension(0, 5)));
	add(p1);
	setBorder(BorderFactory.createEmptyBorder(5,5,5,5));

	dlgAddSegment = new AddSegmentDialog(null);
	init();
    }

    /**
     * Handles a click event
     * @param evt The event
     */
    public void actionPerformed(ActionEvent evt) {
	if(evt.getSource() == btnAddSegment) {
	    DefaultMutableTreeNode node = (DefaultMutableTreeNode)batchTree.getLastSelectedPathComponent();
	    if(node != null && (BatchModel.isFHS(node) || BatchModel.isBHS(node) || BatchModel.isMSG(node) || BatchModel.isBTS(node) || BatchModel.isFTS(node))) {
		dlgAddSegment.init(node);
		dlgAddSegment.pack();
		// Center the window
		Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
		Dimension windowSize = dlgAddSegment.getSize();
		dlgAddSegment.setLocation(Math.max(0,(screenSize.width -windowSize.width)/2), 
					  Math.max(0,(screenSize.height-windowSize.height)/2));
		dlgAddSegment.setVisible(true);
		DefaultTreeModel model = new DefaultTreeModel(bModel.getTree());
		batchTree.setModel(model);
		btnAddSegment.setEnabled(false);
		btnRemoveSegment.setEnabled(false);
	    }
	}	       
	else if(evt.getSource() == btnRemoveSegment) {
	    DefaultMutableTreeNode node = (DefaultMutableTreeNode)batchTree.getLastSelectedPathComponent();
	    if(node != null) {
		bModel.removeSegment(node);
		DefaultTreeModel model = new DefaultTreeModel(bModel.getTree());
		batchTree.setModel(model);
	    }
	    btnAddSegment.setEnabled(false);
	    btnRemoveSegment.setEnabled(false);
	}
    }

    /**
     * Handles a tree click event
     * @param evt The event
     */
    public void valueChanged(TreeSelectionEvent evt) {
	btnAddSegment.setEnabled(false);
	btnRemoveSegment.setEnabled(false);

	DefaultMutableTreeNode node = (DefaultMutableTreeNode)batchTree.getLastSelectedPathComponent();
	if(node == null || (!node.isLeaf() && node.getLevel() == 1)) {
	    btnAddSegment.setEnabled(true);
	    btnRemoveSegment.setEnabled(true);
	    bModel.fireEvent(new MVCArg("nodeBlank", ""));
	    return;
	}

	if(BatchModel.isMSG(node)) {
	    btnAddSegment.setEnabled(true);
	    btnRemoveSegment.setEnabled(true);
	    bModel.setMessageNode((MessageNode)node.getUserObject());
	}

	Object nodeInfo = node.getUserObject();
	bModel.fireEvent(new MVCArg("node", nodeInfo));
    }

    /**
     * MVC
     * @param o The observable object
     * @param arg An argument passed to the notifyObservers method
     */
    public void update(Observable o, Object arg) {
	if(o instanceof BatchModel) {
	    if(arg instanceof MVCArg) {
		if("structure".equals(((MVCArg)arg).getName())) {
		    try {
			DefaultTreeModel model = new DefaultTreeModel(bModel.getTree(((MVCArg)arg).getValue()));
			batchTree.setModel(model);
		    }
		    catch(ParserConfigurationException pce) {
			System.err.println("ParserConfigurationException " + pce.getMessage());
			pce.printStackTrace();
		    }
		    catch(SAXException saxe) {
			System.err.println("SAXException " + saxe.getMessage());
			saxe.printStackTrace();
		    }
		    catch(IOException ioe) {
			System.err.println("IOException " + ioe.getMessage());
			ioe.printStackTrace();
		    }
		}
		else if("init".equals(((MVCArg)arg).getName())) {
		    init();
		}
	    }
	}
    }

    /**
     * Set the model
     * @param model The model
     */
    public void setBatchModel(BatchModel model) {
	bModel = model;
	dlgAddSegment.setBatchModel(model);
    }

    /**
     * Init the panel
     */
    private void init() {
	btnAddSegment.setEnabled(false);
	btnRemoveSegment.setEnabled(false);
    }
}
