<?xml version="1.0"?>
<!-- 
    An xsl transformation for reading a Version 2.5 HL7 profile and 
    generating a message based on the profile description.
    
    Author:   Len Gebase, NIST
    
    Profiles look like this: 
          ...
      <StaticDef MsgStructId="ADT_A01" ...>
        <Metadata ... />
        <Segment Name="PID" ...> 
            <Field dataType="XPN" ...> 
              <Component dataType="FN" ...> 
                <SubComponent dataType="ST" ... />
              </Component>
              <Component dataType="ST" ... />
                  ...
             </Field>
                ...
         </Segment>
              ....
              
    The output message for the above might look like:
    
      <ADT_01>
        <PID>
          <PID.5>
            <XPN.1>
              <FN.1>Smith</FN.1>
            </XPN.1>
            <XPN.2>Joe</XPN.2>
            ...
          </PID.5>
          ...
        <PID>
        ...
      </ADT_01>
            
    Loosely, the mapping looks like:
    
    <StaticDef>   ==>   <ADT_A01>     (MsgStructID)
    <Segment>     ==>   <PID>         (Segment Name)
    <Field>       ==>   <PID.5>       (where 5 is the sequence number
                                    obtained from looking up the 
                                    ItemNo in the Data Element table)
    <Component>   ==>   <XPN.1>       (Field dataType, nested 
                                    elements are numbered sequentially)
    <SubComponent> ==>  <FN.1>        (Component dataType.  When
                                    a primitive dataType is encountered
                                    a value may be generated - usually
                                    looked up in some table we've
                                    populated with data for use when
                                    generating test cases)
    <Component>   ==>   <XPN.2>       (Field dataType, but this time
                                    the Component dataType is 
                                    primitive and we generate a 
                                    value for it)

    The output msg should include values (that we generate or lookup)
    for elements in the profile whose data type is primitive (dataType =
    ST, NM, ID, DT, SI, TN, or IS).
    
    Optionally, Segment Group elements (<SegGroup>), which contain 
    Segments and may also include nested Segment Groups, may also be 
    included in the profile along with Segments.  They appear at the same 
    level as Segments.  Segment Group elements are not reflected in the
    output message, except that the cardinality of the group will impact
    the number of times the enclosed Segments are repeated.
    
    Global Paramters.
    
Option                   Allowed Values           Default Value
================================================================
valid                    true/false                true        
segmentCardinality       0 <= number <= 1           0.5         
fieldCardinality         0 <= number <= 1           0.5         
RE-Usage                 0 <= number <= 1           0.5         
CE-Usage                 0 <= number <= 1           0.5  
cardErrorMinSeg"        -1 <= number <= 0            0
cardErrorMinField"      -1 <= number <= 0            0
cardErrorMaxSeg"        -1 <= number <= 0            0
cardErrorMaxField"      -1 <= number <= 0            0
RusageError             -1 <= number <= 0            0
XusageError             -1 <= number <= 0            0
fieldLength   (Error)   -1 <= number <= 0            0
compLength    (Error)   -1 <= number <= 0            0
subCompLength (Error)   -1 <= number <= 0            0
lengthConstraints           0 or 1                   0
compositeLength             0 or 1                   0
namedSegment							  String									empty
namedField								  String									empty
namedComponent              String									empty
namedSubComponent           String									empty
elementCardinality         number >= 0							empty
elementREinclude           true/false								empty
=================================================================

Note that all R-Usage elements are included unless an error is being generated
and no X-Usage elements are used unless an error is being generated.  C-Usage
and CE-Usage are not currently supported.

- segmentCardinality
- fieldCardinality
- RE-Usage
- CE-Usage:
Parameters that take a non-negative value range generate valid elements in the
output message.  0 = minimum cardinality or don't include element.
1 = maximum cardinality or include all elements.  0.5 means a value roughly
mid-way between the end points, etc.  Only 0, 1, & 0.5 are currently supported.

- cardErrorMinSeg 
- cardErrorMinField 
- cardErrorMaxSeg 
- cardErrorMaxField 
- RusageError 
- XusageError 
- fieldLength 
- compLength 
- subCompLength: 
Parameters that take a non-positive value range generate invalid elements in the
output message.  0 = no error, -1 = all errors, and -0.1 = 1 error.  Only 
zero or 1 error is currently supported.

valid

  If true, then no invalid elements will be created.
  If false, an invalid msg may be generated.
      
lengthConstraints

    if = 1, length constraints are enforced;
    if = 0, length constraints are ignored.
    
compositeLength

    if = 1, for composite types, lengths are checked and a warning issued 
      if they might be violated, but the constraint is NOT enforced.
    if = 0, composite length constraints are ignored.

namedSegment 
namedField 
namedComponent 
namedSubComponent 
elementCardinality 
elementREinclude 

	Identify a specific element, the element's cardinality or if the
  element should be included.
  
................................................................
To handle msg descriptions generated from Roch's tool:  
we'll control all option values before this transform is invoked:

 messageControlId value="1" 
 msgCount value="we can get this from the input" 
 offset value="-1" 
 validity value="VALID" 
 lengthConstraints value="false" 
 compositeLength value="false" 
 segmentCardinality & fieldCardinality we'll have to set here to 1
 RE_Usage value="ALL" 
     ...............................................................
    -->  
<xsl:stylesheet version="1.0"
  xmlns:de="http://www.nist.gov/hl7/dataElements"
  xmlns:config="http://www.nist.gov/hl7/Configuration"
  xmlns:refData="http://www.nist.gov/hl7/ReferenceData"
  xmlns:HL7Tables="http://www.nist.gov/hl7/hl7TableData"
  xmlns:usrTables="http://www.nist.gov/hl7/userTableData"
  xmlns:extTables="http://www.nist.gov/hl7/externalTableData"
  xmlns:local="http://www.nist.gov/hl7/localTableData" 
  xmlns:errorCodes="http://www.nist.gov/hl7/errorCodes" 
  xmlns:xsl="http://www.w3.org/1999/XSL/Transform">

<!-- ...Global Parameters... -->
<xsl:param name="dateOfMessage"      select="'0'"/>
<!-- input tables -->
<xsl:param name="configTable" select="'../../data/ConfigData.xml'"/>
<xsl:param name="hl7Tables" select="'../../data/hl7tables.xml'"/>
<xsl:param name="userTables" select="'../../data/usertables.xml'"/>
<xsl:param name="externalTables" select="'../../data/externaltables.xml'"/>
<xsl:param name="localTables" select="'../../data/localTableData.xml'"/>
<xsl:param name="referenceTable" select="'../../data/ReferenceData.xml'"/>
<xsl:param name="dataElementsTable" select="'../../data/dataElements.xml'"/>
<xsl:param name="errorCodesTable" select="'../../data/errorCodes.xml'"/>

<!-- potential error locations -->
<xsl:param name="cardErrorMinSeg"      select="'0'"/>
<xsl:param name="cardErrorMinField"    select="'0'"/>
<xsl:param name="cardErrorMaxSeg"      select="'0'"/>
<xsl:param name="cardErrorMaxField"    select="'0'"/>
<xsl:param name="rUErrorSegment"       select="'0'"/>
<xsl:param name="rUErrorField"         select="'0'"/>
<xsl:param name="rUErrorComponent"     select="'0'"/>
<xsl:param name="rUErrorSubcomp"       select="'0'"/>
<xsl:param name="xUErrorField"         select="'0'"/>
<xsl:param name="xUErrorSegment"       select="'0'"/>
<xsl:param name="xUErrorComponent"     select="'0'"/>
<xsl:param name="xUErrorSubcomp"       select="'0'"/>
<xsl:param name="fLengthErrorLoc"      select="'0'"/>
<xsl:param name="cLengthErrorLoc"      select="'0'"/>
<xsl:param name="sLengthErrorLoc"      select="'0'"/>
<!-- options: -->
<xsl:param name="valid"                select="'false'"/>
<xsl:param name="starValue"                select="3"/>
<xsl:param name="segmentCardinality"   select="0.5"/>
<xsl:param name="fieldCardinality"     select="0.5"/>
<xsl:param name="RE_Usage"             select="0.5"/>
<xsl:param name="minSegCardError"   select="'false'"/>
<xsl:param name="maxSegCardError"   select="'false'"/>
<xsl:param name="minFieldCardError"     select="'false'"/>
<xsl:param name="maxFieldCardError"     select="'false'"/>
<xsl:param name="RusageError"        select="'false'"/>
<xsl:param name="XusageError"        select="'false'"/>
<xsl:param name="lengthConstraints"    select="0"/>
<xsl:param name="compositeLength"      select="0"/>
<xsl:param name="fieldLength"          select="0"/>
<xsl:param name="compLength"           select="0"/>
<xsl:param name="subCompLength"        select="0"/>
<xsl:param name="namedSegment"        select="''"/>
<xsl:param name="namedField"        select="''"/>
<xsl:param name="namedComponent"        select="''"/>
<xsl:param name="namedSubComponent"        select="''"/>
<xsl:param name="elementCardinality"        select="''"/>
<xsl:param name="elementREinclude"        select="''"/>
<xsl:param name="elementErrorCode"        select="'false'"/>
<!-- not supported yet: -->
<xsl:param name="C_Usage"              select="0.5"/>
<xsl:param name="CE_Usage"             select="0.5"/>

<xsl:param name="messageControlID"      select="0"/>
<xsl:param name="messageControlPrefix"  select="''"/>
<xsl:param name="offset"                select="0"/>
<xsl:param name="msgDescription"        select="''"/>

<!-- max number of times we'll look for a data value for the same element -->
<xsl:variable name="MAX_TRIES" select="10"/>

<xsl:variable name="DEBUG">
  <xsl:value-of select="0"/>            <!-- debug off -->
  <!-- <xsl:value-of select="1"/> -->  <!-- params -->
  <!-- <xsl:value-of select="2"/> -->  <!-- table values -->
  <!-- <xsl:value-of select="3"/> -->  <!-- table index value -->
	<!--  <xsl:value-of select="4"/>-->  <!-- cardinality and usage -->
  <!-- <xsl:value-of select="10"/> --> <!-- all -->
</xsl:variable>

<xsl:variable name="INCLUDE_DUMMY_VALS">
  <!-- <xsl:value-of select="0"/> -->       <!-- don't include -->
  <xsl:value-of select="1"/>      <!-- include -->
</xsl:variable>
<xsl:variable name="DT_DUMMY">
  <xsl:value-of select="'10000101'"/>
</xsl:variable>
<xsl:variable name="ID_DUMMY">
    <xsl:value-of select="'XX'"/> 
</xsl:variable>
<xsl:variable name="IS_DUMMY">
    <xsl:value-of select="99"/>   
</xsl:variable>
<xsl:variable name="ST_DUMMY">
   <xsl:value-of select="'XX'"/>    
</xsl:variable>
<xsl:variable name="NM_DUMMY">
    <xsl:value-of select="'10000101'"/> 
</xsl:variable>
<xsl:variable name="SI_DUMMY">
    <xsl:value-of select="'99'"/>   
</xsl:variable>
<xsl:variable name="TN_DUMMY">
  	<xsl:value-of select="'555555555'"/> 
</xsl:variable>
<xsl:variable name="TM_DUMMY">
   <xsl:value-of select="'000000'"/> 
</xsl:variable>
  
<xsl:variable name="ANNOTATE">
  <xsl:value-of select="0"/> 
<!-- <xsl:value-of select="1"/> -->  
  <!-- <xsl:value-of select="2"/> -->
</xsl:variable>  
<xsl:variable name="WARN">
  <!-- <xsl:value-of select="0"/> -->
  <xsl:value-of select="1"/>
</xsl:variable>  

<xsl:output method="xml" indent="yes"/>

<xsl:strip-space elements="*"/>
  
<xsl:variable name="newline"> 
  <xsl:text>
  </xsl:text>
</xsl:variable>

<xsl:variable name="refDataKey" select="'refDataKey'"/>

<!-- define keys for indexing input data tables -->
<xsl:key name="fieldItem" match="de:dataElement" use="@itemNo"/>
<xsl:key name="configTableKey" match="config:config" use="@Name"/>
<xsl:key name="hl7TableKey" match="HL7Tables:Table" use="@ID"/>
<xsl:key name="userTableKey" match="usrTables:Table" use="@ID"/>
<xsl:key name="externalTableKey" match="extTables:Table" use="@ID"/>
<xsl:key name="localTableKey" match="local:Table" use="@ID"/>
<xsl:key name="errorCodesKey" match="errorCodes:Table" use="@ID"/>
<xsl:key name="refDataKey" match="refData:ReferenceData" use="@Name"/>
<!--
<xsl:key name="refDataKey" match="refData:ReferenceData" use="@Name"/>
-->

<!-- ............................................................. -->
<!-- ...................  ROOT  .................................. -->
<xsl:template match="/">
  <xsl:if test="$DEBUG=1 or $DEBUG &gt; 9">
    Transform Parameters:
    valid=<xsl:value-of select="$valid"/>
    segmentCardinality =<xsl:value-of select= "$segmentCardinality"   />
    fieldCardinality   =<xsl:value-of select= "$fieldCardinality"     />
    RE_Usage           =<xsl:value-of select= "$RE_Usage"             />
    lengthConstraints  =<xsl:value-of select= "$lengthConstraints"    />
    C_Usage            =<xsl:value-of select= "$C_Usage"              />
    CE_Usage           =<xsl:value-of select= "$CE_Usage"             />
    messageControlID   =<xsl:value-of select= "$messageControlID"     />
    messageControlPrefix =<xsl:value-of select= "$messageControlPrefix"     />
    offset             =<xsl:value-of select= "$offset"               />
    fieldLength             =<xsl:value-of select= "$fieldLength"     />
    compLength             =<xsl:value-of select= "$compLength"       />
    subCompLength             =<xsl:value-of select= "$subCompLength" />
    compositeLength        =<xsl:value-of select= "$compositeLength"  />
    cardErrorMinSeg    =<xsl:value-of select= "$cardErrorMinSeg"      />
    cardErrorMinField  =<xsl:value-of select= "$cardErrorMinField"    />
    cardErrorMaxSeg    =<xsl:value-of select= "$cardErrorMaxSeg"      />
    cardErrorMaxField  =<xsl:value-of select= "$cardErrorMaxField"    />
    rUErrorSegment    =<xsl:value-of select= "$rUErrorSegment"       />
    rUErrorField  =<xsl:value-of select= "$rUErrorField"         />
    rUErrorComponent    =<xsl:value-of select= "$rUErrorComponent"     />
    rUErrorSubcomp  =<xsl:value-of select= "$rUErrorSubcomp"       />
    xUErrorField    =<xsl:value-of select= "$xUErrorField"         />
    xUErrorSegment  =<xsl:value-of select= "$xUErrorSegment"       />
    xUErrorComponent    =<xsl:value-of select= "$xUErrorComponent"     />
    xUErrorSubcomp  =<xsl:value-of select= "$xUErrorSubcomp"       />
    minSegCardError  = <xsl:value-of select="$minSegCardError"/>
    maxSegCardError  = <xsl:value-of select="$maxSegCardError"/>
    minFieldCardError  = <xsl:value-of select="$minFieldCardError"/>
    maxFieldCardError  = <xsl:value-of select="$maxFieldCardError"/>
    RusageError  = <xsl:value-of select="$RusageError"/>
    XusageError  = <xsl:value-of select="$XusageError"/>
    lengthConstraints  = <xsl:value-of select="$lengthConstraints"/>
    compositeLength  = <xsl:value-of select="$compositeLength"/>
    fieldLength  = <xsl:value-of select="$fieldLength"/>
    compLength  = <xsl:value-of select="$compLength"/>
    subCompLength  = <xsl:value-of select="$subCompLength"/>
    namedSegment  = <xsl:value-of select="$namedSegment"/>
    namedField  = <xsl:value-of select="$namedField"/>
    namedComponent  = <xsl:value-of select="$namedComponent"/>
    namedSubComponent  = <xsl:value-of select="$namedSubComponent"/>
    elementCardinality  = <xsl:value-of select="$elementCardinality"/>
    elementREinclude  = <xsl:value-of select="$elementREinclude"/>
    fLengthErrorLoc  = <xsl:value-of select="$fLengthErrorLoc"/>
    cLengthErrorLoc  = <xsl:value-of select="$cLengthErrorLoc"/>
    sLengthErrorLoc  = <xsl:value-of select="$sLengthErrorLoc"/>
    starValue  = <xsl:value-of select="$starValue"/>
  </xsl:if>

  <!-- check options before processing starts  
    	(HL7 root node should be HL7v2xConformanceProfile, we also want to handle
      VA, which uses 'Specification' for root. -->
  <xsl:choose>
    <xsl:when test="$valid='true'">
      <xsl:choose>
        <xsl:when test = "$minSegCardError &lt; 0        
            or  $maxSegCardError &lt; 0
            or  $minFieldCardError &lt; 0
            or  $maxFieldCardError &lt; 0
            or  $RusageError &lt; 0
            or  $XusageError &lt; 0
            or  $compLength  &lt; 0
						or  $subCompLength &lt; 0
            or  $fieldLength &lt; 0">
          ERROR: Invalid Option/s with 'valid = true'!              
        </xsl:when>
        <xsl:otherwise> <!-- OK. generate valid msg -->
<!--          <xsl:apply-templates select="HL7v2xConformanceProfile"/>
		      <xsl:apply-templates/> -->
          <xsl:apply-templates select="HL7v2xConformanceProfile"/>
          <xsl:apply-templates select="Specification"/>
       </xsl:otherwise>        
      </xsl:choose>
    </xsl:when>
    <xsl:otherwise>   <!-- valid = false (may be creating invalid msg) -->
<!--      <xsl:apply-templates select="HL7v2xConformanceProfile"/>
      <xsl:apply-templates/> -->
      <xsl:apply-templates select="HL7v2xConformanceProfile"/>
      <xsl:apply-templates select="Specification"/>
    </xsl:otherwise>
  </xsl:choose>
  
</xsl:template>

<!-- hl7 uses 'HL7v2xConformanceProfile', but the VA uses 'Specification'  -->
<xsl:template match="HL7v2xConformanceProfile | Specification">
<!--  <xsl:apply-templates select="MetaData"/>
  <xsl:apply-templates select="Encodings"/>
  <xsl:apply-templates select="DynamicDef"/>
  <xsl:apply-templates select="HL7v2xStaticDef"/>
  <xsl:apply-templates/> -->
  <xsl:apply-templates select="HL7v2xStaticDef"/>
  <xsl:apply-templates select="Message"/>
</xsl:template>

<!--<xsl:template match="MetaData">
</xsl:template>

<xsl:template match="ImpNote">
</xsl:template>

<xsl:template match="UseCase">
</xsl:template>

<xsl:template match="Encodings">
</xsl:template>

<xsl:template match="DynamicDef">
</xsl:template>

<!-#- VA element -#->
<xsl:template match="ReviewNote">
</xsl:template>-->

<!-- ...................................................... 
      processing of static def (or Message) starts here. 
    ..................................................... -->    
<xsl:template match="HL7v2xStaticDef | Message">
  <xsl:param name="index" select="$offset"/>
  <xsl:variable name="msgType" select="@MsgType"/>
  <xsl:variable name="eventType" select="@EventType"/>
  <xsl:variable name="msgStructID" select="@MsgStructID"/>
  <xsl:variable name="eventDescription" select="@EventDesc"/>

  <xsl:apply-templates select="MetaData"/>
  
  <!-- Along with meta data that's being ignored (at least for now),
        the static def is a sequence of Segment and SegGroup elements.
        These can be intermixed in any order.  We invoke apply-templates 
        without the select attribute so that the specific template that gets
        invoked will depend on what's seen in the input stream.
        The msgStructId defines the root element of the output msg. -->
  <xsl:element name="{$msgStructID}">
    <xsl:apply-templates>
      <xsl:with-param name="msgType" select="$msgType"/>
      <xsl:with-param name="eventType" select="$eventType"/>
      <xsl:with-param name="msgStructID" select="$msgStructID"/>
      <xsl:with-param name="loopCounter" select="1"/>
      <xsl:with-param name="index" select="$offset"/>
    </xsl:apply-templates>   
  </xsl:element>  
  
  <xsl:value-of select="$newline"/>
</xsl:template>

<!-- ............................................................. -->
<xsl:template match="SegGroup" name="segGroup">
  <xsl:param name="msgType"/>
  <xsl:param name="eventType"/>
  <xsl:param name="msgStructID"/>
  <xsl:param name="loopCounter" select="1"/>
  <xsl:param name="index" select="0"/>
  <xsl:param name="init" select="true()"/>
  <xsl:param name="reps" select="0"/>

  <xsl:variable name="grpName" select="@Name"/>
  <xsl:variable name="grpLongName" select="@LongName"/>
  <xsl:variable name="grpUsage" select="@Usage"/>

  <!-- there are no options for controlling groups, so this has to be special caseed -->
  <xsl:variable name="includeGrp">
    <xsl:choose>
      <xsl:when test="$grpUsage = 'R'">
        <xsl:value-of select="'true'"/>        
      </xsl:when>
      <xsl:when test="$grpUsage = 'RE'">
        <xsl:choose>
          <xsl:when test="$offset + position() mod 3 = 0">
            <xsl:value-of select="'false'"/>        
          </xsl:when>
          <xsl:otherwise>
            <xsl:value-of select="'true'"/>        
          </xsl:otherwise>
        </xsl:choose>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="'false'"/>        
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>

  <!-- if min isn't a number (should be though), set = 0 -->
  <xsl:variable name="grpMin">
    <xsl:choose>
      <xsl:when test="number(@Min)">
        <xsl:value-of select="@Min"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="0"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>
  <!-- max might = *, we'll set to min+3 in this case -->
  <xsl:variable name="grpMax">
    <xsl:choose>
      <xsl:when test="number(@Max)">
        <xsl:value-of select="@Max"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="$grpMin + 3"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>
  
  <!-- set the number of times the group will repeat.
      (there's no option for controlling group repetition.) -->
  <xsl:variable name="grpReps">
    <xsl:choose>	
      <xsl:when test="$init">  <!-- not recursing  -->	
        <xsl:choose>
          <xsl:when test="$grpMin > $grpMax">	
            <xsl:value-of select="0"/>  <!-- max or min is set incorrectly -->
          </xsl:when>
          <xsl:when test="$grpMax = $grpMin">
            <xsl:value-of select="$grpMin"/>
          </xsl:when>
          <xsl:otherwise>	  <!-- grpMax > grpMin -->
            <xsl:choose>	
              <xsl:when test="$offset + position() mod 7 = 0">
                <xsl:value-of select="$grpMin"/>
              </xsl:when>
              <xsl:when test="$offset + position() mod 3 = 0">
                <xsl:value-of select="$grpMax"/>
              </xsl:when>
              <xsl:otherwise>
                <xsl:value-of select=
                    "$grpMin + floor(($grpMax - $grpMin) div 2)"/>
              </xsl:otherwise>
            </xsl:choose>
          </xsl:otherwise>
        </xsl:choose>
      </xsl:when>
      <xsl:otherwise>   <!-- value's already set if recursing -->
        <xsl:value-of select="$reps"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>
  
  <!-- process Segments or nested Segment Groups -->
  <xsl:if test="$includeGrp = 'true'">
    <xsl:if test="$grpReps &gt; 0">	
      <xsl:element name="{$msgStructID}.{$grpName}">
  	    <xsl:apply-templates>
  	      <xsl:with-param name="msgType" select="$msgType"/>
  	      <xsl:with-param name="eventType" select="$eventType"/>
  	      <xsl:with-param name="msgStructID" select="$msgStructID"/>
  	      <xsl:with-param name="loopCounter" select="1"/>
  	      <xsl:with-param name="index" select="$offset"/>
  	    </xsl:apply-templates>    
      </xsl:element>
    </xsl:if>
    
    <!-- recurse through this template if counter < reps -->
    <xsl:if  test="$loopCounter &lt; $grpReps">	
      <xsl:call-template name="segGroup">
        <xsl:with-param name="loopCounter" select="$loopCounter+1"/>
        <xsl:with-param name="msgType" select="$msgType"/>
        <xsl:with-param name="eventType" select="$eventType"/>
        <xsl:with-param name="msgStructID" select="$msgStructID"/>
        <xsl:with-param name="index" select="$index"/>
        <xsl:with-param name="init" select="false()"/>
        <xsl:with-param name="reps" select="$grpReps"/>
      </xsl:call-template>
    </xsl:if>
  </xsl:if>
</xsl:template>

<!-- ............................................................. -->
<xsl:template match="Segment" name="segment">
  <xsl:param name="msgType"/>
  <xsl:param name="eventType"/>
  <xsl:param name="msgStructID"/>
  <xsl:param name="loopCounter" select="1"/>
  <xsl:param name="index" select="0"/>
  <xsl:param name="init" select="true()"/>
  <xsl:param name="reps" select="0"/>
  <xsl:param name="descendants" select="0"/>
  <xsl:param name="siblingDescendants" select="0"/>

  <xsl:if test="$DEBUG=1 or $DEBUG &gt; 9">
	    Segment Parameters:
	    msgType =<xsl:value-of select= "$msgType"   />
	    eventType =<xsl:value-of select= "$eventType"   />
	    msgStructID =<xsl:value-of select= "$msgStructID"   />
  </xsl:if>
              
  <xsl:variable name="segName" select="@Name"/>
  <xsl:variable name="longName" select="@LongName"/>
 
  <!-- VA uses 'Optionality' in place of 'Usage' attribute -->
  <xsl:variable name="segUsage">
    <xsl:choose>
      <xsl:when test="@Usage!=''">
        <xsl:value-of select="@Usage"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="@Optionality"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>

  <xsl:variable name="segNumber">	
    <xsl:number	level="multiple" count="Segment" />
  </xsl:variable>
  
  <!-- see if the element will be included -->
  <xsl:variable name="includeSegment">
    <xsl:choose>
      <xsl:when test="$init">	
        <xsl:call-template name="includeElement">
          <xsl:with-param name="usage" select="$segUsage"/>
          <xsl:with-param name="nodeNumber" select="$segNumber"/>
          <xsl:with-param name="segmentName" select="$segName"/>
          <xsl:with-param name="rUErrorLocation" select="$rUErrorSegment"/>
          <xsl:with-param name="xUErrorLocation" select="$xUErrorSegment"/>
        </xsl:call-template>
      </xsl:when>
      <xsl:otherwise>	<!-- recursing, just set to true -->
        <xsl:value-of select="'true'"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>

  <!-- set cardinality -->    
  <xsl:variable name="cardinality">
    <xsl:choose>
      <!-- if there's a msg description we're following, set card = 1 -->
      <xsl:when test="$msgDescription != ''">
        <xsl:value-of select="1"/>
      </xsl:when>
      <xsl:otherwise>  <!-- no msg description -->
        <xsl:call-template name="setCardinalityOption">
          <xsl:with-param name="segName" select="$segName"/>
          <xsl:with-param name="cardinality" select="$segmentCardinality"/>
          <xsl:with-param name="elementNumber" select="$segNumber"/>
          <xsl:with-param name="errorMax" select="$maxSegCardError"/>      
          <xsl:with-param name="errorMin" select="$minSegCardError"/>            
          <xsl:with-param name="errorLocationMax" select="$cardErrorMaxSeg"/>
          <xsl:with-param name="errorLocationMin" select="$cardErrorMinSeg"/>
        </xsl:call-template>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>
  
  <!-- for debugging err segments -->
  <xsl:if test="$ANNOTATE=1">
    <xsl:choose>	
      <xsl:when test="$includeSegment='false' and $segUsage='R'">
        <xsl:comment>
          *SEGMENT ERROR*  
          SEGEMENT <xsl:value-of select="$segName"/> Required but not included!	
        </xsl:comment>
      </xsl:when>
      <xsl:when test="$cardinality = -1">
        <xsl:comment>
          *SEGMENT ERROR*  
          SEGEMENT <xsl:value-of select="$segName"/> 
          Maximum Segment Cardinality Violation!	
        </xsl:comment>
      </xsl:when>
      <xsl:when test="$cardinality = -0.1">
        <xsl:comment>
          *SEGMENT ERROR*  
          SEGEMENT <xsl:value-of select="$segName"/> 
          Minimum Segment Cardinality Violation!	
        </xsl:comment>
      </xsl:when>
    </xsl:choose>
  </xsl:if>
      
  <!-- if min isn't a number (should be though), set = 0 -->
  <xsl:variable name="segMin">
    <xsl:choose>
      <xsl:when test="number(@Min)">
        <xsl:value-of select="@Min"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="0"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>
  <!-- max might = *, we'll set to min+3 in this case -->
  <xsl:variable name="segMax">
    <xsl:choose>
      <xsl:when test="number(@Max)">
        <xsl:value-of select="@Max"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="$segMin + $starValue"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>

  <!-- set the number of times this seg will repeat -->
  <xsl:variable name="segReps">
    <xsl:choose>	
      <xsl:when test="$init">  <!-- not recursing  -->	
        <xsl:choose>
          <xsl:when test="$segMin > $segMax">	
            <xsl:value-of select="0"/>  <!-- max or min is set incorrectly -->
          </xsl:when>
          <!-- set to specific number of repetitions (see below) -->
          <xsl:when test="$cardinality &gt; 10 or $cardinality = 10">
            <xsl:choose>
	            <xsl:when test="$cardinality - 10 &gt; $segMax">
		            <xsl:value-of select="$segMax"/>
	            </xsl:when>
	            <xsl:when test="$cardinality - 10 &lt; $segMin ">
		            <xsl:value-of select="$segMin"/>
	            </xsl:when>
              <xsl:otherwise>
		            <xsl:value-of select="$cardinality - 10"/>
              </xsl:otherwise>
            </xsl:choose>	
          </xsl:when>
          <xsl:when test="$cardinality=1">	
            <xsl:value-of select="$segMax"/>
          </xsl:when>
          <xsl:when test="$cardinality=0">	
            <xsl:value-of select="$segMin"/>
          </xsl:when>
          <xsl:when test="$cardinality &gt; 0 and $cardinality &lt; 1">	
            <xsl:value-of select="$segMax"/>
          </xsl:when>
          <!-- generate invalid msg if card < 0 -->
          <xsl:when test="$cardinality = -1">	
            <xsl:value-of select="$segMax + 1"/>
          </xsl:when>
          <xsl:when test="$cardinality = -0.1">
            <xsl:choose>	
              <xsl:when test="$segMin > 0">
                <xsl:value-of select="$segMin - 1"/>
              </xsl:when>
              <xsl:otherwise>	
                <xsl:value-of select="$segMin"/>
              </xsl:otherwise>
            </xsl:choose>
          </xsl:when>
          <xsl:otherwise>   <!-- generate a mix of valid msgs -->	
            <xsl:choose>	
              <xsl:when test="((position() + ($offset mod 3)) mod 3) = 0">
                <xsl:value-of select="$segMax"/>
              </xsl:when>
              <xsl:when test="((position() + ($offset mod 3)) mod 2) = 0">
                <xsl:value-of select="$segMin"/>
              </xsl:when>
              <xsl:otherwise>
                <xsl:value-of select=
                    "$segMin + floor(($segMax - $segMin) div 2)"/>
              </xsl:otherwise>
            </xsl:choose>
          </xsl:otherwise>
        </xsl:choose>
      </xsl:when>
      <xsl:otherwise>   <!-- value's already set if recursing -->
        <xsl:value-of select="$reps"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>

  <xsl:if test="$DEBUG=4 or $DEBUG &gt; 9">
			includeSegment   =<xsl:value-of select= "$includeSegment"     />
			cardinality   =<xsl:value-of select= "$cardinality"     />
			segName    =<xsl:value-of select= "$segName"     />
			segReps    =<xsl:value-of select= "$segReps"     />
			segUsage    =<xsl:value-of select= "$segUsage"     />
	</xsl:if>
 
  <!--  we'll keep a count of all preceding segment descendants plus all 
      descendants of this segment and use the values for setting indexSeg 
      below. -->
  <xsl:variable name="childCount">
    <xsl:choose>
      <xsl:when test="$init">	
      <xsl:value-of select="count(child::*/descendant::Field)
          + count(child::*/descendant::Component)
          + count(child::*/descendant::SubComponent)"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="$descendants"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>
  <xsl:variable name="sibChildCount">
    <xsl:choose>
      <xsl:when test="$init">	
        <xsl:value-of select="count(preceding-sibling::*/descendant::Field)
            + count(preceding-sibling::*/descendant::Component)
            + count(preceding-sibling::*/descendant::SubComponent)"/>
      </xsl:when>
      <xsl:otherwise>	
        <xsl:value-of select="$siblingDescendants"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>
  
  <!-- we calculate a value that will generally be different for each 
      Segment.  adding this value to indexSeg will help to produce a 
      different value for index each time it's used to look up a table 
      value.-->
  <xsl:variable name="p">
  <xsl:value-of select="position()"/>
  </xsl:variable>
  <xsl:variable name="delta">
  <xsl:value-of select="$sibChildCount mod ($p+1)"/>
  </xsl:variable>

  <!-- set indexSeg. the count is used as a starting location for grabbing data
    values from the tables and the location is further modified by a similar
    scheme below. this should help to minimize the amount of data duplication,
    but not eliminate it. (we can't track how many times we've repeated prior
    segments.) 
    (we'd like to avoid duplicating the data values we use as much as
    possible. grabbing data values sequentially from the tables after starting
    at an arbitrary location is too difficult without being able to change
    variable values.  the index doesn't need to change any more frequently
    than we encounter items of the same dataType, but as long as the index
    does change each time we're at a leaf node, we should avoid duplicating
    values.)    -->
  <xsl:variable name="indexSeg">
    <xsl:value-of select=
        "$offset + $delta + ($loopCounter - 1)*$childCount + $sibChildCount"/>
  </xsl:variable>
  
  <xsl:variable name="setID">
  	<xsl:value-of select="Field[starts-with(
      		translate(normalize-space(@Name), 'setid', 'SETID'), 'SET ID')]/@Name"/>
  </xsl:variable>
  <xsl:variable name="sequenceNumber">
    <xsl:choose>
      <xsl:when test="$setID != ''">
        <xsl:value-of select="1"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="''"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>
  <xsl:variable name="segTrailer">
    <xsl:choose>
      <xsl:when test="$sequenceNumber != ''">
        <xsl:value-of select="'.'"/><xsl:value-of select="$sequenceNumber"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="''"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>

<!--setID=<xsl:value-of select="$setID"/>
sequenceNumber=<xsl:value-of select="$sequenceNumber"/>
trailer=<xsl:value-of select="$segTrailer"/>
-->
  <!--  each segment is made up of a sequence of fields.
      the names of the children of the output msg's root element are 
      taken from the seg names. -->
  <xsl:if test="$includeSegment = 'true'">
    <xsl:if test="$segReps &gt; 0">
      <xsl:element name="{$segName}">
          <xsl:apply-templates select="Field">
            <xsl:with-param name="msgType" select="$msgType"/>
            <xsl:with-param name="eventType" select="$eventType"/>
            <xsl:with-param name="segName" select="$segName"/>
            <xsl:with-param name="msgStructID" select="$msgStructID"/>
            <xsl:with-param name="loopCounter" select="1"/>
            <xsl:with-param name="index" select="$indexSeg"/>
          </xsl:apply-templates>
      </xsl:element>
    </xsl:if>
   
    <!-- call this template again if the Segment's being repeated -->
    <xsl:if  test="$loopCounter &lt; $segReps">	
      <xsl:call-template name="segment">
        <xsl:with-param name="loopCounter" select="$loopCounter+1"/>
        <xsl:with-param name="msgType" select="$msgType"/>
        <xsl:with-param name="eventType" select="$eventType"/>
        <xsl:with-param name="msgStructID" select="$msgStructID"/>
        <xsl:with-param name="index" select="$indexSeg"/>
        <xsl:with-param name="init" select="false()"/>
        <xsl:with-param name="reps" select="$segReps"/>
        <xsl:with-param name="descendants" select="$childCount"/>
        <xsl:with-param name="siblingDescendants" select="$sibChildCount"/>
      </xsl:call-template>
    </xsl:if>
  </xsl:if>
</xsl:template>

<!-- ............................................................. -->
<xsl:template match="Field" name="field">
  <xsl:param name="msgType"/>
  <xsl:param name="eventType"/>
  <xsl:param name="msgStructID"/>
  <xsl:param name="loopCounter" select="1"/>
  <xsl:param name="segName"/>
  <xsl:param name="index" select="0"/>
  <xsl:param name="init" select="true()"/>
  <xsl:param name="reps" select="0"/>
  <xsl:param name="descendants" select="0"/>
  <xsl:param name="siblingDescendants" select="0"/>

  <xsl:if test="$DEBUG=1 or $DEBUG &gt; 9">
	    Field Parameters:
	    msgType =<xsl:value-of select= "$msgType"   />
	    eventType =<xsl:value-of select= "$eventType"   />
	    msgStructID =<xsl:value-of select= "$msgStructID"   />
  </xsl:if>
              
  <xsl:variable name="fieldName" select="@Name"/>
  <xsl:variable name="fieldDataType" select="@Datatype"/>
  <xsl:variable name="maxFieldLength" select="@Length"/>
  <xsl:variable name="itemNo">
    <xsl:choose>
      <xsl:when test="number(@ItemNo)">
        <xsl:value-of select="@ItemNo"/>        
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="''"/>        
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>
  <xsl:variable name="tableId" select="@Table"/>
  <xsl:variable name="reference" select="@xmlRef"/>
  <xsl:variable name="exValue" select="DataValues/@ExValue"/>
  <xsl:variable name="constantValue" select="@ConstantValue"/>

	<!-- VA uses 'Optionality' in place of 'Usage' attribute -->
  <xsl:variable name="fieldUsage">
    <xsl:choose>
      <xsl:when test="@Usage!=''">
        <xsl:value-of select="@Usage"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="@Optionality"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>

  <xsl:variable name="fieldNumber">	
    <xsl:number	level="multiple" count="Segment | Field" />
  </xsl:variable>
  
  <!-- if we're not recursing, call includeElement to determine if this 
      field will be included based on the field's usage setting and the 
      values of the usage options, otherwise include. 
      -->
  <xsl:variable name="includeField"> 
    <xsl:choose>
      <xsl:when test="$init">	
        <xsl:call-template name="includeElement">
          <xsl:with-param name="usage" select="$fieldUsage"/>
          <xsl:with-param name="nodeNumber" select="$fieldNumber"/>
          <xsl:with-param name="segmentName" select="$segName"/>
          <xsl:with-param name="fieldName" select="$fieldName"/>
          <xsl:with-param name="rUErrorLocation" select="$rUErrorField"/>
          <xsl:with-param name="xUErrorLocation" select="$xUErrorField"/>
        </xsl:call-template>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="'true'"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>
 
  <xsl:variable name="useErrorCode"> 
    <xsl:choose>
      <xsl:when test="$elementErrorCode = 'true'">	
        <xsl:call-template name="matchesElementName">
          <xsl:with-param name="segmentName" select="$segName"/>
          <xsl:with-param name="fieldName" select="$fieldName"/>
        </xsl:call-template>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="'false'"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>
  
  <!-- set cardinality, same strategy as for segment -->
  <xsl:variable name="cardinality">
    <xsl:choose>
      <!-- if there's a msg description we're following, set card = 1 -->
      <xsl:when test="$msgDescription != ''">
        <xsl:value-of select="1"/>
      </xsl:when>
      <xsl:otherwise>  <!-- no msg description -->
        <xsl:call-template name="setCardinalityOption">
          <xsl:with-param name="segName" select="$segName"/>
          <xsl:with-param name="fieldName" select="$fieldName"/>
      		<xsl:with-param name="cardinality" select="$fieldCardinality"/>
          <xsl:with-param name="elementNumber" select="$fieldNumber"/>
          <xsl:with-param name="errorMax" select="$maxFieldCardError"/>      
          <xsl:with-param name="errorMin" select="$minFieldCardError"/>            
          <xsl:with-param name="errorLocationMax" 
            		select="$cardErrorMaxField"/>
          <xsl:with-param name="errorLocationMin" 
            		select="$cardErrorMinField"/>
        </xsl:call-template>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>

<!-- set lengthError if this is the right field and we want a
      field length error generated. 
      (-0.1, 1 error, only case we handle right now) -->
  <xsl:variable name="lengthError">
    <xsl:choose>
    <!-- 
      <xsl:when test="$init
                    and $fieldNumber = $fLengthErrorLoc
                    and $fieldLength = -0.1">
     -->
        <xsl:when test="$init
            and $segName = $namedSegment
            and $fieldName = $namedField
                    and $fieldLength = -0.1">
        <xsl:value-of select="'true'"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="'false'"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>
 
  <xsl:if test="$ANNOTATE=1">
    <xsl:choose>	
      <xsl:when test="$includeField='false' and $fieldUsage='R'">
        <xsl:comment>
          *FIELD ERROR*  
          FIELD '<xsl:value-of select="$fieldName"/>' Required but not included!	
        </xsl:comment>
      </xsl:when>
      <xsl:when test="$cardinality = -1">
        <xsl:comment>
          *FIELD ERROR*  
          SEGMENT <xsl:value-of select="$segName"/>
          FIELD <xsl:value-of select="$fieldName"/>
          Maximum Field Cardinality Violation!	
        </xsl:comment>
      </xsl:when>
      <xsl:when test="$cardinality = -0.1">
        <xsl:comment>
          *FIELD ERROR*  
          SEGMENT <xsl:value-of select="$segName"/>
          FIELD <xsl:value-of select="$fieldName"/>
          Minimum Field Cardinality Violation!	
        </xsl:comment>
      </xsl:when>
    </xsl:choose>
  </xsl:if>

  <!-- if min isn't a number (should be though), set = 0 -->
  <xsl:variable name="fieldMin">
    <xsl:choose>
      <xsl:when test="number(@Min)">
        <xsl:value-of select="@Min"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="0"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>
  <!-- max might = *, we'll set to min+3 in this case -->
  <xsl:variable name="fieldMax">
    <xsl:choose>
      <xsl:when test="number(@Max)">
        <xsl:value-of select="@Max"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="$fieldMin + $starValue"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>
  
  <!-- set the number of times this field will repeat -->
  <xsl:variable name="fieldReps">
    <xsl:choose>	
      <xsl:when test="$init">  <!-- not recursing  -->	
        <xsl:choose>
          <xsl:when test="$fieldMin > $fieldMax">	
            <xsl:value-of select="0"/>  <!-- max or min is set incorrectly -->
          </xsl:when>
          <!-- set to specific number of repetitions (see below) -->
          <xsl:when test="$cardinality &gt; 10 or $cardinality = 10">
            <xsl:choose>
	            <xsl:when test="$cardinality - 10 &gt; $fieldMax">
		            <xsl:value-of select="$fieldMax"/>
	            </xsl:when>
	            <xsl:when test="$cardinality - 10 &lt; $fieldMin ">
		            <xsl:value-of select="$fieldMin"/>
	            </xsl:when>
              <xsl:otherwise>
		            <xsl:value-of select="$cardinality - 10"/>
              </xsl:otherwise>
            </xsl:choose>	
          </xsl:when>          
          <xsl:when test="$cardinality=1">	
            <xsl:value-of select="$fieldMax"/>
          </xsl:when>
          <xsl:when test="$cardinality=0">	
            <xsl:value-of select="$fieldMin"/>
          </xsl:when>
          <!-- generate invalid msg if card < 0 -->
          <xsl:when test="$cardinality = -1">	
            <xsl:value-of select="$fieldMax + 1"/>
          </xsl:when>
          <xsl:when test="$cardinality = -0.1">
            <xsl:choose>	
              <xsl:when test="$fieldMin > 0">
                <xsl:value-of select="$fieldMin - 1"/>
              </xsl:when>
              <xsl:otherwise>	
                <xsl:value-of select="$fieldMin"/>
              </xsl:otherwise>
            </xsl:choose>
          </xsl:when>
          <xsl:otherwise>   <!-- generate a mix of valid msgs -->	
            <xsl:choose>
              <xsl:when test="((position() + ($offset mod 4)) mod 4) = 0">
                <xsl:value-of select="$fieldMax"/>
              </xsl:when>
              <xsl:when test="((position() + ($offset mod 4)) mod 3) = 0">
                <xsl:value-of select="$fieldMin"/>
              </xsl:when>
              <xsl:otherwise>
                <xsl:value-of select=
                    "$fieldMin + floor(($fieldMax - $fieldMin) div 2)"/>
              </xsl:otherwise>
            </xsl:choose>
          </xsl:otherwise>
        </xsl:choose>
      </xsl:when>
      <xsl:otherwise>	
        <xsl:value-of select="$reps"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>

  <xsl:if test="$DEBUG=4 or $DEBUG &gt; 9">
			includeField   =<xsl:value-of select= "$includeField"     />  
			cardinality   =<xsl:value-of select= "$cardinality"     />
			segName    =<xsl:value-of select= "$segName"     />
			fieldName    =<xsl:value-of select= "$fieldName"     />
			fieldReps    =<xsl:value-of select= "$fieldReps"     />      
			fieldUsage    =<xsl:value-of select= "$fieldUsage"     />      
			loopCounter    =<xsl:value-of select= "$loopCounter"     />      
	</xsl:if>

	<!-- keep a count of descendants of this Field and a count of the preceding 
      Fields descendants including the Fields themselves since they can be 
      leaves. (we change the childCount to 1 when it's 0 because otherwise 
      for a repeating Field with no children the index value will remain the
      same for each repetition.) -->
  <xsl:variable name="childCount">
    <xsl:choose>
      <xsl:when test="$init">
        <xsl:variable name="cnt">		
          <xsl:value-of select="count(child::*/descendant::Component)
              + count(child::*/descendant::SubComponent)"/>
        </xsl:variable>
        <xsl:choose>
          <xsl:when test="$cnt &gt; 0">
            <xsl:value-of select="$cnt"/>
          </xsl:when>
          <xsl:otherwise>
            <xsl:value-of select="1"/>
          </xsl:otherwise>
        </xsl:choose>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="$descendants"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>
  <xsl:variable name="sibChildCount">
    <xsl:choose>
      <xsl:when test="$init">	
        <xsl:value-of select="count(preceding-sibling::Field)
            + count(preceding-sibling::Field/descendant::Component)
            + count(preceding-sibling::Field/descendant::SubComponent)"/>
      </xsl:when>
      <xsl:otherwise>	
        <xsl:value-of select="$siblingDescendants"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>
  
  <!-- set indexField.  -->
  <xsl:variable name="indexField">
    <xsl:value-of select=
        "$index + ($loopCounter - 1)*$childCount + $sibChildCount"/>
  </xsl:variable>

  <!-- some item nums are repeated, 
      so we have to check the name before selecting the sequence num -->  
  <xsl:variable name="fieldSeqNum">
    <xsl:choose>
  		<xsl:when test="starts-with($segName, 'Z')">
			  <xsl:number level="single"/>
  		</xsl:when>    
  		<xsl:when test="$itemNo = ''">
			  <xsl:number level="single"/>
  		</xsl:when>    
		<xsl:otherwise>
		    <xsl:for-each select="document($dataElementsTable)">
		      <xsl:for-each select="key('fieldItem', $itemNo)">
		        <xsl:if test="@name=$segName">
		          <xsl:value-of select="@sequenceNum"/>
		        </xsl:if>
		      </xsl:for-each>
		    </xsl:for-each>
		</xsl:otherwise>	
    </xsl:choose>
  </xsl:variable>

  <xsl:if test="$includeField='true'">
    <xsl:if test="$fieldReps &gt; 0">
      <xsl:variable name="primitive">	
        <xsl:call-template name="isPrimitive">
          <xsl:with-param name="dataType"	select="$fieldDataType"/>
        </xsl:call-template>
      </xsl:variable>
      
      <!--  the output msg elements built from the Fields are named based on 
          the parent Segment name and the field's sequence number. -->
      <xsl:element name="{$segName}.{$fieldSeqNum}"> 
        <!-- get value if primitive or CE -->
        <xsl:variable name="dataValue">
        <xsl:choose>		<!-- if there's a constant value, it has to be used-->
          <xsl:when test="$constantValue!=''">
            <xsl:value-of select="$constantValue"/>
          </xsl:when>
          <xsl:otherwise>	<!-- no const value -->
  	        <xsl:if test="$primitive='true' or $fieldDataType='CE'">
  	          <xsl:call-template name="getValue">
  	            <xsl:with-param name="dataType" select="$fieldDataType"/>
  	            <xsl:with-param name="elementName" select=     "$fieldName"/>
  	            <xsl:with-param name="elementQualifier" select="$segName"/>
  	            <xsl:with-param name="tableId" select=         "$tableId"/>
  	            <xsl:with-param name="useErrorCode" select=    "$useErrorCode"/>
  	            <xsl:with-param name="reference"  select=      "$reference"/>
  	            <xsl:with-param name="indexTable" select=      "$indexField"/>
  	            <xsl:with-param name="messageControlID" select="$messageControlID"/>
  	            <xsl:with-param name="maxLength" select="$maxFieldLength"/>
  	            <xsl:with-param name="lengthError" select="$lengthError"/>
  	            <xsl:with-param name="exValue" select="$exValue"/>
  	          </xsl:call-template>
  	        </xsl:if>
          </xsl:otherwise>
         </xsl:choose>
        </xsl:variable>
          
        <xsl:choose>    <!-- primitive or composite (including CE) --> 
          <xsl:when test="$primitive='true'">   <!-- just add value -->
            <xsl:value-of select="$dataValue"/>
          </xsl:when>
          <xsl:otherwise>   <!-- Components or CE value -->
            <!-- 1st see if the Field's composite length should be checked -->        
            <xsl:if test="$compositeLength = 1">	
              <!-- get the Field's Components whose use hasn't been 
                  excluded -->
              <xsl:variable name="components">
                <xsl:value-of select="child::Component[@Usage!='X']"/>
              </xsl:variable>
    
              <!-- get the sum of the max lengths of each Component -->          
              <xsl:variable name="maxChildLengths">
                <xsl:call-template name="totalLengths">
                  <xsl:with-param name="nodeList" select="$components"/>
                </xsl:call-template>
              </xsl:variable>
              
              <!-- if the Component lengths can exceed the max allowed Field
                  length, issue a warning -->
              <xsl:if test="$WARN=1 and $maxChildLengths > $maxFieldLength">
                <xsl:variable name="msg">
                Total Length of components may violate Field's length constraint.
                </xsl:variable>
                <xsl:call-template name="Warn">
                  <xsl:with-param name="segment" select="$segName"/>
                  <xsl:with-param name="field" select="$fieldName"/>
                  <xsl:with-param name="message" select="$msg"/>
                </xsl:call-template>
              </xsl:if>
            </xsl:if>
  
            <xsl:choose>  <!-- dump CE value or process SubComps -->
              <xsl:when test="$fieldDataType='CE'">
              
                <xsl:variable name="idUsage">
                  <xsl:value-of select="Component[@Name='identifier']/@Usage"/>
                </xsl:variable>
                <xsl:variable name="textUsage">
                  <xsl:value-of select="Component[@Name='text']/@Usage"/>
                </xsl:variable>
                <xsl:variable name="csUsage">
                  <xsl:value-of select="Component[@Name='name of coding system']/@Usage"/>
                </xsl:variable>
                <xsl:variable name="aidUsage">
                  <xsl:value-of select="Component[@Name='alternative identifier']/@Usage"/>
                </xsl:variable>
                <xsl:variable name="atextUsage">
                  <xsl:value-of select="Component[@Name='alternative text']/@Usage"/>
                </xsl:variable>
                <xsl:variable name="acsUsage">
                  <xsl:value-of select="Component[@Name='name of alternate coding system']/@Usage"/>
                </xsl:variable>
                
                <!-- (don't assign to a variable, the xml tags will be
                    lost if this is done unless buildCEelement is changed) -->	
                <xsl:call-template name="buildCEelement">
                  <xsl:with-param name="ceValue"	select="$dataValue"/>
                  <xsl:with-param name="elementName" select="$fieldDataType"/>
                  <xsl:with-param name="idUsage" select="$idUsage"/>
                  <xsl:with-param name="textUsage" select="$textUsage"/>
                  <xsl:with-param name="csUsage" select="$csUsage"/>
                  <xsl:with-param name="aidUsage" select="$aidUsage"/>
                  <xsl:with-param name="atextUsage" select="$atextUsage"/>
                  <xsl:with-param name="acsUsage" select="$acsUsage"/>
                </xsl:call-template>
              </xsl:when>
              <xsl:otherwise>	
                <xsl:apply-templates select="Component">
                  <xsl:with-param name="msgType" select="$msgType"/>
                  <xsl:with-param name="eventType" select="$eventType"/>
                  <xsl:with-param name="segName" select="$segName"/>
                  <xsl:with-param name="msgStructID" select="$msgStructID"/>
                  <xsl:with-param name="fieldDataType" select="$fieldDataType"/>
                  <xsl:with-param name="fieldUsage" select="$fieldUsage"/>
                  <xsl:with-param name="fieldName" select="$fieldName"/>
                  <xsl:with-param name="index" select="$indexField"/>
                </xsl:apply-templates>
              </xsl:otherwise>
            </xsl:choose>
          </xsl:otherwise>
        </xsl:choose>
      </xsl:element>
    </xsl:if>

    <!-- call this template again if the Field's being repeated 
      fieldReps is set when this template is initially invoked
      (not recursing) on subsequent calls (from here) fieldReps
      is set to the value of 'reps', so it doesn't change -->
    <xsl:if  test="$loopCounter &lt; $fieldReps">
      <xsl:call-template name="field">
        <xsl:with-param name="loopCounter" select="$loopCounter+1"/>
        <xsl:with-param name="segName" select="$segName"/>
        <xsl:with-param name="index" select="$indexField"/>
        <xsl:with-param name="init" select="false()"/>
        <xsl:with-param name="reps" select="$fieldReps"/>
        <xsl:with-param name="descendants" select="$childCount"/>
        <xsl:with-param name="siblingDescendants" select="$sibChildCount"/>
      </xsl:call-template>
    </xsl:if>
  </xsl:if>
</xsl:template>

<!-- ............................................................. -->
<xsl:template match="Component">
  <xsl:param name="msgType"/>
  <xsl:param name="eventType"/>
  <xsl:param name="msgStructID"/>
  <xsl:param name="segName"/>
  <xsl:param name="fieldDataType"/>
  <xsl:param name="fieldUsage"/>
  <xsl:param name="fieldName"/>
  <xsl:param name="index" select="0"/>

  <xsl:variable name="compName" select="@Name"/>
  <xsl:variable name="compDataType" select="@Datatype"/>
  <xsl:variable name="maxCompLength" select="@Length"/>
  <xsl:variable name="tableId" select="@Table"/>
  <xsl:variable name="reference" select="@xmlRef"/>
 	<xsl:variable name="exValue" select="DataValues/@ExValue"/>
  <xsl:variable name="constantValue" select="@ConstantValue"/>
 
	<!-- VA uses 'Optionality' in place of 'Usage' attribute -->
  <xsl:variable name="compUsage">
    <xsl:choose>
      <xsl:when test="@Usage!=''">
        <xsl:value-of select="@Usage"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="@Optionality"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>

  <xsl:if test="$DEBUG=1 or $DEBUG &gt; 9">
     Component:
    msgType = <xsl:value-of select="$msgType"/>
    eventType = <xsl:value-of select="$eventType"/>
    msgStructID = <xsl:value-of select="$msgStructID"/>
    segName = <xsl:value-of select="$segName"/>...
    fieldDataType = <xsl:value-of select="$fieldDataType"/>
    fieldUsage = <xsl:value-of select="$fieldUsage"/>
    fieldName = <xsl:value-of select="$fieldName"/>...
    index = <xsl:value-of select="$index"/>
    compName = <xsl:value-of select="$compName"/>...    
    ................................................
  </xsl:if>
  
  <xsl:variable name="compNumber">	
    <xsl:number	level="multiple" count="Segment | Field | Component" />
  </xsl:variable>

  <xsl:variable name="compCount">
    <xsl:number count="Component" level="single"/>
  </xsl:variable>

  <!-- set includeComponent (we don't recurse thru this template) -->      
  <xsl:variable name="includeComponent">
    <xsl:call-template name="includeElement">
      <xsl:with-param name="usage" select="$compUsage"/>
      <xsl:with-param name="nodeNumber" select="$compNumber"/>
          <xsl:with-param name="segmentName" select="$segName"/>
          <xsl:with-param name="fieldName" select="$fieldName"/>
          <xsl:with-param name="componentName" select="$compName"/>
      <xsl:with-param name="rUErrorLocation" select="$rUErrorComponent"/>
      <xsl:with-param name="xUErrorLocation" select="$xUErrorComponent"/>
    </xsl:call-template>
  </xsl:variable>
  
  <xsl:variable name="useErrorCode"> 
    <xsl:choose>
      <xsl:when test="$elementErrorCode = 'true'">	
        <xsl:call-template name="matchesElementName">
          <xsl:with-param name="segmentName" select="$segName"/>
          <xsl:with-param name="fieldName" select="$fieldName"/>
          <xsl:with-param name="componentName" select="$compName"/>
        </xsl:call-template>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="'false'"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>
     
  <xsl:if test="$DEBUG=4 or $DEBUG &gt; 9">  
			includeComponent   =<xsl:value-of select= "$includeComponent"     />  
			segName    =<xsl:value-of select= "$segName"     />
			fieldName    =<xsl:value-of select= "$fieldName"     />
			compName    =<xsl:value-of select= "$compName"     />
			compUsage    =<xsl:value-of select= "$compUsage"     />
  </xsl:if>

  <xsl:variable name="lengthError">
    <xsl:choose>
      <xsl:when test="$segName = $namedSegment
            and $fieldName = $namedField
            and $compName = $namedComponent
            and $compLength = -0.1">
        <xsl:value-of select="'true'"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="'false'"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>

  <!-- keep count of descendants of preceding Component and the 
      Components themselves since they can be leaves. (no recursing, so
      don't need childCount)  -->
  <xsl:variable name="sibChildCount">
    <xsl:value-of 
        select="count(preceding-sibling::Component)
        + count(preceding-sibling::Component/descendant::SubComponent)"/>
  </xsl:variable>
  
  <!-- set indexField.  -->
  <xsl:variable name="indexComp">
    <xsl:value-of select="$index + $sibChildCount"/>
  </xsl:variable>
  
  <!-- the component elements in the output msg are numbered 
      sequentially and named using the parent field data type. -->
  <xsl:if test="$includeComponent = 'true'">
    <xsl:element name="{$fieldDataType}.{$compCount}">
      <xsl:variable name="primitive">	
        <xsl:call-template name="isPrimitive">
          <xsl:with-param name="dataType"	select="$compDataType"/>
        </xsl:call-template>
      </xsl:variable>
      
      <!-- Message Type is special cased,
        	if there's a constant value, we use it,
          otherwise get value if primitive or CE -->
      <xsl:variable name="dataValue">
        <xsl:choose>
          <xsl:when test="$segName='MSH' and $fieldName='Message Type'">
            <xsl:if test="$compName='message type'">
              <xsl:value-of select="$msgType"/>
            </xsl:if>
            <xsl:if test="$compName='trigger event'">
              <xsl:value-of select="$eventType"/>
            </xsl:if>
            <xsl:if test="$compName='message structure'">
              <xsl:value-of select="$msgStructID"/>
            </xsl:if>
          </xsl:when>
	        <xsl:when test="$constantValue!=''">
	          <xsl:value-of select="$constantValue"/>
	        </xsl:when>
          <xsl:otherwise>
            <xsl:if test="$primitive='true' or $compDataType='CE'">
              <xsl:call-template name="getValue">
                <xsl:with-param name="dataType" select="$compDataType"/>
                <xsl:with-param name="elementName" select=     "$compName"/>
                <xsl:with-param name="elementQualifier" select="$fieldName"/>
                <xsl:with-param name="tableId" select=         "$tableId"/>
                <xsl:with-param name="useErrorCode" select=    "$useErrorCode"/>
                <xsl:with-param name="reference"  select=      "$reference"/>
                <xsl:with-param name="indexTable" select=      "$indexComp"/>
                <xsl:with-param name="messageControlID" select="$messageControlID"/>
                <xsl:with-param name="maxLength" select="$maxCompLength"/>
                <xsl:with-param name="lengthError" select="$lengthError"/>
		            <xsl:with-param name="exValue" select="$exValue"/>                
              </xsl:call-template>
            </xsl:if>
          </xsl:otherwise>
        </xsl:choose>
      </xsl:variable>
        
      <xsl:choose>    <!-- primitive or composite (including CE) --> 
        <xsl:when test="$primitive='true'">   <!-- just add value -->
          <xsl:value-of select="$dataValue"/>
        </xsl:when>
        <xsl:otherwise>   <!-- SubComponents or CE value -->
          <!-- 1st see if the Comp's composite length should be checked -->        
          <xsl:if test="$compositeLength = 1">	
            <!-- get this Comp's child SubComponents whose use is in effect -->
            <xsl:variable name="subComps">
              <xsl:value-of select="child::SubComponent[@Usage!='X']"/>
            </xsl:variable>
  
            <!-- get the sum of the max lengths of each SubComponent -->          
            <xsl:variable name="maxChildLengths">
              <xsl:call-template name="totalLengths">
                <xsl:with-param name="nodeList" select="$subComps"/>
              </xsl:call-template>
            </xsl:variable>
            
            <!-- if the SubComponent lengths can exceed the max allowed 
                Component length, issue a warning -->
            <xsl:if test="$WARN=1 and $maxChildLengths > $maxCompLength">
              <xsl:variable name="msg">
              Total Length of SubComponents may violate Component's length constraint.
              </xsl:variable>
              <xsl:call-template name="Warn">
                <xsl:with-param name="segment" select="$segName"/>
                <xsl:with-param name="field" select="$fieldName"/>
                <xsl:with-param name="component" select="$compName"/>
                <xsl:with-param name="message" select="$msg"/>
              </xsl:call-template>
            </xsl:if>
          </xsl:if>

          <xsl:choose>  <!-- dump CE value or process SubComps -->
            <xsl:when test="$compDataType='CE'">
              <xsl:variable name="idUsage">
                <xsl:value-of select="SubComponent[@Name='identifier']/@Usage"/>
              </xsl:variable>
              <xsl:variable name="textUsage">
                <xsl:value-of select="SubComponent[@Name='text']/@Usage"/>
              </xsl:variable>
              <xsl:variable name="csUsage">
                <xsl:value-of select="SubComponent[@Name='name of coding system']/@Usage"/>
              </xsl:variable>
              <xsl:variable name="aidUsage">
                <xsl:value-of select="SubComponent[@Name='alternative identifier']/@Usage"/>
              </xsl:variable>
              <xsl:variable name="atextUsage">
                <xsl:value-of select="SubComponent[@Name='alternative text']/@Usage"/>
              </xsl:variable>
              <xsl:variable name="acsUsage">
                <xsl:value-of select="SubComponent[@Name='name of alternate coding system']/@Usage"/>
              </xsl:variable>

              <!-- (don't assign to a variable, the xml tags will be
                  lost if this is done unless buildCEelement is changed) -->	
              <xsl:call-template name="buildCEelement">
                <xsl:with-param name="ceValue"	select="$dataValue"/>
                <xsl:with-param name="elementName" select="$compDataType"/>
                <xsl:with-param name="idUsage" select="$idUsage"/>
                <xsl:with-param name="textUsage" select="$textUsage"/>
                <xsl:with-param name="csUsage" select="$csUsage"/>
                <xsl:with-param name="aidUsage" select="$aidUsage"/>
                <xsl:with-param name="atextUsage" select="$atextUsage"/>
                <xsl:with-param name="acsUsage" select="$acsUsage"/>
              </xsl:call-template>
            </xsl:when>
            <xsl:otherwise>	
              <xsl:apply-templates select="SubComponent">
                <xsl:with-param name="compDataType" select="$compDataType"/>
                <xsl:with-param name="compName" select="$compName"/>
                <xsl:with-param name="compCount" select="$compCount"/>
                <xsl:with-param name="fieldUsage" select="$fieldUsage"/>
                <xsl:with-param name="compUsage" select="$compUsage"/>
                <xsl:with-param name="fieldName" select="$fieldName"/>
                <xsl:with-param name="segName" select="$segName"/>
                <xsl:with-param name="index" select="$indexComp"/>
              </xsl:apply-templates>
            </xsl:otherwise>
          </xsl:choose>
        </xsl:otherwise>
      </xsl:choose>
    </xsl:element>
  </xsl:if>
</xsl:template>

<!-- ............................................................. -->
<xsl:template match="SubComponent">
  <xsl:param name="compDataType"	/>
  <xsl:param name="compName"	/>
  <xsl:param name="compCount"	/>
  <xsl:param name="fieldName"	/>
  <xsl:param name="fieldUsage"	/>
  <xsl:param name="segName"	/>
  <xsl:param name="compUsage"	/>
  <xsl:param name="index" select="0"/>

  <xsl:variable name="subName" select="@Name"/>
  <xsl:variable name="subDataType" select="@Datatype"/>
  <xsl:variable name="maxSubLength" select="@Length"/>
  <xsl:variable name="tableId" select="@Table"/>
  <xsl:variable name="reference" select="@xmlRef"/>
  <xsl:variable name="exValue" select="DataValues/@ExValue"/>
  <xsl:variable name="constantValue" select="@ConstantValue"/>
  
	<!-- VA uses 'Optionality' in place of 'Usage' attribute -->
  <xsl:variable name="subUsage">
    <xsl:choose>
      <xsl:when test="@Usage!=''">
        <xsl:value-of select="@Usage"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="@Optionality"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>

  <xsl:variable name="subNumber">	
    <xsl:number	level="multiple" count="Segment | Field | Component 
                      | SubComponent" />
  </xsl:variable>

  <xsl:variable name="subCount">
    <xsl:number count="SubComponent" level="single"/>
  </xsl:variable>

  <!-- set includeSubcomponent (we don't recurse thru this template) -->      
  <xsl:variable name="includeSubcomponent">
    <xsl:call-template name="includeElement">
      <xsl:with-param name="usage" select="$subUsage"/>
      <xsl:with-param name="nodeNumber" select="position()"/>
          <xsl:with-param name="segmentName" select="$segName"/>
          <xsl:with-param name="fieldName" select="$fieldName"/>
          <xsl:with-param name="componentName" select="$compName"/>
          <xsl:with-param name="subComponentName" select="$subName"/>
      <xsl:with-param name="rUErrorLocation" select="$rUErrorSubcomp"/>
      <xsl:with-param name="xUErrorLocation" select="$xUErrorSubcomp"/>
    </xsl:call-template>
  </xsl:variable>
  
  <xsl:variable name="useErrorCode"> 
    <xsl:choose>
      <xsl:when test="$elementErrorCode = 'true'">	
        <xsl:call-template name="matchesElementName">
          <xsl:with-param name="segmentName" select="$segName"/>
          <xsl:with-param name="fieldName" select="$fieldName"/>
          <xsl:with-param name="componentName" select="$compName"/>
          <xsl:with-param name="subComponentName" select="$subName"/>
        </xsl:call-template>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="'false'"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>
    
  <xsl:if test="$DEBUG=4 or $DEBUG &gt; 9">  
			includeSubcomponent    =<xsl:value-of select= "$includeSubcomponent"     />
			segName    =<xsl:value-of select= "$segName"     />
			fieldName    =<xsl:value-of select= "$fieldName"     />
			compName    =<xsl:value-of select= "$compName"     />
			subName    =<xsl:value-of select= "$subName"     />
			subUsage    =<xsl:value-of select= "$subUsage"     />
  </xsl:if>

  <xsl:variable name="lengthError">
    <xsl:choose>
      <xsl:when test="$segName = $namedSegment
          and $fieldName = $namedField
          and $compName = $namedComponent
          and $subName = $namedSubComponent
          and $subCompLength = -0.1">
        <xsl:value-of select="'true'"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="'false'"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>
 
  <!-- When we look for a subcomponent value, in addition to being qualified
      by the field name, the subcomp also has to be qualified by the 
      component name to ensure it's distinct. -->
  <xsl:variable name="qualifier">
    <xsl:value-of select="$compName"/>
  </xsl:variable>
  <xsl:variable name="qualifier2">
    <xsl:value-of select="$fieldName"/>
  </xsl:variable>

  <xsl:variable name="sibChildCount">
    <xsl:value-of 
        select="count(preceding-sibling::SubComponent)"/>
  </xsl:variable>
  
  <xsl:variable name="indexSub">
    <xsl:value-of select="$index + $sibChildCount"/>
  </xsl:variable>
  
  <!-- sub component elements in the output msg are named based upon
      the parent component's dataType and a sequential count -->
  <xsl:if test="$includeSubcomponent = 'true'">
    <xsl:element name="{$compDataType}.{$subCount}">
      <xsl:choose>		<!-- if there's a constant value, it has to be used-->
        <xsl:when test="$constantValue!=''">
          <xsl:value-of select="$constantValue"/>
        </xsl:when>
        <xsl:otherwise>	<!-- no const value, so get a value -->
		      <xsl:call-template name="getValue">	
		        <xsl:with-param name="dataType" select="$subDataType"/>
		        <xsl:with-param name="elementName" select=     "$subName"/>
		        <xsl:with-param name="elementQualifier" select="$qualifier"/>
		        <xsl:with-param name="elementQualifier2" select="$qualifier2"/>
		        <xsl:with-param name="tableId" select=         "$tableId"/>
               <xsl:with-param name="useErrorCode" select=      "$useErrorCode"/>
		        <xsl:with-param name="reference"  select=      "$reference"/>
		        <xsl:with-param name="indexTable" select=      "$indexSub"/>
		        <xsl:with-param name="messageControlID" select="$messageControlID"/>
		        <xsl:with-param name="maxLength" select="$maxSubLength"/>
		        <xsl:with-param name="lengthError" select="$lengthError"/>
		        <xsl:with-param name="exValue" select="$exValue"/>
		      </xsl:call-template>
 	      </xsl:otherwise>
      </xsl:choose>
    </xsl:element>
  </xsl:if>
</xsl:template>

<!-- ............................................................. --> 
<xsl:template	name="getValue">
  <xsl:param name="dataType" select=        "''"/>
  <xsl:param name="elementName" select=     "''"/>
  <xsl:param name="elementQualifier" select="''"/>
  <xsl:param name="elementQualifier2" select="''"/>
  <xsl:param name="tableId" select=         "0"/>
  <xsl:param name="useErrorCode" select=      "'false'"/>
  <xsl:param name="reference"  select=      "''"/>
  <xsl:param name="indexTable" select=      "0"/>
  <xsl:param name="messageControlID" select=      "0"/>
  <xsl:param name="maxLength" select="0"/>
  <xsl:param name="lengthError" select="0"/>
  <xsl:param name="exValue" select="''"/>
            
  <xsl:variable name="primitive">	
    <xsl:call-template name="isPrimitive">
      <xsl:with-param name="dataType"	select="$dataType"/>
    </xsl:call-template>
  </xsl:variable>
  
  <!-- we only handle primitive and CE types -->
  <xsl:if test="$primitive='true' or $dataType='CE'">
    <xsl:choose>
      <!-- 1st see if it's the time of msg that we want -->	  
      <xsl:when test="$elementName='Date/Time'
                  and $elementQualifier='Date/Time Of Message'">
        <xsl:value-of select="$dateOfMessage"/>
      </xsl:when>
      <!-- if this is the MSH segment, we have to special 
          case the Message Control ID -->  
      <xsl:when test="$elementQualifier = 'MSH' 
          and $elementName = 'Message Control ID'">
      	<xsl:choose>
          <xsl:when test="$messageControlPrefix=''">
            <xsl:value-of select="$messageControlID"/>
          </xsl:when>
          <xsl:otherwise>	
            <xsl:value-of select="$messageControlPrefix"/>.<xsl:value-of 
                                          select="$messageControlID"/>
          </xsl:otherwise>
        </xsl:choose>
      </xsl:when>
      <xsl:otherwise>   <!-- it's not the Msg Cntrl id -->
        <!-- see if there's a configuration value for this element -->
        <xsl:variable name="configValue">
          <xsl:for-each select="document($configTable)">
            <xsl:call-template	name="getConfigValue">
              <xsl:with-param name="dataType" select="$dataType"/>
              <xsl:with-param name="elementName" select="$elementName"/>
              <xsl:with-param name="elementQualifier" 
                                select="$elementQualifier"/>
              <xsl:with-param name="tableId" select="$tableId"/>
              <xsl:with-param name="reference"  select="$reference"/> 
              <xsl:with-param name="indexTable" select="$indexTable"/>
            </xsl:call-template>
          </xsl:for-each>
        </xsl:variable>

        <xsl:choose>  <!-- config value or primitive or CE value -->
          <!-- we have to use the config value if we found one -->
          <xsl:when test="$configValue != ''">
            <xsl:value-of select="$configValue"/>
          </xsl:when>
          <!-- if there wasn't a config value, look for a value in one
              of our other tables/files -->
          <xsl:otherwise>
            <xsl:choose>
              <!-- primitive values have to be checked (depending on
                  what options are set) before they can be used.
                  if it's a CE value, we can directly grab the value,
                  since no further checking is done -->
              <xsl:when test="$primitive='true'">
                <xsl:call-template name="selectPrimitiveValue">
                  <xsl:with-param name="tableId" select="$tableId"/>              
                  <xsl:with-param name="useErrorCode" select="$useErrorCode"/>              
                  <xsl:with-param name="dataType" select="$dataType"/>
                  <xsl:with-param name="elementName" select="$elementName"/>
                  <xsl:with-param name="elementQualifier" 
                                          select="$elementQualifier"/>
                  <xsl:with-param name="elementQualifier2" 
                                          select="$elementQualifier2"/>
                  <xsl:with-param name="reference" select="$reference"/>
                  <xsl:with-param name="indexTable" select="$indexTable"/>
                  <xsl:with-param name="maxLength" select="$maxLength"/>
                  <xsl:with-param name="lengthError" select="$lengthError"/>
                  <xsl:with-param name="exValue" select="$exValue"/>
                </xsl:call-template>
              </xsl:when>
              <xsl:otherwise>
                <xsl:call-template name="getTableValue">
                  <xsl:with-param name="tableId" select="$tableId"/>              
                  <xsl:with-param name="useErrorCode" select="$useErrorCode"/>              
                  <xsl:with-param name="dataType" select="$dataType"/>
                  <xsl:with-param name="elementName" select="$elementName"/>
                  <xsl:with-param name="elementQualifier" 
                                          select="$elementQualifier"/>
                  <xsl:with-param name="elementQualifier2" 
                                          select="$elementQualifier2"/>
                  <xsl:with-param name="reference" select="$reference"/>
                  <xsl:with-param name="indexTable" select="$indexTable"/>
                </xsl:call-template>
              </xsl:otherwise>
            </xsl:choose>
          </xsl:otherwise>
        </xsl:choose>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:if>
</xsl:template>

<!-- ............................................................. -->
<!-- try to find a suitable data value. -->
<xsl:template name="selectPrimitiveValue"> 
  <xsl:param name="dataType"/>
  <xsl:param name="elementName"/>
  <xsl:param name="elementQualifier"/>
  <xsl:param name="elementQualifier2" select="''"/>
  <xsl:param name="tableId"/>
  <xsl:param name="useErrorCode" select=      "'false'"/>
  <xsl:param name="reference"/>
  <xsl:param name="maxLength" select="0"/>  
  <xsl:param name="indexTable" select="0"/>
  <xsl:param name="lengthError" select="'false'"/>
  <xsl:param name="loopCount" select="0"/>
  <xsl:param name="exValue" select="''"/>
    
  <!-- get a candidate value -->
  <xsl:variable name="candidateValue">
    <xsl:call-template name="getTableValue">		
      <xsl:with-param name="tableId" select="$tableId"/> 
      <xsl:with-param name="useErrorCode" select="$useErrorCode"/>                       
      <xsl:with-param name="dataType" select="$dataType"/>
      <xsl:with-param name="elementName" select="$elementName"/>
      <xsl:with-param name="elementQualifier" select="$elementQualifier"/>
      <xsl:with-param name="elementQualifier2" select="$elementQualifier2"/>
      <xsl:with-param name="reference" select="$reference"/>
      <xsl:with-param name="indexTable" select="$indexTable"/>
    </xsl:call-template>
  </xsl:variable>
   
  <!-- if we found a value, then: 
     if length constraints are being enforced, 
       if the value doesn't violate the length constraint, assign value.
       if it does, look for another value.
     (if we don't find a value after trying MAX_TRIES times, we give up.)             
     if we're not generating an error and length constraints are not
     being enforced,
        we can just use the value.

			if we didn't find a value, (for now at least)
      we'll use one of our dummy values.        
    -->
    <xsl:variable name="dataValue">
  <xsl:choose>
		<xsl:when test="$candidateValue!=''">	
	    <xsl:choose>
	      <!-- if we're not enforcing length constraints and there wasn't an 
	          error to generate, dump value -->
	      <xsl:when test="$lengthConstraints = 0">
	        <xsl:value-of select="$candidateValue"/>
	      </xsl:when>
	      <!-- enforcing length constraints -->
	      <xsl:otherwise> 
	        <xsl:choose>
	          <!-- check length -->
	          <xsl:when test="string-length($candidateValue) &gt; $maxLength">
	            <!-- we'll try looking for another value up to MAX_TRIES
	                times before we give up -->
	            <xsl:choose>	
	              <xsl:when test="$loopCount &lt; $MAX_TRIES">
	                <xsl:call-template name="selectPrimitiveValue">
	                  <xsl:with-param name="tableId" select="$tableId"/>          
	                  <xsl:with-param name="useErrorCode" select="$useErrorCode"/>              
	                  <xsl:with-param name="dataType" select="$dataType"/>
	                  <xsl:with-param name="elementName" select="$elementName"/>
	                  <xsl:with-param name="elementQualifier" select="$elementQualifier"/>
	                  <xsl:with-param name="reference" select="$reference"/>
	                  <xsl:with-param name="maxLength" select="$maxLength"/>
	                  <xsl:with-param name="indexTable" select="$indexTable + 1"/>
	                  <xsl:with-param name="lengthError" select="'false'"/>
	                  <xsl:with-param name="loopCount" select="$loopCount + 1"/>
	                </xsl:call-template>
	              </xsl:when>
	              <xsl:otherwise>
	                <xsl:if test="$WARN=1"> 
	                  <xsl:comment>
	                    **WARNING**
	                    Was not able to find value satisfying length constraint:
	                    dataType=         <xsl:value-of select="$dataType"/>
	                    elementName=      <xsl:value-of select="$elementName"/>
	                    elementQualifier= <xsl:value-of select="$elementQualifier"/>
	                    tableId=          <xsl:value-of select="$tableId"/>          
	                  </xsl:comment>
	                </xsl:if>
	                <xsl:value-of select="$candidateValue"/>
	              </xsl:otherwise>
	            </xsl:choose>
	          </xsl:when>
	          <!-- value length is okay -->
	          <xsl:otherwise>
	            <xsl:value-of select="$candidateValue"/>
	          </xsl:otherwise>
	        </xsl:choose>
	      </xsl:otherwise>
	    </xsl:choose>
	  </xsl:when>
	  <xsl:otherwise>		<!-- no value found-->
      <xsl:choose>		<!-- use example value if we have one -->
        <xsl:when test="$exValue!=''">
		      	<xsl:value-of select="$exValue"/>  
        </xsl:when>
        <xsl:otherwise>
        <!-- if no example value, use dummy value if they're being included  -->
		      <xsl:if test="$INCLUDE_DUMMY_VALS=1">
						<xsl:choose>
				      <xsl:when test="$dataType='DT'">
				      	<xsl:value-of select="$DT_DUMMY"/>  
				      </xsl:when>
				      <xsl:when test="$dataType='ID'">
				      	<xsl:value-of select="$ID_DUMMY"/>    
				      </xsl:when>
				      <xsl:when test="$dataType='IS'">
				      	<xsl:value-of select="$IS_DUMMY"/>    
				      </xsl:when>
				      <xsl:when test="$dataType='ST'">
				      	<xsl:value-of select="$ST_DUMMY"/>    
				      </xsl:when>
				      <xsl:when test="$dataType='NM'">
				        <xsl:value-of select="$NM_DUMMY"/>
				      </xsl:when> 
				      <xsl:when test="$dataType='SI'">
				        <xsl:value-of select="$SI_DUMMY"/>
				      </xsl:when> 
				      <xsl:when test="$dataType='TN'">
				        <xsl:value-of select="$TN_DUMMY"/>
				      </xsl:when>
				      <xsl:when test="$dataType='TM'">
				        <xsl:value-of select="$TM_DUMMY"/>
				      </xsl:when>
						</xsl:choose>
		      </xsl:if>
        </xsl:otherwise>
      </xsl:choose>
		</xsl:otherwise>      
  </xsl:choose>
    </xsl:variable>

  <!-- if we want a length error, geneate a value, otherwise use what we have. -->
  <xsl:choose>
    <xsl:when test="$lengthError = 'true'">
      <xsl:call-template name="generateLengthError">
        <xsl:with-param name="dataType" select="$dataType"/>
        <xsl:with-param name="maxLength" select="$maxLength"/>
        <xsl:with-param name="dataValue" select="$dataValue"/>
      </xsl:call-template>
    </xsl:when>
    <xsl:otherwise>
      <xsl:value-of select="$dataValue"/>
    </xsl:otherwise>
  </xsl:choose>  
  
</xsl:template>  

<!-- ............................................................. -->
<!-- Get a value from one of our table/files:
    config, hl7, local, user, external, and reference.  The tables are searched
    in the order listed based on name, qualifier, and possibly table id.
    Generally, if name & qualifier don't match, we look for a matching
    name before giving up and going to the next table.
    -->
<xsl:template name="getTableValue">
  <xsl:param name="dataType"/>
  <xsl:param name="elementName"/>
  <xsl:param name="elementQualifier"/>
  <xsl:param name="elementQualifier2" select="''"/>
  <xsl:param name="tableId"/>
  <xsl:param name="useErrorCode" select=      "'false'"/>
  <xsl:param name="reference"/>
  <xsl:param name="indexTable" select="0"/>

  <xsl:if test="$DEBUG=1 or $DEBUG &gt; 9">
    getTableValue:
      dataType = <xsl:value-of select="$dataType"/>
      elementName = <xsl:value-of select="$elementName"/>
      table id = <xsl:value-of select="$tableId"/>
      reference = <xsl:value-of select="$reference"/>  
      indexTable = <xsl:value-of select="$indexTable"/>
    ................................................
  </xsl:if>
  
  <xsl:if test="$DEBUG=3 or $DEBUG &gt; 9">	
    <xsl:comment>
      TAB_INDX = <xsl:value-of select="$indexTable"/>..
    </xsl:comment>	
  </xsl:if>

  <xsl:variable name="primitive">	
    <xsl:call-template name="isPrimitive">
      <xsl:with-param name="dataType"	select="$dataType"/>
    </xsl:call-template>
  </xsl:variable>

    <!-- 1st see if we want an error code -->
    <xsl:variable name="errorCodeValue">
      <xsl:choose>
      <xsl:when test="$useErrorCode='true'">
        <xsl:for-each select="document($errorCodesTable)"> 
<!-- <xsl:for-each select="document($hl7Tables)">
-->
          <xsl:call-template	name="getErrorTableValue">
            <xsl:with-param name="tableId" select="$tableId"/>
          </xsl:call-template>
        </xsl:for-each>
      </xsl:when>
        <xsl:otherwise>
          <xsl:value-of select="''"/>
        </xsl:otherwise>
      </xsl:choose>
    </xsl:variable>
  
    <!-- if we're not looking for an error, check hl7 tables 1st -->
    <xsl:variable name="hl7TableValue">
      <xsl:choose>
        <xsl:when test="$errorCodeValue=''">
          <xsl:for-each select="document($hl7Tables)">
            <xsl:call-template	name="getHL7TableValue">
              <xsl:with-param name="dataType" select="$dataType"/>
              <xsl:with-param name="elementName" select="$elementName"/>
              <xsl:with-param name="elementQualifier" select="$elementQualifier"/>
              <xsl:with-param name="tableId" select="$tableId"/>
              <xsl:with-param name="reference"  select="$reference"/> 
              <xsl:with-param name="indexTable" select="$indexTable"/>
            </xsl:call-template>
          </xsl:for-each>
        </xsl:when>
        <xsl:otherwise>
          <xsl:value-of select="''"/>
        </xsl:otherwise>
      </xsl:choose>
    </xsl:variable>

    <!-- if we haven't found a value yet, check the local tables -->
    <xsl:variable name="localTableValue">
      <xsl:choose>
        <xsl:when test="$hl7TableValue=''">
          <xsl:for-each select="document($localTables)">
            <xsl:call-template	name="getLocalTableValue">
              <xsl:with-param name="dataType" select="$dataType"/>
              <xsl:with-param name="elementName" select="$elementName"/>
              <xsl:with-param name="elementQualifier" select="$elementQualifier"/>
              <xsl:with-param name="tableId" select="$tableId"/>
              <xsl:with-param name="reference"  select="$reference"/> 
              <xsl:with-param name="indexTable" select="$indexTable"/>
            </xsl:call-template>
          </xsl:for-each>
        </xsl:when>
        <xsl:otherwise>
          <xsl:value-of select="''"/>	
        </xsl:otherwise>
      </xsl:choose>
    </xsl:variable>

    <!-- if we haven't found a value yet, check the user tables -->
    <xsl:variable name="userTableValue">
      <xsl:choose>
        <xsl:when test="$localTableValue=''">
          <xsl:for-each select="document($userTables)">
            <xsl:call-template	name="getUserTableValue">
              <xsl:with-param name="dataType" select="$dataType"/>
              <xsl:with-param name="elementName" select="$elementName"/>
              <xsl:with-param name="elementQualifier" select="$elementQualifier"/>
              <xsl:with-param name="tableId" select="$tableId"/>
              <xsl:with-param name="reference"  select="$reference"/> 
              <xsl:with-param name="indexTable" select="$indexTable"/>
            </xsl:call-template>
          </xsl:for-each>
        </xsl:when>
        <xsl:otherwise>
          <xsl:value-of select="''"/>	
        </xsl:otherwise>
      </xsl:choose>
    </xsl:variable>

    <!-- if we're still looking, try the external tables -->
    <xsl:variable name="externalTableValue">
      <xsl:choose>
        <xsl:when test="$userTableValue=''">
          <xsl:for-each select="document($externalTables)">
            <xsl:call-template	name="getExternalTableValue">
              <xsl:with-param name="dataType" select="$dataType"/>
              <xsl:with-param name="elementName" select="$elementName"/>
              <xsl:with-param name="elementQualifier" select="$elementQualifier"/>
              <xsl:with-param name="tableId" select="$tableId"/>
              <xsl:with-param name="reference"  select="$reference"/> 
              <xsl:with-param name="indexTable" select="$indexTable"/>
            </xsl:call-template>
          </xsl:for-each>
        </xsl:when>
        <xsl:otherwise>
          <xsl:value-of select="''"/>	
        </xsl:otherwise>
      </xsl:choose>
    </xsl:variable>

    <!-- finally, try our reference data (this is the only table that uses a 2nd qualifier -->
    <xsl:variable name="refValue">
      <xsl:choose>
        <xsl:when test="$externalTableValue=''">
          <xsl:for-each select="document($referenceTable)">
            <xsl:call-template	name="getRefValue">
              <xsl:with-param name="dataType" select="$dataType"/>
              <xsl:with-param name="elementName" select="$elementName"/>
              <xsl:with-param name="elementQualifier" select="$elementQualifier"/>
              <xsl:with-param name="elementQualifier2" select="$elementQualifier2"/>
              <xsl:with-param name="tableId" select="$tableId"/>
              <xsl:with-param name="reference"  select="$reference"/> 
              <xsl:with-param name="indexTable" select="$indexTable"/>
            </xsl:call-template>
          </xsl:for-each>
        </xsl:when>
        <xsl:otherwise>
          <xsl:value-of select="''"/>	
        </xsl:otherwise>
      </xsl:choose>
    </xsl:variable>

    <xsl:choose>
      <xsl:when test="$errorCodeValue != ''">
        <xsl:value-of select="$errorCodeValue"/>
      </xsl:when>
      <xsl:when test="$hl7TableValue != ''">
        <xsl:value-of select="$hl7TableValue"/>
      </xsl:when>
      <xsl:when test="$userTableValue != ''">
        <xsl:value-of select="$userTableValue"/>
      </xsl:when>
      <xsl:when test="$localTableValue != ''">
        <xsl:value-of select="$localTableValue"/>
      </xsl:when>
      <xsl:when test="$externalTableValue != ''">
        <xsl:value-of select="$externalTableValue"/>
      </xsl:when>
      <xsl:when test="$refValue != ''">
        <xsl:value-of select="$refValue"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="''"/>
      </xsl:otherwise>
    </xsl:choose>
</xsl:template>

<xsl:template	name="getErrorTableValue">
  <xsl:param name="tableId"/>

  <xsl:value-of select="key('errorCodesKey', $tableId)/errorCodes:Item[1]
      /@Code"/>
</xsl:template>    
    
<!-- ............................................................. -->
<!-- look for a value in the config table to put in the output stream  -->
<xsl:template name="getConfigValue">
  <xsl:param name="dataType"/>
  <xsl:param name="elementName"/>
  <xsl:param name="elementQualifier"/>
  <xsl:param name="tableId"/>
  <xsl:param name="reference"/>  
  <xsl:param name="indexTable" select="0"/>

  <xsl:choose>
    <!-- first check Config table...
        see if both name & qualifier match -->
    <xsl:when test="key('configTableKey', $elementName)
        [@Qualifier=$elementQualifier]">
      <xsl:value-of select="key('configTableKey', $elementName)
          [@Qualifier=$elementQualifier]/@Value"/>
    </xsl:when>
    <!-- if both didn't match and the Qualifier's empty or missing,
          then see if name matches. --> 
    <xsl:when test="key('configTableKey', $elementName)
        [@Qualifier='' or not(@Qualifier)]">
      <xsl:value-of select="key('configTableKey', $elementName)/@Value"/>
    </xsl:when>
  </xsl:choose>
</xsl:template>

<!-- ............................................................. -->
<!-- look for a value in an hl7 table to put in the output stream.  
      The hl7 data file is an xml file made up of a set of Table 
      elements.  Each Table element has a Name and ID attribute that is
      unique. The Table we select is determined by the ID, but we are 
      ignoring the Name.  Each Table element is made up of a set of Item 
      elements with a Code and Value attribute.  We select an Item based 
      on the index parameter and, for primitive types, return a value from 
      the Code or Value attribute.
-->
<xsl:template name="getHL7TableValue">
  <xsl:param name="dataType"/>
  <xsl:param name="elementName"/>
  <xsl:param name="elementQualifier"/>
  <xsl:param name="tableId"/>
  <xsl:param name="useErrorCode" select=      "'false'"/>
  <xsl:param name="reference"/>  
  <xsl:param name="indexTable" select="0"/>

  <!-- first make sure there's a table id -->
  <xsl:if test="boolean($tableId)">
  
    <xsl:variable name="primitive">	
      <xsl:call-template name="isPrimitive">
        <xsl:with-param name="dataType"	select="$dataType"/>
      </xsl:call-template>
    </xsl:variable>
  
    <xsl:variable name="tCount">
      <xsl:value-of select="count(key('hl7TableKey', $tableId)
          /HL7Tables:Item)"/>
    </xsl:variable>
    
    <xsl:choose>
      <xsl:when test="$primitive='true'">  <!-- we only handle primitive and CE types -->	
        <!-- grap the Code attr if it exists, otherwise Value -->
        <xsl:choose>
          <xsl:when test="key('hl7TableKey', $tableId)/HL7Tables:Item/@Code">
            <xsl:value-of select="key('hl7TableKey', $tableId)/HL7Tables:Item
                                  [($indexTable mod $tCount) + 1]
                                  /@Code"/>
          </xsl:when>
          <xsl:otherwise>
            <xsl:value-of select="key('hl7TableKey', $tableId)/HL7Tables:Item
                                  [($indexTable mod $tCount) + 1]
                                  /@Value"/>
          </xsl:otherwise>
        </xsl:choose>
      </xsl:when>
      <!-- for now at least, we're only maintaining the Identifier and 
          the Text components of a CE value in our tables.  The value of 
          the Code attribute holds the Identifier and the value of the
          Value attribute holds the Text (the problem of defining a 
          a representation that supports maintaining all CE components
          will be addressed later).  here, we'll use a representation for
          CE values here that accounts for all components anyway. 
          -->
      <xsl:when test="$dataType='CE'">
        <xsl:variable name="ID">	
          <xsl:value-of select="key('hl7TableKey', $tableId)/HL7Tables:Item
                                [($indexTable mod $tCount) + 1]
                                /@Code"/>
        </xsl:variable>
        <xsl:variable name="Text">	
          <xsl:value-of select="key('hl7TableKey', $tableId)/HL7Tables:Item
                                [($indexTable mod $tCount) + 1]
                                /@Value"/>
        </xsl:variable>
        <!-- we'll just hard-code the coding sys for now, it's the only
            one we support right now -->
        <xsl:variable name="CodingSystem">	
          <xsl:value-of select="'HL7'"/><xsl:value-of select="$tableId"/>
        </xsl:variable>
        <xsl:variable name="AiD">	
          <xsl:value-of select="''"/>
        </xsl:variable>
        <xsl:variable name="AText">	
          <xsl:value-of select="''"/>        
        </xsl:variable>
        <xsl:variable name="ACodingSystem">
          <xsl:value-of select="''"/>        
        </xsl:variable>

        <!-- return the CE value -->
        ID=<xsl:value-of select="$ID"/>
        Text=<xsl:value-of select="$Text"/>
        CS=<xsl:value-of select="$CodingSystem"/>
        AiD=<xsl:value-of select="$AiD"/>
        AText=<xsl:value-of select="$AText"/>
        ACS=<xsl:value-of select="$ACodingSystem"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="''"/>	
      </xsl:otherwise>
    </xsl:choose>
  </xsl:if>
</xsl:template>

<!-- ............................................................. -->                                      
<xsl:template name="getLocalTableValue">
  <xsl:param name="dataType"/>
  <xsl:param name="elementName"/>
  <xsl:param name="elementQualifier"/>
  <xsl:param name="tableId"/>
  <xsl:param name="useErrorCode" select=      "'false'"/>
  <xsl:param name="reference"/>  
  <xsl:param name="indexTable" select="0"/>

  <!-- first make sure there's a table id -->
  <xsl:if test="boolean($tableId)">
  
    <xsl:variable name="primitive">	
      <xsl:call-template name="isPrimitive">
        <xsl:with-param name="dataType"	select="$dataType"/>
      </xsl:call-template>
    </xsl:variable>
  
    <xsl:variable name="tCount">
      <xsl:value-of select="count(key('localTableKey', $tableId)
          /local:Item)"/>
    </xsl:variable>
    
    <xsl:choose>  <!-- we only handle primitive and CE types -->
      <xsl:when test="$primitive='true'">	
        <!-- grap the Code attr if it exists, otherwise Value -->
        <xsl:choose>
          <xsl:when test="key('localTableKey', $tableId)/local:Item/@Code">
            <xsl:value-of select="key('localTableKey', $tableId)/local:Item
                                  [($indexTable mod $tCount) + 1]
                                  /@Code"/>
          </xsl:when>
          <xsl:otherwise>
            <xsl:value-of select="key('localTableKey', $tableId)/local:Item
                                  [($indexTable mod $tCount) + 1]
                                  /@Value"/>
          </xsl:otherwise>
        </xsl:choose>
      </xsl:when>
      <!-- for now at least, we're only maintaining the Identifier and 
          the Text components of a CE value in our tables.  The value of 
          the Code attribute holds the Identifier and the value of the
          Value attribute holds the Text (the problem of defining a 
          a representation that supports maintaining all CE components
          will be addressed later).  here, we'll use a representation for
          CE values here that accounts for all components anyway. 
          -->
      <xsl:when test="$dataType='CE'">
        <xsl:variable name="ID">	
          <xsl:value-of select="key('localTableKey', $tableId)/local:Item
                                [($indexTable mod $tCount) + 1]
                                /@Code"/>
        </xsl:variable>
        <xsl:variable name="Text">	
          <xsl:value-of select="key('localTableKey', $tableId)/local:Item
                                [($indexTable mod $tCount) + 1]
                                /@Value"/>
        </xsl:variable>
        <!-- we'll just hard-code the coding sys for now, it's the only
            one we support right now -->
        <xsl:variable name="CodingSystem">	
          <xsl:value-of select="'HL7'"/><xsl:value-of select="$tableId"/>
        </xsl:variable>
        <xsl:variable name="AiD">	
          <xsl:value-of select="''"/>
        </xsl:variable>
        <xsl:variable name="AText">	
          <xsl:value-of select="''"/>        
        </xsl:variable>
        <xsl:variable name="ACodingSystem">
          <xsl:value-of select="''"/>        
        </xsl:variable>

        <!-- return the CE value -->
        ID=<xsl:value-of select="$ID"/>
        Text=<xsl:value-of select="$Text"/>
        CS=<xsl:value-of select="$CodingSystem"/>
        AiD=<xsl:value-of select="$AiD"/>
        AText=<xsl:value-of select="$AText"/>
        ACS=<xsl:value-of select="$ACodingSystem"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="''"/>	
      </xsl:otherwise>
    </xsl:choose>
  </xsl:if>
</xsl:template>
                                      
<!-- ............................................................. -->
<xsl:template name="getUserTableValue">
  <xsl:param name="dataType"/>
  <xsl:param name="elementName"/>
  <xsl:param name="elementQualifier"/>
  <xsl:param name="tableId"/>
  <xsl:param name="useErrorCode" select=      "'false'"/>
  <xsl:param name="reference"/>  
  <xsl:param name="indexTable" select="0"/>

  <!-- first make sure there's a table id -->
  <xsl:if test="boolean($tableId)"> 
    <xsl:variable name="primitive">	
      <xsl:call-template name="isPrimitive">
        <xsl:with-param name="dataType"	select="$dataType"/>
      </xsl:call-template>
    </xsl:variable>
  
    <xsl:variable name="tCount">
      <xsl:value-of select="count(key('userTableKey', $tableId)
          /usrTables:Item)"/>
    </xsl:variable>
    
    <xsl:choose>  <!-- we only handle primitive and CE types -->
      <xsl:when test="$primitive='true'">	
        <!-- grap the Code attr if it exists, otherwise Value -->
        <xsl:choose>
          <xsl:when test="key('userTableKey', $tableId)/usrTables:Item/@Code">
            <xsl:value-of select="key('userTableKey', $tableId)/usrTables:Item
                                  [($indexTable mod $tCount) + 1]
                                  /@Code"/>
          </xsl:when>
          <xsl:otherwise>
            <xsl:value-of select="key('userTableKey', $tableId)/usrTables:Item
                                  [($indexTable mod $tCount) + 1]
                                  /@Value"/>
          </xsl:otherwise>
        </xsl:choose>
      </xsl:when>
      <!-- for now at least, we're only maintaining the Identifier and 
          the Text components of a CE value in our tables.  The value of 
          the Code attribute holds the Identifier and the value of the
          Value attribute holds the Text (the problem of defining a 
          a representation that supports maintaining all CE components
          will be addressed later).  here, we'll use a representation for
          CE values here that accounts for all components anyway. 
          -->
      <xsl:when test="$dataType='CE'">
        <xsl:variable name="ID">	
          <xsl:value-of select="key('userTableKey', $tableId)/usrTables:Item
                                [($indexTable mod $tCount) + 1]
                                /@Code"/>
        </xsl:variable>
        <xsl:variable name="Text">	
          <xsl:value-of select="key('userTableKey', $tableId)/usrTables:Item
                                [($indexTable mod $tCount) + 1]
                                /@Value"/>
        </xsl:variable>
        <!-- we'll just hard-code the coding sys for now, it's the only
            one we support right now -->
        <xsl:variable name="CodingSystem">	
          <xsl:value-of select="'HL7'"/><xsl:value-of select="$tableId"/>
        </xsl:variable>
        <xsl:variable name="AiD">	
          <xsl:value-of select="''"/>
        </xsl:variable>
        <xsl:variable name="AText">	
          <xsl:value-of select="''"/>        
        </xsl:variable>
        <xsl:variable name="ACodingSystem">
          <xsl:value-of select="''"/>        
        </xsl:variable>
        
        <!-- return the CE value -->
        ID=<xsl:value-of select="$ID"/>
        Text=<xsl:value-of select="$Text"/>
        CS=<xsl:value-of select="$CodingSystem"/>
        AiD=<xsl:value-of select="$AiD"/>
        AText=<xsl:value-of select="$AText"/>
        ACS=<xsl:value-of select="$ACodingSystem"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="''"/>	
      </xsl:otherwise>
    </xsl:choose>
  </xsl:if>
</xsl:template>
                                      
<!-- ............................................................. -->
<!-- look for a value in the external tables to put in the output stream  -->
<xsl:template name="getExternalTableValue">
  <xsl:param name="dataType"/>
  <xsl:param name="elementName"/>
  <xsl:param name="elementQualifier"/>
  <xsl:param name="tableId"/>
  <xsl:param name="reference"/>  
  <xsl:param name="indexTable" select="0"/>

  <!-- first make sure there's a table id -->
  <xsl:if test="boolean($tableId)">
  
    <xsl:variable name="primitive">	
      <xsl:call-template name="isPrimitive">
        <xsl:with-param name="dataType"	select="$dataType"/>
      </xsl:call-template>
    </xsl:variable>
  
    <xsl:variable name="tCount">
      <xsl:value-of select="count(key('externalTableKey', $tableId)
          /extTables:Item)"/>
    </xsl:variable>
    
    <xsl:choose>  <!-- we only handle primitive and CE types -->
      <xsl:when test="$primitive='true'">	
        <!-- grap the Code attr if it exists, otherwise Value -->
        <xsl:choose>
          <xsl:when test="key('externalTableKey', $tableId)/extTables:Item/@Code">
            <xsl:value-of select="key('externalTableKey', $tableId)/extTables:Item
                                  [($indexTable mod $tCount) + 1]
                                  /@Code"/>
          </xsl:when>
          <xsl:otherwise>
            <xsl:value-of select="key('externalTableKey', $tableId)/extTables:Item
                                  [($indexTable mod $tCount) + 1]
                                  /@Value"/>
          </xsl:otherwise>
        </xsl:choose>
      </xsl:when>
      <!-- for now at least, we're only maintaining the Identifier and 
          the Text components of a CE value in our tables.  The value of 
          the Code attribute holds the Identifier and the value of the
          Value attribute holds the Text (the problem of defining a 
          a representation that supports maintaining all CE components
          will be addressed later).  here, we'll use a representation for
          CE values here that accounts for all components anyway. 
          -->
      <xsl:when test="$dataType='CE'">
        <xsl:variable name="ID">	
          <xsl:value-of select="key('externalTableKey', $tableId)/extTables:Item
                                [($indexTable mod $tCount) + 1]
                                /@Code"/>
        </xsl:variable>
        <xsl:variable name="Text">	
          <xsl:value-of select="key('externalTableKey', $tableId)/extTables:Item
                                [($indexTable mod $tCount) + 1]
                                /@Value"/>
        </xsl:variable>
        <!-- we'll just hard-code the coding sys for now, it's the only
            one we support right now -->
        <xsl:variable name="CodingSystem">	
          <xsl:value-of select="'HL7'"/><xsl:value-of select="$tableId"/>
        </xsl:variable>
        <xsl:variable name="AiD">	
          <xsl:value-of select="''"/>
        </xsl:variable>
        <xsl:variable name="AText">	
          <xsl:value-of select="''"/>        
        </xsl:variable>
        <xsl:variable name="ACodingSystem">
          <xsl:value-of select="''"/>        
        </xsl:variable>
        
        <!-- return the CE value -->
        ID=<xsl:value-of select="$ID"/>
        Text=<xsl:value-of select="$Text"/>
        CS=<xsl:value-of select="$CodingSystem"/>
        AiD=<xsl:value-of select="$AiD"/>
        AText=<xsl:value-of select="$AText"/>
        ACS=<xsl:value-of select="$ACodingSystem"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="''"/>	
      </xsl:otherwise>
    </xsl:choose>
  </xsl:if>
</xsl:template>
                                      
<!-- ............................................................. -->
<!-- put a value from the reference data table into the output stream  -->
<xsl:template name="getRefValue">
  <xsl:param name="dataType"/>
  <xsl:param name="elementName"/>
  <xsl:param name="elementQualifier"/>
  <xsl:param name="elementQualifier2" select="''"/>
  <xsl:param name="tableId"/>
  <xsl:param name="reference"/>  
  <xsl:param name="indexTable" select="0"/>
 
  <xsl:choose>
    <!-- if the 2nd qualifier is present in the db, then it has to match -->
    <xsl:when test="key($refDataKey, $elementName)[@Qualifier2]">
      <xsl:if test="key($refDataKey, $elementName)
          [@Qualifier=$elementQualifier and @Qualifier2=$elementQualifier2]">
        <xsl:variable name="iCount">
          <xsl:value-of select=
          "count(key($refDataKey, $elementName)
            [@Qualifier=$elementQualifier and @Qualifier2=$elementQualifier2]
              /refData:Item)"/>
        </xsl:variable>
        <xsl:value-of select="key($refDataKey, $elementName)
          [@Qualifier=$elementQualifier and @Qualifier2=$elementQualifier2]
              /refData:Item[($indexTable mod $iCount) + 1]
              /@Value"/>
      </xsl:if>
    </xsl:when>
    <!-- only match one qualifier if 2nd wasn't present -->
    <xsl:when test="key($refDataKey, $elementName)[@Qualifier]">
      <xsl:if test="key($refDataKey, $elementName)
          [@Qualifier=$elementQualifier]">
        <xsl:variable name="iCount">
          <xsl:value-of select=
          "count(key($refDataKey, $elementName)
              [@Qualifier=$elementQualifier]
              /refData:Item)"/>
        </xsl:variable>
        <xsl:value-of select="key($refDataKey, $elementName)
              [@Qualifier=$elementQualifier]
              /refData:Item[($indexTable mod $iCount) + 1]
              /@Value"/>
      </xsl:if>
    </xsl:when>
    <xsl:otherwise>  <!-- only name has to match -->
      <xsl:variable name="iCount">
        <xsl:value-of select=
        "count(key($refDataKey, $elementName)/refData:Item)"/>
      </xsl:variable>
      <xsl:value-of select="key($refDataKey, $elementName)
          /refData:Item[($indexTable mod $iCount) + 1]/@Value"/>
    </xsl:otherwise>
  </xsl:choose>
</xsl:template>

<!-- ............................................................. -->
<xsl:template name="buildCEelement">
  <xsl:param name="ceValue"	select="''"/>
  <xsl:param name="elementName" select="''"/>
  <xsl:param name="idUsage" select=   "''"/>
  <xsl:param name="textUsage" select= "''"/>
  <xsl:param name="csUsage" select=   "''"/>
  <xsl:param name="aidUsage" select=  "''"/>
  <xsl:param name="atextUsage" select="''"/>
  <xsl:param name="acsUsage" select=  "''"/>

  <xsl:variable name="idPlus">
    <xsl:value-of select="substring-after($ceValue, 'ID=')"/>
  </xsl:variable>                                   
  <xsl:variable name="textPlus">                 
    <xsl:value-of select="substring-after($ceValue, 'Text=')"/>
  </xsl:variable>                                   
  <xsl:variable name="CSPlus">                   
    <xsl:value-of select="substring-after($ceValue, 'CS=')"/>
  </xsl:variable>                                   
  <xsl:variable name="aIdPlus">                  
    <xsl:value-of select="substring-after($ceValue, 'AiD=')"/>
  </xsl:variable>                                   
  <xsl:variable name="aTextPlus">                
    <xsl:value-of select="substring-after($ceValue, 'AText=')"/>
  </xsl:variable>                                   
                                           
  <xsl:variable name="ACS">                         
    <xsl:value-of select="substring-after($ceValue, 'ACS=')"/>
  </xsl:variable>

  <xsl:variable name="ID">
    <xsl:value-of select="substring-before($idPlus, 'Text=')"/>
  </xsl:variable>
  <xsl:variable name="text">
    <xsl:value-of select="substring-before($textPlus, 'CS=')"/>
  </xsl:variable>
  <xsl:variable name="CS">
    <xsl:value-of select="substring-before($CSPlus, 'AiD=')"/>
  </xsl:variable>
  <xsl:variable name="AiD">
    <xsl:value-of select="substring-before($aIdPlus, 'AText=')"/>
  </xsl:variable>
  <xsl:variable name="aText">
    <xsl:value-of select="substring-before($aTextPlus, 'ACS=')"/>
  </xsl:variable>
  
  <xsl:variable name="seqNumber" select="1"/>

	<!-- if any elements are empty, we'll use these values -->
	<xsl:variable name="ID_DUMMY">
    <xsl:value-of select="'MM_DUMMY_ID'"/>
	</xsl:variable>
	<xsl:variable name="TXT_DUMMY">
    <xsl:value-of select="'MM_DUMMY_TEXT'"/>
	</xsl:variable>
	<xsl:variable name="CS_DUMMY">
    <xsl:value-of select="'MM_DUMMY_CS'"/>
	</xsl:variable>
	<xsl:variable name="AID_DUMMY">
    <xsl:value-of select="'MM_DUMMY_AID'"/>
	</xsl:variable>
	<xsl:variable name="ATXT_DUMMY">
    <xsl:value-of select="'MM_DUMMY_ATEXT'"/>
	</xsl:variable>
	<xsl:variable name="ACS_DUMMY">
    <xsl:value-of select="'MM_DUMMY_ACS'"/>
	</xsl:variable>

  <!-- we'll include all elements whose usage isn't excluded -->
  <xsl:if test="$idUsage!='X'">	
    <xsl:element name="{$elementName}.{$seqNumber}">
      <xsl:choose>
        <xsl:when test="$ID!=''">
          <xsl:value-of select="normalize-space($ID)"/>
        </xsl:when>
        <xsl:otherwise>
          <xsl:if test="$INCLUDE_DUMMY_VALS">
	          <xsl:value-of select="$ID_DUMMY"/>
          </xsl:if>
        </xsl:otherwise>
      </xsl:choose>
    </xsl:element>
  </xsl:if>
  <xsl:if test="$textUsage!='X'">
    <xsl:element name="{$elementName}.{$seqNumber+1}">
      <xsl:choose>
        <xsl:when test="$text!=''">
          <xsl:value-of select="normalize-space($text)"/>
        </xsl:when>
        <xsl:otherwise>
          <xsl:if test="$INCLUDE_DUMMY_VALS">
	          <xsl:value-of select="$TXT_DUMMY"/>
          </xsl:if>
        </xsl:otherwise>
      </xsl:choose>
    </xsl:element>
   </xsl:if>
   <xsl:if test="$csUsage!='X'">
    <xsl:element name="{$elementName}.{$seqNumber+2}">
      <xsl:choose>
        <xsl:when test="$CS!=''">
          <xsl:value-of select="normalize-space($CS)"/>
        </xsl:when>
        <xsl:otherwise>
          <xsl:if test="$INCLUDE_DUMMY_VALS">          
          	<xsl:value-of select="$CS_DUMMY"/>
          </xsl:if>
        </xsl:otherwise>
      </xsl:choose>
    </xsl:element>
  </xsl:if>
  <xsl:if test="$aidUsage!='X'">
    <xsl:element name="{$elementName}.{$seqNumber+3}">
      <xsl:choose>
        <xsl:when test="$AiD!=''">
          <xsl:value-of select="normalize-space($AiD)"/>
        </xsl:when>
        <xsl:otherwise>
          <xsl:if test="$INCLUDE_DUMMY_VALS">
	          <xsl:value-of select="$AID_DUMMY"/>
          </xsl:if>
        </xsl:otherwise>
      </xsl:choose>
    </xsl:element>
  </xsl:if>
  <xsl:if test="$atextUsage!='X'">
    <xsl:element name="{$elementName}.{$seqNumber+4}">
      <xsl:choose>
        <xsl:when test="$aText!=''">
          <xsl:value-of select="normalize-space($aText)"/>
        </xsl:when>
        <xsl:otherwise>
          <xsl:if test="$INCLUDE_DUMMY_VALS">
          	<xsl:value-of select="$ATXT_DUMMY"/>
          </xsl:if>
        </xsl:otherwise>
      </xsl:choose>
    </xsl:element>
  </xsl:if>
  <xsl:if test="$acsUsage!='X'">
    <xsl:element name="{$elementName}.{$seqNumber+5}">
      <xsl:choose>
        <xsl:when test="$ACS!=''">
          <xsl:value-of select="normalize-space($ACS)"/>
        </xsl:when>
        <xsl:otherwise>
          <xsl:if test="$INCLUDE_DUMMY_VALS">
          	<xsl:value-of select="$ACS_DUMMY"/>
					</xsl:if>
        </xsl:otherwise>
      </xsl:choose>
    </xsl:element>
  </xsl:if>
</xsl:template>

<!-- ............................................................. -->
<!-- check to see if element should be included in output based on
    the element's Usage setting  -->
<xsl:template	name="includeElement">
  <xsl:param name="usage" />
  <xsl:param name="nodeNumber"/>
  <xsl:param name="segmentName" select="''"/>
  <xsl:param name="fieldName" select="''"/>
  <xsl:param name="componentName" select="''"/>
  <xsl:param name="subComponentName" select="''"/>
  <xsl:param name="rUErrorLocation"/>
  <xsl:param name="xUErrorLocation"/>

 <xsl:choose>
    <!-- all R-Usage elements are included unless an error is being generated -->
    <xsl:when test="$usage = 'R'">
      <xsl:choose>
        <xsl:when test="$RusageError &lt; 0"> <!-- error -->
          <xsl:choose>	
            <!-- if this is the designated err location (name or 
              number match), don't include it -->
            <xsl:when test="$nodeNumber = $rUErrorLocation">
              <xsl:value-of select="'false'" />
            </xsl:when>
            <xsl:when test="$segmentName = $namedSegment
              		and $fieldName = $namedField
                  and $componentName = $namedComponent
                  and $subComponentName = $namedSubComponent">
              <xsl:value-of select="'false'" />
            </xsl:when>
            <xsl:otherwise> <!-- include element -->	  
              <xsl:value-of select="'true'" />
            </xsl:otherwise>
          </xsl:choose>
        </xsl:when>
        <xsl:otherwise> <!-- valid -->
          <xsl:value-of select="'true'" />
        </xsl:otherwise>
      </xsl:choose>
    </xsl:when>   <!-- end usage=R -->
    <!-- no X-Usage elements are included unless an error is being generated -->    
    <xsl:when test="$usage = 'X'">
      <xsl:choose>	
        <xsl:when test="$XusageError &lt; 0">
          <xsl:choose>	
            <!-- if this is the designated err location (name or 
              	number match), include it -->
            <xsl:when test="$nodeNumber = $xUErrorLocation">
              <xsl:value-of select="'true'" />
            </xsl:when>
            <xsl:when test="$segmentName = $namedSegment
              		and $fieldName = $namedField
                  and $componentName = $namedComponent
                  and $subComponentName = $namedSubComponent">
              <xsl:value-of select="'true'" />
            </xsl:when>
            <xsl:otherwise> <!-- don't include element -->	  
              <xsl:value-of select="'false'" />
            </xsl:otherwise>
          </xsl:choose>
        </xsl:when>
        <xsl:otherwise>	
          <xsl:value-of select="'false'" />
        </xsl:otherwise>
      </xsl:choose>
    </xsl:when>   <!-- end usage=X -->
    <xsl:when test="$usage = 'RE'">
      <xsl:choose>
        <!-- if elementREinclude isn't empty and this is the designated
          	element, then its value determines if the element is
            included or omitted -->	
        <xsl:when test="$elementREinclude != ''
          			and $segmentName = $namedSegment
              	and $fieldName = $namedField
                and $componentName = $namedComponent
                and $subComponentName = $namedSubComponent">
          <xsl:value-of select="$elementREinclude" />                          
        </xsl:when>
        <xsl:when test="$RE_Usage = 1">   <!-- include all -->
          <xsl:value-of select="'true'" />          
        </xsl:when>
        <xsl:when test="$RE_Usage = 0">   <!-- include none -->
          <xsl:value-of select="'false'" />          
        </xsl:when>
        <xsl:when test="$RE_Usage = 0.5"> <!-- include 1/5 to 1/2 -->
          <xsl:variable name="modNum">
            <xsl:value-of select="($offset mod 4) + 2"/>
          </xsl:variable>
          <xsl:choose>
            <xsl:when test="((position() + ($offset mod 6)) mod $modNum) = 0">              
		          <xsl:value-of select="'true'" />
            </xsl:when>
            <xsl:otherwise>	
              <xsl:value-of select="'false'" />
            </xsl:otherwise>
          </xsl:choose>
        </xsl:when>
        <xsl:otherwise>   <!-- should catch all cases above, but ... -->	
          <xsl:value-of select="'false'" />
        </xsl:otherwise>
      </xsl:choose>
    </xsl:when>   <!-- end usage=RE -->
    <xsl:otherwise>   <!-- not including C/CE usage segs right now -->
      <xsl:value-of select="'false'" />
    </xsl:otherwise>
  </xsl:choose>
</xsl:template>

<!-- ............................................................. -->
<xsl:template	name="matchesElementName">
  <xsl:param name="segmentName" select="''"/>
  <xsl:param name="fieldName" select="''"/>
  <xsl:param name="componentName" select="''"/>
  <xsl:param name="subComponentName" select="''"/>
  
  <xsl:choose>
    <xsl:when test="$segmentName = $namedSegment
          and $fieldName = $namedField
          and $componentName = $namedComponent
          and $subComponentName = $namedSubComponent">
      <xsl:value-of select="'true'" />
    </xsl:when>
    <xsl:otherwise>	  
      <xsl:value-of select="'false'" />
    </xsl:otherwise>
  </xsl:choose>
</xsl:template>

<!-- ............................................................. -->
<xsl:template	name="setCardinalityOption">
  <xsl:param name="segName" select="''"/>
  <xsl:param name="fieldName" select="''"/>
  <xsl:param name="cardinality" select="0"/>
  <xsl:param name="elementNumber" select="0"/>
  <xsl:param name="errorMax" select="0"/>      
  <xsl:param name="errorMin" select="0"/>            
  <xsl:param name="errorLocationMax" select="0"/>
  <xsl:param name="errorLocationMin" select="0"/>

  <!-- we either set card to elementCardinality (value for just 1 element), 
    	an error value (only 1 error), or 'cardinality' (all but 1 element) -->  
	<xsl:choose>
	  <xsl:when test="$elementCardinality != ''
      		and $segName = $namedSegment 
          and $fieldName = $namedField">
    	<!-- segment and field cardinality options take on values from 0 to 1.
        	0 = set card to min for all elements,
          1 = set card to max for all elements.
          values between 0 & 1 are for setting a mix of card
          values between min and max.
          'elementCardinality' is for setting the exact number
          of repetitions for a specific element.  we're adding 10 
          for a specific count, to distinguish between tbe 2 cases. -->
      <xsl:value-of select="$elementCardinality+10"/>
	</xsl:when>        
    <xsl:when test="$errorMax &lt; 0" >
      <xsl:choose>
        <xsl:when test="$segName = $namedSegment
        			and $fieldName = $namedField">
          <xsl:value-of select="-1"/>  <!-- max card error -->
        </xsl:when>
        <xsl:when test="$elementNumber = $errorLocationMax">
          <xsl:value-of select="-1"/>
        </xsl:when>
        <xsl:otherwise>
          <xsl:value-of select="$cardinality"/>          
        </xsl:otherwise>
      </xsl:choose>
    </xsl:when>
    <xsl:when test="$errorMin &lt; 0" >
      <xsl:choose>
        <xsl:when test="$segName = $namedSegment
        			and $fieldName = $namedField">
          <xsl:value-of select="-0.1"/>		<!-- min card error -->
        </xsl:when>
        <xsl:when test="$elementNumber = $errorLocationMin">
          <xsl:value-of select="-0.1"/>
        </xsl:when>
        <xsl:otherwise>
          <xsl:value-of select="$cardinality"/>          
        </xsl:otherwise>
      </xsl:choose>
    </xsl:when>
    <xsl:otherwise>
      <xsl:value-of select="$cardinality"/>
    </xsl:otherwise>
  </xsl:choose>
</xsl:template>

<!-- ............................................................. -->
<xsl:template	name="totalLengths">
  <xsl:param name="nodeList" select="''"/>

  <xsl:choose>
    <xsl:when test="$nodeList">
      <xsl:variable name="head">
        <xsl:value-of select="$nodeList[1]"/>
      </xsl:variable>
      <xsl:variable name="subTotal">
        <xsl:call-template name="totalLengths">
          <xsl:with-param name="nodeList" select="$nodeList[position() != 1]"/>
        </xsl:call-template>
      </xsl:variable>
     
      <xsl:value-of select="$head/@Length + $subTotal"/>
    </xsl:when>
    <xsl:otherwise>0</xsl:otherwise>
  </xsl:choose>
</xsl:template>

<!-- ............................................................. -->
<xsl:template	name="Warn">
  <xsl:param name="segment" select="''"/>
  <xsl:param name="field" select="''"/>
  <xsl:param name="component" select="''"/>
  <xsl:param name="message" select="''"/>

  <xsl:message>
    WARNING: Segment, <xsl:value-of select="$segment"/>
    Field, <xsl:value-of select="$field"/>
    <xsl:if test="$component != ''">
      Component, <xsl:value-of select="$component"/>
    </xsl:if>
    <xsl:value-of select="$message"/>	
  </xsl:message>
  <xsl:comment>
    WARNING: Field, <xsl:value-of select="$field"/>
    <xsl:if test="$component != ''">
      Component, <xsl:value-of select="$component"/>
    </xsl:if>
    <xsl:value-of select="$message"/>	
  </xsl:comment>	
</xsl:template>
      
<!-- ............................................................. -->
<xsl:template	name="isPrimitive">
  <xsl:param name="dataType"/>
  
  <xsl:choose>	
    <xsl:when test="$dataType='ST'
          or $dataType='NM' 
          or $dataType='ID'
          or $dataType='DT'
          or $dataType='SI'
          or $dataType='TN'
          or $dataType='TM'
          or $dataType='TX'
          or $dataType='IS'">
      <xsl:value-of select="'true'"/>
    </xsl:when>
    <xsl:otherwise>	
      <xsl:value-of select="'false'"/>
    </xsl:otherwise>
  </xsl:choose>
</xsl:template>

<!-- ............................................................. -->
<xsl:template name="generateLengthError">
  <xsl:param name="dataType" select="''"/>
  <xsl:param name="maxLength" select="0"/>
  <xsl:param name="dataValue" select="''"/>
  
  <!-- we're only generating length errors for String (ST) and
      numeric values (NM).  (The data type should always be set to
      one of these since these are the only elements designated for 
      length errors in the java module XsltTranform.) -->
  <xsl:choose>
    <xsl:when test="string-length($dataValue) &gt; $maxLength">
      <xsl:value-of select="$dataValue"/>
    </xsl:when>
    <xsl:otherwise>
      <xsl:choose>
        <xsl:when test="$dataType='NM'">
          <xsl:call-template name="generateLengthError">
            <xsl:with-param name="dataType" select="$dataType"/>
            <xsl:with-param name="maxLength" select="$maxLength"/>
            <xsl:with-param name="dataValue" 
                              select="concat($dataValue, '9')"/>
          </xsl:call-template>
        </xsl:when>
        <xsl:otherwise>
        <!-- <xsl:when test="$dataType='ST'"> -->
          <xsl:call-template name="generateLengthError">
            <xsl:with-param name="dataType" select="$dataType"/>
            <xsl:with-param name="maxLength" select="$maxLength"/>
            <xsl:with-param name="dataValue" 
                              select="concat($dataValue, 'X')"/>
          </xsl:call-template>
        </xsl:otherwise>
      </xsl:choose>
    </xsl:otherwise>
  </xsl:choose>
</xsl:template>

</xsl:stylesheet>
